/*
3d.c

Functions for doing some 3D transformations

By Andrei Ellman.
*/

/*
TODO:
+ Fixed-point versions of reverse_persp_project_f() and clip_poly_to_frustum_f()
+ More efficient version of clip_poly_to_frustum() (see Foley and VanDam)
+ clip_object_to_frustum()
*/

#include <math.h>
#include <allegro.h>

#include "3d.h"

#define RAD2HDEG (128.0 / M_PI)	// Multiply by this to convert radians to 'Hexadecimal-degrees' (256 hdegrees in a circle)
#define HDEG2RAD (M_PI / 128.0) // The reverse of the above

/*
Find the Z component of the normal vector of the polygon (must be convex).
This is used mainly in back-face culling.
The back-faces of closed polyhedra are never visible to the viewer.
Therefore, they need never be drawn. This can cull on average half the polygons from a scene.
If the vertices of the polygon form a clockwise path, then the normal faces the front (the Z of the cross-product is negative/coming toward the viewer) and the polygon must be culled. The normal points in the opposite direction that the polygon is facing.
This assumes that vertex1 is directly connected to vertex2 which is directly connected to vertex3.
For a single convex polyhedron, this is sufficient for calculating the complete set of polygons to be rendered.
This function returns the Z value of the cross-product to cull backfaces.
The X and Y components of the crosss-product need not be calculated.
Reccomended use of function in right-handed co-ordinate system (which Allegro uses):
	if (z_normal_polygon(&vtx1, &vtx2, &vtx3) < 0) {
		// The back-face is not going to be drawn.
		// Do something to prevent that polygon from being drawn
	}
Note: if should_cull_backface returns 0, then the polygon is perpendicular to the screen.
Therefore, polygons with normal's Z = 0 can be culled too.
*/
fixed z_normal_polygon(V3D *vertex1, V3D *vertex2, V3D *vertex3)
{
	return (fmul(vertex2->x-vertex1->x, vertex3->y-vertex2->y) - fmul(vertex3->x-vertex2->x, vertex2->y-vertex1->y));
}

/* Float version of above */
float z_normal_polygon_f(V3D_f *vertex1, V3D_f *vertex2, V3D_f *vertex3)
{
	return ( (vertex2->x-vertex1->x) * (vertex3->y-vertex2->y) - (vertex3->x-vertex2->x) * (vertex2->y-vertex1->y) );
}



/*
Gets the latitude on a sphere from a point in 3D space.
The angle between the point's position vector
and the point's position vector projected to the y=0 plane
As the pole is the Y axis (the line {0,1,0}) the plane normal to the pole is y=0.
North = 0 - +64, South = 0 - -64
*/
fixed get_latitude(fixed Px, fixed Py, fixed Pz)
{
	// The vector projected to the plane y=0
	fixed Ptempx = Px;
	fixed Ptempy = 0;
	fixed Ptempz = Pz;

	/*
	Do these special cases when at the poles
	to prevent a division by 0 when doing the normalised vector
	*/
	if ( (Px == 0) && (Pz == 0) ) {
		if (Py > 0) {
			// Is the point at the North pole?
			return(0x40<<16);
		} else {
			// Otherwise, it's at the South pole
			// Also for when the point is {0,0,0}
			return(0xFFC0<<16);
		}
	}

	/* Find the point on the position vector that intercepts the sphere of radius 1. */
	normalize_vector(&Px, &Py, &Pz);

	/* Find point on the position vector projected to the plane y=0 that intercepts the sphere of radius 1. */
	normalize_vector(&Ptempx, &Ptempy, &Ptempz);

	/* Return the angle between the two vectors - the latitude	*/
	return(facos(dot_product(Px,Py,Pz,Ptempx,Ptempy,Ptempz)) * SGN(Py));
}

/* float version of above */
float get_latitude_f(float Px, float Py, float Pz)
{
	// The vector projected to the plane y=0
	float Ptempx = Px;
	float Ptempy = 0.0;
	float Ptempz = Pz;

	/*
	Do these special cases when at the poles
	to prevent a division by 0 when doing the normalised vector
	*/
	if ( (Px == 0.0) && (Pz == 0.0) ) {
		if (Py > 0.0) {
			// Is the point at the North pole?
			return(64.0);
		} else {
			// Otherwise, it's at the South pole
			// Also for when the point is {0,0,0}
			return(-64.0);
		}
	}

	/* Find the point on the position vector that intercepts the sphere of radius 1. */
	normalize_vector_f(&Px, &Py, &Pz);

	/* Find point on the position vector projected to the plane y=0 that intercepts the sphere of radius 1. */
	normalize_vector_f(&Ptempx, &Ptempy, &Ptempz);

	/* Return the angle between the two vectors - the latitude	*/
	return(acos(dot_product_f(Px,Py,Pz,Ptempx,Ptempy,Ptempz)) * SGN(Py) * RAD2HDEG);
}


/*
Gets the longitude on a sphere from a point in 3D space.

The angle between the point's position vector projected to the y=0 plane
and the line {0,0,-1} (This is the meridian, so longitude=0)
As the pole is the Y axis (the line {0,1,0}) the plane normal to the pole is y=0.
Note: Here, longitude increases when going east.
East = 0 - +128, West = 0 - -128
*/
// Is it necessary to pass the y value? Only to maintain consistncy, and also as a preparation in case I get to specify alternative poles and meridians.
fixed get_longitude(fixed Px, fixed Py, fixed Pz)
{
	/*
	Do the special case when at the poles to prevent a division by 0
	Is this necesary, or does atan2() take care of this for us?
	*/
	if ( (Px == 0) && (Pz == 0) ) {
		return(0);	/* At the pole or the middle, the longitude is irrelevant. */
	}

	/*
	return the angle between {0,0,-1} and {px,0,pz} - the longitude
	The angle on the y=0 plane made by the line of gradient px/pz with respect to the meridian {0,0,-1} is the longitude.
	*/
	return(fatan2(-Px,-Pz));
}

/* float version of above */
float get_longitude_f(float Px, float Py, float Pz)
{

	/*
	Do the special case when at the poles to prevent a division by 0 when doing the normalised vector
	*/
	if ( (Px == 0.0) && (Pz == 0.0) ) {
		return(0.0);	/* At the pole, the longitude is irrelevant. */
	}

	/*
	return the angle between {0,0,-1} and {px,0,pz} - the longitude
	The angle made by the line of gradient px/pz with respect to the meridian {0,0,-1} is the longitude.
	*/
	return(atan2(-Px,-Pz) * RAD2HDEG);
}


/*
Get a matrix that rotates a point around the origin to the specified 3D polar co-ordinates at the same altitude from the origin.
Use returned matrix to rotate the vector (0,0,-altitude (ie. the vector facing the viewer) to the correct longitude and latitude)
Uses latitude and longitude model of spherical positioning with the Y axis being the pole.
To rotate around a point that is not the origin, first rotate around the origin, and then displace rotated vector by the vector that goes from the origin to the point.
The longitude and latitude values wrap around properly if the input is outside the normal range, so no range checking needs to be done.
Note: The same longitude can be two places (determined by latitude), but if latitude restricted to { -64 - +64} (0x00-0x40 and 0xC0-0xFF), then that isn't so.
*/
void get_polar_positioning_matrix(MATRIX *polar_position_matrix, fixed lat, fixed lon)
{
	MATRIX tmp_rotation_matrix_1;
	MATRIX tmp_rotation_matrix_2;

	get_x_rotate_matrix(&tmp_rotation_matrix_1, lat);	// First do the angular distance from poles
	get_y_rotate_matrix(&tmp_rotation_matrix_2, lon);	// Then rotate around the pole
	matrix_mul(&tmp_rotation_matrix_1, &tmp_rotation_matrix_2, polar_position_matrix);
}

/* Float version of above */
void get_polar_positioning_matrix_f(MATRIX_f *polar_position_matrix, float lat, float lon)
{
	MATRIX_f tmp_rotation_matrix_1;
	MATRIX_f tmp_rotation_matrix_2;

	get_x_rotate_matrix_f(&tmp_rotation_matrix_1, lat);	// First do the angular distance from poles
	get_y_rotate_matrix_f(&tmp_rotation_matrix_2, lon);	// Then rotate around the pole
	matrix_mul_f(&tmp_rotation_matrix_1, &tmp_rotation_matrix_2, polar_position_matrix);
}


/*
Rotate the camera around a point (xmiddle,ymiddle,zmiddle) using the 3D polar (ie. latitude longitude and altitude) model of spherical positioning (the pole is the Y axis). the latitude, longitude, and altitude paramaters are distances relative to the current position in 3D polar space.
Also points camera to face (xmiddle,ymiddle,zmiddle) and changes 'up' vector accordingly.
Camera altitude (distance from middle) and camera spin (clockwise rotation of camera) are applied as well.
A new camera matrix is constructed for the camera when it's rotated. (Maybe this need not be in the function. For example, if you displace the camera after it has been pointed?)

Note: At the two poles, changing longitude is equivalent to changing spin in the other direction.

Note: If you are just changing the position of the point which the camera is pointing at, just use get_camera_matrix with the same rotations but a different position.

Note: To adjust altitude, find vector from middle to camera, multiply this by new_altitude/old_altitude, and add the vector to the middle to get new displacement. See above to change displacement.
*/

void polar_rotate_camera(camera *the_camera, fixed dlat, fixed dlon, fixed dalt, fixed dspin, fixed xmiddle, fixed ymiddle, fixed zmiddle)
{
	MATRIX polar_positioning_matrix, spin_rotation_matrix;

	/*
	 * First adjust the latitude, longitude, spin, and altitude
	 */

	the_camera->lon += dlon;
	the_camera->lon &= 0x00FFFFFF;
	the_camera->lat += dlat;
	the_camera->lat &= 0x00FFFFFF;
	the_camera->spin += dspin;
	the_camera->spin &= 0x00FFFFFF;
	the_camera->alt += dalt;

	/*
	 * Calculate the coresponding camera position in worldspace.
	 */

	/*
	Find the vector 0,0,-the_camra.alt and rotate it with respect to lat and lon
	This is the camera's position vector relative to (xmiddle, ymiddle, zmiddle)
	*/
	get_polar_positioning_matrix(&polar_positioning_matrix, the_camera->lat, the_camera->lon);
	apply_matrix(&polar_positioning_matrix, 0,0,-the_camera->alt, &the_camera->x,&the_camera->y,&the_camera->z);

	/*
	Make the camera's front vector face the middle
	The direction is the reverse direction of the camera's position vector.
	*/

	the_camera->xf= - the_camera->x;
	the_camera->yf= - the_camera->y;
	the_camera->zf= - the_camera->z;

	/*
	 * Change the 'up' vector
	 */

	/*
	Rotate the 'up' vector to remain perpendicular to the front vector
	*/
	// Perhaps this isn't necessary if camera's direction vector is never paralell to the pole and spin isn't used.
	apply_matrix(&polar_positioning_matrix, 0,1<<16,0, &the_camera->xup,&the_camera->yup,&the_camera->zup);

	/*
	SPIN:
	Rotate the 'up' vector around the 'front' vector
	*/

	get_vector_rotation_matrix(&spin_rotation_matrix, the_camera->x, the_camera->y, the_camera->z, the_camera->spin);
	apply_matrix(&spin_rotation_matrix, the_camera->xup,the_camera->yup,the_camera->zup, &the_camera->xup,&the_camera->yup,&the_camera->zup);
	// Is two applies slower than matrix mul?

	/*
	 * Because the camera wasn't rotated around the origin, it needs displacing
	 * Transalate from 'middle' space to world space.
	 */

	the_camera->x+=xmiddle;
	the_camera->y+=ymiddle;
	the_camera->z+=zmiddle;

	/*
	 * And now, change the camera matrix.
	 */

	get_camera_matrix(&the_camera->camera_matrix, the_camera->x, the_camera->y, the_camera->z, the_camera->xf, the_camera->yf, the_camera->zf, the_camera->xup, the_camera->yup, the_camera->zup, the_camera->fov, the_camera->aspect);

}


/* Float version of above */
void polar_rotate_camera_f(camera_f *the_camera, float dlat, float dlon, float dalt, float dspin, float xmiddle, float ymiddle, float zmiddle)
{
	MATRIX_f polar_positioning_matrix, spin_rotation_matrix;

	/*
	 * First adjust the latitude, longitude, spin, and altitude
	 */

	the_camera->lon += dlon;
	if (the_camera->lon>=256.0) {
		the_camera->lon-=256.0;
	} else if (the_camera->lon<0.0) {
		the_camera->lon+=256.0;
	}

	the_camera->lat += dlat;
	if (the_camera->lat>=256.0) {
		the_camera->lat-=256.0;
	} else if (the_camera->lat<0.0) {
		the_camera->lat+=256.0;
	}

	the_camera->spin += dspin;
	if (the_camera->spin>=256.0) {
		the_camera->spin-=256.0;
	} else if (the_camera->spin<0.0) {
		the_camera->spin+=256.0;
	}

	the_camera->alt += dalt;

	/*
	 * Calculate the coresponding camera position in worldspace.
	 */

	/*
	Find the vector 0.0,0.0,-the_camra.alt and rotate it with respect to lat and lon
	This is the camera's position vector relative to (xmiddle, ymiddle, zmiddle)
	*/
	get_polar_positioning_matrix_f(&polar_positioning_matrix, the_camera->lat, the_camera->lon);
	apply_matrix_f(&polar_positioning_matrix, 0.0,0.0,-the_camera->alt, &the_camera->x,&the_camera->y,&the_camera->z);

	/*
	Make the camera's front vector face the middle
	The direction is the reverse direction of the camera's position vector.
	*/

	the_camera->xf= - the_camera->x;
	the_camera->yf= - the_camera->y;
	the_camera->zf= - the_camera->z;

	/*
	 * Change the 'up' vector
	 */

	/*
	Rotate the 'up' vector to remain perpendicular to the front vector
	*/
	// Perhaps this isn't necessary if camera's direction vector is never paralell to the pole and spin isn't used.
	apply_matrix_f(&polar_positioning_matrix, 0.0,1.0,0.0, &the_camera->xup,&the_camera->yup,&the_camera->zup);

	/*
	SPIN:
	Rotate the 'up' vector around the 'front' vector
	*/

	get_vector_rotation_matrix_f(&spin_rotation_matrix, the_camera->x, the_camera->y, the_camera->z, the_camera->spin);
	apply_matrix_f(&spin_rotation_matrix, the_camera->xup,the_camera->yup,the_camera->zup, &the_camera->xup,&the_camera->yup,&the_camera->zup);
	// Is two applies slower than matrix mul?

	/*
	 * Because the camera wasn't rotated around the origin, it needs displacing
	 * Transalate from 'middle' space to world space.
	 */

	the_camera->x+=xmiddle;
	the_camera->y+=ymiddle;
	the_camera->z+=zmiddle;

	/*
	 * And now, change the camera matrix.
	 */

	get_camera_matrix_f(&the_camera->camera_matrix, the_camera->x, the_camera->y, the_camera->z, the_camera->xf, the_camera->yf, the_camera->zf, the_camera->xup, the_camera->yup, the_camera->zup, the_camera->fov, the_camera->aspect);

}


/*
Does the reverse of Allegro's persp_project_f()
If this was part of Allegro's 3D.c, then 1.0/_persp_[x|y]scale_f can be pre-calculated each time the projection viewport changes.
*/
void reverse_persp_project_f(float x, float y, float z, float *xout, float *yout)
{
	/*
	These are defind in Allegro's 3d.c file.
	There is a possible hazard if variables with the same name are used elsewhere
	*/

   extern float _persp_xscale_f;
   extern float _persp_yscale_f;
   extern float _persp_xoffset_f;
   extern float _persp_yoffset_f;

	*xout = (z * (x-_persp_xoffset_f)) / _persp_xscale_f;
	*yout = (z * (y-_persp_yoffset_f)) / _persp_yscale_f;
}


/*
The next two functions ease converting code between using fixed point variables
and floating point variables by making the transition easier by alowing you
to mix fixed and float matrices.
*/

/*
Converts a fixed point matrix (out_matrix) to a floating point matrix (in_matrix_f).
*/
void matrix_to_matrix_f(MATRIX *in_matrix, MATRIX_f *out_matrix_f)
{
	int i,j;

	for(i=0;i<3;i++) {
		for(j=0;j<3;j++) {
			out_matrix_f->v[i][j]=fixtof(in_matrix->v[i][j]);	/* Scaling and rotation */
		}
		out_matrix_f->t[i]=fixtof(in_matrix->t[i]);	/* Translation */
	}

}


/*
Converts a floating point matrix (in_matrix_f) to a fixed point (out_matrix) matrix.
*/
void matrix_f_to_matrix(MATRIX_f *in_matrix_f, MATRIX *out_matrix)
{
	int i,j;

	for(i=0;i<3;i++) {
		for(j=0;j<3;j++) {
			out_matrix->v[i][j]=ftofix(in_matrix_f->v[i][j]);	/* Scaling and rotation */
		}
		out_matrix->t[i]=ftofix(in_matrix_f->t[i]);	/* Translation */
	}

}


/* ----------------------------------------------------------------- */


/*
There now follows several functions to help
clip polygons to the viewing frustum.
This is designed just to clip polygons, and not an object composed of several polygons (unless none of the polygons shared vertices with otgher polygons).
*/


/*
Helper function for clip_poly_to_frustum_f()
Function for finding if a point is in behind the plane
Should I use function pointers instead of case statements and passing the plane number, or should I manually inline and unroll the loop?
*/
int is_point_behind_plane_f(V3D_f *the_point, float near_z, int plane)
{
	switch (plane) {

		case 0: {
			/* Z=near_z */
			return (the_point->z < near_z);
		};break;
		case 1: {
			/* Z=X */
			return (the_point->z < the_point->x);
		};break;
		case 2: {
			/* Z=-X */
			return (the_point->z < -the_point->x);
		};break;
		case 3: {
			/* Z=Y */
			return (the_point->z < the_point->y);
		};break;
		case 4: {
			/* Z=-Y */
			return (the_point->z < -the_point->y);
		};break;

	}

	return(0); // This should stop the compiler whingeing.
}


/*
Helper function for clip_poly_to_frustum_f()
Function for finding the position on the line-vector connecting the 2 passed vertices that intercepts the the plane.
All lines passed to this function intercept the plane.
Should I use function pointers instead of case statements and passing the plane number, or should I manually inline and unroll the loop?

Can a divide_by_zero occur when one point is behind and the other is on? No.

Are there any instances where position can become a very large number?

For Z=Z_CLIP_NEAR_PLANE:
We know that the_z_value > v_behind_adr->z and z <= v_infront_adr->z
v_behind_adr->z < v_infront_adr->z	=> Division by 0 not possible.
Therefore the_z_value-v_behind_adr->z <= v_infront_adr->z-v_behind_adr->z
position = ( (the_z_value-v_behind_adr->z) / (v_infront_adr->z-v_behind_adr->z) )
the_z_value = v_behind_adr->z + position*(v_infront_adr->z-v_behind_adr->z)
*/
float find_position_on_line_f(V3D_f *v_behind_adr, V3D_f *v_infront_adr, float near_z, int plane)
{

	switch (plane) {

		case 0: {
			/* Z=near_z */
			return (near_z-v_behind_adr->z) / (v_infront_adr->z-v_behind_adr->z);
		};break;
		case 1: {
			/* Z=X */
			/*
			let v = p2 - p1
			z=x => p.z + K*v.z = p.x + K*v.x => K = (p.z - p.x) / (v.x - v.z)
			=> K = (p1.z - p1.x) / (p2.x - p1.x - p2.z + p1.z)
			*/
			return (v_behind_adr->z - v_behind_adr->x) / (v_infront_adr->x-v_behind_adr->x-v_infront_adr->z+v_behind_adr->z);
		};break;
		case 2: {
			/* Z=-X */
			/*
			let v = p2 - p1
			z=-x => p.z + K*v.z = - (p.x + K*v.x) => K = (p.z + p.x) / (v.z + v.x)
			=> K = (p1.z + p1.x) / ( - p2.z + p1.z - p2.x + p1.x)
			*/
			return (v_behind_adr->z + v_behind_adr->x) / (-v_infront_adr->z+v_behind_adr->z-v_infront_adr->x+v_behind_adr->x);
		};break;
		case 3: {
			/* Z=Y */
			/*
			let v = p2 - p1
			z=y => p.z + K*v.z = p.y + K*v.y => K = (p.z - p.y) / (v.y - v.z)
			=> K = (p1.z - p1.y) / (p2.y - p1.y - p2.z + p1.z)
			*/
			return (v_behind_adr->z - v_behind_adr->y) / (v_infront_adr->y-v_behind_adr->y-v_infront_adr->z+v_behind_adr->z);
		};break;
		case 4: {
			/* Z=-Y */
			/*
			let v = p2 - p1
			z=-y => p.z + K*v.z = - (p.y + K*v.y) => K = (p.z + p.y) / (v.z + v.y)
			=> K = (p1.z + p1.y) / ( - p2.z + p1.z - p2.y + p1.y)
			*/
			return (v_behind_adr->z + v_behind_adr->y) / (-v_infront_adr->z+v_behind_adr->z-v_infront_adr->y+v_behind_adr->y);
		};break;
	}

	return(0); // This should stop the compiler whingeing.
}



/*
Helper function for clip_poly_to_frustum_f()
Given two vertices, a 3rd vertex is created along the line-vector at 'position', and written to the address passed to the function.
'position' was calculated in find_position_on_line_f(), and is where the line intercepts.
As this is general-purpose, this also clips u,v,c, which may be un-necessary.

Perhaps, I could pass the value of the type paramater of the polygon drawer (POLYTYPE_FLAT, etc.), and it would clip the appropriate u,v,c co-ords (if PTEX, I may need a special texture-co-ord clipper).
*/
void clip_line_to_plane_f(V3D_f *v_behind_adr, V3D_f *v_infront_adr, float position, V3D_f *new_vertex_adr)
{

	/*
	If this function wasn't general-purpose, then for when clipping to Z plane, use:
	// new_vertex_adr->z = Z_CLIP_NEAR_PLANE;	// new Z is just the z plane

	For clipping to z=x, z=-x, z=y, z=-y, we can let x=+-z or y=+-z.
	*/

	new_vertex_adr->z = v_behind_adr->z + (position * (v_infront_adr->z-v_behind_adr->z));
	new_vertex_adr->y = v_behind_adr->y + (position * (v_infront_adr->y-v_behind_adr->y));
	new_vertex_adr->x = v_behind_adr->x + (position * (v_infront_adr->x-v_behind_adr->x));

	/*
	Unless the polygon is gouraud-shaded and texture-mapped,
	some of these will be un-necessary

	QUESTION: Are the equations for u and v sufficient for Perspective-correct texture-mapped polygons?
	ANSWER: Yes, because the poly is clipped in viewspace, and PTEX is done in screenspace.
	*/

	new_vertex_adr->u = v_behind_adr->u + (position * (v_infront_adr->u-v_behind_adr->u));
	new_vertex_adr->v = v_behind_adr->v + (position * (v_infront_adr->v-v_behind_adr->v));
	new_vertex_adr->c = v_behind_adr->c + (position * (v_infront_adr->c-v_behind_adr->c));

}





/*
Clips a poly to the viewing frustum and returns resulting number of vertices in the clipped poly.
Does not do the far Z=K plane.

This is a general purpose routine that clips the x,y,z co-ords and also clips the texture co-ords (u,v) and the goraud-lighting co-ord (c).

out_vertices_1 contains the address of the first vertex of the unclipped polygon (use the output of the camera transformations).
out_vertices_2 contains the address the vertices of the clipped polygon will be written to.

Note: Both the source address and the destination address must contain room for enough V3D_f structures for the source polygon (vertex_count) + NUM_CLIPPING_PLANES (defined by 3d.h to be 5 clipping planes in this case).

For example, say we have V3D_f arrays source_vertices[] and dest_vertices[] of size NUM_VERTICES_IN_POLYGON+NUM_CLIPPING_PLANES, we call this function like this.
	source_vertices contains vertices after the camera-matrix has been applied.
	vertex_count=clip_poly_to_frustum_f(source_vertices, dest_vertices, NUM_VERTICES_IN_POLYGON, Z_CLIP_NEAR_PLANE);
	// Some function for perspective-correction goes here.
	// The returned vertex_count can be used as the vertex_count of the polygon/triangle/quad3d_f function.

// Should I have a startc tmp_vertex_array so I don't have to worry about having a source array of size NUM_VERTICES_IN_POLYGON+NUM_CLIPPING_PLANES. I would have to make this array take into account the size of the polygon with the largest number of verices.

// If I expand to do 6 planes, it would output the results into the source, and require a static temporary vertex array.

// Note: It could be possible to make the source address have room for one less vertex by re-writing this code a bit.


Would this would have been better with a circular doubly-linked-list structure?
Allegro's polygon3D function takes an array and a vertex_count, which is best produced by using arrays. With a linked-list, I would have had to copy each node into an array in the order they go round the polygon.
*/
int clip_poly_to_frustum_f(V3D_f *out_vertices_1, V3D_f *out_vertices_2, int vertex_count, float near_z)
{
	int plane;
	int unclipped_i, next_unclipped_i, out_i;

	V3D_f *tmp_V3D_ptr;	// Used for swapping out_vertioces_[1|2]

	float position;	/* A posotion along the line-vector to clip */

	int clipped_vertex_count;	/* Number of vertices after a round of clipping */


	for (plane=0;plane<NUM_CLIPPING_PLANES;plane++) {

		for (unclipped_i=0,clipped_vertex_count=0;unclipped_i<vertex_count;unclipped_i++) {

			next_unclipped_i = + ((unclipped_i+1)%vertex_count);
			out_i = clipped_vertex_count;

			/* With no clipping, it copies the poly info like this
			// out_vertices_2[out_i] = out_vertices_1[unclipped_i];
			// clipped_vertex_count++;
			*/

			/* Should vertex be unchanged, moved onto clip-plane or deleted? */
			if (is_point_behind_plane_f(&out_vertices_1[unclipped_i], near_z, plane)) {
				/*
				This vertex is behind,
				so we move it along the line onto the plane, or not include it
				*/

				if (!is_point_behind_plane_f(&out_vertices_1[next_unclipped_i], near_z, plane)) {
					/* This vertex is behind, but next isn't, so we move it to the plane */

					position=find_position_on_line_f(&out_vertices_1[unclipped_i],&out_vertices_1[next_unclipped_i],near_z,plane);
					clip_line_to_plane_f(&out_vertices_1[unclipped_i],&out_vertices_1[next_unclipped_i], position, &out_vertices_2[out_i]);
					clipped_vertex_count++;
				}

			} else {
				/* This vertex isn't behind */
				out_vertices_2[out_i] = out_vertices_1[unclipped_i];
				clipped_vertex_count++;

				/* Should a vertex be added? */
				if (is_point_behind_plane_f(&out_vertices_1[next_unclipped_i], near_z, plane)) {
					/*
					The next vertex is behind
					so we must create a new vertex that intercepts plane
					*/

					position=find_position_on_line_f(&out_vertices_1[next_unclipped_i], &out_vertices_1[unclipped_i],near_z,plane);
					clip_line_to_plane_f(&out_vertices_1[next_unclipped_i], &out_vertices_1[unclipped_i], position, &out_vertices_2[out_i+1]);
					clipped_vertex_count++;
				}
			}

		}

		/* Swap the two pointers for the next iteration */
		tmp_V3D_ptr = out_vertices_1;
		out_vertices_1 = out_vertices_2;
		out_vertices_2 = tmp_V3D_ptr;

		vertex_count=clipped_vertex_count;
	}

	return vertex_count;
}


/* ----------------------------------------------------------------- */
