/*
RGB <-> HSV and RGB <-> HLS
Algorithms taken from Chapter 13 of "Principles of Computer Graphics" by Foley And VanDam
Andrei has corrected a bug that appeared in the original HLS -> RGB function.
*/

/*
TODO:
HLS <-> HSV
Fixed point versions of these algorithms. While I'm there, make the RGB space 0-1 instead of 0-63.
*/

#include <math.h>	// Needed for floor
#include <allegro.h>	// Needed for RGB struct

#include "hsv_hls.h"

#define MAX_INTENSITY 63.0

/* hsv_to_rgb:
 *  Converts from HSV colorspace to RGB values.
 *  Taken From Foley & VanDam Fig. 13.34 with a few modifications.
 *  Input: h[0.0,6.0 or UNDEFINED_COLOUR] s[0.0,1.0 or UNDEFINED_COLOUR] v[0.0,1.0]
 *  (s must be 0.0 when h is UNDEFINED_COLOUR, v must be 0.0 when s is UNDEFINED_COLOUR)
 *  Output: An RGB structure. each RGB in [0,MAX_INTENSITY]
 *  Note: This does not check paramaters for the valid ranges,
 *  but h=6.0 becomes h=0.0
 */
void xhsv_to_rgb (float h, float s, float v, RGB *rgb)
{
   float i, f, p, q, t;

   if (s == 0.0 || s == UNDEFINED_COLOUR) {   /* greyscale (h should ideally be UNDEFINED_COLOUR) */
      rgb->r = rgb->g = rgb->b = (int)(v * MAX_INTENSITY);
   }
   else {
      if (h == 6.0)
	 h = 0.0;	/* Hue is a cyclic paramater measued from 0-6 */

      i = floor(h);	/* i is largest integer <= h */
      f = h - i;	/* f is fractional part of h */
      p = (v * (1.0 - s) * MAX_INTENSITY);
      q = (v * (1.0 - (s * f)) * MAX_INTENSITY);
      t = (v * (1.0 - (s * (1.0 - f))) * MAX_INTENSITY);

      switch ((int)i) {
	 case 0:
	 	rgb->r = (unsigned char)(v * MAX_INTENSITY);
	 	rgb->g = (unsigned char)(t * MAX_INTENSITY);
		rgb->b = (unsigned char)(p * MAX_INTENSITY);
		break;
	 case 1:
	 	rgb->r = (unsigned char)(q * MAX_INTENSITY);
		rgb->g = (unsigned char)(v * MAX_INTENSITY);
		rgb->b = (unsigned char)(p * MAX_INTENSITY);
		break;
	 case 2:
	 	rgb->r = (unsigned char)(p * MAX_INTENSITY);
		rgb->g = (unsigned char)(v * MAX_INTENSITY);
		rgb->b = (unsigned char)(t * MAX_INTENSITY);
		break;
	 case 3:
	 	rgb->r = (unsigned char)(p * MAX_INTENSITY);
		rgb->g = (unsigned char)(q * MAX_INTENSITY);
		rgb->b = (unsigned char)(v * MAX_INTENSITY);
		break;
	 case 4:
	 	rgb->r = (unsigned char)(t * MAX_INTENSITY);
		rgb->g = (unsigned char)(p * MAX_INTENSITY);
		rgb->b = (unsigned char)(v * MAX_INTENSITY);
		break;
	 case 5:
	 	rgb->r = (unsigned char)(t * MAX_INTENSITY);
		rgb->g = (unsigned char)(p * MAX_INTENSITY);
		rgb->b = (unsigned char)(q * MAX_INTENSITY);
		break;
      }
   }
}	/* End of hsv_to_rgb */


/* rgb_to_hsv:
 *  Converts an RGB value into the HSV colorspace.
 *  Taken From Foley & VanDam Fig. 13.33 with a few modifications.
 *  Input: An RGB structure. each RGB in [0,MAX_INTENSITY]
 *  Output: h[0.0,6.0 or UNDEFINED_COLOUR] s[0.0,1.0 or UNDEFINED_COLOUR] v[0.0,1.0]
 * 	h is UNDEFINED_COLOUR when s=0.0, s is UNDEFINED_COLOUR when v=0.0
 *  Note: This does not check paramaters for the valid ranges,
 */
void xrgb_to_hsv(RGB *rgb, float *h, float *s, float *v)
{
   float min, max, delta, r, g, b;

   r = (float)(rgb->r) / MAX_INTENSITY;
   g = (float)(rgb->g) / MAX_INTENSITY;
   b = (float)(rgb->b) / MAX_INTENSITY;

   max = MAX(r, MAX(g, b));
   min = MIN(r, MIN(g, b));

   *v = max;	/* The V component is the maximum of r,g,b */

   if (*v==0.0) {
		/* At bottom apex of hex-cone */
		*h = UNDEFINED_COLOUR;        /* colour has no hue */
		*s = UNDEFINED_COLOUR;        /* colour has no saturation */
   } else {
	   delta = max - min;

	   /* Calculate Saturation: 0.0 if r,g,b all 0 */
	   *s = (max != 0.0) ? (delta / max) : 0.0;

	   if (*s == 0.0)
		  *h = UNDEFINED_COLOUR;        /* colour has no hue */
	   else {
		  if (r == max)
		 *h = (g-b) / delta;  /* Resulting colour between yellow & magenta */
		  else if (g == max)
		 *h = 2.0 + (b-r) / delta;  /* Resulting colour between cyan & yellow */
		  else if (b == max)
		 *h = 4.0 + (r-g) / delta;  /* Resulting colour between magenta & cyan */
		  if (*h < 0.0)
		 *h += 6.0;	/* Make sure hue is non-negative */
		}
	}
}	/* End of rgb_to_hsv */


/* hls_to_rgb:
 *  Converts from HLS colorspace to RGB values.
 *  Taken From Foley & VanDam Fig. 13.37 with a few modifications.
 *  Input: h[0.0,6.0 or UNDEFINED_COLOUR] s[0.0,1.0 or UNDEFINED_COLOUR] v[0.0,1.0]
 *  (s must be 0.0 when h is UNDEFINED_COLOUR, l must be 0.0 or 1.0 when s is UNDEFINED_COLOUR )
 *  Output: An RGB structure. each RGB in [0,MAX_INTENSITY]
 *  Note: This does not check paramaters for the valid ranges,
 */

/* First, a helper function for working out a value for a colour chanel */
float hls_to_rgb_value (float n1, float n2, float hue)
{
	if (hue > 6.0)
		hue -= 6.0;
	else if (hue < 0.0)
		hue += 6.0;

	if (hue < 1.0)
		return n1 + (n2-n1)*hue;
	else if (hue < 3.0)
		return n2;
	else if (hue < 4.0)
		return n1 + (n2-n1)*(4.0-hue);
	else
		return n1;
}

/* The function proper */
void hls_to_rgb (float h, float l, float s, RGB *rgb)
{
	float m1, m2;

	/*
	AE: The commented out line appeared in Foley&VanDam
	The line below is it my attempt to correct it. I think it works.
	*/
	/* m2 = (l<=0.5) ? (l*(l+s)) : (l+s - (l*s)); */
	m2 = (l<=0.5) ? (l+(l*s)) : (l+s - (l*s));
	m1 = 2.0 * l - m2;

	if (s == 0.0 || s== UNDEFINED_COLOUR) {   /* greyscale (h should ideally be UNDEFINED_COLOUR here) */
    	rgb->r = rgb->g = rgb->b = (unsigned char)(l * MAX_INTENSITY);
	} else {
		rgb->r = (unsigned char)(hls_to_rgb_value(m1, m2, h+2.0) * MAX_INTENSITY);
		rgb->g = (unsigned char)(hls_to_rgb_value(m1, m2, h) * MAX_INTENSITY);
		rgb->b = (unsigned char)(hls_to_rgb_value(m1, m2, h-2.0) * MAX_INTENSITY);
	}
}	/* End of hls_to_rgb */


/* rgb_to_hls:
 *  Converts an RGB value into the HLS colorspace.
 *  Taken From Foley & VanDam Fig. 13.36 with a few modifications.
 *  Input: An RGB structure. each RGB in [0,MAX_INTENSITY]
 *  Output: h[0.0,6.0 or UNDEFINED_COLOUR] s[0.0,1.0 or UNDEFINED_COLOUR] v[0.0,1.0]
 * 	h is UNDEFINED_COLOUR when s=0.0, s is UNDEFINED_COLOUR when l is 0.0 or when l is 1.0
 *  Note: This does not check paramaters for the valid ranges,
 */
void rgb_to_hls(RGB *rgb, float *h, float *l, float *s)
{
   float min, max, delta, r, g, b;

   r = (float)(rgb->r) / MAX_INTENSITY;
   g = (float)(rgb->g) / MAX_INTENSITY;
   b = (float)(rgb->b) / MAX_INTENSITY;

   max = MAX(r, MAX(g, b));
   min = MIN(r, MIN(g, b));

   *l = (max+min)/2.0;	/* L is the average of maximum and minimum r,g,b */

	if (*l==0.0 || *l==1.0) {
		/* At one of the two apeces of double-hex-cone */
		*h = UNDEFINED_COLOUR;        /* colour has no hue */
		*s = UNDEFINED_COLOUR;        /* colour has no saturation */
	} else {

	   delta = max - min;

	   if (delta == 0.0) {
		   *s = 0.0;
		   *h = UNDEFINED_COLOUR;        /* colour has no hue */

	   } else {

		  /* Calculate Saturation */
		  *s = (*l<=0.5) ? (delta/(max+min)) : (delta/(2.0-(max+min)));

		  /* Calculate Hue */
		  if (r == max)
		 *h = (g-b) / delta;  /* Resulting colour between yellow & magenta */
		  else if (g == max)
		 *h = 2.0 + (b-r) / delta;  /* Resulting colour between cyan & yellow */
		  else if (b == max)
		 *h = 4.0 + (r-g) / delta;  /* Resulting colour between magenta & cyan */
		  if (*h < 0.0)
		 *h += 6.0;	/* Make sure hue is non-negative */
		}
	}
}	/* End of rgb_to_hls */

