/*
main3dsp.c

Main file for testing the 3D sprites testing demo (3DSPDEMO)

This is a demo of AllegroPak's additions to Allegro.

By Andrei Ellman (e-mail: ellman@xs4all.nl)
*/

/*
Currently at Version 1.01b

Changes:

1.01b
+ Uses AllegroPak 1.0 beta 1a
+ Better feedback
+ Has a seperate help-screen leaving the main screen less cluttered.
+ You can now access the display of the key commands by pressing F12.

1.01a
+ First version made available to the general public.
+ Uses AllegroPak 1.0 beta 1
+ Re-arranged the API for drawing zprites. The ZPRITE structure is no longer part of the API, but part of the demo (although the ZPRITE definitions are in their own files (zpriteds.[c|h]). This should allow people to write alternatives to the ZPRITE structure and the way it's initialised. The zprite drawing functions take a bitmap and zbuffer instead of a ZPRITE.
+ Other changes (mostly cosmetic) to the 3D sprite drawing API

1.01
+ Uses a new alpha of AllegroPak with signifiacntly updated 3D and Zprite engine. Amongst it's many improvements are that the bug causing a 1/2^32 chance of the displace_[trans_]zprite function crashing has been eliminated. And also the bug causing the zb_polygon3d_f() function to occasionally crash when using the PTEX modes has been eliminated.
+ Now has full viewing-frustum 3D clipping (near-z, z=x, z=-x, z=y, z=-y, but not the far-z plane)
+ When working out the camera matrix, the affects of yaw, pitch and roll are now applied correctly (yaw, pitch, roll).

1.0
+ First version to be distributed to various people.
+ Uses an alpha of AllegroPak.

Keys not mentioned on the screen:
ALT+S: Save a screendump as a PCX file.
B: Holding down B removes Z-buffering. As well as being useful for tracing bugs, it also shows how much things are slowed down by Z-buffering.
*/

/*
KNOWN BUGS:
Sometimes, if LENGTH_RATIO is too large, the program may crash with an FPE in the polygon drawer's PTEX_* scanline filler. This also happens in Shawn's original code as well as my Z-buffered modification, but I have yet to conclusively prove that this one's not my fault.

When using PTEX_LIT (and I think ATEX_LIT too), the light value wraps back to 255 when it has passed 0. This also happens in Shawn's original code, but as above, I've yet to find out who'se bug it is (Although the wrapping of the light is probarbly the fault of Shawn who originally wrote the polygon drawer). I have worked-around this by making the minimum light value > 0.

All zprites (flat and non-flat) must be flipped when the camera is rotated past the Z=0 plane in worldspace. If the yaw faces behind, the zprites must be flipped horizontally, and if the pitch has rotated past the z=0 plane, the zprites must be flipped vertically.

'Zpheres' are currently not rotated when the camera tilts.
*/

/*
FUTURE:
Have a demo of zb_blit and zb_clear that uses dirty-rectangles in a z-buffered world (eg. some zprites passing through a background of mountains made of spheres). I could also spawn off a third demo of AllegroPak.
Question: How does one best make a 32-bit 1/z depthmap anyway? Perhaps have function for converting linear-level 256 greyscale depth-map to the correct format.

Perhaps I could even demonstrate 2D bump-mapping? (but then, I would have to work out the orientation of each pixel wrt. a lightsource).
+ Draw background using PLIT spheres.
+ When drawing sprite, use the bump-map as a way of reflecting light. However, I would have to write a new zprite rendering routine. This one would work out the angle the sprite makes with the light from the depth-map.

3D sprites should always be able to be blitted directly to screen memory (dirty rectangles), but with polygons, I suggest only writing to memory bitmaps, and in future, will probarbly no longer support drawing polygons directly to the screen.

frustum_CLIPPING: Improve the function (in 3d.c) to clip to all 5 (or 6) planes at once instead of each plane in turn. Have a function that given a closed convex polygon, clips the polygon to the frustum. Returns the resulting vertex count and stores the new points in an array. Should I use a linked-list instead of an array?

Investigate having more accurate light calculation.
Perhaps break up each polygon into triangles that radiate from the point on the wall that's nearest to the lightsource, or the point in the visible wall that's the lightest.
Also, do I need to calculate light according to Z or 1/Z? What are some good gamma values?

When I rotate past the z=0 plane, I should implement [flat]_lit_zprite v,h,vh flip

Fow when the polygons are not moving (ie. the camera isn't moving) , if all zprites are depth-sorted, there needn't be any need to record the new Z values in the Z-buffer, and the z-buffer needn't be cleared (but a copy of the screen-background must exist). I think that this would only work with flat_zprites, unless they don't share any space with other zprites.

Optimise translucent zprite drawer by not re-writing offset each line (re-write it once per zprite) (Check Shawn's e-mail about read-write offsets for older cards first).

Make a better draw_angle() function (show_angle() ) that uses a rotated sprite instead of a line to show the orientation.
Also, have a graphical representation of the displacement.

I should read the keyboard inside the loop that updates the positions instead of just inside the main loop. I do however lose some accuracy in positioning if the framerate is too low.

Have an option to hide the on-screen feedback (F10).
*/


#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include <allegro.h>

#include "global.h"
#include "zbuf.h"
#include "zprite.h"
#include "zbpoly.h"
#include "3d.h"

#include "zpriteds.h"

#include "data.h"

#define SMILEY_XSIZE 48
#define SMILEY_YSIZE 48
#define SMILEY_XSCALE 0.7
#define SMILEY_YSCALE 0.7
#define SPHERE_XSIZE 128
#define SPHERE_YSIZE 128
#define SPHERE_XSCALE 0.14
#define SPHERE_YSCALE 0.14
#define MAXOBJECT_SIZE 128

#define NUM_POLYS 4
#define NUM_VERTICES_IN_POLYGON 4
#define MAX_NUM_VERTICES (NUM_VERTICES_IN_POLYGON+NUM_CLIPPING_PLANES)

#define NUM_FOGS 9
#define NUM_FOG_PALETTES 2

#define NUM_SMILEYS 5	// Used to be 10
#define NUM_SPHERES 5	// Used to be 10

#define NUM_CHARACTERS (NUM_SMILEYS+NUM_SPHERES)

/* Define positions for the textual output */
#define text_yheight 8
#define text_xstart 23
#define textblock_ysize text_xstart
#define text_moreinfo_ystart (text_yheight*2)


#define LENGTH_RATIO 47.8	/* Ratio of tunnel L to tunnel H/W */

#define TUNNEL_SIZE 22.0	// Tunnel diamater
#define TUNNEL_COMPONENT_SIZE 22.0	// width/height of component.
// Actual values for above two are double (due to -X..+X).

/* Z should never be <= 0. If it's > max_depth, the colour is always 0 */
#define MAX_DEPTH (TUNNEL_COMPONENT_SIZE*LENGTH_RATIO)
// Note: before z-clipping was implemented, tunnel stretched back twice this far due to extra Z-displacement

#define Z_CLIP_NEAR_PLANE 3.0

#define MIN_DIST_TO_WALL (Z_CLIP_NEAR_PLANE*2.0)	/* The closet you can displace towards a wall */

#define GROUND_TEXTURE_XMAX 128.0
#define GROUND_TEXTURE_YMAX (GROUND_TEXTURE_XMAX*LENGTH_RATIO)
#define GROUND_XSIZE TUNNEL_COMPONENT_SIZE
#define GROUND_YSIZE (GROUND_XSIZE*LENGTH_RATIO)

#define TOP_TEXTURE_XMAX 128.0
#define TOP_TEXTURE_YMAX (TOP_TEXTURE_XMAX*LENGTH_RATIO)
#define TOP_XSIZE TUNNEL_COMPONENT_SIZE
#define TOP_YSIZE (TOP_XSIZE*LENGTH_RATIO)

#define SIDE_TEXTURE_XMAX 512.0
#define SIDE_TEXTURE_YMAX (SIDE_TEXTURE_XMAX*LENGTH_RATIO)
#define SIDE_XSIZE TUNNEL_COMPONENT_SIZE
#define SIDE_YSIZE (SIDE_XSIZE*LENGTH_RATIO)

/* The limits of the flying characters */
#define CHARACTER_ZMIN (Z_CLIP_NEAR_PLANE+5.0)
#define CHARACTER_ZMAX (MAX_DEPTH-(SPHERE_XSIZE*2.0))

#define CHARACTER_XYLIMITS_RATIO 1.3	// This is the maximum distance a character can travel from the center measured in lengths of TUNNEL_SIZE

/*
Maximum rotation angle (measured from 0-255)
Used to clamp the rotation if needed.
*/
// If MAXIMUM_ANGLE >= 128.0, then there's no clamping.
// If Maximum Angle < 128.0 - maximum value for change_value, then there's always clamping.
#define MAXIMUM_ANGLE 999
// An angle of 999 implies no clamping is done.

#define MINIMUM_LIGHT 3.0
#define MAXIMUM_LIGHT 256.0


/*
This lighting model assumes the lightsource is a plane normal to the viewer.
Perhaps MAX_DEPTH should be replaced by sqrt(MAX_DEPTH^2+2*(TUNNEL_SIZE^2))
This looked better when I only clipped to the near Z plane and not the rest of the frustum.
*/
#define LIGHT_OF_POINT(P)  (int)(MINIMUM_LIGHT+252.45*get_light(P.z,Z_CLIP_NEAR_PLANE,MAX_DEPTH))

/*
If I have frustum clipping and divide the polygon into several triangles that radiate from the nearest point on the polygon to the lightsource, this would be the best lighting formula (It works best with radiating triangles, but should work fairly well when MAX_DEPTH is large).
Perhaps MAX_DEPTH should be replaced by sqrt(MAX_DEPTH^2+2*(TUNNEL_SIZE^2))
*/
// #define LIGHT_OF_POINT(P) (int)(252.45*get_light(vector_length_f(P.x,P.y,P.z),Z_CLIP_NEAR_PLANE,MAX_DEPTH))



DATAFILE *dat;
BITMAP *wall, *coral, *ground, *wood;	// Textures

/* The screen and it's Z-buffer */
BITMAP *buffer;
Z_BUFFER *the_zbuffer;

ZPRITE *outzprite_1, *outzprite_2;	// Temporary zprites used when pipelining zprite drawing.

BITMAP *smiley;	/* The current smiley sprite */
ZPRITE *zphere;	/* The current sphere_zprite */

ZPRITE *zphere_array[NUM_FOG_PALETTES];	// A 'zphere' for each fog palette

char tmp_text[80];			// Used as output for sprintf when using textout()

int last_retrace_count;		/* Used to keep the speed constant */

float change_value;	/* When changing things with the keys, this is how much they should change */

int use_zbuffering;	/* Used when de-bugging or to see how-much Z-buffering slows things down. */


/*
The text colours used to display the feedback
I have can have a different set of text colours for each fog
*/
typedef struct feedback_colours {
	int text_colour;			/* Colour of main text */
	int stimulus_colour;		/* Colour that will be used to print a stimlus */
	int value_colour;			/* Colour that will be used to print a value */
} feedback_colours;

/*
Information about each fog-palette
*/
typedef struct FOG_PALETTES_INFO {
	int fog_palette_created;	// Has this palette been mapped yet?
	PALLETE *pal;
	RGB_MAP *rgbmap;
	COLOR_MAP *trans_table;	// I could have multiple tranlucency tables for fading effects.
} FOG_PALETTES_INFO;
// I only need a seperate RGB map when doing GRGB shading.

FOG_PALETTES_INFO fog_palettes_info[NUM_FOG_PALETTES];

/*
Information about each type of fog.
*/
typedef struct FOGSTRUCT {
	int fog_palette_index;	/* Which palette to use */
	float gamma;			/* Gamma correction for this fog */
	RGB	colour;				/* The fog's colour */
	feedback_colours text_colours;	/* Which text colours in this fog */
	COLOR_MAP *light_table;		// Each fog has a unique lighting table
	// I could have a pointer to the appropriate fog_palette_info structure to reduce complexity.
	int background_colour_index;	/* Which colour to use to erase the background */
} FOGSTRUCT;

/*
Fog Data
The number of entries must equal NUM_FOGS
*/
FOGSTRUCT fog [] = {
	{0, 0.21, {0x00,0x00,0x00}, {2,3,4} },
	{0, 0.34, {0x00,0x00,0x00}, {2,3,4} },
	{0, 0.58, {0x00,0x00,0x00}, {2,3,4} },
	{0, 0.40, {0x0C,0x06,0x00}, {2,3,4} },
	{0, 0.41, {0x06,0x08,0x01}, {2,3,4} },
	{0, 0.60, {0x0C,0x0C,0x0C}, {156,159,183} },
	{1, 0.40, {0x2A,0x34,0x3A}, {13,32,128} },
	{1, 0.60, {0x2A,0x34,0x3A}, {13,32,128} },
	{1, 0.55, {0x2A,0x37,0x33}, {13,32,128} },
};

int fog_number=0;	// Index into array of fogs.


typedef struct QUAD {
	V3D_f vertices[4];
	float tex_ymax;
	float tex_dy;
	BITMAP **tex;
} QUAD;

#define MAX_NUM_VERTICES (NUM_VERTICES_IN_POLYGON+NUM_CLIPPING_PLANES)


/*
This is where the transformed V3Ds are stored before being clipped.
The last few NUM_CLIPPING_PLANES are used as temporary storage by the clipper. Hence NUM_POLYS*MAX_NUM_VERTICES instead of NUM_POLYS*NUM_VERTICES_IN_POLYGON (should this be "NUM_POLYS*(MAX_NUM_VERTICES-1)"? )
*/
V3D_f unclipped_vertices[NUM_POLYS*MAX_NUM_VERTICES];

/* Where the transformed and clipped V3Ds are stored */
V3D_f out_vertices[NUM_POLYS*MAX_NUM_VERTICES];

/*
An array of pointers to the above V3D_f's.
Used by zb_polygon3d_f() which needs an array of pointers
*/
V3D_f *pointers_to_out_vertices[NUM_POLYS*MAX_NUM_VERTICES];


QUAD background_poly[] = {
	{
		{
			{-GROUND_XSIZE, -GROUND_YSIZE,0 ,0 ,0, MINIMUM_LIGHT},
			{GROUND_XSIZE , -GROUND_YSIZE,0 ,GROUND_TEXTURE_XMAX ,0 , MINIMUM_LIGHT},
			{GROUND_XSIZE , GROUND_YSIZE ,0 ,GROUND_TEXTURE_XMAX ,GROUND_TEXTURE_YMAX, MINIMUM_LIGHT},
			{-GROUND_XSIZE, GROUND_YSIZE ,0 ,0 ,GROUND_TEXTURE_YMAX, MINIMUM_LIGHT}
		}, GROUND_TEXTURE_YMAX, 1.0/8.0
	},

	{
		{
			{-TOP_XSIZE, -TOP_YSIZE,0 ,0 ,0, MINIMUM_LIGHT},
			{TOP_XSIZE , -TOP_YSIZE,0 ,TOP_TEXTURE_XMAX ,0 , MINIMUM_LIGHT},
			{TOP_XSIZE , TOP_YSIZE ,0 ,TOP_TEXTURE_XMAX ,TOP_TEXTURE_YMAX, MINIMUM_LIGHT},
			{-TOP_XSIZE, TOP_YSIZE ,0 ,0 ,TOP_TEXTURE_YMAX, MINIMUM_LIGHT}
		}, TOP_TEXTURE_YMAX, 1.0/8.0
	},

	{
		{
			{-SIDE_XSIZE, -SIDE_YSIZE,0 ,0 , 0, MINIMUM_LIGHT},
			{SIDE_XSIZE , -SIDE_YSIZE,0 ,SIDE_TEXTURE_XMAX ,0 , MINIMUM_LIGHT},
			{SIDE_XSIZE , SIDE_YSIZE ,0 ,SIDE_TEXTURE_XMAX ,SIDE_TEXTURE_YMAX, MINIMUM_LIGHT},
			{-SIDE_XSIZE, SIDE_YSIZE ,0 ,0 ,SIDE_TEXTURE_YMAX , MINIMUM_LIGHT}
		}, SIDE_TEXTURE_YMAX, 1.0/2.0
	},

	{
		{
			{-SIDE_XSIZE, -SIDE_YSIZE,0 ,0 , 0, MINIMUM_LIGHT},
			{SIDE_XSIZE , -SIDE_YSIZE,0 ,SIDE_TEXTURE_XMAX ,0 , MINIMUM_LIGHT},
			{SIDE_XSIZE , SIDE_YSIZE ,0 ,SIDE_TEXTURE_XMAX ,SIDE_TEXTURE_YMAX, MINIMUM_LIGHT},
			{-SIDE_XSIZE, SIDE_YSIZE ,0 ,0 ,SIDE_TEXTURE_YMAX , MINIMUM_LIGHT}
		}, SIDE_TEXTURE_YMAX, 1.0/2.0
	},

};


typedef struct camera_stats {
	float xr;		/* camera yaw */
	float yr;		/* camera pitch */
	float tilt;		/* camera roll */
	float xdisp;
	float ydisp;
	MATRIX_f matrix;	// We are not using get_camera_matrix.
} camera_stats;

camera_stats the_camera = {0.0,0.0,0.0,0.0,0.0};

typedef struct CHARACTER {
	ZPRITE **zprite;	// For sprites with depth-maps
	BITMAP **sprite;	// For sprites that are just flat
	float x, y, z;
	float outx, outy, outz;	// Transformed vertices
	float dx, dy, dz;
	int xsize, ysize;	// Size of original object in pixels;
	float xscale, yscale;	// Ammount by which object will be scaled when stretched.
	void (*character_drawer)();
} CHARACTER;

CHARACTER characters[NUM_CHARACTERS];


void create_gamma_light_table(COLOR_MAP *table, PALLETE pal, int r, int g, int b, float gamma, void (*callback)(int pos))
{
   int x, y, gx;
   RGB c;

   for (x=0; x<PAL_SIZE; x++) {
      gx = (int)(pow( ((float)x)/255.0, gamma) * 255.0);

      for (y=0; y<PAL_SIZE; y++) {
	 c.r = (r * (255 - gx) / 255) + ((int)pal[y].r * gx / 255);
	 c.g = (g * (255 - gx) / 255) + ((int)pal[y].g * gx / 255);
	 c.b = (b * (255 - gx) / 255) + ((int)pal[y].b * gx / 255);

	 if (rgb_map)
	    table->data[x][y] = rgb_map->data[c.r>>1][c.g>>1][c.b>>1];
	 else
	    table->data[x][y] = 0; /* bestfit_color(pal, c.r, c.g, c.b); */
      }


      if (callback)
	 (*callback)(x);
   }
}



/*
Finds the correct x and y size to stretch a zprite given a Z value.
Try to use values other than the screen size for the viewport size.
( see persp_project() )
// SCREEN_W and SCREEN_H are the viewport size (can I get this from elsewhere?) Perhaps try a method similar to the one I used in reverse_persp_project() to access Allegro's viewport.
*/
void find_zprite_size(int xsize, int ysize, float z, float xscale, float yscale, float *proj_w, float *proj_h)
{
	/* Project rectangle 0,0,xsize,ysize in viewspace to screenspace */
	persp_project_f( ((float)(xsize<<1)) * xscale, ((float)(ysize<<1)) * ((float)SCREEN_W)/((float)SCREEN_H) * yscale, z, proj_w, proj_h);

	/* Make the rectangle start at 0,0 in screenspace */
	*proj_w -= (float) (SCREEN_W>>1);
	*proj_h -= (float) (SCREEN_H>>1);

}


/*
Returns the light value as a float from 0 to 1
This assumes that the lightsource is the z=0 plane.

minz or maxz must not be zero
At the mo' Values outside 0-1 get truncated.

IDEA for lighting tables: Have a gamma correction component.
*/
float get_light(float z, float minz, float maxz)
{

	/* Prevent division by 0 */
	if (z==0.0 || minz==maxz) {
		return(1.0);	/* Maximum illumination */
	}

	/* Force values from 0 to 1 */
	if (z<minz) {
		z=minz;
	} else if (z>maxz) {
		z=maxz;
	}


	/* Light varies linearly with 1/z */
	z = 1.0/z;
	minz = 1.0/minz;
	maxz = 1.0/maxz;

	z = (z-maxz) / (minz-maxz);	/* Normalise z from 0 to 1 */

	/* Force values from 0 to 1 (not needed, but makes it easier to trace bugs) */
	if (z<0.0) {
		z=0.0;
	} else if (z>1.0) {
		z=1.0;
	}

	return (z);
}


/*
Gets a rotation matrix based on the given camera and
stores it in the given camera's matrix
This is simpler than Allegro's equivalent function (or at least it was until I started doing fancy things to it).
*/
void get_matrix_f(camera_stats *camera_to_change)
{

	MATRIX_f tmp_matrix_1, tmp_matrix_2, tmp_matrix_3;

	// If I do the displacements before the rotations, I don't have to worry about reaching the end of the tunnel, but it then looks less realistic when walking.

	/* Yaw: moving head side to side - rotation around the y axis */
	get_translation_matrix_f(&tmp_matrix_1,camera_to_change->xdisp, camera_to_change->ydisp, 0.0);
	get_y_rotate_matrix_f(&tmp_matrix_2, camera_to_change->xr);
	matrix_mul_f(&tmp_matrix_1, &tmp_matrix_2, &tmp_matrix_3);

	/* Pitch: moving head up and down - rotation around the x axis */
	get_x_rotate_matrix_f(&tmp_matrix_1, camera_to_change->yr);
	matrix_mul_f(&tmp_matrix_3, &tmp_matrix_1, &camera_to_change->matrix);

	/* Roll: tilting head - rotation around z axis */
	get_z_rotate_matrix_f(&tmp_matrix_1, camera_to_change->tilt);
	matrix_mul_f(&camera_to_change->matrix, &tmp_matrix_1, &tmp_matrix_3);

	get_scaling_matrix_f(&tmp_matrix_1, 1.0, ((float) SCREEN_W)/((float)SCREEN_H), 1.0);	// When converting from World to Camera space, use this to correct aspect distortions caused by persp_project().
	matrix_mul_f(&tmp_matrix_3, &tmp_matrix_1, &camera_to_change->matrix);	// Scaling done last.

}


/*
Used when creating rgb_table
This is a function that is pointed to.
*/
void print_progress(int pos)
{
	if ((pos & 3) == 3) {
		printf("*");
		fflush(stdout);
	}
}


/*
The smiley drawing pipeline.
This is a function that is pointed to.
*/
void draw_smiley(CHARACTER *out_character, camera_stats *camera_to_modify)
{
	float projected_sprite_w, projected_sprite_h;

	/* Use the temporary zprite's BITMAPs as temporary bitmaps */
	clear(outzprite_1->bitmap);
	clear(outzprite_2->bitmap);

	/* Rotate the zprite accoring to the tilt */
	rotate_sprite(outzprite_1->bitmap, *out_character->sprite, 0, 0, ftofix(camera_to_modify->tilt));

	// TODO: Flip the sprite if Camera has been rotated beyond pi/2

	/* Illuminate the zprite */
	color_map = fog[fog_number].light_table;
	draw_lit_sprite(outzprite_2->bitmap, outzprite_1->bitmap, 0, 0, (int)(255.49*get_light(out_character->outz, Z_CLIP_NEAR_PLANE, MAX_DEPTH)));

	color_map = fog_palettes_info[fog[fog_number].fog_palette_index].trans_table;

	/* Determine size when stretched. */
	find_zprite_size(out_character->xsize, out_character->ysize, out_character->outz, out_character->xscale, out_character->yscale, &projected_sprite_w, &projected_sprite_h);

	if (use_zbuffering) {
		displace_flat_trans_zprite(buffer, the_zbuffer, outzprite_2->bitmap, (int)(out_character->outx-projected_sprite_w/2.0), (int)(out_character->outy-projected_sprite_h/2.0), (int)projected_sprite_w, (int)projected_sprite_h, 1.0/out_character->outz);
	} else {
		stretch_sprite(buffer, outzprite_2->bitmap, (int)(out_character->outx-projected_sprite_w/2.0), (int)(out_character->outy-projected_sprite_h/2.0), (int)projected_sprite_w, (int)projected_sprite_h);
	}
}


/*
The zphere drawing pipeline
This is a function that is pointed to.
*/
void draw_zphere(CHARACTER *out_character, camera_stats *camera_to_modify)
{
	float projected_sprite_w, projected_sprite_h;

	clear_zprite(outzprite_1);

	// TODO: Rotate the zprite
	// TODO: Flip the zprite if Camera has been rotated beyond pi/2

	/* Illuminate the zprite */
	color_map = fog[fog_number].light_table;

	if (use_zbuffering) {
		draw_lit_zprite(outzprite_1->bitmap, outzprite_1->z_buffer, (*(out_character->zprite))->bitmap, (*(out_character->zprite))->z_buffer, 0, 0, (int)(255.49*get_light(out_character->outz, Z_CLIP_NEAR_PLANE, MAX_DEPTH)));	// Z buffer unchanged (but it won't be if I flip and light the zprite in one command).
	} else {
		draw_lit_sprite(outzprite_1->bitmap, (*(out_character->zprite))->bitmap, 0, 0, (int)(255.49*get_light(out_character->outz, Z_CLIP_NEAR_PLANE, MAX_DEPTH)));
	}

	color_map = fog_palettes_info[fog[fog_number].fog_palette_index].trans_table;

	/* Determine size when stretched. */
	find_zprite_size(out_character->xsize, out_character->ysize, out_character->outz, out_character->xscale, out_character->yscale, &projected_sprite_w, &projected_sprite_h);

	if (use_zbuffering) {

		// These two commented-out lines used for testing purposes
		// draw_zprite(buffer, the_zbuffer, (*(out_character->zprite))->bitmap, (*(out_character->zprite))->z_buffer, (int)out_character->outx, (int)out_character->outy);
		// rotate_flat_zprite(buffer, the_zbuffer, (*(out_character->zprite))->bitmap, (int)out_character->outx, (int)out_character->outy, ftofix(camera_to_modify->tilt), 1.0/10.0);

		displace_trans_zprite(buffer, the_zbuffer, outzprite_1->bitmap, outzprite_1->z_buffer, (int)(out_character->outx-projected_sprite_w/2), (int)(out_character->outy-projected_sprite_h/2), (int)projected_sprite_w, (int)projected_sprite_h, out_character->outz);
	} else {
		stretch_sprite(buffer, outzprite_1->bitmap, (int)(out_character->outx-projected_sprite_w/2), (int)(out_character->outy-projected_sprite_h/2), (int)projected_sprite_w, (int)projected_sprite_h);
	}
}


/*
Position the background shapes in place.
*/
void init_background()
{
	int c;
	MATRIX_f translation_matrix, rotation_matrix, out_matrix, out_matrix_2;

	translation_matrix = identity_matrix_f;

	/* The ground */
	background_poly[0].tex = &ground;

	/* Rotate and translate the ground poly */
	get_x_rotate_matrix_f(&rotation_matrix, 64.0);
	get_translation_matrix_f(&translation_matrix, 0.0, TUNNEL_SIZE, 0.0);
	matrix_mul_f(&rotation_matrix, &translation_matrix, &out_matrix);

	/* Place the ground in position */
	for (c=0;c<4;c++) {
		apply_matrix_f(&out_matrix, background_poly[0].vertices[c].x, background_poly[0].vertices[c].y, background_poly[0].vertices[c].z, &background_poly[0].vertices[c].x, &background_poly[0].vertices[c].y, &background_poly[0].vertices[c].z);
	}

	/* The ceiling */
	background_poly[1].tex = &coral;

	/* Rotate and translate the top poly */
	get_x_rotate_matrix_f(&rotation_matrix, 64.0);
	get_translation_matrix_f(&translation_matrix, 0.0, -TUNNEL_SIZE, 0.0);
	matrix_mul_f(&rotation_matrix, &translation_matrix, &out_matrix);

	/* Place the top in position */
	for (c=0;c<4;c++) {
		apply_matrix_f(&out_matrix, background_poly[1].vertices[c].x, background_poly[1].vertices[c].y, background_poly[1].vertices[c].z, &background_poly[1].vertices[c].x, &background_poly[1].vertices[c].y, &background_poly[1].vertices[c].z);
	}

	/* The right side */
	background_poly[2].tex = &wall;

	/* Rotate and translate the right poly */
	get_x_rotate_matrix_f(&rotation_matrix, 64.0);
	get_translation_matrix_f(&translation_matrix, 0.0, -TUNNEL_SIZE, 0.0);
	matrix_mul_f(&rotation_matrix, &translation_matrix, &out_matrix);
	get_z_rotate_matrix_f(&rotation_matrix, 64.0);
	matrix_mul_f(&out_matrix, &rotation_matrix, &out_matrix_2);

	/* Place the right side in position */
	for (c=0;c<4;c++) {
		apply_matrix_f(&out_matrix_2, background_poly[2].vertices[c].x, background_poly[2].vertices[c].y, background_poly[2].vertices[c].z, &background_poly[2].vertices[c].x, &background_poly[2].vertices[c].y, &background_poly[2].vertices[c].z);
	}

	/* The left side */
	background_poly[3].tex = &wall;

	/* Rotate and translate the left poly */
	get_x_rotate_matrix_f(&rotation_matrix, 64.0);
	get_translation_matrix_f(&translation_matrix, 0.0, TUNNEL_SIZE, 0.0);
	matrix_mul_f(&rotation_matrix, &translation_matrix, &out_matrix);
	get_z_rotate_matrix_f(&rotation_matrix, 64.0);
	matrix_mul_f(&out_matrix, &rotation_matrix, &out_matrix_2);

	/* Place the left side in position */
	for (c=0;c<4;c++) {
		apply_matrix_f(&out_matrix_2, background_poly[3].vertices[c].x, background_poly[3].vertices[c].y, background_poly[3].vertices[c].z, &background_poly[3].vertices[c].x, &background_poly[3].vertices[c].y, &background_poly[3].vertices[c].z);
	}


	/*
	Make an array of pointers to the vertices in the array of transformed vertices.
	This is so that they can be taken by zb_polygon3d_f()
	*/
	for (c=0;c<NUM_POLYS*MAX_NUM_VERTICES;c++) {
		pointers_to_out_vertices[c] = &out_vertices[c];
	}


}



void init_zprites()
{
	int i;

	/* Smileys */

	smiley=(BITMAP *)dat[SMILEY_1+fog[fog_number].fog_palette_index].dat;

	for (i=0;i<NUM_SMILEYS;i++) {

		characters[i].x=0.0;
		characters[i].y=0.0;
		characters[i].z=CHARACTER_ZMAX;

		characters[i].dx=fixtof( (random()%(1<<16)) - (1<<15) );
		characters[i].dy=fixtof( (random()%(1<<16)) - (1<<15) );
		characters[i].dz=-fixtof(random()%(2<<16));

		characters[i].sprite = &smiley;
		characters[i].zprite = NULL;

		characters[i].xsize=SMILEY_XSIZE;
		characters[i].ysize=SMILEY_YSIZE;
		characters[i].xscale=SMILEY_XSCALE;
		characters[i].yscale=SMILEY_YSCALE;

		characters[i].character_drawer = draw_smiley;
	}


	/* Zpheres */

	zphere = zphere_array[fog[fog_number].fog_palette_index];

	for (i=0;i<NUM_SPHERES;i++) {

		characters[i+NUM_SMILEYS].x=0.0;
		characters[i+NUM_SMILEYS].y=0.0;
		characters[i+NUM_SMILEYS].z=CHARACTER_ZMAX;

		characters[i+NUM_SMILEYS].dx=fixtof( (random()%(1<<15)) - (1<<14) );
		characters[i+NUM_SMILEYS].dy=fixtof( (random()%(1<<15)) - (1<<14) );
		characters[i+NUM_SMILEYS].dz=-fixtof(random()%(2<<16));


		characters[i+NUM_SMILEYS].sprite = NULL;
		characters[i+NUM_SMILEYS].zprite = &zphere;

		characters[i+NUM_SMILEYS].xsize=SPHERE_XSIZE;
		characters[i+NUM_SMILEYS].ysize=SPHERE_YSIZE;
		characters[i+NUM_SMILEYS].xscale=SPHERE_XSCALE;
		characters[i+NUM_SMILEYS].yscale=SPHERE_YSCALE;

		characters[i+NUM_SMILEYS].character_drawer = draw_zphere;
	}

}


void init()
{
	int c, w, h;
	Z_BUFFER *tmp_zb_ptr;	// Used to point to the z-buffer that will be shared by all the zpheres.
	MATRIX_f zphere_rotation_matrix;


	allegro_init();
	install_keyboard();
	install_mouse();
	install_timer();

	dat = load_datafile("3dspdemo.dat");
	if (!dat) {
		printf("ERROR: 3DSPDEMO.DAT not loaded");
		exit(1);
	}

	/* Set the palette info to point to each palette in the datafile */
	for(w=0;w<NUM_FOG_PALETTES;w++) {
		fog_palettes_info[w].pal=(PALLETE *) dat[PALETTE_1+w].dat;
	}


	/* set up a greycale in the bottom quater of the pallete */
	// for (c=0; c<64; c++) {
	// 	pal[0][c].r = pal[0][c].g = pal[0][c].b = c;
	// }


	/* Make a Z-buffer for all the zpheres to use */
	tmp_zb_ptr = create_zbuffer(SPHERE_XSIZE,SPHERE_YSIZE);

	/* Get a texture rotation matrix for the zpheres of each palette */
	get_transformation_matrix_f(&zphere_rotation_matrix, 1.0, 24.0,0.0,32.0, 0.0,0.0,0.0);

	/*
	Build up a fog structure for each colour of fog
	And while we're at it, make a zphere for each fog palette
	*/
	for (h=0;h<NUM_FOGS;h++) {

		if (fog_palettes_info[fog[h].fog_palette_index].fog_palette_created) {
			/* This fog's palette has been taken care of */

			rgb_map = fog_palettes_info[fog[h].fog_palette_index].rgbmap;

			/* build a lighting table */
			printf("\nGenerating lighting table for fog #%d:\n", h+1);
			printf("<................................................................>\r<");
			fog[h].light_table = malloc(sizeof(COLOR_MAP));
			create_gamma_light_table(fog[h].light_table, *fog_palettes_info[fog[h].fog_palette_index].pal, fog[h].colour.r, fog[h].colour.g, fog[h].colour.b, fog[h].gamma, print_progress);

		} else {
			/* We are dealing with a fog who'se palette hasn't been worked on */

			/* Build a rgb mapping table for speeding translucency and lighting. */
			// Should this RGB table be recorded into the struct?
			// Advantage, can use GRGB, [bi|tri]-linear interpolation (future), and perhaps do a few more FX.
			// Also, I can generate colour tables faster if lack of coherency in list of palette indexes (ie 0, 1, 0)
			// Dis-advantage: 32K per table.
			printf("\nGenerating RGB table for fog #%d:\n", h+1);
			printf("<................................................................>\r<");
			rgb_map = fog_palettes_info[fog[h].fog_palette_index].rgbmap = malloc(sizeof(RGB_MAP));
			create_rgb_table(fog_palettes_info[fog[h].fog_palette_index].rgbmap, *fog_palettes_info[fog[h].fog_palette_index].pal, print_progress);

			/* build a lighting table */
			printf("\nGenerating lighting table for fog #%d:\n", h+1);
			printf("<................................................................>\r<");
			color_map = fog[h].light_table = malloc(sizeof(COLOR_MAP));
			create_gamma_light_table(fog[h].light_table, *fog_palettes_info[fog[h].fog_palette_index].pal, fog[h].colour.r, fog[h].colour.g, fog[h].colour.b, fog[h].gamma, print_progress);

			/* build a translucency table */
			printf("\nGenerating translucency table for fog #%d:\n", h+1);
			printf("<................................................................>\r<");
			fog_palettes_info[fog[h].fog_palette_index].trans_table = malloc(sizeof(COLOR_MAP));
			create_trans_table(fog_palettes_info[fog[h].fog_palette_index].trans_table, *fog_palettes_info[fog[h].fog_palette_index].pal, 0xA0, 0xA0, 0xA0, print_progress);

			fog_palettes_info[fog[h].fog_palette_index].fog_palette_created=true;

			/*
			Each palette requires a different bitmap.
			Make a zphere for this palette
			*/

			zphere = zphere_array[fog[h].fog_palette_index] = set_zprite_to_zbuffer(tmp_zb_ptr);
			clear_zprite(zphere);

			wood = (BITMAP *)dat[WOOD_1+fog[h].fog_palette_index].dat;

			sphere_f(zphere->bitmap, zphere->z_buffer, wood, SPHERESHADE_ILIT, SPHERETEXTURE_ROTATED_MTEX, SPHERE_XSIZE/2.0, SPHERE_YSIZE/2.0, (SPHERE_XSIZE/2.0)+1, (SPHERE_XSIZE/2.0)-1, 0,0,(float)wood->w,(float)wood->h, &zphere_rotation_matrix, 255<<16, 80<<16, -4.0, -6.0, -3.0);

			/*
			The following line is used when testing zprite_make_z()
			If it works, the zpheres should be flat,
			// zprite_make_z(zphere->bitmap, zphere->z_buffer, 0.25 ,0);	// Used to test zprite_make_z
			*/

		}

		/*
		Determine the colour to clear the screen with.
		Use the nearest colour in the palette to that of the fog colour.
		*/
		fog[h].background_colour_index = makecol8(fog[h].colour.r<<2,fog[h].colour.g<<2,fog[h].colour.b<<2);
	}


	rgb_map = fog_palettes_info[fog[fog_number].fog_palette_index].rgbmap;
	color_map = fog[fog_number].light_table;

	set_gfx_mode(GFX_VGA, 320, 200, 0, 0);
	set_pallete(desktop_pallete);

	coral = (BITMAP *)dat[CORAL_1+fog[fog_number].fog_palette_index].dat;
	ground = (BITMAP *)dat[GROUND_1+fog[fog_number].fog_palette_index].dat;
	wall = (BITMAP *)dat[WALL_1+fog[fog_number].fog_palette_index].dat;

	/* Used for testing MASKED modes */
	/*
	circlefill(wall, 64, 64, 45, 0);
	circlefill(coral, 64, 64, 45, 0);
	circlefill(ground, 64, 64, 45, 0);
	circlefill(wood, 64, 64, 45, 0);
	*/


	if (!gfx_mode_select(&c, &w, &h)) {
		allegro_exit();
		printf("Graphics mode not selected\n\n");
		exit(1);
	}

	if (set_gfx_mode(c, w, h, 0, 0) != 0) {
		allegro_exit();
		printf("Error setting graphics mode\n%s\n\n", allegro_error);
		exit(1);
	}

	/* double buffer the animation */
	if ( (buffer=create_bitmap(SCREEN_W, SCREEN_H)) == NULL) {
		allegro_exit();
		printf("Error setting up the double buffer\n");
		printf("Possibly low on memory\n\n");
		exit(1);
	}

	/* The Z buffer */
	if ( (the_zbuffer = create_zbuffer(SCREEN_W, SCREEN_H)) == NULL ) {
		allegro_exit();
		printf("Error setting up the Z buffer\n");
		printf("Possibly low on memory\n\n");
		exit(1);
	}

	/* Temporary zprites for combining effects */
	/* Would it effect things too much if I wrote to where I read from in order to save an extra zprite? */

	if ( (outzprite_1=create_zprite(MAXOBJECT_SIZE,MAXOBJECT_SIZE)) == NULL || (outzprite_2=create_zprite(MAXOBJECT_SIZE,MAXOBJECT_SIZE)) == NULL) {
		printf("ERROR: Possibly low on memory\n\n");
	}

	set_pallete(*fog_palettes_info[fog[fog_number].fog_palette_index].pal);

	set_projection_viewport(0, 0, SCREEN_W, SCREEN_H);
	/* set up a SQUARE viewport for the perspective projection */
	// Remember to set the sprite clipping rectangle to these dimensions.
	/*
	if (SCREEN_W>SCREEN_H) {
		set_projection_viewport(0, 0, SCREEN_H, SCREEN_H);
	} else {
		set_projection_viewport(0, 0, SCREEN_W, SCREEN_W);
	}
	*/

	get_matrix_f(&the_camera);

	use_zbuffering = true;

	init_background();
	init_zprites();

	text_mode(-1);
	srandom(retrace_count);

}


/* Give the viewer the illusion tat they are traveling through the corridor */
void animate_textures()
{
	int i;

	for (i=0;i<NUM_POLYS;i++) {
		background_poly[i].vertices[0].v += background_poly[i].tex_dy;
		if (background_poly[i].vertices[0].v >= (float) (background_poly[i].tex_ymax-1)) {
			background_poly[i].vertices[0].v =- (float) (background_poly[i].tex_ymax);
		}
		background_poly[i].vertices[1].v += background_poly[i].tex_dy;
		if (background_poly[i].vertices[1].v >= (float) (background_poly[i].tex_ymax-1)) {
			background_poly[i].vertices[1].v =- (float) (background_poly[i].tex_ymax);
		}
		background_poly[i].vertices[2].v = background_poly[i].vertices[1].v + background_poly[i].tex_ymax;
		background_poly[i].vertices[3].v = background_poly[i].vertices[0].v + background_poly[i].tex_ymax;
	}

}





/*
Translates the polys making up the ground, walls and ceiling,
clips them to the viewing frustum, projects them, and draws them.

Question: Do I need an array for the vertex count, or would a single integer suffice? Methinx a single integer would do.
*/
void transform_and_draw_polys(camera_stats *camera_to_modify)
{

	int i,j;
	int unclipped_k, out_k;

	int vertex_count[NUM_POLYS];	//	Number of vertices in poly (do we need an array?).


	/*
	When debugging the frustum clipping, scale down the results to see if it works properly

	// MATRIX_f scaling_matrix;	// DEBUG

	// get_scaling_matrix_f(&scaling_matrix, 0.5, 0.5, 1.0);	// DEBUG
	*/


	color_map=fog[fog_number].light_table;

	for (i=0;i<NUM_POLYS;i++) {

		/* Apply the camera to the poly
		*/
		for (j=0;j<NUM_VERTICES_IN_POLYGON;j++) {

			unclipped_k=(i*MAX_NUM_VERTICES)+j;

			unclipped_vertices[unclipped_k]=background_poly[i].vertices[j];	// Copy vertex info into the unclipped vertex array. This also copies u,v info.

			apply_matrix_f(&camera_to_modify->matrix, unclipped_vertices[unclipped_k].x, unclipped_vertices[unclipped_k].y, unclipped_vertices[unclipped_k].z, &unclipped_vertices[unclipped_k].x, &unclipped_vertices[unclipped_k].y, &unclipped_vertices[unclipped_k].z);	// Note, this does not copy U,V info.

		}


		/* Clip the Poly to the viewing frustum */
		vertex_count[i]=clip_poly_to_frustum_f(&unclipped_vertices[i*MAX_NUM_VERTICES], &out_vertices[i*MAX_NUM_VERTICES], NUM_VERTICES_IN_POLYGON, Z_CLIP_NEAR_PLANE);


		/* Project the poly */

		for (j=0;j<vertex_count[i];j++) {

			out_k=(i*MAX_NUM_VERTICES)+j;

			/*
			When debugging the frustum clipping, scale down the results to see if it works properly
			// apply_matrix_f(&scaling_matrix, out_vertices[out_k].x, out_vertices[out_k].y, out_vertices[out_k].z, &out_vertices[out_k].x, &out_vertices[out_k].y, &out_vertices[out_k].z);	// DEBUG
			*/

			/* Find the light on each point */
			out_vertices[out_k].c = LIGHT_OF_POINT(out_vertices[out_k]);

			persp_project_f(out_vertices[out_k].x, out_vertices[out_k].y, out_vertices[out_k].z, &out_vertices[out_k].x, &out_vertices[out_k].y);

		}


		/* Draw the poly */

		if (vertex_count[i]>2) {

			// The KEY_B bit is used to switch off Z-buffering and 3D sprites.
			if (use_zbuffering) {
				zb_polygon3d_f(buffer, the_zbuffer, POLYTYPE_PTEX_LIT, *background_poly[i].tex, vertex_count[i], pointers_to_out_vertices+(i*MAX_NUM_VERTICES));
			} else {
				polygon3d_f(buffer, POLYTYPE_PTEX_LIT, *background_poly[i].tex, vertex_count[i], pointers_to_out_vertices+(i*MAX_NUM_VERTICES));
			}

		}

	}
}


/* Move the zprites in the 3D world */
void animate_zprites()
{
	int i;

	for (i=0;i<NUM_CHARACTERS;i++) {

		characters[i].x += characters[i].dx;
		if ( (characters[i].x < -(TUNNEL_SIZE*CHARACTER_XYLIMITS_RATIO)) || (characters[i].x > (TUNNEL_SIZE*CHARACTER_XYLIMITS_RATIO)) ) {
			characters[i].dx = -characters[i].dx;
			characters[i].x += characters[i].dx;
		}

		characters[i].y += characters[i].dy;
		if ( (characters[i].y < -TUNNEL_SIZE*CHARACTER_XYLIMITS_RATIO) || (characters[i].y > TUNNEL_SIZE*CHARACTER_XYLIMITS_RATIO) ) {
			characters[i].dy = -characters[i].dy;
			characters[i].y += characters[i].dy;
		}

		characters[i].z += characters[i].dz;
		if (characters[i].z < -CHARACTER_ZMAX) {
			characters[i].z = CHARACTER_ZMAX;
		}

	}

}


/*
callback for qsort()
The greater the Z value, the further back the object is.
The objects with the greatest Zs (furthest) become the earliest in the array and get drawn first.
See libc.inf for what qsort does.
*/
int character_cmp(const void *e1, const void *e2)
{
	return ( (int)( ((CHARACTER *)e2)->z - ((CHARACTER *)e1)->z ) );
}

/*
Transforms the zprites from worldspace to viewspace and screenspace.
This also translates and projects the zprites.
BEWARE: Lighting part of a sprite to 0 makes trans_sprite think it's the masked colour.
*/
void transform_and_draw_zprites(camera_stats *camera_to_modify)
{
	int i;

	/* Apply the camera matrix to the zprites. */
	for (i=0;i<NUM_CHARACTERS;i++) {
		apply_matrix_f(&camera_to_modify->matrix, characters[i].x, characters[i].y, characters[i].z, &characters[i].outx, &characters[i].outy, &characters[i].outz);

		persp_project_f(characters[i].outx, characters[i].outy, characters[i].outz, &characters[i].outx, &characters[i].outy);
	}


	/*
	Apply depth-sort so that the translucent sprites that are the furthest back get drawn first. This will enable translucent sprites to work properly with Z-buffering. However, non translucent sprites should be drawn before the translucent ones are drawn.
	*/
	qsort(characters, NUM_CHARACTERS, sizeof(CHARACTER), character_cmp);


	/* Draw the zprites */
	for (i=0;i<NUM_CHARACTERS;i++) {

		/* Don't draw zprites the viewer cannot see */
		if (!(characters[i].outz<CHARACTER_ZMIN || characters[i].outx > (float)(SCREEN_W+characters[i].xsize) || characters[i].outx < (float)(-characters[i].xsize) || characters[i].outy > (float)(SCREEN_H+characters[i].ysize) || characters[i].outy < (float)(-characters[i].ysize))) {

			(*characters[i].character_drawer)(&characters[i], camera_to_modify);

		}

	}

}




/*
Display commands for the demos.
*/

/* Waits for a key to be pressed while flashing a message */
void wait_while_user_reads_instructions()
{
	clear_keybuf();

	while(!keypressed()) {
		vsync();
		textout(screen, font, "Press any key to continue", 0, SCREEN_H-text_yheight, makecol8(255,255,255)*((retrace_count&32)==0) );
	}

	/* Make sure that we don't instantly return to the help-screen if holding down the help key */
	while(key[KEY_F12]) {
	}
}


void display_keys_for_3dsp()
{
	clear(screen);

	textout_centre(screen, font, "Demo of Z-buffered sprites", SCREEN_W/2, 0, makecol8(255,255,255));

	textout(screen, font, "Press       ,     /     to tilt camera", 0, text_yheight*2, fog[fog_number].text_colours.text_colour);
	textout(screen, font, "ARROWS  PGUP PGDN", 48, text_yheight*2, fog[fog_number].text_colours.stimulus_colour);

	textout(screen, font, "Press  ,  ,  ,   to displace the camera", 0, text_yheight*3, fog[fog_number].text_colours.text_colour);
	textout(screen, font, "Q  A  Z  X", 48, text_yheight*3, fog[fog_number].text_colours.stimulus_colour);

	textout(screen, font, "Hold down       and         for speed", 0,text_yheight*5,fog[fog_number].text_colours.text_colour);
	textout(screen, font, "SHIFT",80,text_yheight*5,fog[fog_number].text_colours.stimulus_colour);
	textout(screen, font, "CONTROL",160,text_yheight*5,fog[fog_number].text_colours.stimulus_colour);

	textout(screen, font, "Press keys     to change fog", 0,text_yheight*7,fog[fog_number].text_colours.text_colour);
	sprintf(tmp_text, "1-%d", (NUM_FOGS%10));
	textout(screen, font, tmp_text, 88,text_yheight*7,fog[fog_number].text_colours.stimulus_colour);

	textout(screen, font, "Press     to see keyboard commands", 0, text_yheight*9, fog[fog_number].text_colours.text_colour);
	textout(screen, font, "F12", 48, text_yheight*9, fog[fog_number].text_colours.stimulus_colour);

	textout(screen, font, "Press ESC to quit", 0, text_yheight*11, fog[fog_number].text_colours.stimulus_colour);

	wait_while_user_reads_instructions();
}


/*
Feedback display functions
*/

/*
Draws a diagram on the screen representing the angle
c1 and c2 are thee two colours to use. c1 for grid, and c2 for the angle.
Angle is a float from 0-255
*/
void draw_angle(BITMAP *buffer, float angle, int x, int y, int size, int c1, int c2)
{
	vline (buffer, x+(size>>1), y, y+size-1, c1);
	hline (buffer, x, y+(size>>1), x+size-1, c1);

	line (buffer, x+(size>>1), y+(size>>1), (int)(sin(angle*(M_PI/128.0))*((float)size)/2.0)+x+(size>>1), (int)(-cos(angle*(M_PI/128.0))*((float)size)/2.0)+y+(size>>1), c2);

}

void display_feedback(int retraces)
{
	textout(buffer, font, "Vertical retraces since last update: ", 0, 0, fog[fog_number].text_colours.text_colour);
	sprintf(tmp_text, "%d", retraces);
	textout(buffer, font, tmp_text, 296, 0, fog[fog_number].text_colours.value_colour);

	/* Display info about the viewer's orientation */

	textout(buffer, font, "X-displacement", text_xstart, text_moreinfo_ystart+textblock_ysize*0, fog[fog_number].text_colours.text_colour);
	sprintf (tmp_text, "%+6.2f", -the_camera.xdisp);
	textout(buffer, font, tmp_text, text_xstart, text_moreinfo_ystart+textblock_ysize*0+text_yheight, fog[fog_number].text_colours.value_colour);

	textout(buffer, font, "Y-displacement", text_xstart, text_moreinfo_ystart+textblock_ysize*1, fog[fog_number].text_colours.text_colour);
	sprintf (tmp_text, "%+6.2f", the_camera.ydisp);
	textout(buffer, font, tmp_text, text_xstart, text_moreinfo_ystart+textblock_ysize*1+text_yheight, fog[fog_number].text_colours.value_colour);

	textout(buffer, font, "ROLL", text_xstart, text_moreinfo_ystart+textblock_ysize*2, fog[fog_number].text_colours.text_colour);
	sprintf (tmp_text, "0x%02X.%01X", (int)(-the_camera.tilt) & 0xFF, ((int)(-the_camera.tilt * 16.0)) & 0xF);
	textout(buffer, font, tmp_text, text_xstart, text_moreinfo_ystart+textblock_ysize*2+text_yheight, fog[fog_number].text_colours.value_colour);
	draw_angle (buffer, -the_camera.tilt, 0, text_moreinfo_ystart+textblock_ysize*2-3, textblock_ysize-4, fog[fog_number].text_colours.text_colour, fog[fog_number].text_colours.value_colour);

	textout(buffer, font, "YAW", text_xstart, text_moreinfo_ystart+textblock_ysize*3, fog[fog_number].text_colours.text_colour);
	sprintf (tmp_text, "0x%02X.%01X", (int)(-the_camera.xr) & 0xFF, ((int)(-the_camera.xr * 16.0)) & 0xF);
	textout(buffer, font, tmp_text, text_xstart, text_moreinfo_ystart+textblock_ysize*3+text_yheight, fog[fog_number].text_colours.value_colour);
	draw_angle (buffer, -the_camera.xr, 0, text_moreinfo_ystart+textblock_ysize*3-3, textblock_ysize-4, fog[fog_number].text_colours.text_colour, fog[fog_number].text_colours.value_colour);

	textout(buffer, font, "PITCH", text_xstart, text_moreinfo_ystart+textblock_ysize*4, fog[fog_number].text_colours.text_colour);
	sprintf (tmp_text, "0x%02X.%01X", (int)the_camera.yr & 0xFF, ((int)(the_camera.yr * 16.0)) & 0xF);
	textout(buffer, font, tmp_text, text_xstart, text_moreinfo_ystart+textblock_ysize*4+text_yheight, fog[fog_number].text_colours.value_colour);
	draw_angle (buffer, the_camera.yr, 0, text_moreinfo_ystart+textblock_ysize*4-3, textblock_ysize-4, fog[fog_number].text_colours.text_colour, fog[fog_number].text_colours.value_colour);

}


/*
Reads keys and acts accordingly
*/
int read_keys(camera_stats *camera_to_modify, int fogchange)
{
	/* Press shift to speed up all movements */
	/* Press control to further up all movements */
	/* Press shift and control together for lots of speed */

	int i;

	if (key[KEY_CONTROL]) {
		change_value=2.0;
	} else {
		change_value=0.25;
	}
	if (key[KEY_LSHIFT] || key[KEY_RSHIFT]) {
		change_value*=2.0;
	}
	if (key[KEY_X]) {
		if (camera_to_modify->xdisp > -TUNNEL_SIZE+MIN_DIST_TO_WALL) {
			camera_to_modify->xdisp -= change_value;
			get_matrix_f(camera_to_modify);
		}
	}
	if (key[KEY_Z]) {
		if (camera_to_modify->xdisp < TUNNEL_SIZE-MIN_DIST_TO_WALL ) {
			camera_to_modify->xdisp += change_value;
			get_matrix_f(camera_to_modify);
		}
	}
	if (key[KEY_A]) {
		if (camera_to_modify->ydisp > -TUNNEL_SIZE+MIN_DIST_TO_WALL ) {
			camera_to_modify->ydisp -= change_value;
			get_matrix_f(camera_to_modify);
		}
	}
	if (key[KEY_Q]) {
		if (camera_to_modify->ydisp < TUNNEL_SIZE-MIN_DIST_TO_WALL ) {
			camera_to_modify->ydisp += change_value;
			get_matrix_f(camera_to_modify);
		}
	}
	if (key[KEY_LEFT]) {
		camera_to_modify->xr += change_value;
		/* Wrap around */
		if (camera_to_modify->xr >= 128.0) {
			camera_to_modify->xr -= 256.0;
		}
		/* Clamp */
		if (camera_to_modify->xr >= MAXIMUM_ANGLE) {
			camera_to_modify->xr = MAXIMUM_ANGLE;
		}
		get_matrix_f(camera_to_modify);
	}
	if (key[KEY_RIGHT]) {
		camera_to_modify->xr -= change_value;
		/* Wrap around */
		if (camera_to_modify->xr <= -128) {
			camera_to_modify->xr += 256.0;
		}
		/* Clamp */
		if (camera_to_modify->xr <= -MAXIMUM_ANGLE) {
			camera_to_modify->xr = -MAXIMUM_ANGLE;
		}
		get_matrix_f(camera_to_modify);
	}
	if (key[KEY_UP]) {
		camera_to_modify->yr -= change_value;
		/* Wrap around */
		if (camera_to_modify->yr <= -128) {
			camera_to_modify->yr += 256.0;
		}
		/* Clamp */
		if (camera_to_modify->yr <= -MAXIMUM_ANGLE) {
			camera_to_modify->yr = -MAXIMUM_ANGLE;
		}
		get_matrix_f(camera_to_modify);
	}
	if (key[KEY_DOWN]) {
		camera_to_modify->yr += change_value;
		/* Wrap around */
		if (camera_to_modify->yr >= 128.0) {
			camera_to_modify->yr -= 256.0;
		}
		/* Clamp */
		if (camera_to_modify->yr >= MAXIMUM_ANGLE) {
			camera_to_modify->yr = MAXIMUM_ANGLE;
		}
		get_matrix_f(camera_to_modify);
	}
	if (key[KEY_PGUP]) {
		camera_to_modify->tilt -= change_value;
		if (camera_to_modify->tilt <= -128) {
			camera_to_modify->tilt += 256.0;
		}
		/* Clamp */
		if (camera_to_modify->tilt <= -MAXIMUM_ANGLE) {
			camera_to_modify->tilt = -MAXIMUM_ANGLE;
		}
		get_matrix_f(camera_to_modify);
	}
	if (key[KEY_PGDN]) {
		camera_to_modify->tilt += change_value;
		if (camera_to_modify->tilt >= 128.0) {
			camera_to_modify->tilt -= 256.0;
		}
		if (camera_to_modify->tilt >= MAXIMUM_ANGLE) {
			camera_to_modify->tilt = MAXIMUM_ANGLE;
		}
		get_matrix_f(camera_to_modify);
	}
	// Has a new fog been selected?
	for (i=KEY_1;i<KEY_1+NUM_FOGS;i++) {
		if (key[i]) {
			fog_number=i-KEY_1;
			fogchange=true;
			break;
		}
	}
	if (key[KEY_F12]) {
		display_keys_for_3dsp();
	}
	/* Dump screen */
	if (key[KEY_ALT] && key[KEY_S]) {
		save_bitmap ("dump.pcx", buffer, *fog_palettes_info[fog[fog_number].fog_palette_index].pal);
	}

	/* When de-bugging, or seing how much Z-buffering slows things down, Hold down B in order not to use Z_buffering */
	use_zbuffering = !key[KEY_B];

	return(fogchange);
}


/******************************
The main procedure of 3DSPDEMO
*******************************/

void main()
{
	int num_retraces_since_previous=0;	/* Used to display number of retraces */
	int fogchange=false;	/* This is so the changing of the background colour is in sync with the changing of the fog */

	init();

	display_keys_for_3dsp();

	last_retrace_count = retrace_count;

	while(!key[KEY_ESC]){

		clear_to_color(buffer, fog[fog_number].background_colour_index);
		if (use_zbuffering) {
			clear_zbuffer(the_zbuffer);	// Draw shapes stretching back to infinity.
			// fill_zbuffer(the_zbuffer,0.003);
		}

		/*
		Wait for the object positions to catch up with the frame counter
		before any transformation or rendering is done.
		This is to keep a constant speed when the number of vbls per render is unpredictable.
		Also, record the number of retraces.
		*/
		num_retraces_since_previous=0;
		while (last_retrace_count < retrace_count) {
			animate_textures();
			animate_zprites();
			last_retrace_count++;
			num_retraces_since_previous++;
			// Could the main keyboard commands be read here instead? It would make it smother, but a bit harder to debug.
			if (key[KEY_ESC]) {
				return;
			}
		}


		if (fogchange) {
			coral = (BITMAP *)dat[CORAL_1+fog[fog_number].fog_palette_index].dat;
			ground = (BITMAP *)dat[GROUND_1+fog[fog_number].fog_palette_index].dat;
			wall = (BITMAP *)dat[WALL_1+fog[fog_number].fog_palette_index].dat;
			wood = (BITMAP *)dat[WOOD_1+fog[fog_number].fog_palette_index].dat;

			smiley = (BITMAP *)dat[SMILEY_1+fog[fog_number].fog_palette_index].dat;
			zphere = zphere_array[fog[fog_number].fog_palette_index];

			transform_and_draw_polys(&the_camera);
			transform_and_draw_zprites(&the_camera);
			display_feedback(num_retraces_since_previous);

			rgb_map = fog_palettes_info[fog[fog_number].fog_palette_index].rgbmap;

			set_pallete(*fog_palettes_info[fog[fog_number].fog_palette_index].pal);	/* Implies a vsync */

			fogchange=false;
		} else {
			transform_and_draw_polys(&the_camera);
			transform_and_draw_zprites(&the_camera);
			display_feedback(num_retraces_since_previous);

			vsync();
		}

		blit(buffer, screen, 0, 0, 0, 0, SCREEN_W, SCREEN_H);

		fogchange=read_keys(&the_camera, fogchange);
	}


}
