/*
mainsph.c for sphere test program (SPH_DEMO).
Tests real-time (on in 320*200) textured and lit z-buffered sphere drawing. The texture can be rotated.
Also tests masked z-buffered perspective-correct texture-mapped gouraud shaded pyramids.
Also shows off the hls2rgb function.

This is a demo of AllegroPak's additions to Allegro.

By Andrei Ellman (e-mail: ellman@xs4all.nl)
*/

/*
Currently at Version 1.11a

Changes:

1.11a
+ Uses AllegroPak 1.0 beta 1a
+ The "de-referencing a NULL pointer" bug crept back in in 1.11 in sphere_f() when I tried to convert a NULL MATRIX_f into a MATRIX. It has been removed and flogged to death.
+ A bug involving drawing the light-sphere too many times has been removed. It is now only drawn once.
+ The spheres are now depth-sorted so that the frontmost gets drawn first. In Z-buffered systems, this is generally a good idea because it drastically reduces the probability that the same pixel's value is calculated twice. The speed improvements are most noticed when using fancy lighting and texturing modes.
+ You can now access the display of the key commands by pressing F12.

1.11
+ First version made available to the general public.
+ Uses AllegroPak 1.0 beta 1
+ When demonstrating the SPHERETEXTURE_ROTATED texture modes, the spheres's texture rotates according to the position of the camera relative to the middle sphere.
+ When demonstrating the SPHERESHADE_PLIT lighting mode, you can now position the light-sphere using the following co-ordinate spaces:	viewspace, screenspace, polar viewspace.
+ The program now creates four texture-bitmaps for the spheres instead of just one, which can be noticed by the fact that there are spheres with different combinations of texture colour.

1.1a
+ sphere() only de-references the texture bitmap to get w and h if it is not NULL. This should prevent it from crashing under CWSDPMI.

1.1
+ Added psychadelic pyramid.
+ Uses a new alpha of AllegroPak with signifiacntly updated 3D engine. One noticable difference is that the bug causing odd-pixels to poke through the pyramid when they weren't supposed to has been eliminated.
+ Everything except the sphere-drawing function has changed from fixed-point to floating-point.
+ When demonstrating the SPHERESHADE_PLIT lighting mode, you can now see where the lightsource is by the extra 'light-sphere' that appears.

1.0
+ First version to be distributed to various people.
+ Uses an alpha of AllegroPak.


Keys not mentioned on the screen:
ALT+S: Save a screendump as a PCX file.
*/

/*
Known bugs:
With texture-mapped spheres, whenever the light vector in ILIT space is {0,0,0} or in PLIT space, the light-sphere is exactly in the middle of a sphere, the program crashes with a page fault.
*/

/*
TODO:

Make a FLOAT version of the sphere.c file.

Make the pyramids zoomable.
Make the lightsource for the pyramids movable.

Add a feature to the sphere drawer so that it only draws the back of a sphere (or only the front if masked texture is used). Then, I can use both sides of the sphere with translucent drawing.

Have a non-Z-buffered version of sphere().

For the MLIT demonstration, have 4 PLIT lightpoints in a pyramid (with global_zmiddle at the center), and you can control the ILIT point. Or have multiple light-spheres.

Warn about light sphere getting too close to viewer. At the mo, the Z co-ord is truncated to MIN_LIGHT_Z (in VECTOR_f light), but the internal co-ords (lightcoords) remain unchanged. When the space to display and control the co-ordinates is changed, the internal co-ords are the ones that are translated. When switching to ILIT, the light vector is restored. There will be a problem if Z a lot less than MAX_Z_LIGHT, as user will see no change when decreasing Z. Try to use a warning indicator.
I could also change the code so that it only truncates the sphere's Z position, and not the actual Z component of the light vector. This should be easy to implement, but what about the warning that Z is too small?

With PLIT, have lightsphere in WORLDSPACE and POLARWORLDSPACE. I would not only have to rotate the co-ordinates round the central sphere, but apply camera FOV and aspect effects as well.

When demonstrating rotated texture, perhaps orient texture accodring to positions of each sphere instead of orienting the texture just to the center sphere. Is this necessary with isometric spheres?

With SPHERETEXTURE_ROTATED texture modes, have ability to change speed of aditional rotation, and also ability to restore the texture's rotation to the identity matrix.

If I decide to control the aditional rotation:
keypad + and keypad - : change speed of aditional texture rotation (infinity to -infinity). Perhaps I could even have keys to change the x, y, and z rotation components seperately.
ENTER to reset texture rotation angles and also to set the rotation speed to 0.
In the demo, for rotated_texture, display some info on texture rotation speed.

Perhaps have a more consistent control mechanism for all demos. Then, I could have one main screen with keys for all demos (eg. object rotation), and a seperate screen for each sub-demo.
Rotate pyramids / spheres: Arrows,PGUP/PGDN.
Light (Soon, I can move the lightsource with the pyramids when I do this): light-x: ZX lighty: QA lightZ: SE
Shift+Control for speed
F? to toggle rotation. Perhaps this is local to pyramids part.
Perhaps I could even have a single function to process common key-functions?

Should I have seperate helpscreens for global commands (F12) and local commands (F11)?

In all parts, I should read the keyboard inside the loop that updates the positions instead of just inside the main loop. I do however lose some accuracy and smoothness in manual-positioning if the framerate is too low.

In the psychadelic pyramid part when I do the above, make sure that changing the colours does not get the demo out of sync.

Have an option to hide the on-screen feedback (F10).
*/

/*
Idea:
Perhaps I could do the small spheres normally, and for the big one, make it translucent. I would however need to draw the big one last. This would require re-arranging the palette.

// Idea: I could also have a macro to expand "vector" to "vector.x, vector.y, vector.z", or add an & to the front as well, or with ->, or with &...->
*/


#include <stdlib.h>
#include <stdio.h>
#include <math.h>	// draw_angle() uses this

#include <allegro.h>

#include "global.h"
#include "zbuf.h"
#include "3d.h"
#include "zbpoly.h"
#include "hsv_hls.h"


#define NUM_SHAPES			4	/* number of spinning pyramids */

#define NUM_VERTICES		4	/* a pyramid has four corners */
#define NUM_FACES			4	/* a pyramis has four faces */

#define NUM_BALLS 16	// Main ball plus small balls

#define NUM_CAMERAS 		3	/* How many cameras there are in the sphere demo */

#define NUM_TEXTURES		4	/* Howmany different texture bitmaps for pyramids and spheres */

/* Define camera properties (could these be alterable?) */
#define camera_fov 40.0
#define camera_aspect 1.0

/* The position in world-space of the point that the camera is rotated around */
#define global_xmiddle 0.0
#define global_ymiddle 0.0
#define global_zmiddle 256.0
#define start_camera_alt (global_zmiddle+50.0)	/* Camera's altitude at start */

#define PY_SCALE 100.0	// 83 would make it equivalent to the fixed version.

#define START_PY_LIGHTX 6.0
#define START_PY_LIGHTY 8.0
#define START_PY_LIGHTZ -7.0

#define START_SPH_LIGHTX -6.0
#define START_SPH_LIGHTY -9.0
#define START_SPH_LIGHTZ -8.0

#define START_SAT 0.085
#define START_HUESTART 0.0
#define START_HUETWIST 21.0

/*
Minimum Altitude: This should prevent z values <= 0 from occuring as a result of zooming too close to the center.
*/
#define min_camera_alt (global_zmiddle+16.0)

#define LIGHTSPHERE_RADIUS 8.0

#define MIN_LIGHT_Z (global_zmiddle/4.0)	// This is the miomimum Z-distance in screenspace of the light-sphere.

// Define values for the small balls
#define ball_zstart	(global_zmiddle * 1.5)	// was: fmul(global_zmiddle, 0x3<<15)
#define ball_zmin 100.0	// Must not be less than (max_radius+max_dz).
#define ball_zmax (global_zmiddle*2.0-ball_zmin)	// Mirrored from zmiddle.


/* Define positions for the textual output */
#define text_xstart 23
#define textblock_ysize text_xstart
#define text_yheight 8
#define text_lightvector_ystart (text_yheight*6)
#define text_moreinfo_ystart (text_yheight*10)

/* Define text colours for user-interface */
#define value_colour 56
#define stimulus_colour 48
#define text_colour 32

/* Define some of the keys for the sphere user-interface */
#define DEC_LON KEY_Z
#define INC_LON KEY_X
#define DEC_LAT KEY_A
#define INC_LAT KEY_Q
#define DEC_ALT KEY_K
#define INC_ALT KEY_L
#define DEC_SPIN KEY_H
#define INC_SPIN KEY_J


/*
Data structures for the pyramid part
*/

/*
Structiure to store a single point (vertex)
This is similar to a V3D but only contains the position.
The other things from V3D are worked out in the draw_triangle() function that draws a triangle.
*/
typedef struct VTX
{
	float x, y, z;
	float nx, ny, nz;	/* Vertex Normal: Used to calculate light at vertex. */
} VTX;

typedef struct TRIANGLE				/* three vertices makes a triangle */
{
	VTX *vtxlist;				/* Address of start of array of possible vertices */
	int v1, v2, v3;				/* Offsets into the array of vertices */
} TRIANGLE;

typedef struct SHAPE				/* store position of a shape */
{
	float x, y, z;					/* x, y, z position */
	float rx, ry, rz;				/* rotations */
	float dz;						/* speed of movement */
	float drx, dry, drz;			/* speed of rotation */
} SHAPE;


VTX points[] =	/* a pyramid (tetrahedron in this case), centered on the origin */
{
	/* vertices of the pyramid (regular unit-tetrahedron) scaled by PY_SCALE */
	{ PY_SCALE * -0.500000000 , PY_SCALE * -0.288675134 , PY_SCALE * -0.288675134 },
	{ PY_SCALE *  0.0         , PY_SCALE * +0.577350269 , PY_SCALE *  0.0         },
	{ PY_SCALE * +0.500000000 , PY_SCALE * -0.288675134 , PY_SCALE * -0.288675134 },
	{ PY_SCALE *  0.0         , PY_SCALE * -0.288675134 , PY_SCALE * +0.577350269 },
};
// 0.288675134 = tan(30)/2
// 0.577350269 = (tan(60)-tan(30))/2 or tan(60)/3



TRIANGLE faces[] =
/*
In a right-handed co-ordinate system (Z going into the screen),
anti-clockwise polygons face the front, and clockwise polygons face the back.
*/
{
	{ points, 0, 2, 1 },
	{ points, 0, 1, 3 },
	{ points, 3, 1, 2 },
	{ points, 0, 3, 2 },
};


camera_f camera_array[NUM_CAMERAS];	/* An array of the cameras */
int camera_number;					/* Offset into the array of cameras */

/* somewhere to put translated vertices */
SHAPE shapes[NUM_SHAPES];		 	/* an array of the shapes */
VTX output_points[NUM_VERTICES * NUM_SHAPES];
TRIANGLE output_faces[NUM_FACES * NUM_SHAPES];


/*
Data structures for the sphere part
*/

// FIX -> FLOAT for the light values
typedef struct ball {
	float x, y, z;
	float radius;
	float x_tex_start, y_tex_start;
	float x_tex_size, y_tex_size;

	/* Variables to be used in the three models of lighting */
	fixed light_shade;	/* Notexture-col */
	fixed dark_shade;
	fixed light_RGB;	/* Notexture-RGB */
	fixed dark_RGB;
	fixed light_value;	/* Lit Texture */
	fixed dark_value;

	float dz;	/* Z velocity of ball */

	MATRIX_f rotation_matrix;	/* The rotation of the poles of the sphere */
	float trx, try, trz;	/* Rotation of texture  (rotation around pole) */
	float dtrx, dtry, dtrz;		/* Texture rotation speed */
} ball;

ball balls[NUM_BALLS];


PALLETE pal;			/* Global Palette */
PALLETE psychpal;		/* Palette used for the psychadelic pyramids */


int last_retrace_count;		/* Used to keep the speed constant */


char tmp_text[50];			// Used as output for sprintf when using textout()

BITMAP *buffer;	/* Where the screen is composed before being blasted to the screen */
Z_BUFFER *the_zbuffer;
BITMAP *texture[NUM_TEXTURES];		/* Where the sphere textures are stored */
BITMAP *pyramid_texture[NUM_TEXTURES];	/* A shaded texture for White, R, G, B */


char *shade_mode_desc[] = {
	"Flat (no shading)",
	"Infinite light shading",
	"Point light shading",
	"Multiple light sources (not yet implemented)"
};

#define last_shade_mode 4	// This must correspond to shade_mode_desc[]
int shade_mode=SPHERESHADE_FLAT;

char *texture_mode_desc[] = {
	"No texture. Uses colour palette for light",
	"No texture. Uses RGB table for light",
	"Texture",
	"Mercartor Texture",
	"Masked Texture",
	"Masked Mercartor Texture",
	"Rotated Texture",
	"Rotated Mercartor Texture",
	"Rotated Masked Texture",
	"Rotated Masked Mercartor Texture"
};

#define last_texture_mode 10	// This must correspond to shade_mode_desc[]
int texture_mode=SPHERETEXTURE_NOTEXTURE_COL;

char *coords_space_desc[] = {
	"viewspace",
	"screenspace",
	"polar viewspace"
};

#define last_coords_space 3	// This must correspond to shade_mode_desc[]
int coords_space=SCREENSPACE;



/*
Macro Expansions
*/

#define floatrand(SIZE,OFS) (((float)random()/((float)RAND_MAX+1.0)*SIZE)+OFS)


int press_key(int keycode)
/*
Returns true only if a key is pressed.
If a key is held down, this will not return true until the key is released and then pressed again.
This is better than readkey(), because it allows you to mix reading the key just once when it's pressed (toggle-keys, use press_key()) and reading it continuously when it's pressed (movement keys, use key[]). Also, you don't have to wait for the keys to repeat.
*/
{
	/*
	The values of the array correspond to wether a key is pressed, ordered by scancode.
	*/

	static int _pressed[128];


	if (key[keycode]) {
		if (!_pressed[keycode]) {
			_pressed[keycode]=true;
			return(true);
		} else {
			return(false);
		}
	} else {
		_pressed[keycode]=false;
		return(false);
	}
}


void print_progress(int pos)	/* Used when creating lighting_table */
{
	if ((pos & 3) == 3) {
		printf("*");
		fflush(stdout);
	}
}


int get_light_value(VTX *v, VECTOR_f *light)
/*
Work out the illumination of a point (unit vector normal to the point) with a unit-vector lightsource using the dot product.
Return value from 0-255
!!!***Make sure that >255 is never returned.
*/
{
	return (int) ( (dot_product_f(v->nx,v->ny,v->nz, light->x,light->y,light->z)+1.0) * 128.0);
}




void setup()
/* SETUP */
{
	int c, w, h;
	VECTOR_f tmp_vector;

	allegro_init();
	install_keyboard();
	install_mouse();
	install_timer();

	/* set up a greyscale in bottom quater of the pallete */
	for (c=0; c<64; c++) {
		pal[c].r = pal[c].g = pal[c].b = c;
	}

	/* set up red in next quater of the pallete */
	for (c=64; c<128; c++) {
		pal[c].r = c-64;
		pal[c].g = pal[c].b = 0;
	}

	/* set up green in next quater of the pallete */
	for (c=128; c<192; c++) {
		pal[c].g = c-128;
		pal[c].r = pal[c].b = 0;
	}

	/* set up blue in next quater of the pallete */
	for (c=192; c<256; c++) {
		pal[c].b = c-192;
		pal[c].r = pal[c].g = 0;
	}


	printf("Generating RGB mapping table:\n");
	printf("<................................................................>\r<");
	rgb_map = malloc(sizeof(RGB_MAP));
	create_rgb_table(rgb_map, pal, print_progress);

	printf("\nGenerating lighting table:\n");
	printf("<................................................................>\r<");
	color_map = malloc(sizeof(COLOR_MAP));
	create_light_table(color_map, pal, 0, 0, 0, print_progress);


	/* set the graphics mode */
	set_gfx_mode(GFX_VGA, 320, 200, 0, 0);
	set_pallete(desktop_pallete);

	if (!gfx_mode_select(&c, &w, &h)) {
		allegro_exit();
		printf("Graphics mode not selected\n\n");
		exit(1);
	}

	if (set_gfx_mode(c, w, h, 0, 0) != 0) {
		allegro_exit();
		printf("Error setting graphics mode\n%s\n\n", allegro_error);
		exit(1);
	}

	set_pallete(pal);

	/* double buffer the animation */
	if ( (buffer=create_bitmap(SCREEN_W, SCREEN_H)) == NULL) {
		allegro_exit();
		printf("Error setting up the double buffer\n");
		printf("Possibly low on memory\n\n");
		exit(1);
	}

	/* The Z buffer */
	if ( (the_zbuffer = create_zbuffer(SCREEN_W, SCREEN_H)) == NULL ) {
		allegro_exit();
		printf("Error setting up the Z buffer\n");
		printf("Possibly low on memory\n\n");
		exit(1);
	}


	text_mode(-1);	// Transparent text background.
	srandom(retrace_count);


	/* Make some Textures */

	/* make texture maps for the pyramids */
	for (w=0;w<=3;w++) {
		pyramid_texture[w] = create_bitmap(128, 128);

		rectfill(pyramid_texture[w],0,0,127,127,63+(w<<6));
		rect(pyramid_texture[w],24,13,104,23,63+(((w+3)%NUM_TEXTURES)<<6));
		rect(pyramid_texture[w],25,14,103,22,63+(((w+3)%NUM_TEXTURES)<<6));
		for (c=31;c>=0;c--) {
			circlefill(pyramid_texture[w],63,55,c,(w<<6)+(c<<1));
		}
		circlefill(pyramid_texture[w],63,55,17,0);	// Hole in the middle.
		textout(pyramid_texture[w], font, "Some texture", 16, 2, 63+(((w+1)%NUM_TEXTURES)<<6));
	}

	/* make a bitmap for use as a sphere texture map */
	for (w=0;w<=3;w++) {
		texture[w] = create_bitmap(64, 64);
		clear(texture[w]);
		rectfill(texture[w], 22, 22, 42, 42, 63+(((w+3)%NUM_TEXTURES)<<6));
		rectfill(texture[w], 32, 32, 63, 63, 63+(w<<6));
		line(texture[w], 0, 0, 31, 31, 63+(((w+1)%NUM_TEXTURES)<<6));
		line(texture[w], 0, 31, 31, 0, 63+(((w+1)%NUM_TEXTURES)<<6));
		rect(texture[w], 0, 0, 31, 31, 63+(((w+1)%NUM_TEXTURES)<<6));
		line(texture[w], 1, 0, 63, 62, 63+(((w+1)%NUM_TEXTURES)<<6));
		line(texture[w], 0, 1, 62, 63, 63+(((w+1)%NUM_TEXTURES)<<6));
		line(texture[w], 0, 0, 63, 63, 63+(((w+1)%NUM_TEXTURES)<<6));
		line(texture[w], 0, 63, 63, 0, 63+(((w+1)%NUM_TEXTURES)<<6));
		line(texture[w], 1, 63, 63, 1, 63+(((w+1)%NUM_TEXTURES)<<6));
		line(texture[w], 0, 62, 62, 0, 63+(((w+1)%NUM_TEXTURES)<<6));
		rect(texture[w], 0, 0, 63, 63, 63+(((w+1)%NUM_TEXTURES)<<6));
		line(texture[w], 47, 0, 47, 63, 63+(((w+1)%NUM_TEXTURES)<<6));
		textout(texture[w], font, "Dead", 0, 0, 63+(((w+2)%NUM_TEXTURES)<<6));
		textout(texture[w], font, "pigs", 0, 8, 63+(((w+2)%NUM_TEXTURES)<<6));
		textout(texture[w], font, "cant", 0, 16, 63+(((w+2)%NUM_TEXTURES)<<6));
		textout(texture[w], font, "fly.", 0, 24, 63+(((w+2)%NUM_TEXTURES)<<6));
		textout(texture[w], font, "Four", 32, 32, 63+(((w+2)%NUM_TEXTURES)<<6));
		textout(texture[w], font, "cows", 32, 40, 63+(((w+2)%NUM_TEXTURES)<<6));
		textout(texture[w], font, "cant", 32, 48, 63+(((w+2)%NUM_TEXTURES)<<6));
		textout(texture[w], font, "cry.", 32, 56, 63+(((w+2)%NUM_TEXTURES)<<6));
	}


	/* set up the viewport for the perspective projection */
	set_projection_viewport(0, 0, SCREEN_W, SCREEN_H);

	/*
	PYRAMIDS
	*/

	for (c=0; c<NUM_SHAPES; c++) {
		shapes[c].x = fixtof((random() & 0x1FFFFFF) - 0x1000000);
		shapes[c].y = fixtof((random() & 0xFFFFFF) - 0x800000);
		// shapes[c].z = shape_zstart;
		shapes[c].rx = 0.0;
		shapes[c].ry = 0.0;
		shapes[c].rz = 0.0;
		shapes[c].dz =  fixtof((random() & 0xFFFFF) - 0x80000);
		shapes[c].drx = fixtof((random() & 0x7FFF) - 0x4000);
		shapes[c].dry = fixtof((random() & 0x7FFF) - 0x4000);
		shapes[c].drz = fixtof((random() & 0x7FFF) - 0x4000);

		// FOR NOW: Pyramids always remain in the middle.
		shapes[c].x = 0.0;
		shapes[c].y = 0.0;
		shapes[c].z = global_zmiddle - 100.0;
		shapes[c].dz = 0.0;
	}


	/* Calculate the vertex normals for use with gouraud shading */
	for (c=0;c<NUM_VERTICES;c++) {
		/*
		Add normalised edge vectors and normalise the result.
		We are using regular tetrahedrons (pyramids with triangualr bases),
		therefore, all edges are of equal length,
		therefore, we do not need to normalise the edges before we add them.
		*/
		// If the shapes remain the same (ie. do not de-form), the normals need only be calculated once.
		// Do we need tmp_vector?

		tmp_vector.x=NUM_VERTICES*points[c].x-points[(c+1)%NUM_VERTICES].x-points[(c+2)%NUM_VERTICES].x-points[(c+3)%NUM_VERTICES].x;
		tmp_vector.y=NUM_VERTICES*points[c].y-points[(c+1)%NUM_VERTICES].y-points[(c+2)%NUM_VERTICES].y-points[(c+3)%NUM_VERTICES].y;
		tmp_vector.z=NUM_VERTICES*points[c].z-points[(c+1)%NUM_VERTICES].z-points[(c+2)%NUM_VERTICES].z-points[(c+3)%NUM_VERTICES].z;

		points[c].nx=tmp_vector.x;
		points[c].ny=tmp_vector.y;
		points[c].nz=tmp_vector.z;

		normalize_vector_f(&points[c].nx,&points[c].ny,&points[c].nz);
	}



	/*
	BALLS
	*/

	// Positions are in objectspace.

	/* The big white ball */
	balls[0].x=global_xmiddle;
	balls[0].y=global_ymiddle;
	balls[0].z=global_zmiddle;
	balls[0].radius=global_zmiddle/3.0;
	balls[0].light_shade=63<<16;
	balls[0].dark_shade=10<<16;
	balls[0].light_RGB = 0x000000BF;
	balls[0].dark_RGB = 0x00000009;
	balls[0].light_value = 255<<16;
	balls[0].dark_value = 30<<16;
	balls[0].dz=0.0;
	/* Texture */
	balls[0].x_tex_start=0.0;
	balls[0].y_tex_start=0.0;
	balls[0].x_tex_size=63.0;
	balls[0].y_tex_size=63.0;

	/* Texture position */
	balls[0].trx=balls[0].try=balls[0].trz=0.0;
	/* Texture velocity */
	balls[0].dtrx = fixtof((random() & 0x00001FFF) - 0x00001000);
	balls[0].dtry = fixtof((random() & 0x00001FFF) - 0x00001000);
	balls[0].dtrz = fixtof((random() & 0x00001FFF) - 0x00001000);


	/* The small balls */
	for (c=1;c<NUM_BALLS;c++) {

		/* Position */
		balls[c].x = fixtof((random() & 0x7FFFFF) - 0x400000);
		balls[c].y = fixtof((random() & 0x3FFFFF) - 0x200000);
		balls[c].z = ball_zstart;

		/* Radius */
		balls[c].radius = fixtof((random() & 0x1FFFFF) + 0x80000);

		/* Colour / brightness */
		w=64*((random()%3)+1);
		balls[c].light_shade = (63+w)<<16;
		balls[c].dark_shade = (10+w)<<16;
		balls[c].light_RGB = 0x00FFFFFF;
		balls[c].dark_RGB = 0x000A0A0A;
		balls[c].light_value = 255<<16;
		balls[c].dark_value = 30<<16;

		/* Velocity */
		balls[c].dz=fixtof((random() & 0x3FFFF) - 0x20000);

		/* Texture */
		balls[c].x_tex_start=0.0;
		balls[c].y_tex_start=0.0;
		balls[c].x_tex_size=63.0;
		balls[c].y_tex_size=63.0;

		/* Texture position */
		balls[c].trx=balls[c].try=balls[c].trz=0.0;
		/* Texture velocity */
		balls[c].dtrx = fixtof((random() & 0x00001FFF) - 0x00001000);
		balls[c].dtry = fixtof((random() & 0x00001FFF) - 0x00001000);
		balls[c].dtrz = fixtof((random() & 0x00001FFF) - 0x00001000);

	}

	/* Init cameras */

	camera_number=0;

	for (c=0;c<NUM_CAMERAS;c++) {

		// Front vector
		camera_array[c].xf=0.0;
		camera_array[c].yf=0.0;
		camera_array[c].zf=1.0;
		// Position
		camera_array[c].x = global_xmiddle - start_camera_alt*camera_array[c].xf;
		camera_array[c].y = global_ymiddle - start_camera_alt*camera_array[c].yf;
		camera_array[c].z = global_zmiddle - start_camera_alt*camera_array[c].zf;
		// Up vector
		camera_array[c].xup=0.0;
		camera_array[c].yup=1.0;
		camera_array[c].zup=0.0;

		camera_array[c].lon=0.0;
		camera_array[c].lat=0.0;
		camera_array[c].spin=0.0;
		camera_array[c].alt = start_camera_alt;

		camera_array[c].fov=camera_fov;
		camera_array[c].aspect=camera_aspect;

		get_camera_matrix_f(&camera_array[c].camera_matrix, camera_array[c].x, camera_array[c].y, camera_array[c].z, camera_array[c].xf, camera_array[c].yf, camera_array[c].zf, camera_array[c].xup, camera_array[c].yup, camera_array[c].zup, camera_array[c].fov, camera_array[c].aspect);

	}

}



/*
For the psychedelic pyramid, Make a new palette
// Have a start_hue paramater
// Note: The computer waits for a vertical-retrace before the palette is changed.
*/
void set_psychedelic_colours(float sat, float huestart, float huetwist)
{
	int c;
	float h, l;

	RGB output_colour;


	for(c=0,h=huestart,l=0.0;c<256;c++) {
		hls_to_rgb(h, l, sat, &output_colour);
		psychpal[c]=output_colour;


		/* Update h and l for next iteration */

		l+=(1.0/255.0);
		// optimise huetwist/255.0
		h+=(huetwist/255.0);
		if (h>=6.0) {
			h-=6.0;
		} else if (h<0.0) {
			h+=6.0;
		}

	}

	set_pallete(psychpal);
}


/* update shape positions */

void animate_shapes(num_shapes)
/* Pyramids */
{
	int c;

	for (c=0; c<num_shapes; c++) {
		// shapes[c].z += shapes[c].dz;

		/*
		if ((shapes[c].z > shape_zmax) || (shapes[c].z < shape_zmin)) {
			shapes[c].dz = -shapes[c].dz;
		}
		*/


		shapes[c].rx += shapes[c].drx;
		shapes[c].ry += shapes[c].dry;
		shapes[c].rz += shapes[c].drz;

	}
}


void animate_balls(int rotate_flag)
/* Spheres */
{
	int c;

	MATRIX_f polar_rotation_matrix, spin_rotation_matrix, texture_rotation_matrix, additional_rotation_matrix;

	/*
	Get the texture rotation to be consistent with viewpoint orientation.
	The texture is fixed in Worldspace, so we need to rotate it
	in the opposite direction to get the texture correct in viewspace.
	In isometric space, this would be the same for all balls. Ideally, we should use the camera's position in ball-object-space.
	*/

	/* First find the effect of latitude and longitude */
	get_polar_positioning_matrix_f(&polar_rotation_matrix, -camera_array[camera_number].lat, -camera_array[camera_number].lon);

	/* Next, find the effect of spin */
	get_z_rotate_matrix_f(&spin_rotation_matrix, camera_array[camera_number].spin);

	/* Get the matrix for polar rotation and spin */
	matrix_mul_f(&spin_rotation_matrix, &polar_rotation_matrix, &texture_rotation_matrix);


	for (c=0;c<NUM_BALLS;c++) {
		balls[c].z += balls[c].dz;

		if ((balls[c].z > ball_zmax) || (balls[c].z < ball_zmin)) {
			balls[c].dz = -balls[c].dz;
		}

		/*
		WAS ORIGINALLY:
		// FIX -> FLOAT
		balls[c].x_tex_start += balls[c].dtrx;
		balls[c].x_tex_start &= (((texture->h-1)<<16)+0xFFFF);
		// This technique could rotate the texture horizontally in non-rotated sphere-texture modes.
		*/


		if (rotate_flag) {
			/*
			And now add some additional rotation to the texture
			Each ball spins differently.
			*/

			balls[c].trx+=balls[c].dtrx;
			balls[c].try+=balls[c].dtry;
			balls[c].trz+=balls[c].dtrz;

			get_transformation_matrix_f(&additional_rotation_matrix, 1.0, balls[c].trx, balls[c].try, balls[c].trz, 0.0, 0.0, 0.0);

			/* Combine these effects to produce the matrix that is used to rotate the texture */
			matrix_mul_f(&texture_rotation_matrix, &additional_rotation_matrix, &balls[c].rotation_matrix);
		} else {
			/*
			No aditional rotation
			Only rotate the texture according to the orientation of the camera.
			*/
			balls[c].rotation_matrix = texture_rotation_matrix;
		}
	}

}


/*
For the pyramids:
translate shapes from 3d object-space to 2d screen-space
Uses the camera pointed to by camera_array[camera_number]
Do I really need to have the camera movements in the pyramids part?
*/
void translate_shapes(int num_shapes)
{
	int c, d;
	MATRIX_f matrix_o2w, matrix_o2v, normal_rotation_matrix;
	VTX *outpoint = output_points;
	TRIANGLE *outface = output_faces;


	for (c=0; c<num_shapes; c++) {

		/* build a transformation matrix (object-space -> world-space) */
		get_transformation_matrix_f(&matrix_o2w, 1.0, shapes[c].rx, shapes[c].ry, shapes[c].rz, shapes[c].x, shapes[c].y, shapes[c].z);

		/* build a rotation matrix for the vertex normals */
		get_transformation_matrix_f(&normal_rotation_matrix, 1.0, shapes[c].rx, shapes[c].ry, shapes[c].rz, 0.0, 0.0, 0.0);

		/* Combine Obj->World and World->View(camera) to produce Obj->View matrix */
		// You get the greatest speed increase with objects with lots of vertices.
		matrix_mul_f(&matrix_o2w, &camera_array[camera_number].camera_matrix, &matrix_o2v);

		/* output the vertices */
		for (d=0; d<NUM_VERTICES; d++) {

			/**************************************************
			// This is how it would be done without multiplying the 2 matrices

			// Can do Obj -> View by multiplying 2 matrices before the loop.
			// Saves time when object has lots of vertices.

			// Object to World
			apply_matrix_f(&matrix_o2w, points[d].x, points[d].y, points[d].z, &outpoint[d].x, &outpoint[d].y, &outpoint[d].z);

			// World to View
			apply_matrix_f(&camera_array[camera_number].camera_matrix, outpoint[d].x, outpoint[d].y, outpoint[d].z, &outpoint[d].x, &outpoint[d].y, &outpoint[d].z);
			**************************************************/

			/* Object to View */
			apply_matrix_f(&matrix_o2v, points[d].x, points[d].y, points[d].z, &outpoint[d].x, &outpoint[d].y, &outpoint[d].z);

			/* Transform Vertex Normals to worldspace (light source remains constant in viewspace) */
			apply_matrix_f(&normal_rotation_matrix, points[d].nx, points[d].ny, points[d].nz, &outpoint[d].nx, &outpoint[d].ny, &outpoint[d].nz);

			/* View to Screen */
			persp_project_f(outpoint[d].x, outpoint[d].y, outpoint[d].z, &outpoint[d].x, &outpoint[d].y);

		}

		/* output the faces */
		for (d=0; d<NUM_FACES; d++) {
			outface[d] = faces[d];	/* The positions of the faces in the vertexlist are the same. */
			outface[d].vtxlist = outpoint;	/* Set vtxlist to the translated vertices */
		}

		/* Update the start in the arrays for the next shape */
		outpoint += NUM_VERTICES;	// Translated points
		outface += NUM_FACES;		// Faces of the polyhedra
	}
}


/* draw a triangle */
void draw_triangle(BITMAP *b, VTX *v1, VTX *v2, VTX *v3, VECTOR_f *light, int texture_number, int cull_back_faces_and_use_no_zbuffer, int textpos)
// textpos just needed for feedback
{
	/* four vertices */
	V3D_f vtx1 = { v1->x, v1->y, v1->z, 127.0,	0.0,	get_light_value(v1, light) };
	V3D_f vtx2 = { v2->x, v2->y, v2->z, 63.0,	110.8512517,	get_light_value(v2, light) };
	V3D_f vtx3 = { v3->x, v3->y, v3->z, 0.0,	0.0, 	get_light_value(v3, light) };

	// tan(60)*(128/2) == 110.8512517

	/*
	Cull back faces
	Note use of lazy-evaluation to prevent z_normal_polygon_f being
	un-necessarily called.
	*/

	if (cull_back_faces_and_use_no_zbuffer && z_normal_polygon_f(&vtx1, &vtx2, &vtx3) < 0.0) {
		return;
	}


	/* draw the triangle */
	if (cull_back_faces_and_use_no_zbuffer) {
		triangle3d_f(b, POLYTYPE_GCOL, pyramid_texture[texture_number], &vtx1, &vtx2, &vtx3);
	} else {
		zb_triangle3d_f(b, the_zbuffer, POLYTYPE_PTEX_MASK_LIT, pyramid_texture[texture_number], &vtx1, &vtx2, &vtx3);
	}

	/*
	Verify that the veritces are drawn correctly.
	*/
	// Note: Z not shown here.
	// Note: No perspective-correction applied to the direction
	/*
	line(b, (int)(v1->x), (int)(v1->y), (int)(v1->x+(v1->nx*8.0)), (int)(v1->y+(v1->ny*8.0)), 63);
	line(b, (int)(v2->x), (int)(v2->y), (int)(v2->x+(v2->nx*8.0)), (int)(v2->y+(v2->ny*8.0)), 63);
	line(b, (int)(v3->x), (int)(v3->y), (int)(v3->x+(v3->nx*8.0)), (int)(v3->y+(v3->ny*8.0)), 63);
	*/

	/* Print stuff on the screen */
	/*
	sprintf (tmp_text, "%d", vtx1.c);
	textout(buffer, font, tmp_text, 100*textpos, 400+text_yheight*0, value_colour);
	sprintf (tmp_text, "%d", vtx2.c);
	textout(buffer, font, tmp_text, 100*textpos, 400+text_yheight*1, value_colour);
	sprintf (tmp_text, "%d", vtx3.c);
	textout(buffer, font, tmp_text, 100*textpos, 400+text_yheight*2, value_colour);
	*/
}


/* draw the shapes (pyramids) calculated by translate_shapes() */
void draw_shapes(BITMAP *b, VECTOR_f *light, int num_shapes, int cull_back_faces_and_use_no_zbuffer)
{
	int c;
	TRIANGLE *face = output_faces;
	VTX *v1, *v2, *v3;
	VECTOR_f light_n = *light;

	normalize_vector_f(&light_n.x,&light_n.y,&light_n.z);

	for (c=0; c < NUM_FACES * num_shapes; c++) {
		/* find the vertices used by the face */
		v1 = face->vtxlist + face->v1;
		v2 = face->vtxlist + face->v2;
		v3 = face->vtxlist + face->v3;

		/* draw the face */
		draw_triangle(b, v1, v2, v3, &light_n, (c/NUM_FACES)%num_shapes, cull_back_faces_and_use_no_zbuffer, c);
		// (c/NUM_FACES)%num_shapes) makes sure that polygons from the same shape are textured the same.

		face++;	/* Advance face to point to the next triangle structure */
	}
}



/*
And now, for the balls.
*/

/* callback for qsort()
Sorts outballs[] using Z value of all balls.
The greater the Z value, the further back the ball is.
The objects with the smallest Zs (nearest) become the earliest in the array and get drawn first.
See libc.inf for what qsort does.
*/
int ball_cmp(const void *e1, const void *e2)
{
	// This only compares the centres of the balls.
	// Should I subtract the balls' radii, or subtract sqrt(0.5)*radius?
	return ( (int)( ((ball *)e1)->z - ((ball *)e2)->z ) );
}


void translate_and_draw_spheres(VECTOR_f *light)
{
	int c;
	static ball outballs[NUM_BALLS];	// Output for transformed co-ordinates.

	// FIX -> FLOAT
	fixed *light_ptr, *dark_ptr;	// Points to which version of the light for the sphere we are using.

	/* transform the co-ordinates of the balls */
	for (c=0;c<NUM_BALLS;c++) {

		/*
		Transform the balls from World-space to view-space.
		*/
		apply_matrix_f(&camera_array[camera_number].camera_matrix, balls[c].x, balls[c].y, balls[c].z, &outballs[c].x, &outballs[c].y, &outballs[c].z);

		/*
		Transform the balls from View-space to screen-space.
		*/
		persp_project_f(outballs[c].x, outballs[c].y, outballs[c].z, &outballs[c].x, &outballs[c].y);

		/*
		Find radius in screenspace
		Radius in screenspace = (radius_in_worldspace/z_in_screenspace)*SCREEN_H
		WARNING: VERY DODGY. X values get stretched differently than Y values... Unless the projection viewport is square. Maybe I could make sphere() stretch the circle. Have radius1 (same as below), and radius2 (use SCREEN_W).

		Could I then use persp_project (r1,r2,z,r1out,r2out)?
		No, coz persp_project adds an offset.

		See sphere.c's rems for more info.
		*/

		outballs[c].radius = (balls[c].radius/outballs[c].z) * ((float)SCREEN_H)/2.0;	// See Allegro's 3D maths code for how I derived that (perspective projection).
		// Do I need to divide by 2.0? Couldn't I just update the random initializers?


		/* Copy stuff that remains unchanged to outballs */
		// I could pass directly from balls if I decide never to modify these things during transformation.
		outballs[c].light_shade=balls[c].light_shade;
		outballs[c].dark_shade=balls[c].dark_shade;
		outballs[c].light_RGB=balls[c].light_RGB;
		outballs[c].dark_RGB=balls[c].dark_RGB;
		outballs[c].light_value=balls[c].light_value;
		outballs[c].dark_value=balls[c].dark_value;
		outballs[c].x_tex_start=balls[c].x_tex_start;
		outballs[c].y_tex_start=balls[c].y_tex_start;
		outballs[c].x_tex_size=balls[c].x_tex_size;
		outballs[c].y_tex_size=balls[c].y_tex_size;
		outballs[c].rotation_matrix=balls[c].rotation_matrix;	/* Pass pointers instead of copying */
	}

	/* First, draw the light-sphere */
	if (shade_mode==SPHERESHADE_PLIT) {
		sphere_f(buffer, the_zbuffer, NULL, SPHERESHADE_ILIT, SPHERETEXTURE_NOTEXTURE_COL, light->x, light->y, light->z ,(LIGHTSPHERE_RADIUS/light->z)*((float)SCREEN_H), 0, 0, 0, 0, NULL, 63<<16, 50<<16, 0.0, 0.0, -1.0);
	}

	/*
	Sort the other spheres by depth so that the frontmost gets drawn first.
	This is because we are using Z-buffering, and if the frontmost balls are drawn first, then we reduce the probarbility that the same pixel will be re-drawn. Every time a pixel must be re-drawn, the illumination and the texture must be calculated each time.
	The best speedup is with the fancier lighting and texturing modes.
	*/
	qsort(outballs, NUM_BALLS, sizeof(ball), ball_cmp);

	/* And now draw the other spheres */
	for (c=0;c<NUM_BALLS;c++) {

		/* Which lighting model are we using */
		if (texture_mode==SPHERETEXTURE_NOTEXTURE_COL) {
			light_ptr = &outballs[c].light_shade;
			dark_ptr = &outballs[c].dark_shade;
		} else if (texture_mode==SPHERETEXTURE_NOTEXTURE_RGB) {
			light_ptr = &outballs[c].light_RGB;
			dark_ptr = &outballs[c].dark_RGB;
		} else {
			light_ptr = &outballs[c].light_value;
			dark_ptr = &outballs[c].dark_value;
		}

		/*
		And now, do the drawing.
		To calculate which texture to use, take the colour of the sphere in SPHERETEXTURE_NOTEXTURE_COL from the light_shade member. Use the top two bits of the shade value for the texture as this is the way the palette is organised (top two bits represent colour, bottom six represent brightness). As this is a fixed-point value, shift right an additional 16 places (total = 16 + 6 = 22)
		*/
		sphere_f(buffer, the_zbuffer, texture[outballs[c].light_shade>>22], shade_mode, texture_mode, outballs[c].x, outballs[c].y, outballs[c].z ,outballs[c].radius, outballs[c].x_tex_start, outballs[c].y_tex_start, outballs[c].x_tex_size, outballs[c].y_tex_size, &outballs[c].rotation_matrix, *light_ptr, *dark_ptr, light->x, light->y, light->z);

	}

}



/*
Display commands for the demos.
*/

/* Waits for a key to be pressed while flashing a message */
void wait_while_user_reads_instructions()
{
	clear_keybuf();

	while(!keypressed()) {
		vsync();
		textout(screen, font, "Press any key to continue", 0, SCREEN_H-text_yheight, 63*((retrace_count&32)==0) );
	}

	/* Make sure that if SPACE is pressed, we don't skip a part of the demo */
	if (press_key(KEY_SPACE)) {
		while(key[KEY_SPACE]) {
		}
	}

	/* Make sure that we don't instantly return to the help-screen if holding down the help key */
	while(key[KEY_F12]) {
	}
}


void display_keys_for_pyramids()
{
	clear(screen);

	textout_centre(screen, font, "Polygon Z-buffering", SCREEN_W/2, 0, 63);

	textout(screen, font, "Press        and                 to rotate", 0,text_yheight*2,text_colour);
	textout(screen, font, "ARROWS", 48,text_yheight*2,stimulus_colour);
	textout(screen, font, "PageUp/PageDown", 136,text_yheight*2,stimulus_colour);

	textout(screen, font, "Hold down       and         for speed" ,0,text_yheight*3,text_colour);
	textout(screen, font, "SHIFT",80,text_yheight*3,stimulus_colour);
	textout(screen, font, "CONTROL",160,text_yheight*3,stimulus_colour);

	textout(screen, font, "Press a number key to select pyramid:" ,0,text_yheight*5,text_colour);

	textout(screen, font, "Press    to toggle pyramid rotation control", 0,text_yheight*6,text_colour);
	textout(screen, font, "F1" ,48,text_yheight*6,stimulus_colour);

	textout(screen, font, "Press       to reset pyramid rotation" ,0,text_yheight*8,text_colour);
	textout(screen, font, "ENTER" ,48,text_yheight*8,stimulus_colour);

	textout(screen, font, "Press     to see keyboard commands", 0, text_yheight*10, text_colour);
	textout(screen, font, "F12", 48, text_yheight*10, stimulus_colour);

	wait_while_user_reads_instructions();
}


void display_keys_for_psychadelic_pyramid()
{
	clear(screen);

	set_pallete(pal);	/* Restore palette to the 64 levels of Grey/Red/Green/Blue */

	textout_centre(screen, font, "Psychadelic Pyramid", SCREEN_W/2, 0, 63);

	textout(screen, font, "Press        and                 to rotate", 0,text_yheight*2,text_colour);
	textout(screen, font, "ARROWS", 48,text_yheight*2,stimulus_colour);
	textout(screen, font, "PageUp/PageDown", 136,text_yheight*2,stimulus_colour);

	textout(screen, font, "Hold down       and         for speed" ,0,text_yheight*3,text_colour);
	textout(screen, font, "SHIFT",80,text_yheight*3,stimulus_colour);
	textout(screen, font, "CONTROL",160,text_yheight*3,stimulus_colour);

	textout(screen, font, "Press     to change saturation" ,0,text_yheight*5,text_colour);
	textout(screen, font, "Q/A", 48, text_yheight*5, stimulus_colour);
	textout(screen, font, "Press     to change starting hue" ,0,text_yheight*6,text_colour);
	textout(screen, font, "Z/X", 48, text_yheight*6, stimulus_colour);
	textout(screen, font, "Press     to change hue-twist" ,0,text_yheight*7,text_colour);
	textout(screen, font, "C/V", 48, text_yheight*7, stimulus_colour);

	textout(screen, font, "Press    to toggle pyramid rotation control", 0,text_yheight*9,text_colour);
	textout(screen, font, "F1" ,48,text_yheight*9,stimulus_colour);

	textout(screen, font, "Press       to reset pyramid rotation" ,0,text_yheight*10,text_colour);
	textout(screen, font, "ENTER" ,48,text_yheight*10,stimulus_colour);

	textout(screen, font, "Press     to see keyboard commands", 0, text_yheight*12, text_colour);
	textout(screen, font, "F12", 48, text_yheight*12, stimulus_colour);

	wait_while_user_reads_instructions();

	/* The psychadelic palette is restored after this has exited */
}


void display_keys_for_spheres()
{
	clear(screen);

	textout_centre(screen, font, "Real-time sphere rendering", SCREEN_W/2, 0, 63);

	textout(screen, font, "During the demo:", 0, text_yheight*2, value_colour);

	textout(screen, font, "Press    to change shading mode", 0, text_yheight*4, text_colour);
	textout(screen, font, "F1", 48, text_yheight*4, stimulus_colour);

	textout(screen, font, "Press    to change texturing mode", 0, text_yheight*5, text_colour);
	textout(screen, font, "F2", 48, text_yheight*5, stimulus_colour);

	textout(screen, font, "Press    to change the light's co-ordinate space", 0, text_yheight*6, text_colour);
	textout(screen, font, "F3", 48, text_yheight*6, stimulus_colour);

	textout(screen, font, "Press    to toggle aditional texture rotation", 0, text_yheight*7, text_colour);
	textout(screen, font, "F4", 48, text_yheight*7, stimulus_colour);

	textout(screen, font, "Use the       ,        and         ", 0, text_yheight*9, text_colour);
	textout(screen, font, "arrows", 64, text_yheight*9, stimulus_colour);
	textout(screen, font, "PageUp", 128, text_yheight*9, stimulus_colour);
	textout(screen, font, "PageDown", 216, text_yheight*9, stimulus_colour);
	textout(screen, font, "to change light direction/source.", 0, text_yheight*10, text_colour);

	textout(screen, font, "Use     to change Latitude", 0, text_yheight*12, text_colour);
	textout(screen, font, "Q/A", 32, text_yheight*12, stimulus_colour);
	textout(screen, font, "Use     to change longitude", 0, text_yheight*13, text_colour);
	textout(screen, font, "Z/X", 32, text_yheight*13, stimulus_colour);
	textout(screen, font, "Use     to change Camera Spin", 0, text_yheight*14, text_colour);
	textout(screen, font, "H/J", 32, text_yheight*14, stimulus_colour);
	textout(screen, font, "Use     to change Altitude", 0, text_yheight*15, text_colour);
	textout(screen, font, "K/L", 32, text_yheight*15, stimulus_colour);
	textout(screen, font, "Press a number to change the camera number", 0, text_yheight*16, text_colour);

	textout(screen, font, "Hold down       and         for speed",0,text_yheight*18,text_colour);
	textout(screen, font, "SHIFT",80,text_yheight*18,stimulus_colour);
	textout(screen, font, "CONTROL",160,text_yheight*18,stimulus_colour);

	textout(screen, font, "Press     to see keyboard commands", 0, text_yheight*20, text_colour);
	textout(screen, font, "F12", 48, text_yheight*20, stimulus_colour);

	textout(screen, font, "Press ESC to quit", 0, text_yheight*22, stimulus_colour);

	wait_while_user_reads_instructions();
}



/*
Feedback display functions
*/

/*
Draws a diagram on the screen representing the angle
c1 and c2 are thee two colours to use. c1 for grid, and c2 for the angle.
Angle is a float from 0-255
*/
void draw_angle(BITMAP *buffer, float angle, int x, int y, int size, int c1, int c2)
{
	vline (buffer, x+(size>>1), y, y+size-1, c1);
	hline (buffer, x, y+(size>>1), x+size-1, c1);

	line (buffer, x+(size>>1), y+(size>>1), (int)(sin(angle*(M_PI/128.0))*((float)size)/2.0)+x+(size>>1), (int)(-cos(angle*(M_PI/128.0))*((float)size)/2.0)+y+(size>>1), c2);

}


void display_feedback_pyramids(int pyramid_number, int retraces)
{
	// Next version of Allegro has a textout that takes printf formatted strings.

	textout(buffer, font, "Vertical retraces since last update: ", 0, 0, text_colour);
	sprintf(tmp_text, "%d", retraces);
	textout(buffer, font, tmp_text, 296, 0, value_colour);

	textout(buffer, font, "Now rotating pyramid number:",0,text_yheight*2,text_colour);
	sprintf(tmp_text, "%d", pyramid_number+1);
	textout(buffer, font, tmp_text, 232, text_yheight*2, (pyramid_number<<6) + value_colour);

	textout(buffer, font, "Press SPACE for the next part of the demo" ,0,SCREEN_H-text_yheight,63);
}


void display_feedback_psychadelic_pyramid(float sat, float huestart, float huetwist, int retraces)
{
	// Next version of Allegro has a textout that takes printf formatted strings.

	textout(buffer, font, "Vertical retraces since last update: ", 0, 0, text_colour<<2);
	sprintf(tmp_text, "%d", retraces);
	textout(buffer, font, tmp_text, 296, 0, value_colour<<2);

	textout(buffer, font, "Saturation:  Hue-start:   Hue-twist:" ,0,text_yheight*2,text_colour<<2);
	sprintf(tmp_text, "%+8.5f     %+7.4f    %+10.3f", sat, huestart, huetwist);
	textout(buffer, font, tmp_text, 0, text_yheight*3, value_colour<<2);

	textout(buffer, font, "Press SPACE for the next part of the demo" ,0,SCREEN_H-text_yheight,255);
}


void display_feedback_spheres(float *lightcoords, int rotate_flag, int retraces)
/*
Prints stuff on the screen
// retraces could be FPS if framerate/retraces.
// Question, how do I obtain a value for the framerate?
*/
{
	// Next version of Allegro has a textout that takes printf formatted strings.

	textout(buffer, font, "Vertical retraces since last update: ", 0, 0, text_colour);
	sprintf(tmp_text, "%d", retraces);
	textout(buffer, font, tmp_text, 296, 0, value_colour);

	textout(buffer, font, texture_mode_desc[texture_mode], 0, text_yheight*2, value_colour);

	/* When doing the rotated texture, say wether the aditional rotation is being used */
	if(texture_mode >= SPHERETEXTURE_ROTATED_TEX) {
		textout(buffer, font, "Additional texture rotation: ", 0, text_yheight*3, text_colour);
		if (rotate_flag) {
			textout(buffer, font, "ON", 232, text_yheight*3, value_colour);
		} else {
			textout(buffer, font, "OFF", 232, text_yheight*3, value_colour);
		}
	}

	textout(buffer, font, shade_mode_desc[shade_mode], 0, text_yheight*4, value_colour);

	/* Display feedback on the lightsource according to lighting mode */
	switch(shade_mode) {

		case SPHERESHADE_FLAT: {
			textout(buffer, font, "No light", 0, text_lightvector_ystart, text_colour);
		};break;
		case SPHERESHADE_ILIT: {
			textout(buffer, font, "Light direction vector:", 0, text_lightvector_ystart, text_colour);
			sprintf(tmp_text, "%+8.3f %+8.3f %+8.3f", lightcoords[0], lightcoords[1], lightcoords[2]);
			textout(buffer, font, tmp_text, 0, text_lightvector_ystart+text_yheight, value_colour);
		};break;
		case SPHERESHADE_PLIT: {
			textout(buffer, font, "Light position:", 0, text_lightvector_ystart, text_colour);
			textout(buffer, font, coords_space_desc[coords_space], 128, text_lightvector_ystart, value_colour);
			sprintf(tmp_text, "%+8.2f %+8.2f %+8.2f", lightcoords[0], lightcoords[1], lightcoords[2]);
			// Should the number of decimal places change sepending on which space the co-ordinates are in?
			textout(buffer, font, tmp_text, 0, text_lightvector_ystart+text_yheight, value_colour);
		};break;
		case SPHERESHADE_MLIT: {
			textout(buffer, font, "Not yet implemented!", 0, text_lightvector_ystart, text_colour);
		};break;

	}

	/* Display info about the camera's position in polar form */

	textout(buffer, font, "Latitude", text_xstart, text_moreinfo_ystart+textblock_ysize*0, text_colour);
	sprintf (tmp_text, "0x%02X.%01X", (int)camera_array[camera_number].lat, ((int)(camera_array[camera_number].lat * 16.0)) & 0xF);
	textout(buffer, font, tmp_text, text_xstart, text_moreinfo_ystart+textblock_ysize*0+text_yheight, value_colour);
	draw_angle (buffer, camera_array[camera_number].lat, 0, text_moreinfo_ystart+textblock_ysize*0-3, textblock_ysize-4, text_colour, value_colour);

	textout(buffer, font, "longitude", text_xstart, text_moreinfo_ystart+textblock_ysize*1, text_colour);
	sprintf (tmp_text, "0x%02X.%1X", (int)camera_array[camera_number].lon, ((int)(camera_array[camera_number].lon * 16.0))& 0xF);
	textout(buffer, font, tmp_text, text_xstart, text_moreinfo_ystart+textblock_ysize*1+text_yheight, value_colour);
	draw_angle (buffer, camera_array[camera_number].lon, 0, text_moreinfo_ystart+textblock_ysize*1-3, textblock_ysize-4, text_colour, value_colour);

	textout(buffer, font, "Camera spin", text_xstart, text_moreinfo_ystart+textblock_ysize*2, text_colour);
	sprintf (tmp_text, "0x%02X.%1X", (int)camera_array[camera_number].spin, ((int)(camera_array[camera_number].spin * 16.0))& 0xF);
	textout(buffer, font, tmp_text, text_xstart, text_moreinfo_ystart+textblock_ysize*2+text_yheight, value_colour);
	draw_angle (buffer, camera_array[camera_number].spin, 0, text_moreinfo_ystart+textblock_ysize*2-3, textblock_ysize-4, text_colour, value_colour);

	textout(buffer, font, "Camera altitude:", text_xstart, text_moreinfo_ystart+textblock_ysize*3, text_colour);
	sprintf (tmp_text, "0x%04X.%1X", (int)camera_array[camera_number].alt, ((int)(camera_array[camera_number].alt * 16.0))& 0xF);
	textout(buffer, font, tmp_text, text_xstart, text_moreinfo_ystart+textblock_ysize*3+text_yheight, value_colour);

	/* Camera number */
	textout(buffer, font, "Camera number:", text_xstart, text_moreinfo_ystart+textblock_ysize*4, text_colour);
	sprintf(tmp_text, "%d", camera_number+1);
	textout(buffer, font, tmp_text, text_xstart, text_moreinfo_ystart+textblock_ysize*4+text_yheight, value_colour);

}



/*
Reads keys for the pyramids and acts accordingly
*/
void read_keys_pyramids (int *pyramid_number, int *spin_flag)
{
	int i;
	float change_value;	/* When changing things with the keys, this is how much they should change */

	if (key[KEY_CONTROL]) {
		change_value=0.5;
	} else {
		change_value=0.0625;
	}
	if (key[KEY_LSHIFT] || key[KEY_RSHIFT]) {
		change_value*=2.0;
	}
	if (key[KEY_UP]) {
		shapes[*pyramid_number].rx -= change_value;
	}
	if (key[KEY_DOWN]) {
		shapes[*pyramid_number].rx += change_value;
	}
	if (key[KEY_LEFT]) {
		shapes[*pyramid_number].ry += change_value;
	}
	if (key[KEY_RIGHT]) {
		shapes[*pyramid_number].ry -= change_value;
	}
	if (key[KEY_PGUP]) {
		shapes[*pyramid_number].rz -= change_value;
	}
	if (key[KEY_PGDN]) {
		shapes[*pyramid_number].rz += change_value;
	}
	if (key[KEY_ENTER]) {
		shapes[*pyramid_number].rx = shapes[*pyramid_number].ry = shapes[*pyramid_number].rz = 0.0;
	}
	// Has a new pyramid been selected?
	for (i=KEY_1;i<KEY_1+NUM_SHAPES;i++) {
		if (key[i]) {
			*pyramid_number=i-KEY_1;
			break;
		}
	}
	if (press_key(KEY_F1)) {
		// The computer takes or releases control of the pyramids
		*spin_flag^=1;
	}
	if (key[KEY_F12]) {
		display_keys_for_pyramids();
	}
	/* Dump screen */
	if (key[KEY_ALT] && key[KEY_S]) {
		save_bitmap ("dump_pym.pcx", buffer, pal);
	}
}


/*
Reads keys for the psychadelic pyramid and acts accordingly
*/
void read_keys_psychadelic_pyramid(float *sat, float *huestart, float *huetwist, int *spin_flag)
{
	float change_value;	/* When changing things with the keys, this is how much they should change */

	if (key[KEY_CONTROL]) {
		change_value=0.5;
	} else {
		change_value=0.0625;
	}
	if (key[KEY_LSHIFT] || key[KEY_RSHIFT]) {
		change_value*=2.0;
	}
	if (key[KEY_UP]) {
		shapes[0].rx -= change_value;
	}
	if (key[KEY_DOWN]) {
		shapes[0].rx += change_value;
	}
	if (key[KEY_LEFT]) {
		shapes[0].ry += change_value;
	}
	if (key[KEY_RIGHT]) {
		shapes[0].ry -= change_value;
	}
	if (key[KEY_PGUP]) {
		shapes[0].rz -= change_value;
	}
	if (key[KEY_PGDN]) {
		shapes[0].rz += change_value;
	}
	if (key[KEY_Q]) {
		*sat += change_value/16.0;
		if (*sat>1.0) {
			*sat = 1.0;
		}
		set_psychedelic_colours(*sat, *huestart, *huetwist);
	}
	if (key[KEY_A]) {
		*sat -= change_value/16.0;
		if (*sat<0.0) {
			*sat = 0.0;
		}
		set_psychedelic_colours(*sat, *huestart, *huetwist);
	}
	if (key[KEY_Z]) {
		*huestart -= change_value/8.0;
		if (*huestart<0.0) {
			*huestart += 6.0;
		}
		set_psychedelic_colours(*sat, *huestart, *huetwist);
	}
	if (key[KEY_X]) {
		*huestart += change_value/8.0;
		if (*huestart>=6.0) {
			*huestart -= 6.0;
		}
		set_psychedelic_colours(*sat, *huestart, *huetwist);
	}
	if (key[KEY_C]) {
		*huetwist -= change_value/4.0;
		set_psychedelic_colours(*sat, *huestart, *huetwist);
	}
	if (key[KEY_V]) {
		*huetwist += change_value/4.0;
		set_psychedelic_colours(*sat, *huestart, *huetwist);
	}

	if (key[KEY_ENTER]) {
		shapes[0].rx = shapes[0].ry = shapes[0].rz = 0.0;
	}
	if (press_key(KEY_F1)) {
		// The computer takes or releases control of the pyramids
		*spin_flag^=1;
	}
	if (key[KEY_F12]) {
		display_keys_for_psychadelic_pyramid();
		set_psychedelic_colours(*sat, *huestart, *huetwist);	/* Restore the psychadelic colours */
	}
	/* Dump screen */
	if (key[KEY_ALT] && key[KEY_S]) {
		save_bitmap ("dump_psy.pcx", buffer, psychpal);
	}
}


/*
Reads keys for the spheres and acts accordingly
*/
void read_keys_spheres(float *lightcoords, VECTOR_f *light, int *spin_flag)
{
	int i;

	float change_value;	/* When changing things with the keys, this is how much they should change */
	float light_change_value;	/* Used to set the speed of the lightsource's change */

	float tmp_lat, tmp_lon, tmp_alt;
	MATRIX_f tmp_polar_rotation_matrix;

	// use #defines to define some key charecteristics.

	/* The rendering mode */

	if (press_key(KEY_F1)) {
		shade_mode++;
		if (shade_mode >= last_shade_mode) {
			shade_mode = SPHERESHADE_FLAT;
		} else if (shade_mode == SPHERESHADE_ILIT) {
			lightcoords[0]=START_SPH_LIGHTX;
			lightcoords[1]=START_SPH_LIGHTY;
			lightcoords[2]=START_SPH_LIGHTZ;
		} else if (shade_mode == SPHERESHADE_PLIT) {
			coords_space=SCREENSPACE;
			lightcoords[0]=((float)SCREEN_W)/2.4;
			lightcoords[1]=((float)SCREEN_H)/2.4;
			lightcoords[2]=global_zmiddle-balls[0].radius*1.25;
		}
	}

	if (press_key(KEY_F2)) {
		texture_mode++;
		if (texture_mode >= last_texture_mode) {
			texture_mode = SPHERETEXTURE_NOTEXTURE_COL;
		}
	}

	/* Change the 'space' of the light-sphere */

	if (shade_mode == SPHERESHADE_PLIT) {
		if (press_key(KEY_F3)) {

			/*
			We are changing viewing space.
			Transform the co-ordinates to those for the next space, and advance to the next space
			*/

			switch(coords_space++) {
				case VIEWSPACE : {
					/* Translate co-ordinates from VIEWSPACE to SCREENSPACE*/

					persp_project_f(lightcoords[0], lightcoords[1], lightcoords[2], &lightcoords[0], &lightcoords[1]);

				};break;
				case SCREENSPACE : {
					/* Translate co-ordinates from SCREENSPACE to POLARVIEWSPACE */

					/*
					The reason why latitude and longitude need to be inverted is because in viewspace, the vecror {0,1,0} points downwards (in worldspace, it points upwards).
					Do I need to bother with this inversion, or just tell people that in viewspace, the north pole is facing downwards.
					*/

					/* First translate to viewspace */
					reverse_persp_project_f(lightcoords[0], lightcoords[1], lightcoords[2], &lightcoords[0], &lightcoords[1]);

					/* Now translate to polar co-ordinates with respect to camera's altitude */
					// CAMERA ALTITUDE
					// Note, with the currently implemented code, global_[x,y,z]middle is the same in view and world spaces. Beware when it's possible to point the camera away from the big middle sphere.

					// lightcoords[0] -= global_xmiddle;
					// lightcoords[1] -= global_ymiddle;
					// lightcoords[2] -= global_zmiddle;

					lightcoords[2] -= camera_array[camera_number].alt;

					tmp_lat = -get_latitude_f(lightcoords[0], lightcoords[1], lightcoords[2]);
					tmp_lon = -get_longitude_f(lightcoords[0], lightcoords[1], lightcoords[2]);
					tmp_alt = vector_length_f(lightcoords[0], lightcoords[1], lightcoords[2]);

					lightcoords[0] = tmp_lon;
					lightcoords[1] = tmp_lat;
					lightcoords[2] = tmp_alt;


				};break;
				case POLARVIEWSPACE : {
					/* Translate co-ordinates from POLARVIEWSPACE to VIEWSPACE */

					// In viewspace, latitude and longitude need to be inverted.

					/* Translate to cartesian co-ordinates with respect to camera's altitude */
					get_polar_positioning_matrix_f(&tmp_polar_rotation_matrix, -lightcoords[1], -lightcoords[0]);
					apply_matrix_f(&tmp_polar_rotation_matrix, 0.0, 0.0, -lightcoords[2], &lightcoords[0], &lightcoords[1], &lightcoords[2]);

					/* And now translate to viewspace */
					lightcoords[2] += camera_array[camera_number].alt;

					/* Make the coords_space variable wraparound */
					coords_space -= last_coords_space;	/* Get back to the first space */
				};break;
			}

		}

		light_change_value=8.0;
	} else {
		// Using ILIT: Smaller changes - more accuracy
		light_change_value=1.0;
	}


	/* The lightsource */

	/* Press shift to speed up all movements */
	/* Press control to further up all movements */
	/* Press shift and control together for lots of speed */

	if (key[KEY_CONTROL]) {
		change_value=0.25;
	} else {
		change_value=0.03125;
	}
	if (key[KEY_LSHIFT] || key[KEY_RSHIFT]) {
		change_value*=2.0;
	}
	if (key[KEY_LEFT]) {
		lightcoords[0] -= change_value*light_change_value;
	}
	if (key[KEY_RIGHT]) {
		lightcoords[0] += change_value*light_change_value;
	}
	if (key[KEY_UP]) {
		lightcoords[1] -= change_value*light_change_value;
	}
	if (key[KEY_DOWN]) {
		lightcoords[1] += change_value*light_change_value;
	}
	if (key[KEY_PGUP]) {
		lightcoords[2] -= change_value*light_change_value;
	}
	if (key[KEY_PGDN]) {
		lightcoords[2] += change_value*light_change_value;
	}


	/* The camera */

	if (key[DEC_LON]) {
		// Rotate clockwise / left / decrease longitude

		polar_rotate_camera_f(&camera_array[camera_number], 0.0, -change_value*8.0, 0.0, 0.0, global_xmiddle, global_ymiddle, global_zmiddle);
	}
	if (key[INC_LON]) {
		// Rotate anti-clockwise / right / increase longitude

		polar_rotate_camera_f(&camera_array[camera_number], 0.0, change_value*8.0, 0.0, 0.0, global_xmiddle, global_ymiddle, global_zmiddle);
	}
	if (key[DEC_LAT]) {
		// Rotate camera down / Decrease latitude

		polar_rotate_camera_f(&camera_array[camera_number], -change_value*8.0,0.0, 0.0, 0.0, global_xmiddle, global_ymiddle, global_zmiddle);
	}
	if (key[INC_LAT]) {
		// Rotate camera up / Increase latitude

		polar_rotate_camera_f(&camera_array[camera_number], change_value*8.0,0.0, 0.0, 0.0, global_xmiddle, global_ymiddle, global_zmiddle);
	}
	if (key[DEC_SPIN]) {
		// Decrease spin (rotation of camera around the direction vector it's facing)

		polar_rotate_camera_f(&camera_array[camera_number],0.0, 0.0, 0.0, -change_value*8.0, global_xmiddle, global_ymiddle, global_zmiddle);
	}
	if (key[INC_SPIN]) {
		// Increase spin (rotation of camera around the direction vector it's facing)

		polar_rotate_camera_f(&camera_array[camera_number],0.0, 0.0, 0.0, change_value*8.0, global_xmiddle, global_ymiddle, global_zmiddle);
	}
	if (key[INC_ALT]) {
		// Increase altitude (distance from middle)

		polar_rotate_camera_f(&camera_array[camera_number],0.0, 0.0, change_value*16.0, 0.0, global_xmiddle, global_ymiddle, global_zmiddle);
	}
	if (key[DEC_ALT]) {
		// Decrease altitude (distance from middle), but don't get too close

		polar_rotate_camera_f(&camera_array[camera_number],0.0, 0.0, -change_value*16.0, 0.0, global_xmiddle, global_ymiddle, global_zmiddle);

		if (camera_array[camera_number].alt < min_camera_alt) {
			polar_rotate_camera_f(&camera_array[camera_number],0.0, 0.0, min_camera_alt-camera_array[camera_number].alt, 0.0, global_xmiddle, global_ymiddle, global_zmiddle);
		}
	}

	// Has a new camera been selected?
	for (i=KEY_1;i<KEY_1+NUM_CAMERAS;i++) {
		if (key[i]) {
			camera_number=i-KEY_1;
			break;
		}
	}
	if (press_key(KEY_F4)) {
		// The computer takes or releases control of the sphere rotation
		*spin_flag^=1;
	}
	if (key[KEY_F12]) {
		display_keys_for_spheres();
	}
	/* Dump screen */
	if (key[KEY_ALT] && key[KEY_S]) {
		save_bitmap ("dump_sph.pcx", buffer, pal);
	}

	/* Adjust the light according to the light co-ordinates */

	if (shade_mode == SPHERESHADE_PLIT) {
		/*
		In PLIT mode,
		Translate the new co-ordinates of the lightsource from the currently selected space to SCREENSPACE
		*/

		switch(coords_space) {
			case VIEWSPACE : {
				/* Translate co-ordinates from VIEWSPACE to SCREENSPACE */
				persp_project_f(lightcoords[0], lightcoords[1], lightcoords[2], &light->x, &light->y);
				light->z=lightcoords[2];
			};break;
			case SCREENSPACE : {
				/* leave co-ords unchanged */
				light->x=lightcoords[0];
				light->y=lightcoords[1];
				light->z=lightcoords[2];
			};break;
			case POLARVIEWSPACE : {
				/* Wrap longitude */
				if (lightcoords[0]>=128.0) {
					lightcoords[0]-=256.0;
				} else if (lightcoords[0]<-128.0) {
					lightcoords[0]+=256.0;
				}

				/* Wrap latitude */
				if (lightcoords[1]>=128.0) {
					lightcoords[1]-=256.0;
				} else if (lightcoords[1]<-128.0) {
					lightcoords[1]+=256.0;
				}

				/* First, translate co-ordinates from POLARVIEWSPACE to VIEWSPACE */

				// In viewspace, latitude and longitude need to be inverted.

				/* Translate to cartesian co-ordinates with respect to camera's altitude */
				get_polar_positioning_matrix_f(&tmp_polar_rotation_matrix, -lightcoords[1], -lightcoords[0]);
				apply_matrix_f(&tmp_polar_rotation_matrix, 0.0, 0.0, -lightcoords[2], &light->x, &light->y, &light->z);

				/* And now translate to viewspace */
				light->z += camera_array[camera_number].alt;

				/* And now, translate co-ordinates from VIEWSPACE to SCREENSPACE */
				persp_project_f(light->x, light->y, light->z, &light->x, &light->y);

			};break;
		}



		/*
		After the translation, make sure the lightsource doesn't get too close to the viewer
		*/
		if (light->z < MIN_LIGHT_Z) {
			/* Prevent the lightsource from coming too close to the viewer */
			light->z = MIN_LIGHT_Z;
		}

	} else {
		/* If ILIT, the co-ords are just the same */

		light->x=lightcoords[0];
		light->y=lightcoords[1];
		light->z=lightcoords[2];
	}


}


/******************************
The main procedure of SPH_DEMO
*******************************/

void main()
{
	int pyramid_number=0;	/* Which pyramid are we manipulating */


	int num_retraces_since_previous;	/* Used to display number of retraces */
	int computer_spin=false;	/* Should the computer take control of the spinning pyramids? */

	float sat=START_SAT, huestart=START_HUESTART, huetwist=START_HUETWIST;

	/*
	Light Vector
	Must be usable with dot product.
	For infinite light, it's a direction vector (reverse direction of light).
	For point-light, it's a position vector (position of lightsource).
	*/

	/*
	When rendering the spheres, can have multiple ways of representing the co-ordinates.
	lightcoords[] are the co-ordinates in the current space.
	light is the light vector in screenspace
	*/

	float lightcoords[3];
	VECTOR_f light;

	// See rems for an idea about making macros to expand vectors so they can be taken by a function that only takes vector-components.


	setup();

	clear_keybuf();	/* Clear any keys held down while selecting graphics mode */




	/* ***
	Polygon Z-buffering.
	A demonstration of the Z-buffered polygon routines.
	*** */

	display_keys_for_pyramids();

	light.x=START_PY_LIGHTX;
	light.y=START_PY_LIGHTY;
	light.z=START_PY_LIGHTZ;

	last_retrace_count = retrace_count;


	for (;;) {

		clear(buffer);
		clear_zbuffer(the_zbuffer);	// Draw shapes stretching back to infinity.


		/*
		Wait for the shape positions to catch up with the frame counter
		before any transformation or rendering is done.
		This is to keep a constant speed when the number of vbls per render is unpredictable.
		Also, record the number of retraces.
		*/
		num_retraces_since_previous=0;
		while (last_retrace_count < retrace_count) {
			if (computer_spin) {
				animate_shapes(NUM_SHAPES);
			}
			last_retrace_count++;
			num_retraces_since_previous++;
			// Could the main keyboard commands be read here instead?
			if (key[KEY_ESC]) {
				return;
			}
		}


		/* Translate and draw polys */
		translate_shapes(NUM_SHAPES);		// Should I pass the light vector?
		draw_shapes(buffer, &light, NUM_SHAPES, false);


		display_feedback_pyramids(pyramid_number,num_retraces_since_previous);

		vsync();
		blit(buffer, screen, 0, 0, 0, 0, SCREEN_W, SCREEN_H);

		read_keys_pyramids(&pyramid_number, &computer_spin);

		/* Should we go on to the next part? */
		if (press_key(KEY_SPACE)) {
			while(key[KEY_SPACE]) {
			}
			break;
		}


	}



	/* ***
	Psychadelic pyramid
	A demonstration of combining the HLS->RGB function with goraud shading.
	*** */

	// Should I clear the shapes array? //

	display_keys_for_psychadelic_pyramid();

	set_psychedelic_colours(sat, huestart, huetwist);	/* Set the psychadelic colours */

	computer_spin=false;
	shapes[0].rx = shapes[0].ry = shapes[0].rz = 0.0;

	last_retrace_count = retrace_count;


	for (;;) {

		clear(buffer);


		/*
		Wait for the shape positions to catch up with the frame counter
		before any transformation or rendering is done.
		This is to keep a constant speed when the number of vbls per render is unpredictable.
		Also, record the number of retraces.
		*/
		num_retraces_since_previous=0;
		while (last_retrace_count < retrace_count) {
			if (computer_spin) {
				animate_shapes(1);
			}
			last_retrace_count++;
			num_retraces_since_previous++;
			// Could the main keyboard commands be read here instead?
			if (key[KEY_ESC]) {
				return;
			}
		}

		/* Translate and draw polys */
		translate_shapes(1);		// Should I pass the light vector?
		draw_shapes(buffer, &light, 1, true);


		display_feedback_psychadelic_pyramid(sat, huestart, huetwist, num_retraces_since_previous);

		vsync();
		blit(buffer, screen, 0, 0, 0, 0, SCREEN_W, SCREEN_H);

		read_keys_psychadelic_pyramid(&sat, &huestart, &huetwist, &computer_spin); // If I put this inside the loop that waits for positions to catch-up with the framecounter, I should bear in mind that 1 retrace is lost each time the colours change.

		/* Should we go on to the next part? */
		if (press_key(KEY_SPACE)) {
			while(key[KEY_SPACE]) {
			}
			break;
		}

	}

	set_pallete(pal);	/* Restore palette to the 64 levels of Grey/Red/Green/Blue */



	/* ***
	Real-time 3D spheres
	A demonstration of Z-buffered light-sourced textured spheres.
	*** */

	display_keys_for_spheres();

	coords_space = SCREENSPACE;

	lightcoords[0]=light.x=START_SPH_LIGHTX;
	lightcoords[1]=light.y=START_SPH_LIGHTY;
	lightcoords[2]=light.z=START_SPH_LIGHTZ;


	computer_spin=false;

	last_retrace_count = retrace_count;

	for (;;) {
		clear(buffer);
		clear_zbuffer(the_zbuffer);	// Draw shapes stretching back to infinity.

		/*
		Wait for the shape positions to catch up with the frame counter
		before any transformation or rendering is done.
		This is to keep a constant speed when the number of vbls per render is unpredictable.
		Also, record the number of retraces.
		*/
		num_retraces_since_previous=0;
		while (last_retrace_count < retrace_count) {
			animate_balls(computer_spin);
			last_retrace_count++;
			num_retraces_since_previous++;
			// Could the main keyboard commands be read here instead?
			if (key[KEY_ESC]) {
				return;
			}
		}

		/* Draw and translate spheres */
		translate_and_draw_spheres(&light);
		// rectfill(buffer,200,0,300,194,63);	// White reference

		display_feedback_spheres(lightcoords, computer_spin, num_retraces_since_previous); // If I wanted to always display the co-ords in screenspace, pass &light instead of &lightcoords[0] and change the types in the function.

		vsync();
		blit(buffer, screen, 0, 0, 0, 0, SCREEN_W, SCREEN_H);

		read_keys_spheres(lightcoords, &light, &computer_spin);	/* Read keys and act accordingly */

	}
}
