/*
Sphere.c
C code for sphere drawing functions in Allegro.
By Andrei Ellman.

Thanks to Ove Kaaven (ovek@arcticnet.no): For giving me a few hints about texture-mapping spheres.
*/


/*
Known bugs:
With texture-mapped spheres, whenever the light vector in ILIT space is {0,0,0} or in PLIT space, the light-sphere is exactly in the middle of a sphere, the program crashes with a page fault.
I suspect the bug causes the crash in the shaded texture mapping section of plot_sphere_pixel(), and the cause may be a rounding error in the dor_product reurning a fixed-point value outside the range {-1..+1}. This will lead to a ridiculously large index value in plot_sphere_pixel().
Workaround, do not have ILIT light vectors of {0,0,0} or PLIT light sources right in the middel of a sphere.
*/

/*
TODO:
MLIT light mode
Use an incremental algorithm for determiming the circle on the X-Z plane.
Can I optimise even more by having all the circles on each Y=K plane be reduced versions of the sphere on the Y=equator plane?

When wrapping onto the backside, increase x_offset by x_size. (when rotating texture, should this be done before or after the matrix?)

Do I get a noticable speedup if I sort the spheres so the frontmost gets drawn first? I would especially in textured modes (and to a lesser extent in lit modes). As the relative positions of these spheres don't change so often, use another sorting algorithm instead of qsort. A simple insretion sort may be OK if you don't rotate the spheres too often.

Convert cases and ifs to function pointers.
Alternatively, re-write the whole lot in ASM.

Perspective-distorted spheres.
Stretch component for non-square modes.
3D ellipses?
Use FLOATs instead. This will reduce rounding-errors considerably.

Note: with NOTEXTURE_COL, provided I have setup the palette properly, I can do coloured light-sources (this would look best when MLIT is done).
light_RGB is the colour of the lightsource. dark_RGB is the colour of the ambient light. In MLIT mode, there will be several light_RGBs, but just one dark_RGB.

Another idea for the lighting model, when doing the dot-product, instead of converting 1..-1 -> 255..0, convert 1..0 -> 255..0, 0..-1 -> 0..0. This will be a more accurate lighting model, but the spheres won't look as nice as they used to be.

Re-write in Assembler!
*/


/*
NOTE: Due to perspective distortions, this isn't very accurate sphere drawing, but it's close enough. In fact, this is isometric sphere drawing.

The co-ordinates passed to this function are in screenspace. (could a future version with perspective-correct spheres handle viewspace co-ords?)
For now, All lighting is calculated in screenspace.

To translate from view-space:
persp_project(x,y,z, xout, yout, zout)
rout=r/z*SCREEN_H
When stretch is added, r2out=r/z*SCREEN_W, or derive some form of stretch_factor.
stretch=SCREEN_W/SCREEN_H, or the inverse.

to find the radius in screen-space: do this,
Convert from viewspace(cameraspace) to screenspace the line from (center) of sphere to (centre)+(0,radius,0). In screenspace, find length of line. That is the radius.
Note: Don't forget to transform the light-source vector if its relative to the world.

Note: the spheres are not perspective-distorted. This will show up most when using unlit masked texture (lit masked texture could be an improvement). This is a prefect representation for an Isometric 3D world, and almost pefrect for perspective 3D (especially when far away, or near the centre of the screen). In isometric 3D, stretch should not be used (unless non-square pixels). If the sphere is perspective-distorted, as well as being stretched in 2D by SCREEN_W/SCREEN_H with persp_project(), the front is expanded, and the back is compressed. Also, the maxima and minima of Z may not be in the middle of the elipse that is the result of the projection onto the screen. Perhaps I should have proper 3D tilted ellipses that can have slices removed? These slices could be simulated with masked texture mapping and curved surfaces.

Try substitusing projection equasion into the equasion of the sphere. I wonder if I could then directly obtain 1/Z? The solution proparbly involves adjusting Z according to the projection (ie. shifting maximum Z on the Z=K plane, and squashing non-maximum Zs.

Uses GNU C's nested functions. This means the code is less portable.
Without nested functions, I would need to pass masses of paramaters.
Even GNU C++ doesn't support it.

Do not have radius==0 (although radius close to 0 should be OK)

May also behave strangely if light vector is 0,0,0 for ILIT, or for a single pixel if the lightsource is on the pixel.

Also, have an ambient light component.

Idea: using the texture mapping code, make code to map any 2D co-ords to 3D co-ords on a sphere. Then, I could map splines on my sphere, and objects can follow these paths.

RE-WRITE in assembler, but first, make the algorithms work perfectly.
*/


#include <allegro.h>

#include "global.h"
#include "zbuf.h"	// Plotting the pixels.
#include "3d.h" // matrix_f_to_matrix(). Can get rid of this when sphere_f done properly.

/* Gnu C++ does not support nested functions */
#ifdef __cplusplus
extern "C" {
#endif

#define MASKED_COLOUR 0

/* For now, just convert the floats to fixeds and call the fixed sphere function */
void sphere_f(BITMAP *bmp, struct Z_BUFFER *zb, BITMAP *texture_bitmap, int shade_type, int tex_type, float x_middle, float y_middle, float z_middle, float radius, float tex_xoffset, float tex_yoffset, float tex_xsize, float tex_ysize, MATRIX_f *texture_rotation_matrix, fixed light_shade, fixed dark_shade, float x_light, float y_light, float z_light)
// light_shade, dark_shade 	// FIX -> FLOAT
{
	MATRIX fixed_texture_rotation_matrix;	/* Temporary matrix with fixed point values */

	/* Prevent de-referencing of null pointer */
	if (texture_rotation_matrix == NULL) {
		sphere(bmp, zb, texture_bitmap, shade_type, tex_type, ftofix(x_middle), ftofix(y_middle), ftofix(z_middle), ftofix(radius), ftofix(tex_xoffset), ftofix(tex_yoffset), ftofix(tex_xsize), ftofix(tex_ysize), NULL, light_shade, dark_shade, ftofix(x_light), ftofix(y_light), ftofix(z_light));
	} else {
		matrix_f_to_matrix(texture_rotation_matrix, &fixed_texture_rotation_matrix);

		sphere(bmp, zb, texture_bitmap, shade_type, tex_type, ftofix(x_middle), ftofix(y_middle), ftofix(z_middle), ftofix(radius), ftofix(tex_xoffset), ftofix(tex_yoffset), ftofix(tex_xsize), ftofix(tex_ysize), &fixed_texture_rotation_matrix, light_shade, dark_shade, ftofix(x_light), ftofix(y_light), ftofix(z_light));
	}

}

void sphere(BITMAP *bmp, Z_BUFFER *zb, BITMAP *texture_bitmap, int shade_type, int tex_type, fixed x_middle, fixed y_middle, fixed z_middle, fixed radius, fixed tex_xoffset, fixed tex_yoffset, fixed tex_xsize, fixed tex_ysize, MATRIX *texture_rotation_matrix, fixed light_shade, fixed dark_shade, fixed x_light, fixed y_light, fixed z_light)
/*
Draw a sphere.
See global.h for values for shade_type and tex_type.

The co-ordinates (x_middle, y_middle, z_middle) are in screen space.

Light_shade and dark_shade represent the maximum intensity possible for the sphere, and the intensity of the ambient light.
In SPHERETEXTURE_NOTEXTURE_COL lighting mode, light_shade and dark_shade are values of type fixed representing a colour-index.
In SPHERETEXTURE_NOTEXTURE_RGB lighting mode, these two values are 24-bit number (R is bits 16-23, g is bits 8-15, B is bits 0-7).
In all textured drawing modes, the two values are values of type fixed representing the brightness (0-255).

With SPHERETEXTURE_COL (or perhaps any other mode), dark_shade is usually the strength of the ambient light.

Perhaps Light Intensity could be specified from from 0-1 with an ambient component instead of light_shade and dark_shade? Perhaps dark_shade is the ambient light, and light_shade is the colour that such a sphere is lit if light_intensity=100%

With SPHERESHADE_ILIT, x_light, y_light, z_light is the direction vector of the light (distance is infinity).
With SPHERESHADE_PLIT, x_light, y_light, z_light is the position vector of the light.

The light vector given does not need to be normalised. ILIT normalises it once per sphere. PLIT normalises it once per pixel.

BEWARE: Radius must be < sqrt(itofix(MAX_FIX)) (=181). Otherwise rsquared > max_middle (no problem with floats).
Possible soultion: Shift the radius, x, y to the right a bit before it's squared, and then shift Z back left. Better solution: sphere_f

Add a stretch component for non-square modes. This may be a solution to the distorted x&y screenspace co-ords, but spheres in viewspace become squashed in screenspace.
This stretch component is usually the ratio of the persp_project viewport's dimensions.
This may affect sphere_line too, coz of the 2 different radii

tex_size, tex_offset:
These can be any value. The texture bitmap's dimensions needs to have dimensions of a power of two.
Size is the number of texture-pixels to be mapped onto the front of the sphere. Offset is where it starts.
If masked texture is used, the back is mapped with the next tex_size pixels in the texture bitmap. The back is reversed because the texture is being mapped round the sphere.
If the edge of the texture-bitmap is reached, then it wraps around to the start.

texture_rotation_matrix: This matrix is the rotation matrix that is applied to the texture in the ROTATED modes.
*/
{
	/* Pre-calculate some values. */
	int x_mid_int=fixtoi(x_middle);
	int y_mid_int=fixtoi(y_middle);
	int r_int=fixtoi(radius);

	/* For use in the inner loop of sphere_line() */
	fixed rsquared=fmul(radius, radius);
	fixed one_over_radius=fdiv(1<<16,radius);
	int int_2_x_middle=x_middle>>15;
	int int_2_y_middle=y_middle>>15;

	/* Pre-calculate the light colour-difference (is this only useful for SPHERETEXTURE_NOTEXTURE_COL?) */
	fixed light_minus_dark=light_shade-dark_shade;

	/* Masks for the texture bitmap. */
	int texture_height_mask, texture_width_mask;

	int masked_texture_exitflag=false;	// Should the pixel drawing function exit early if using masked textures?


	void sphere_line(int x1, int y_line, int x2)
	/*
	Helper function for sphere()
	Draws a horizontal line of the sphere (and also mirrors it (if needed))

	Minor problem: With back/front-lit spheres, rim may be too bright/dark.
	Solution: Multiply surface_vector_z by some constant. Do it after value for Z buffer is calculated, but before surface_vector_z is normalised.
	Solution2: Pixel Gamma correction (slow)
	Solution3: Gamma correct the Colour mapping table, but then, I can only use SPHERESHADE_RGB instead of SPHERESHADE_COL
	Maybe if perspective-correct spheres were used instead?...

	Idea: use a different function for plotting the point colour on each type of sphere.
	Use function pointers to select the function. Or use a case statement.
	*/
	{
		/*
		Surface vector (from middle to edge) Direction point on surface is from middle.
		Only the Zs facing the viewer are calculated. (unless in *TEX_MASK_* mode)
		I should really put these 3 in a VECTOR struct, but Allegro prefers them on their own.
		*/
		fixed surface_vector_x;
		fixed surface_vector_y=(y_line<<16)-y_middle;	// Y remains constant in inner loop
		fixed surface_vector_z;
		fixed unnormalised_sz;	// Z needs to be re-calc'd in masked texture.

		fixed one_over_z;	/* Place to put the calculated Z value */

		/* Pre-calculate values that are fixed in the inner loop (x_line) */
		fixed ysquared=fmul(surface_vector_y, surface_vector_y);
		fixed rsquared_minus_ysquared=rsquared-ysquared;

		int x_line;	// Loop counter for the piece of sphere.


		void sphere_pixel(int point_x, int point_y, fixed sx, fixed sy, fixed sz)
		/*
		Helper function for sphere_line()
		This function plots the point.
		If necessary, it reads the texture and finds the light intensity at that point.
		The point is checked to see if it can be drawn onscreen, and is then checked against the value in the Z buffer to see if it's visible.

		Re-write to use function pointers instead of case statements and ifs. Function pointers if function is called 4 times (it may or may not be called 4 times when perspective-distortion is used). The function pointers can be worked out once, whereas the cases and ifs need doing for each iteration.
		*/

		{
			static fixed surface_intensity;
			static fixed surface_texture;


			void texture_map()
			/*
			helper function for sphere_pixel()
			This function determines the texture colour for that point and sets surface_texture to it's colour index.
			Have another version for colour modes > 256 colours that returns an RGB triplet.
			This may or may not eventually be referenced by function pointers.
			// Should this modify surface_texture globally, or should it return the calculated value?
			*/
			{

				int determine_texture_pixel(fixed sph_x, fixed sph_y, fixed sph_z)
				/*
				helper finction for texture_map()
				Finds the actual texel that corresponds to the texture pixel on the sphere
				This will eventually be referenced by function pointers.
				*/
				// With non-rotated texture, I could save a bit of time by using global sx,sy,sz instead of passing them to sph_x, sph_y, sph_z
				// Note: For when you want both sides to be the same, the texture wraps to the start of the texture bitmap (the actual bitmap start, and not the point specified by offset).
				{
					static int tx, ty;

					// tex_xsize,tex_ysize can be anything. It's the texture bitmap that must be a power of two.

					// If mapping the backside and I want it to be different, offset+=size. Does this work with rotated texture?

					// sph_x, sph_y, sph_z are already normalised.
					// Due to rounding errors, sph_y can sometimes be outside the range -1.0 - +1.0. Is this true with floats?

					if (tex_type >= SPHERETEXTURE_ROTATED_TEX) {
						/* Rotate {sx,sy,sz} */
						apply_matrix(texture_rotation_matrix, sph_x,sph_y,sph_z, &sph_x,&sph_y,&sph_z);

					}

					// It would be quicker to use a bitfield.
					if (tex_type==SPHERETEXTURE_MTEX || tex_type==SPHERETEXTURE_MASK_MTEX || tex_type==SPHERETEXTURE_ROTATED_MTEX || tex_type==SPHERETEXTURE_ROTATED_MASK_MTEX) {

						/* Mercartor texture mapping (is this what it's called?) */
						// Note similarity with get_latitude() function.

						ty = fixtoi(fmul(facos(-sph_y)>>7,tex_ysize)) & texture_height_mask;

					} else {

						// sph_y -> ty is just a linear mapping.

						ty=fixtoi( fmul(((sph_y+(1<<16))>>1),tex_ysize) + tex_yoffset ) & texture_height_mask;

					}

					/* Find x co-ordinate */
					// Note similarity with get_longitude() function. sph_x and sph_z do not need to be normalised.

					tx=fixtoi( fmul((fatan2(-sph_z,-sph_x)>>7),tex_xsize) + tex_xoffset ) & texture_width_mask;

					return(_getpixel(texture_bitmap, tx, ty));

				}


				if (tex_type==SPHERETEXTURE_NOTEXTURE_COL) {
					surface_texture=light_shade;	// light_shade is A colour index
				} else if (tex_type==SPHERETEXTURE_NOTEXTURE_RGB) {
					surface_texture=light_shade;	// light_shade is An RGB value
					// I may one day make one of these two statements different.
				} else {

					surface_texture=determine_texture_pixel(sx, sy, sz);

					if ( surface_texture==MASKED_COLOUR && (tex_type==SPHERETEXTURE_MASK_TEX || tex_type==SPHERETEXTURE_MASK_MTEX || tex_type==SPHERETEXTURE_ROTATED_MASK_TEX || tex_type==SPHERETEXTURE_ROTATED_MASK_MTEX) ) {

						/*
						Masked pixel. Switch to other side of sphere.
						Should I reverse sz here, or pass a reversed sz to the determine_texture_pixel() function?
						The answer depends on how you want the pixels on the other side to be lit.
						Is it this simple with perspective-correct spheres? No.
						*/
						sz=-sz;	// Used to dot-product with light
						unnormalised_sz=-unnormalised_sz;	// Used to calculate new 1/z
						one_over_z=fdiv(1<<16,(z_middle+unnormalised_sz));

						if (fixtof(one_over_z) > zb->line[point_y][point_x]) {

							surface_texture=determine_texture_pixel(sx, sy, sz);
							// If I didn't reverse sz earlier, but passed it here as a paramater, the inside of the sphere will be illuminated instaed of the outside.

							if (surface_texture==MASKED_COLOUR) {
								/*
								The other side of the sphere is also a masked pixel, so no
								pixel is to be plotted.
								*/
								masked_texture_exitflag=true; // exit the pixel drawing function completely.
								// Is it possible to exit two layers of functions without a goto?
							}

						} else {
							/* Pixel obscured by another object in Z buffer. */
							masked_texture_exitflag=true; // exit the pixel drawing function completely.
							// Is it possible to exit two functions without a goto?
						}

					}
				}

			}


			fixed work_out_surface_intensity()
			/*
			helper function for sphere_pixel()
			This function determines the intensity of light on the surface and returns it.
			This will eventually be referenced by function pointers.
			// When this is a function that's pointed to, should it take the shade type as a paramater?
			*/

			/*
			Get the dot_product of direction vector {middle to point} and the light vector {direction of source} to get the point intensity.
			Isurface=Iincident(surface_vector.-light_vector) (light vector already negated)
			We've already defined the max and min colours, so we assume Iincident=1
			Surface intensity now from -1 - +1. To get 0 - +1, (..)/2+0.5

			See Zen & The art of computer graphics for a better formua
			(including one with ambient light and different light colours).
			*/
			{
				/*
				For the point-source, a light-direction vector needs to be
				calculated for each pixel
				*/
				static fixed light_vector_x;
				static fixed light_vector_y;
				static fixed light_vector_z;

				switch(shade_type) {
					case SPHERESHADE_FLAT: {

						/* No shading */
						return(0);	// Return zero illumination

					}; break;
					case SPHERESHADE_ILIT: {

						/*
						Infinitely far light. All rays are paralell (in screenspace)
						{x_light, y_light, z_light} is the direction of the light source.
						*/

						return(dot_product(sx,sy,sz, x_light,y_light,z_light)+(1<<16)) >>1;
						// NOTE: With fixed, This has been believed to exceed 1<<16 for light vectors close to {0,-1,0}

					}; break;
					case SPHERESHADE_PLIT: {

						/*
						Point-source. Each pixel calculates the light vector (vector from {x,y,z} on sphere to point source.
						Slow. The light vector has to be calculated (and normalised) for each point.
						The light vector is in screen space.
						*/
						// Is it quicker to do the dot-product with the un-normalised light vector and then divide by the length of the light_vector?
						light_vector_x= x_light-(x_middle+sx);
						light_vector_y= y_light-(y_middle+sy);
						light_vector_z= z_light-(z_middle+sz);

						if (light_vector_x==0 && light_vector_y==0 && light_vector_z==0) {
							// Prevent division by zero. Would this be needed if there was a check outside to make sure the lightsource wasn't on the surface?
							return(1<<16);
						} else {
							normalize_vector(&light_vector_x,&light_vector_y,&light_vector_z);

							return(dot_product(sx,sy,sz,light_vector_x,light_vector_y,light_vector_z) + (1<<16) ) >>1;
							// NOTE: With fixed, This has been believed to exceed 1<<16 for light vectors close to {0,-1,0}
						}

					}; break;
					case SPHERESHADE_MLIT: {

						// NOT_YET_IMPLEMENTED
						// See global.h for more info.
						/*
						for(each light vector)
							if(ilit)
								return(work_out_surface_intensity_for_ILIT());
							else
								return(work_out_surface_intensity_for_PLIT());


						*/
						return(1<<16);	// For now, return the light colour.

					}; break;
				}

				return(0);	// This should stop gcc from complaining.
			}


			void plot_sphere_pixel()
			/*
			helper function for sphere_pixel()
			This function plots the pixel.
			This will eventually be referenced by function pointers.
			We are not using modex, and we have established that the pixel is in range, so we can use _putpixel.
			*/
			{
				int r, g, b;

				if (tex_type==SPHERETEXTURE_NOTEXTURE_COL) {

					/*
					Colour indexes to be used spread from fixtoi(dark_shade) to fixtoi(light_shade).
					sphere_hue and sphere_saturation not used.
					*/

					// pixel_value = dark_shade + (light_shade-dark_shade)*intensity.

					_putpixel(bmp, point_x, point_y, fixtoi(dark_shade+fmul(light_minus_dark,surface_intensity)) );

				} else if (tex_type==SPHERETEXTURE_NOTEXTURE_RGB) {

					/* Here, dark_shade and light_shade are 24-bit RRGGBB values */

					// Could I have an HSV or HLS version as well?

					// pixel_value in HSV space: h = sphere_hue, s = sphere_saturation, v = dark_shade + (light_shade-dark_shade)*intensity

					r = dark_shade & 0x00FF0000;
					r = fixtoi( r+fmul(((light_shade & 0x00FF0000)-r), surface_intensity) );
					g = (dark_shade & 0x0000FF00)<<8;
					g = fixtoi( g+fmul((((light_shade & 0x0000FF00)<<8)-g), surface_intensity) );
					b = (dark_shade & 0x000000FF)<<16;
					b = fixtoi( b+fmul((((light_shade & 0x000000FF)<<16)-b), surface_intensity) );

					_putpixel(bmp, point_x, point_y, rgb_map->data[r>>3][g>>3][b>>3]);

				} else if (shade_type==SPHERESHADE_FLAT) {
					/* Direct texture mapping */

					_putpixel(bmp, point_x, point_y, surface_texture);


				} else {

					/* Shaded texture mapping */

					_putpixel(bmp, point_x, point_y, color_map->data[fixtoi(dark_shade+fmul(light_minus_dark,surface_intensity))][surface_texture]);

				}

			}



			/*
			Is pixel in range? (REM: expand to take any set of values for range, and translate them to an appropriately sized z-buffer, or just blit to a bitmap the same size as the zbuffer (this saves memory)).
			Otherwise, when pixel is out of range, memory gets corrupted.
			If better clipping was done (only pixels in sphere drawn), this isn't needed.
			Try to clip the sphere outside of any loops. Dificult because the points are reflected to preserve the Z value.
			However, now that we know it's in range, we can use _putpixel and _getpixel.
			Beware also of negative one_over_z values, or when z==0.
			*/
			if ( (point_x>=0) && (point_x<bmp->w) && (point_y>=0) && (point_y<bmp->h) ) {

				/*
				Compare value with Zbuffer
				If one_over_z > the previous value,
				then work out texture, work out light, plot pixel and record new z value.
				*/

				if (fixtof(one_over_z) > zb->line[point_y][point_x]) {


					/* Work out the texture colour and set surface_texture accordingly */
					texture_map();	// Modifies surface_texture.

					/*
					When texturing,
					were both sides of the sphere maksed
					or was the sphere-front a mask pixel and the sphere-back obscured?
					*/

					if (masked_texture_exitflag) {
						masked_texture_exitflag=false;	// this is set to false whenever sphere() called.
						return;
					}


					/* Work out the surface intensity */
					surface_intensity=work_out_surface_intensity();
					// work_out_surface_intensity() could in the future become a function pointer.

					// ** NOTE: With fixed, This has been believed to exceed 1.0 for light vectors close to {0,-1,0}
					// Use the below if statement to limit it to 1.
					// The float version should be better at handling this?
					if (surface_intensity > 1<<16) {
						surface_intensity = 1<<16;
					}


					/* Plot the pixel */
					plot_sphere_pixel();

					/* Record the new 1/z value */
					zb->line[point_y][point_x] = fixtof(one_over_z);
				}
			}

		}


		surface_vector_y=fdiv(surface_vector_y,radius);	// From now on, we only need y normalised.

		// To impose a limit on the x value, x_middle must be offscreen. Otherwise, the reflected x value will not be drawn.

		/* Find positions of all points on line Y of the sphere */

		for (x_line=x1;x_line<=x_mid_int;x_line++) {

			/*
			circle around origin: x^2 + y^2 + z^2 = r^2;
			=> z^2 = r^2 - y^2 - x^2;
			*/
			// We could use an incremental circle algorithm to eliminate the square-root calculation (Hardenburgh's?), but that only gives discrete Zs. Perhaps try up-scaling x,y and then down-scaling the resulting Z? At least we're using lookup tables for sqrt here.

			surface_vector_x=(x_line<<16)-x_middle;
			surface_vector_z=unnormalised_sz=-fsqrt(rsquared_minus_ysquared-fmul(surface_vector_x,surface_vector_x));
			// Do rounding errors with small values of radius lead to a possible sqrt(-n)?
			// Alternatively, skip small spheres from the drawing.

			/*
			Calculate value for Z buffer
			Z values are more efficiently calculated for isometric spheres, but 1/z values to make it compatible with the future Allegro standard Z-buffer.
			// BEWARE of when z=0 (don't worry if proper object-clipping is done)
			*/
			one_over_z=fdiv(1<<16,(z_middle+surface_vector_z));

			/*
			Normalise the surface vector.
			This can be done quickly, since we know the length of the vector (ie. the radius)
			Y is pre-normalised outside the loop.
			*/
			surface_vector_x=fmul(surface_vector_x,one_over_radius);
			surface_vector_z=fmul(surface_vector_z,one_over_radius);

			/* Plot pixels for the two visible quadrants in each hemisphere */
			// Mirroring may or may not be a good idea when perspective-correct sphere-Z-calculation is used. Maybe stretched mirroring could then be used?
			// Reason for not using four horizontal octants: X varies linearly, and Z doesn't.
			// Is mirroring worth the speedup? I have to make sure the values are in range before the Zbuffer is used. However, there is a square-root in the Z calculation.

			sphere_pixel(x_line, y_line, surface_vector_x, surface_vector_y, surface_vector_z); // No mirroring

			sphere_pixel(int_2_x_middle-x_line, y_line, -surface_vector_x, surface_vector_y, surface_vector_z); // Flip around X axis

			sphere_pixel(x_line, int_2_y_middle-y_line, surface_vector_x, -surface_vector_y, surface_vector_z); // Flip around the Y axis

			sphere_pixel(int_2_x_middle-x_line, int_2_y_middle-y_line, -surface_vector_x, -surface_vector_y, surface_vector_z); // Flip around X and Y axes

		}
	}

	/*
	The below code for determining the vertical lines of a filled circle is was pinched from Allegro's circlefill code.
	Modified to do half a circle. The sphere_line function mirrors this half.
	Could I skip lines where Y is out of bounds? No, but I could skip lines where both Y and Y' are out of bounds (Y' is Y mirrored on the horizontal line that intercepts the center of the circle).
	This implies that y_middle must be offscreen.
	If the code is re-written for the perspective distortions, the original algorithm that mirrors the lises at this stage can be restored.
	*/

	int cx = 0;
	int cy = r_int;
	int df = 1 - r_int;
	int d_e = 3;
	int d_se = -2 * r_int + 5;

	if (shade_type==SPHERESHADE_ILIT) {
		/*
		For infinitely distant light, the light vector remains the same for all pixels, so it's a good idea to normalise it here.
		*/
		normalize_vector(&x_light,&y_light,&z_light);
	}

	/*
	Pre-calculate masks for the texture bitmap
	They are &'d to keep values within range.
	Texture dimensions must be a power of two.
	*/
	if (texture_bitmap!=NULL) {
		texture_height_mask = texture_bitmap->h-1;
		texture_width_mask = texture_bitmap->w-1;
	}


	do {
		// Center to top-middle
		// skip if both y_mid_int-cx and y_mid_int+cx are out of bounds.
		sphere_line(x_mid_int-cy, y_mid_int-cx, x_mid_int+cy);

		if (df < 0)  {
			df += d_e;
			d_e += 2;
			d_se += 2;
		} else {

			if (cx != cy) {
				// Top to top-middle
				// skip if both y_mid_int-cy and y_mid_int+cy are out of bounds.
				sphere_line(x_mid_int-cx, y_mid_int-cy, x_mid_int+cx);
			}

			df += d_se;
			d_e += 2;
			d_se += 4;
			cy--;
		}

		cx++;

	} while (cx <= cy);

}


#ifdef __cplusplus
}
#endif
