/*
zbpscan.s

Z-buffered Polygon scanline filler helpers (gouraud shading, tmapping, etc).

Modified by AE to use Z-buffering.

A lot of this code was originally written by Shawn Hargreaves.
Only the Z-buffer stuff was by Andrei, who also did some further optimisations (self-modifying code and re-arranging a few opcodes to reduce pipeline stalls).
Andrei also added these two scanline fillers: zb_poly_scanline_atex_mask_lit, zb_poly_scanline_ptex_mask_lit

asmdef.h and asmdefs.h not modified.

NOTE: A lot of this code uses self-modifying code, which on Pentiums, is generally not a good idea, as it causes the code-cache to re-load. Instead, try to only use self-modifying code as little as possible (even once per poly is too much). Perhaps use it to hard-code the screensize and the screen address / z-buffer address.

NOTE: I have deliberately avoided using %ebp as a register to help maintain consistency with Shawn's code, but he tells me it's OK to use %ebp (the reason having something to do with a faulty assumption made about Windows).

Note: Writing directly to the screen can be slow, as we are only writing a byte at a time (this is because of z-buffering) (the alternative would be to read an lword, and write an lword back with the appropriate pixls changed). Reading from screen memory can be slow.

KNOWN BUGS:
Sometimes, if polygons go too far in the Z direction (maybe the Z slope in th scanline is too large), the program may crash with an FPE in the polygon drawer's PTEX_* scanline filler. This also happens in Shawn's original code as well as my Z-buffered modification (which is here), but I have yet to conclusively prove that this one's not my fault.

When using PTEX_LIT (and I think ATEX_LIT too), the light value wraps back to 255 when it has passed 0. This also happens in Shawn's original code, but as above, I've yet to find out who'se bug it is. Try also with GCOL and GRGB

QUESTIONS:
+ Which parts of the SRW code need only be modified once per poly instead of once per scanline. Perhaps if I only modify the code once per poly instead of once per scanline, this would mean less code-cache re-loads. In fact, modifying once per poly is still too much.
+ Can I use the remaining FPU regs? How many FPU regs remain?
+ Should I have the jmp +2 to clear the pre-fetch queue, or are there enough bytes between the modidied and the modifier? Does the Pentium's branch-prediction tell the prefetch-queue what to load, or will a jmp always clear it?
+ With PTEX, In the helpers for updating the u, v, and z values, is it best to add DZ before or after storing Z in ONEOVERZ_n?
*/


#include "asmdefs.h"

.text



/* all these functions share the same parameters */

#define ADDR      ARG1
#define W         ARG2
#define INFO      ARG3


/* Thie next define should really be in asmdef.h */
#define POLYSEG_ZBLINE   92

/* Dummy value for use in self-re-writing code */
#define REWRITE_DUMMY 0x12345678


/* Helper for updating the z value (DZn is stored in the FPU stack) */
/* %fs(0)=Z, %fs(1)=DZ */
#define UPDATE_Z_F()															\
	fadd %st(1)   /* Update Z value */ 									; \
	fsts ONEOVERZ /* Question: Is it best to have this before or after the 'fadd' */


#define ONEOVERZ     -4(%ebp)


/* void _zb_poly_scanline_flat(unsigned long addr, int w, POLYGON_SEGMENT *info);
 *  Fills a flat shaded polygon scanline.
 */
.globl __zb_poly_scanline_flat
   .align 4
__zb_poly_scanline_flat:
   pushl %ebp
   movl %esp, %ebp
	subl $4, %esp                /* one local variable: */
    pushl %ebx
   pushl %edi
   pushw %es


   movl INFO, %edi               /* load registers */
   movl POLYSEG_SEG(%edi), %eax
   movw %ax, %es
	flds POLYSEG_DZ(%edi)          /* dz at bottom of fpu stack */
	flds POLYSEG_Z(%edi)          /* z at top of fpu stack */
	movl POLYSEG_ZBLINE(%edi), %ebx
   movl POLYSEG_C(%edi), %eax	/* eax is the colour */
   movl W, %ecx
   movl ADDR, %edi

/* eax = colour,  ebx = ZB_LINE,  ecx = Width,  edx = new_zb_value,  edi = ADDR */

zb_flat_loop:

	UPDATE_Z_F()

	movl ONEOVERZ, %edx
	cmpl %edx, (%ebx)  /* If existing_value >= ONEOVERZ: skip */
	jge zb_flat_skip

   movb %al, %fs:(%edi)          /* write pixels */
	movl %edx, (%ebx)	/* Record the new value in the Z_BUFFER */
   incl %edi
	addl $4, %ebx
   decl %ecx
   jg zb_flat_loop
   jmp zb_flat_done

   .align 4, 0x90
zb_flat_skip:
   incl %edi
 	addl $4, %ebx                 /* skip zero pixels */
   decl %ecx
   jg zb_flat_loop

zb_flat_done:
	fstp %st(0)
	fstp %st(0)
   popw %es
   popl %edi
   popl %ebx
   movl %ebp, %esp
   popl %ebp
   ret                           /* end of _zb_poly_scanline_flat() */

#undef ONEOVERZ



#define ONEOVERZ	-4(%ebp)


/* void _zb_poly_scanline_gcol(unsigned long addr, int w, POLYGON_SEGMENT *info);
 *  Fills a single-color gouraud shaded polygon scanline.
 */
.globl __zb_poly_scanline_gcol
   .align 4
__zb_poly_scanline_gcol:
   pushl %ebp
   movl %esp, %ebp
	subl $4, %esp                /* one local variables: */
   pushl %ebx
   pushl %esi
   pushl %edi

   movl INFO, %edi               /* load registers */
   movl POLYSEG_DC(%edi), %esi
   movl POLYSEG_C(%edi), %edx	/* %edx = C */
	movl %esi, zb_gcol_skip_odd_byte+2
	flds POLYSEG_DZ(%edi)          /* dz at bottom of fpu stack */
	movl %esi, REWRITE_DOWN_zb_gcol_skip_1+2
	flds POLYSEG_Z(%edi)          /* z at top of fpu stack */
	movl %esi, REWRITE_DOWN_zb_gcol_skip_2+2
	movl POLYSEG_ZBLINE(%edi), %ebx	/* ebx = zb_line */
	movl %esi, REWRITE_DOWN_zb_gcol_skip_3+2
   movl W, %ecx
	movl %esi, zb_gcol_skip_4+2
   movl ADDR, %edi
	movl %esi, zb_gcol_skip_leftover_byte+2

/* C and DC are stored as FIXED point numbers */


/* eax (al) = calculated pixel, ebx = ZB_LINE, ecx = W, edx = C, esi = new_zb_value , edi = ADDR */
/* DC is re-written */

   .align 4, 0x90
zb_gcol_odd_byte_loop:
   testl $3, %edi                /* deal with any odd pixels */
   jz zb_gcol_no_odd_bytes

	UPDATE_Z_F()

	movl ONEOVERZ, %esi
	cmpl %esi, (%ebx)	/* If existing_value >= ONEOVERZ: skip */
	jge zb_gcol_skip_odd_byte

   movl %edx, %eax
	movl %esi, (%ebx)	/* Record the new value in the Z_BUFFER */
   shrl $16, %eax
   movb %al, %fs:(%edi)

zb_gcol_skip_odd_byte:
   addl $REWRITE_DUMMY, %edx
	addl $4, %ebx
   incl %edi
   decl %ecx
   jg zb_gcol_odd_byte_loop

   .align 4, 0x90
zb_gcol_no_odd_bytes:
   movl %ecx, W                  /* inner loop expanded 4 times */
   shrl $2, %ecx
   jz zb_gcol_cleanup_leftovers

   .align 4, 0x90
zb_gcol_loop:

	UPDATE_Z_F()

	movl ONEOVERZ, %esi
	cmpl %esi, (%ebx)	/* If existing_value >= ONEOVERZ: skip */
	jge zb_gcol_skip_1

   movl %edx, %eax               /* pixel 1 */
	movl %esi, (%ebx)	/* Record the new value in the Z_BUFFER */
   shrl $16, %eax
	movb %al, %fs:(%edi)

zb_gcol_skip_1:
	UPDATE_Z_F()

	movl ONEOVERZ, %esi
	REWRITE_DOWN_zb_gcol_skip_1:
   addl $REWRITE_DUMMY, %edx
	cmpl %esi, 4(%ebx)	/* If existing_value >= ONEOVERZ: skip */
	jge zb_gcol_skip_2

   movl %edx, %eax               /* pixel 2 */
	movl %esi, 4(%ebx)	/* Record the new value in the Z_BUFFER */
   shrl $16, %eax
	movb %al, %fs:1(%edi)

zb_gcol_skip_2:
	UPDATE_Z_F()

	movl ONEOVERZ, %esi
	REWRITE_DOWN_zb_gcol_skip_2:
   addl $REWRITE_DUMMY, %edx
	cmpl %esi, 8(%ebx)	/* If existing_value >= ONEOVERZ: skip */
	jge zb_gcol_skip_3

   movl %edx, %eax               /* pixel 3 */
	movl %esi, 8(%ebx)	/* Record the new value in the Z_BUFFER */
   shrl $16, %eax
	movb %al, %fs:2(%edi)

zb_gcol_skip_3:
	UPDATE_Z_F()

	movl ONEOVERZ, %esi
 	REWRITE_DOWN_zb_gcol_skip_3:
  addl $REWRITE_DUMMY, %edx
	cmpl %esi, 12(%ebx)	/* If existing_value >= ONEOVERZ: skip */
	jge zb_gcol_skip_4

   movl %edx, %eax               /* pixel 4 */
	movl %esi, 12(%ebx)	/* Record the new value in the Z_BUFFER */
   shrl $16, %eax
	movb %al, %fs:3(%edi)

zb_gcol_skip_4:
   addl $REWRITE_DUMMY, %edx

	addl $16, %ebx
   addl $4, %edi
   decl %ecx
   jg zb_gcol_loop

zb_gcol_cleanup_leftovers:
   movl W, %ecx                  /* deal with any leftover pixels */
   andl $3, %ecx
   jz zb_gcol_done

   .align 4, 0x90
zb_gcol_leftover_byte_loop:
	UPDATE_Z_F()

	movl ONEOVERZ, %esi
	cmpl %esi, (%ebx)	/* If existing_value >= ONEOVERZ: skip */
	jge zb_gcol_skip_leftover_byte

   movl %edx, %eax
	movl %esi, (%ebx)	/* Record the new value in the Z_BUFFER */
   shrl $16, %eax
   movb %al, %fs:(%edi)

zb_gcol_skip_leftover_byte:
   addl $REWRITE_DUMMY, %edx
	addl $4, %ebx
   incl %edi
   decl %ecx
   jg zb_gcol_leftover_byte_loop

zb_gcol_done:
	fstp %st(0)
	fstp %st(0)
   popl %edi
   popl %esi
   popl %ebx
   movl %ebp, %esp
   popl %ebp
   ret                           /* end of _poly_scanline_gcol() */


#undef DU
#undef ONEOVERZ



#define DR     -4(%ebp)
#define DG     -8(%ebp)
#define DB     -12(%ebp)
#define ONEOVERZ	-16(%ebp)
#define ZB_LINE		-20(%ebp)

/* void _zb_poly_scanline_grgb(unsigned long addr, int w, POLYGON_SEGMENT *info);
 *  Fills an RGB gouraud shaded polygon scanline.
 */
.globl __zb_poly_scanline_grgb
   .align 4
__zb_poly_scanline_grgb:
   pushl %ebp
   movl %esp, %ebp
   subl $20, %esp                /* four local variables: */

   pushl %ebx
   pushl %esi
   pushl %edi

	/* Re-write the code with the rgb_mapping table as a displacement */
	movl _rgb_map, %eax
   movl INFO, %esi               /* load registers */
	movl %eax, REWRITE_UP_zb_gcol_rgbmap-4

   movl POLYSEG_DR(%esi), %eax
   movl POLYSEG_DG(%esi), %ebx
   movl POLYSEG_DB(%esi), %ecx

 	movl %eax,REWRITE_DOWN_zb_grgb_r+2
 	movl %ebx,REWRITE_DOWN_zb_grgb_g+2
 	movl %ecx,REWRITE_DOWN_zb_grgb_b+2
 	movl %eax,REWRITE_DOWN_zb_grgb_skip_r+2
 	movl %ebx,REWRITE_DOWN_zb_grgb_skip_g+2
 	movl %ecx,REWRITE_DOWN_zb_grgb_skip_b+2

	flds POLYSEG_DZ(%esi)          /* dz at bottom of fpu stack */
	flds POLYSEG_Z(%esi)          /* z at top of fpu stack */

   movl POLYSEG_R(%esi), %ebx
   movl POLYSEG_G(%esi), %ecx
   movl POLYSEG_B(%esi), %edx

	movl POLYSEG_ZBLINE(%esi), %edi


/* eax = palette colour, %ebx = R, %ecx = G, %edx = B, %esi = RGB colour/ADDR,  %edi = ZB_LINE */
   .align 4, 0x90
zb_grgb_loop:
	UPDATE_Z_F()

	movl ONEOVERZ, %eax
	cmpl %eax, (%edi)  /* If existing_value >= ONEOVERZ: skip */
	jge zb_grgb_skip


   movl %ecx, %eax               /* green */
   movl %ebx, %esi               /* red */
   shrl $19, %eax                /* green */
   shrl $19, %esi                /* red */
   shll $5, %eax                 /* green */
   shll $10, %esi                /* red */
   orl %eax, %esi                /* green */

   movl %edx, %eax               /* blue */
   shrl $19, %eax                /* blue */
   orl %eax, %esi                /* blue */

	movb REWRITE_DUMMY(%esi), %al     /* table lookup */
	REWRITE_UP_zb_gcol_rgbmap:

	movl ADDR, %esi
   movb %al, %fs:(%esi)          /* write the pixel */

	movl ONEOVERZ, %eax
	movl %eax, (%edi)		/* Record 1/z */

   incl ADDR
	addl $4, %edi
 	REWRITE_DOWN_zb_grgb_r:
   addl $REWRITE_DUMMY, %ebx
 	REWRITE_DOWN_zb_grgb_g:
   addl $REWRITE_DUMMY, %ecx
 	REWRITE_DOWN_zb_grgb_b:
   addl $REWRITE_DUMMY, %edx
   decl W
   jg zb_grgb_loop
   jmp zb_grgb_done

	.align 4, 0x90
zb_grgb_skip:
   incl ADDR
	addl $4, %edi
 	REWRITE_DOWN_zb_grgb_skip_r:
   addl $REWRITE_DUMMY, %ebx
 	REWRITE_DOWN_zb_grgb_skip_g:
   addl $REWRITE_DUMMY, %ecx
 	REWRITE_DOWN_zb_grgb_skip_b:
   addl $REWRITE_DUMMY, %edx
   decl W
   jg zb_grgb_loop

zb_grgb_done:
	fstp %st(0)
	fstp %st(0)
   popl %edi
   popl %esi
   popl %ebx
   movl %ebp, %esp
   popl %ebp
   ret                           /* end of _poly_scanline_grgb() */



#undef DR
#undef DG
#undef DB
#undef ONEOVERZ
#undef ZB_LINE


#define DU        -4(%ebp)
#define DV        -8(%ebp)
#define UMASK     -12(%ebp)
#define VMASK     -16(%ebp)
#define VSHIFT    -20(%ebp)
#define TMP       -24(%ebp)
#define COUNT     -28(%ebp)
#define C         -32(%ebp)
#define ONEOVERZ  -36(%ebp)
#define DZ1       -40(%ebp)
#define DZ4       -44(%ebp)


/* helper for setting up an affine texture mapping operation */
#define INIT_ATEX(extra...)                                                  \
   pushl %ebp                                                              ; \
   movl %esp, %ebp                                                         ; \
   subl $44, %esp                /* 11 local variables */                  ; \
									   ; \
   pushl %ebx                                                              ; \
   pushl %esi                                                              ; \
   pushl %edi                                                              ; \
									   ; \
   movl INFO, %esi               /* load registers */                      ; \
									   ; \
   extra                                                                   ; \
									   ; \
   movl POLYSEG_DU(%esi), %eax                                             ; \
   movl POLYSEG_DV(%esi), %ebx                                             ; \
   movl POLYSEG_VSHIFT(%esi), %ecx                                         ; \
   movl POLYSEG_UMASK(%esi), %edx                                          ; \
   movl POLYSEG_VMASK(%esi), %edi                                          ; \
									   ; \
   shll %cl, %edi                /* adjust v mask and shift value */       ; \
   negl %ecx                                                               ; \
   addl $16, %ecx                                                          ; \
									   ; \
   movl %eax, DU                                                           ; \
   movl %ebx, DV                                                           ; \
   movl %ecx, VSHIFT                                                       ; \
   movl %edx, UMASK                                                        ; \
   movl %edi, VMASK                                                        ; \
									   ; \
	flds POLYSEG_DZ(%esi)          /* dz at bottom of fpu stack */          ; \
	flds POLYSEG_Z(%esi)          /* z at top of fpu stack */               ; \
									   ; \
   movl POLYSEG_U(%esi), %ebx                                              ; \
   movl POLYSEG_V(%esi), %edx                                              ; \
									                                       ; \
   movl ADDR, %edi




/* void _zb_poly_scanline_atex(unsigned long addr, int w, POLYGON_SEGMENT *info);
 *  Fills an affine texture mapped polygon scanline.
 */
.globl __zb_poly_scanline_atex
   .align 4
__zb_poly_scanline_atex:
   INIT_ATEX()

    movl POLYSEG_TEXTURE(%esi), %eax
    movl POLYSEG_ZBLINE(%esi), %esi
	movl %eax, REWRITE_UP_texture_atex_z-4

   .align 4, 0x90
atex_z_loop:
   movl %edx, %eax               /* get v */
   movb VSHIFT, %cl
   sarl %cl, %eax
   andl VMASK, %eax


   movl %ebx, %ecx               /* get u */
   sarl $16, %ecx
   andl UMASK, %ecx
   addl %ecx, %eax

	UPDATE_Z_F()

	movl ONEOVERZ, %ecx
	cmpl %ecx, (%esi)  /* If existing_value >= ONEOVERZ: skip */
	jge atex_z_skip

   movb REWRITE_DUMMY(%eax), %al        /* read texel */
	REWRITE_UP_texture_atex_z:

   movb %al, %fs:(%edi)          /* write solid pixels */
	movl %ecx, (%esi)	/* Record the new value in the Z_BUFFER */
   addl DU, %ebx
   addl DV, %edx
	addl $4, %esi
   incl %edi
   decl W
   jg atex_z_loop
   jmp atex_z_done

   .align 4, 0x90
atex_z_skip:
   addl DU, %ebx                 /* skip covered pixels */
   addl DV, %edx
 	addl $4, %esi
  incl %edi
   decl W
   jg atex_z_loop

atex_z_done:
   fstp %st(0)
   fstp %st(0)
   popl %edi
   popl %esi
   popl %ebx
   movl %ebp, %esp
   popl %ebp
   ret                           /* end of _poly_scanline_atex() */





/* void _zb_poly_scanline_atex_mask(ulong addr, int w, POLYGON_SEGMENT *info);
 *  Fills a masked affine texture mapped polygon scanline.
 */
.globl __zb_poly_scanline_atex_mask
   .align 4
__zb_poly_scanline_atex_mask:
   INIT_ATEX()

    movl POLYSEG_TEXTURE(%esi), %eax
    movl POLYSEG_ZBLINE(%esi), %esi
	movl %eax, REWRITE_UP_texture_atex_z_mask-4

   .align 4, 0x90
atex_z_mask_loop:

   movl %edx, %eax               /* get v */
   movb VSHIFT, %cl
   sarl %cl, %eax
   andl VMASK, %eax

   movl %ebx, %ecx               /* get u */
   sarl $16, %ecx
   andl UMASK, %ecx
   addl %ecx, %eax

	UPDATE_Z_F()

	movl ONEOVERZ, %ecx
	cmpl %ecx, (%esi)  /* If existing_value >= ONEOVERZ: skip */
	jge atex_z_mask_skip

   movb REWRITE_DUMMY(%eax), %al        /* read texel */
	REWRITE_UP_texture_atex_z_mask:

   orb %al, %al			/* Is this a zero pixel? */
   jz atex_z_mask_skip

   movb %al, %fs:(%edi)          /* write solid pixels */
	movl %ecx, (%esi)	/* Record the new value in the Z_BUFFER */
   addl DU, %ebx
   addl DV, %edx
	addl $4, %esi
   incl %edi
   decl W
   jg atex_z_mask_loop
   jmp atex_z_mask_done

   .align 4, 0x90
atex_z_mask_skip:
   addl DU, %ebx                 /* skip covered and zero pixels */
   addl DV, %edx
 	addl $4, %esi
  incl %edi
   decl W
   jg atex_z_mask_loop

atex_z_mask_done:
   fstp %st(0)
   fstp %st(0)
   popl %edi
   popl %esi
   popl %ebx
   movl %ebp, %esp
   popl %ebp
   ret                           /* end of _poly_scanline_atex_mask() */




/* void _zb_poly_scanline_atex_lit(unsigned long addr, int w, POLYGON_SEGMENT *info);
 *  Fills an affine texture mapped polygon scanline.
 */
.globl __zb_poly_scanline_atex_lit
   .align 4
__zb_poly_scanline_atex_lit:
   INIT_ATEX(
      movl POLYSEG_C(%esi), %eax
      movl POLYSEG_DC(%esi), %ebx
 		movl _color_map, %edx
      sarl $8, %eax
      sarl $8, %ebx
      movl %eax, C
		movl %ebx, REWRITE_UP_dc_atex_z_lit_1-4
		movl %ebx, REWRITE_UP_dc_atex_z_lit_2-4
		movl %edx, REWRITE_UP_colourmap_atex_z_lit-4
   )
	movl POLYSEG_TEXTURE(%esi), %eax
	movl POLYSEG_ZBLINE(%esi), %esi
	movl %eax, REWRITE_UP_texture_atex_z_lit-4


   .align 4, 0x90
atex_z_lit_loop:
   movl %edx, %eax               /* get v */
   movb VSHIFT, %cl
   sarl %cl, %eax
   andl VMASK, %eax


   movl %ebx, %ecx               /* get u */
   sarl $16, %ecx
   andl UMASK, %ecx
   addl %ecx, %eax

	UPDATE_Z_F()

	movl ONEOVERZ, %ecx
	cmpl %ecx, (%esi)  /* If existing_value >= ONEOVERZ: skip */
	jge atex_z_lit_skip

   movb REWRITE_DUMMY(%eax), %al        /* read texel */
	REWRITE_UP_texture_atex_z_lit:

	movb 1+C, %ah       /* lookup in lighting table */
	andl $0x0000FFFF, %eax
	movb REWRITE_DUMMY(%eax), %al
	REWRITE_UP_colourmap_atex_z_lit:

   movb %al, %fs:(%edi)          /* write solid pixels */
	movl %ecx, (%esi)	/* Record the new value in the Z_BUFFER */

   addl $REWRITE_DUMMY, C	/* Update C with DC */
	REWRITE_UP_dc_atex_z_lit_1:
   addl DU, %ebx
   addl DV, %edx
	addl $4, %esi
   incl %edi
   decl W
   jg atex_z_lit_loop
   jmp atex_z_lit_done

   .align 4, 0x90
atex_z_lit_skip:
   addl $REWRITE_DUMMY, C	/* Update C with DC */
	REWRITE_UP_dc_atex_z_lit_2:
   addl DU, %ebx                 /* skip covered pixels */
   addl DV, %edx
 	addl $4, %esi
  incl %edi
   decl W
   jg atex_z_lit_loop

atex_z_lit_done:
   fstp %st(0)
   fstp %st(0)
   popl %edi
   popl %esi
   popl %ebx
   movl %ebp, %esp
   popl %ebp
   ret                           /* end of _zb_poly_scanline_atex_lit() */



/* void _zb_poly_scanline_atex_mask_lit(unsigned long addr, int w, POLYGON_SEGMENT *info);
 *  Fills an affine texture mapped polygon scanline.
 */
.globl __zb_poly_scanline_atex_mask_lit
   .align 4
__zb_poly_scanline_atex_mask_lit:
   INIT_ATEX(
      movl POLYSEG_C(%esi), %eax
      movl POLYSEG_DC(%esi), %ebx
 		movl _color_map, %edx
      sarl $8, %eax
      sarl $8, %ebx
      movl %eax, C
		movl %ebx, REWRITE_UP_dc_atex_z_mask_lit_1-4
		movl %ebx, REWRITE_UP_dc_atex_z_mask_lit_2-4
		movl %edx, REWRITE_UP_colourmap_atex_z_mask_lit-4
   )
	movl POLYSEG_TEXTURE(%esi), %eax
	movl POLYSEG_ZBLINE(%esi), %esi
	movl %eax, REWRITE_UP_texture_atex_z_mask_lit-4


   .align 4, 0x90
atex_z_mask_lit_loop:
   movl %edx, %eax               /* get v */
   movb VSHIFT, %cl
   sarl %cl, %eax
   andl VMASK, %eax


   movl %ebx, %ecx               /* get u */
   sarl $16, %ecx
   andl UMASK, %ecx
   addl %ecx, %eax

	UPDATE_Z_F()

	movl ONEOVERZ, %ecx
	cmpl %ecx, (%esi)  /* If existing_value >= ONEOVERZ: skip */
	jge atex_z_mask_lit_skip

   movb REWRITE_DUMMY(%eax), %al        /* read texel */
	REWRITE_UP_texture_atex_z_mask_lit:

   orb %al, %al			/* Is this a zero pixel? */
   jz atex_z_mask_lit_skip

	movb 1+C, %ah       /* lookup in lighting table */
	andl $0x0000FFFF, %eax
	movb REWRITE_DUMMY(%eax), %al
	REWRITE_UP_colourmap_atex_z_mask_lit:

   movb %al, %fs:(%edi)          /* write solid pixels */
	movl %ecx, (%esi)	/* Record the new value in the Z_BUFFER */
   addl DU, %ebx
   addl DV, %edx
   addl $REWRITE_DUMMY, C	/* Update C with DC */
	REWRITE_UP_dc_atex_z_mask_lit_1:
	addl $4, %esi
   incl %edi
   decl W
   jg atex_z_mask_lit_loop
   jmp atex_z_mask_lit_done

   .align 4, 0x90
atex_z_mask_lit_skip:
   addl DU, %ebx                 /* skip covered pixels */
   addl DV, %edx
   addl $REWRITE_DUMMY, C	/* Update C with DC */
	REWRITE_UP_dc_atex_z_mask_lit_2:
 	addl $4, %esi
  incl %edi
   decl W
   jg atex_z_mask_lit_loop

atex_z_mask_lit_done:
   fstp %st(0)
   fstp %st(0)
   popl %edi
   popl %esi
   popl %ebx
   movl %ebp, %esp
   popl %ebp
   ret                           /* end of _zb_poly_scanline_atex_mask_lit() */



#undef INIT_ATEX

#undef DU
#undef DV
#undef UMASK
#undef VMASK
#undef VSHIFT
#undef TMP
#undef COUNT
#undef C
#undef ONEOVERZ
#undef DZ1
#undef DZ4






#define TMP       -4(%ebp)
#define UMASK     -8(%ebp)
#define VMASK     -12(%ebp)
#define VSHIFT    -16(%ebp)
#define DU1       -20(%ebp)
#define DV1       -24(%ebp)
#define DZ1       -28(%ebp)
#define DU4       -32(%ebp)
#define DV4       -36(%ebp)
#define DZ4       -40(%ebp)
#define UDIFF     -44(%ebp)
#define VDIFF     -48(%ebp)
#define PREVU     -52(%ebp)
#define PREVV     -56(%ebp)
#define COUNT     -60(%ebp)
#define C         -64(%ebp)
#define ONEOVERZ_0  -68(%ebp)
#define ONEOVERZ_1  -72(%ebp)
#define ONEOVERZ_2  -76(%ebp)
#define ONEOVERZ_3  -80(%ebp)
#define DZ3		-84(%ebp)



/* helper for starting an fpu 1/z division */
#define START_FP_DIV()                                                       \
   fld1                                                                    ; \
   fdiv %st(3), %st(0)



/* helper for ending an fpu division, returning corrected u and v values */
#define END_FP_DIV(ureg, vreg)                                               \
   fld %st(0)                    /* duplicate the 1/z value */             ; \
									   ; \
   fmul %st(3), %st(0)           /* divide u by z */                       ; \
   fxch %st(1)                                                             ; \
									   ; \
   fmul %st(2), %st(0)           /* divide v by z */                       ; \
									   ; \
   fistpl TMP                    /* store v */                             ; \
   movl TMP, vreg                                                          ; \
									   ; \
   fistpl TMP                    /* store u */                             ; \
   movl TMP, ureg




/* Question: In these helpers, is it best to add DZ before or after storing Z? */
/* helper for updating the u, v, and z values */
#define UPDATE_FP_POS_1()                                                    \
   fadds DV1                   /* update v coord */                        ; \
   fxch %st(1)                   /* swap vuz stack to uvz */               ; \
									   ; \
   fadds DU1                   /* update u coord */                        ; \
   fxch %st(2)                   /* swap uvz stack to zvu */               ; \
									   ; \
   fadds DZ1                   /* update z value */                        ; \
	fsts ONEOVERZ_0                                                        ; \
   fxch %st(2)                   /* swap zvu stack to uvz */               ; \
   fxch %st(1)                   /* swap uvz stack back to vuz */

/* helper for updating the u, v, and z values. */
/* Also produces 4 Z values for the unrolled loop */
#define UPDATE_FP_POS_4()                                                    \
   fadds DV4                   /* update v coord */                        ; \
   fxch %st(1)                   /* swap vuz stack to uvz */               ; \
									   ; \
   fadds DU4                   /* update u coord */                        ; \
   fxch %st(2)                   /* swap uvz stack to zvu */               ; \
									   ; \
	fsubs DZ3								   ; \
	fsts ONEOVERZ_0                                                        ; \
   fadds DZ1                   /* update z value */                        ; \
	fsts ONEOVERZ_1                                                        ; \
   fadds DZ1                   /* update z value */                        ; \
	fsts ONEOVERZ_2                                                        ; \
   fadds DZ1                   /* update z value */                        ; \
	fsts ONEOVERZ_3                                                        ; \
	fadds DZ4								   ; \
   fxch %st(2)                   /* swap zvu stack to uvz */               ; \
   fxch %st(1)                   /* swap uvz stack back to vuz */


/* helper for updating the u, v, and z values. */
/* This is meant for the last iteration of the main loop */
#define UPDATE_FP_POS_LAST_1()                                               \
   fadds DV1                   /* update v coord */                        ; \
   fxch %st(1)                   /* swap vuz stack to uvz */               ; \
									   ; \
   fadds DU1                   /* update u coord */                        ; \
   fxch %st(2)                   /* swap uvz stack to zvu */               ; \
									   ; \
	fsubs DZ3								   ; \
	fsts ONEOVERZ_0                                                        ; \
   fadds DZ1                   /* update z value */                        ; \
	fsts ONEOVERZ_1                                                        ; \
   fadds DZ1                   /* update z value */                        ; \
	fsts ONEOVERZ_2                                                        ; \
   fadds DZ1                   /* update z value */                        ; \
	fsts ONEOVERZ_3                                                        ; \
	fadds DZ1								   ; \
   fxch %st(2)                   /* swap zvu stack to uvz */               ; \
   fxch %st(1)                   /* swap uvz stack back to vuz */


/* Get the correct Z value in ONEOVERZ_0 */
/* in the first iteration of the cleanup loop */
#define UPDATE_Z_FIRST_LEFTOVER()                                            \
    flds ONEOVERZ_0                                                         ; \
	fadds DZ4                                                              ; \
	fstps ONEOVERZ_0


/* main body of the perspective-correct texture mapping routine */
#define DO_ZB_PTEX(name)                                                        \
   pushl %ebp                                                              ; \
   movl %esp, %ebp                                                         ; \
   subl $84, %esp                /* 21 local variables */                  ; \
									   ; \
   pushl %ebx                                                              ; \
   pushl %esi                                                              ; \
   pushl %edi                                                              ; \
									   ; \
   movl INFO, %esi               /* load registers */                      ; \
									   ; \
   flds POLYSEG_Z(%esi)          /* z at bottom of fpu stack */            ; \
   flds POLYSEG_FU(%esi)         /* followed by u */                       ; \
   flds POLYSEG_FV(%esi)         /* followed by v */                       ; \
									   ; \
   flds POLYSEG_DFU(%esi)        /* multiply diffs by four */              ; \
   flds POLYSEG_DFV(%esi)                                                  ; \
   flds POLYSEG_DZ(%esi)                                                   ; \
   fxch %st(2)                   /* u v z */                               ; \
   fadd %st(0), %st(0)           /* 2u v z */                              ; \
   fxch %st(1)                   /* v 2u z */                              ; \
   fadd %st(0), %st(0)           /* 2v 2u z */                             ; \
   fxch %st(2)                   /* z 2u 2v */                             ; \
   fadd %st(0), %st(0)           /* 2z 2u 2v */                            ; \
   fxch %st(1)                   /* 2u 2z 2v */                            ; \
   fadd %st(0), %st(0)           /* 4u 2z 2v */                            ; \
   fxch %st(2)                   /* 2v 2z 4u */                            ; \
   fadd %st(0), %st(0)           /* 4v 2z 4u */                            ; \
   fxch %st(1)                   /* 2z 4v 4u */                            ; \
   fadd %st(0), %st(0)           /* 4z 4v 4u */                            ; \
   fxch %st(2)                   /* 4u 4v 4z */                            ; \
   fstps DU4                                                               ; \
   fstps DV4                                                               ; \
   fstps DZ4                                                               ; \
									   ; \
   START_FP_DIV()                /* prime the 1/z calculation */           ; \
									   ; \
   movl POLYSEG_DFU(%esi), %eax  /* copy diff values onto the stack */     ; \
   movl POLYSEG_DFV(%esi), %ebx                                            ; \
   movl POLYSEG_DZ(%esi), %ecx                                             ; \
   movl %eax, DU1                                                          ; \
   movl %ebx, DV1                                                          ; \
   movl %ecx, DZ1                                                          ; \
									   ; \
	/* Make DZ3 */					   ; \
	flds DZ1                                                               ; \
	fadds DZ1                                                              ; \
	fadds DZ1                                                              ; \
	fstps DZ3                                                              ; \
									   ; \
   movl POLYSEG_VSHIFT(%esi), %ecx                                         ; \
   movl POLYSEG_VMASK(%esi), %edx                                          ; \
   movl POLYSEG_UMASK(%esi), %eax                                          ; \
									   ; \
   shll %cl, %edx                /* adjust v mask and shift value */       ; \
   negl %ecx                                                               ; \
   movl %eax, UMASK                                                        ; \
   addl $16, %ecx                                                          ; \
   movl %edx, VMASK                                                        ; \
   movl %ecx, VSHIFT                                                       ; \
									   ; \
   INIT()                                                                  ; \
									   ; \
	movl POLYSEG_ZBLINE(%esi), %esi 									   ; \
   movl ADDR, %edi                                                         ; \
									   ; \
   .align 4, 0x90                                                          ; \
name##_odd_byte_loop:                                                      ; \
   testl $3, %edi                /* deal with any odd pixels */            ; \
   jz name##_no_odd_bytes                                                  ; \
									   ; \
   END_FP_DIV(%ebx, %edx)        /* get corrected u/v position */          ; \
   UPDATE_FP_POS_1()              /* update u/v/z values */                 ; \
   START_FP_DIV()                /* start up the next divide */            ; \
									   ; \
   movb VSHIFT, %cl              /* shift and mask v coordinate */         ; \
   sarl %cl, %edx                                                          ; \
   andl VMASK, %edx                                                        ; \
									   ; \
   sarl $16, %ebx                /* shift and mask u coordinate */         ; \
   andl UMASK, %ebx                                                        ; \
   addl %ebx, %edx                                                         ; \
									   ; \
   SINGLE_PIXEL(1)               /* process the pixel */                   ; \
									   ; \
   decl W                                                                  ; \
   jg name##_odd_byte_loop                                                 ; \
									   ; \
   .align 4, 0x90                                                          ; \
name##_no_odd_bytes:                                                       ; \
   movl W, %ecx                  /* inner loop expanded 4 times */         ; \
   movl %ecx, COUNT                                                        ; \
   shrl $2, %ecx                                                           ; \
   jg name##_prime_x4                                                      ; \
									   ; \
   END_FP_DIV(%ebx, %edx)        /* get corrected u/v position */          ; \
   UPDATE_FP_POS_1()              /* update u/v/z values */                 ; \
   START_FP_DIV()                /* start up the next divide */            ; \
									   ; \
   movl %ebx, PREVU              /* store initial u/v pos */               ; \
   movl %edx, PREVV                                                        ; \
									   ; \
   jmp name##_cleanup_leftovers                                            ; \
									   ; \
   .align 4, 0x90                                                          ; \
name##_prime_x4:                                                           ; \
   movl %ecx, W                                                            ; \
									   ; \
   END_FP_DIV(%ebx, %edx)        /* get corrected u/v position */          ; \
   UPDATE_FP_POS_4()              /* update u/v/z values */                ; \
   START_FP_DIV()                /* start up the next divide */            ; \
									   ; \
   movl %ebx, PREVU              /* store initial u/v pos */               ; \
   movl %edx, PREVV                                                        ; \
									   ; \
   jmp name##_start_loop                                                   ; \
									   ; \
   .align 4, 0x90                                                          ; \
name##_last_time:                                                          ; \
   END_FP_DIV(%eax, %ecx)        /* get corrected u/v position */          ; \
   UPDATE_FP_POS_LAST_1()              /* update u/v/z values */           ; \
   jmp name##_loop_contents                                                ; \
									   ; \
   .align 4, 0x90                                                          ; \
name##_loop:                                                               ; \
   END_FP_DIV(%eax, %ecx)        /* get corrected u/v position */          ; \
   UPDATE_FP_POS_4()              /* update u/v/z values */                ; \
									   ; \
name##_loop_contents:                                                      ; \
   START_FP_DIV()                /* start up the next divide */            ; \
									   ; \
   movl PREVU, %ebx              /* read the previous u/v pos */           ; \
   movl PREVV, %edx                                                        ; \
									   ; \
   movl %eax, PREVU              /* store u/v for the next cycle */        ; \
   movl %ecx, PREVV                                                        ; \
									   ; \
   subl %ebx, %eax               /* get u/v diffs for the four pixels */   ; \
   subl %edx, %ecx                                                         ; \
   sarl $2, %eax                                                           ; \
   sarl $2, %ecx                                                           ; \
   movl %eax, UDIFF                                                        ; \
   movl %ecx, VDIFF                                                        ; \
									   ; \
   PREPARE_FOUR_PIXELS()                                                   ; \
									   ; \
   /* pixel 1 */                                                           ; \
   movl %edx, %eax               /* shift and mask v coordinate */         ; \
   movb VSHIFT, %cl                                                        ; \
   sarl %cl, %eax                                                          ; \
   andl VMASK, %eax                                                        ; \
									   ; \
   movl %ebx, %ecx               /* shift and mask u coordinate */         ; \
   sarl $16, %ecx                                                          ; \
   andl UMASK, %ecx                                                        ; \
   addl %ecx, %eax                                                         ; \
									   ; \
   FOUR_PIXELS(0)                /* process pixel 1 */                     ; \
									   ; \
   addl UDIFF, %ebx                                                        ; \
   addl VDIFF, %edx                                                        ; \
									   ; \
   /* pixel 2 */                                                           ; \
   movl %edx, %eax               /* shift and mask v coordinate */         ; \
   movb VSHIFT, %cl                                                        ; \
   sarl %cl, %eax                                                          ; \
   andl VMASK, %eax                                                        ; \
									   ; \
   movl %ebx, %ecx               /* shift and mask u coordinate */         ; \
   sarl $16, %ecx                                                          ; \
   andl UMASK, %ecx                                                        ; \
   addl %ecx, %eax                                                         ; \
									   ; \
   FOUR_PIXELS(1)                /* process pixel 2 */                     ; \
									   ; \
   addl UDIFF, %ebx                                                        ; \
   addl VDIFF, %edx                                                        ; \
									   ; \
   /* pixel 3 */                                                           ; \
   movl %edx, %eax               /* shift and mask v coordinate */         ; \
   movb VSHIFT, %cl                                                        ; \
   sarl %cl, %eax                                                          ; \
   andl VMASK, %eax                                                        ; \
									   ; \
   movl %ebx, %ecx               /* shift and mask u coordinate */         ; \
   sarl $16, %ecx                                                          ; \
   andl UMASK, %ecx                                                        ; \
   addl %ecx, %eax                                                         ; \
									   ; \
   FOUR_PIXELS(2)                /* process pixel 3 */                     ; \
									   ; \
   addl UDIFF, %ebx                                                        ; \
   addl VDIFF, %edx                                                        ; \
									   ; \
   /* pixel 4 */                                                           ; \
   movl %edx, %eax               /* shift and mask v coordinate */         ; \
   movb VSHIFT, %cl                                                        ; \
   sarl %cl, %eax                                                          ; \
   andl VMASK, %eax                                                        ; \
									   ; \
   movl %ebx, %ecx               /* shift and mask u coordinate */         ; \
   sarl $16, %ecx                                                          ; \
   andl UMASK, %ecx                                                        ; \
   addl %ecx, %eax                                                         ; \
									   ; \
   FOUR_PIXELS(3)                /* process pixel 4 */                     ; \
									   ; \
   WRITE_FOUR_PIXELS()           /* write four pixels at a time */         ; \
									   ; \
name##_start_loop:                                                         ; \
   decl W                                                                  ; \
   jg name##_loop                                                          ; \
   jz name##_last_time                                                     ; \
									   ; \
	UPDATE_Z_FIRST_LEFTOVER()  /* Set ONEOVERZ_0 to the correct value */   ; \
									   ; \
name##_cleanup_leftovers:                                                  ; \
   movl COUNT, %ecx              /* deal with any leftover pixels */       ; \
   andl $3, %ecx                                                           ; \
   jz name##_done                                                          ; \
									   ; \
   movl %ecx, COUNT                                                        ; \
   movl PREVU, %ebx                                                        ; \
   movl PREVV, %edx                                                        ; \
									   ; \
   .align 4, 0x90                                                          ; \
name##_cleanup_loop:                                                       ; \
   movb VSHIFT, %cl              /* shift and mask v coordinate */         ; \
   sarl %cl, %edx                                                          ; \
   andl VMASK, %edx                                                        ; \
									   ; \
   sarl $16, %ebx                /* shift and mask u coordinate */         ; \
   andl UMASK, %ebx                                                        ; \
   addl %ebx, %edx                                                         ; \
									   ; \
   SINGLE_PIXEL(2)               /* process the pixel */                   ; \
									   ; \
   decl COUNT                                                              ; \
   jz name##_done                                                          ; \
									   ; \
   END_FP_DIV(%ebx, %edx)     /* get corrected u/v position */          ; \
   UPDATE_FP_POS_1()              /* update u/v/z values */                 ; \
   START_FP_DIV()                /* start up the next divide */            ; \
   jmp name##_cleanup_loop                                                 ; \
									   ; \
name##_done:                                                               ; \
   fstp %st(0)                   /* pop fpu stack */                       ; \
   fstp %st(0)                                                             ; \
   fstp %st(0)                                                             ; \
   fstp %st(0)                                                             ; \
									   ; \
   popl %edi                                                               ; \
   popl %esi                                                               ; \
   popl %ebx                                                               ; \
   movl %ebp, %esp                                                         ; \
   popl %ebp





/* void _zb_poly_scanline_ptex(ulong addr, int w, POLYGON_SEGMENT *info);
 *  Fills a perspective correct texture mapped polygon scanline.
 */
.globl __zb_poly_scanline_ptex
   .align 4
__zb_poly_scanline_ptex:

   #define INIT()                                                            \
      movl POLYSEG_TEXTURE(%esi), %ecx                                     ; \
		movl %ecx, REWRITE_UP_ptex_z_single_pixel_1-4                     ; \
		movl %ecx, REWRITE_UP_ptex_z_single_pixel_2-4                     ; \
		movl %ecx, REWRITE_UP_ptex_z_four_pixels_0-4                      ; \
		movl %ecx, REWRITE_UP_ptex_z_four_pixels_1-4                      ; \
		movl %ecx, REWRITE_UP_ptex_z_four_pixels_2-4                      ; \
		movl %ecx, REWRITE_UP_ptex_z_four_pixels_3-4

   #define SINGLE_PIXEL(n)                                                   \
																			; \
		movl ONEOVERZ_0, %ecx                                              	; \
		cmpl %ecx, (%esi)  /* If existing_value >= ONEOVERZ: skip */	    ; \
		jge ptex_z_skip_left_##n 									    	; \
		   																	; \
		movb REWRITE_DUMMY(%edx), %al     /* read texel */                  ; \
		REWRITE_UP_ptex_z_single_pixel_##n:							    	; \
		   																	; \
      movb %al, %fs:(%edi) /* write pixel */                                ; \
		movl %ecx, (%esi)  /* Record the new value in the Z_BUFFER */       ; \
		   																	; \
   ptex_z_skip_left_##n:                                                    ; \
      incl %edi                                                             ; \
		addl $4, %esi


   #define PREPARE_FOUR_PIXELS()                                             \
      /* noop */


   #define FOUR_PIXELS(n)													\
		   												                	; \
		movl ONEOVERZ_##n, %ecx                                           	; \
		cmpl %ecx, (%esi)  /* If existing_value >= ONEOVERZ: skip */	    ; \
		jge ptex_z_skip_##n 										    	; \
		   																	; \
		movb REWRITE_DUMMY(%eax), %al     /* read texel */                  ; \
		REWRITE_UP_ptex_z_four_pixels_##n:						            ; \
		   																	; \
      movb %al, %fs:(%edi) /* write pixel */                                ; \
		movl %ecx,(%esi)	/* Record the new value in the Z_BUFFER */		; \
		   																	; \
   ptex_z_skip_##n:                                                         ; \
      incl %edi                                                             ; \
		addl $4, %esi


   #define WRITE_FOUR_PIXELS()                                               \
      /* noop */


   DO_ZB_PTEX(ptex)


   #undef INIT
   #undef SINGLE_PIXEL
   #undef PREPARE_FOUR_PIXELS
   #undef FOUR_PIXELS
   #undef WRITE_FOUR_PIXELS

   ret                           /* end of _poly_scanline_ptex() */



/* void _zb_poly_scanline_ptex_mask(ulong addr, int w, POLYGON_SEGMENT *info);
 *  Fills a masked perspective correct texture mapped polygon scanline.
 */
.globl __zb_poly_scanline_ptex_mask
   .align 4
__zb_poly_scanline_ptex_mask:

   #define INIT()                                                            \
      movl POLYSEG_TEXTURE(%esi), %ecx                                     ; \
		movl %ecx, REWRITE_UP_ptex_z_mask_single_pixel_1-4                 ; \
		movl %ecx, REWRITE_UP_ptex_z_mask_single_pixel_2-4                 ; \
		movl %ecx, REWRITE_UP_ptex_z_mask_four_pixels_0-4                  ; \
		movl %ecx, REWRITE_UP_ptex_z_mask_four_pixels_1-4                  ; \
		movl %ecx, REWRITE_UP_ptex_z_mask_four_pixels_2-4                  ; \
		movl %ecx, REWRITE_UP_ptex_z_mask_four_pixels_3-4


   #define SINGLE_PIXEL(n)                                                   \
																			; \
		movl ONEOVERZ_0, %ecx                                               ; \
		cmpl %ecx, (%esi)  /* If existing_value >= ONEOVERZ: skip */	    ; \
		jge ptex_z_mask_skip_left_##n  										; \
		   																	; \
		movb REWRITE_DUMMY(%edx), %al     /* read texel */                  ; \
		REWRITE_UP_ptex_z_mask_single_pixel_##n:					    	; \
		   																	; \
      orb %al, %al               /* Is there a zero pixel? */               ; \
      jz ptex_z_mask_skip_left_##n                                          ; \
		   																	; \
      movb %al, %fs:(%edi) /* write pixel */                                ; \
		movl %ecx, (%esi)  /* Record the new value in the Z_BUFFER */       ; \
		   																	; \
   ptex_z_mask_skip_left_##n:                                               ; \
      incl %edi                                                             ; \
		addl $4, %esi


   #define PREPARE_FOUR_PIXELS()                                             \
      /* noop */


   #define FOUR_PIXELS(n)													\
		   												                	; \
		movl ONEOVERZ_##n, %ecx                                             ; \
		cmpl %ecx, (%esi)  /* If existing_value >= ONEOVERZ: skip */	    ; \
		jge ptex_z_mask_skip_##n  											; \
		   																	; \
		movb REWRITE_DUMMY(%eax), %al     /* read texel */                  ; \
		REWRITE_UP_ptex_z_mask_four_pixels_##n:						    	; \
		   																	; \
      orb %al, %al               /* Is there a zero pixel? */               ; \
      jz ptex_z_mask_skip_##n                                               ; \
		   																	; \
      movb %al, %fs:(%edi) /* write pixel */                                ; \
		movl %ecx, (%esi)  /* Record the new value in the Z_BUFFER */		; \
		   																	; \
   ptex_z_mask_skip_##n:                                                    ; \
      incl %edi                                                             ; \
		addl $4, %esi


   #define WRITE_FOUR_PIXELS()                                               \
      /* noop */


   DO_ZB_PTEX(ptex_mask)


   #undef INIT
   #undef SINGLE_PIXEL
   #undef PREPARE_FOUR_PIXELS
   #undef FOUR_PIXELS
   #undef WRITE_FOUR_PIXELS

   ret                           /* end of _poly_scanline_ptex_mask() */




/* void _zb_poly_scanline_ptex_lit(ulong addr, int w, POLYGON_SEGMENT *info);
 *  Fills a lit perspective correct texture mapped polygon scanline.
 */
.globl __zb_poly_scanline_ptex_lit
   .align 4
__zb_poly_scanline_ptex_lit:

   #define INIT()                                                            \
      movl POLYSEG_C(%esi), %eax                                           ; \
      movl POLYSEG_DC(%esi), %ebx                                          ; \
      movl POLYSEG_TEXTURE(%esi), %ecx                                     ; \
      sarl $8, %eax                                                        ; \
      sarl $8, %ebx                                                        ; \
 		movl _color_map, %edx                                              ; \
		movl %ebx, REWRITE_UP_dc_ptex_z_lit_single_pixel_1-4          ; \
		movl %ebx, REWRITE_UP_dc_ptex_z_lit_single_pixel_2-4          ; \
		movl %ebx, REWRITE_UP_dc_ptex_z_lit_four_pixels_0-4           ; \
		movl %ebx, REWRITE_UP_dc_ptex_z_lit_four_pixels_1-4           ; \
		movl %ebx, REWRITE_UP_dc_ptex_z_lit_four_pixels_2-4           ; \
		movl %ebx, REWRITE_UP_dc_ptex_z_lit_four_pixels_3-4           ; \
		movl %ecx, REWRITE_UP_texture_ptex_z_lit_single_pixel_1-4         ; \
		movl %ecx, REWRITE_UP_texture_ptex_z_lit_single_pixel_2-4         ; \
		movl %ecx, REWRITE_UP_texture_ptex_z_lit_four_pixels_0-4          ; \
		movl %ecx, REWRITE_UP_texture_ptex_z_lit_four_pixels_1-4          ; \
		movl %ecx, REWRITE_UP_texture_ptex_z_lit_four_pixels_2-4          ; \
		movl %ecx, REWRITE_UP_texture_ptex_z_lit_four_pixels_3-4          ; \
		movl %edx, REWRITE_UP_colourmap_ptex_z_lit_single_pixel_1-4       ; \
		movl %edx, REWRITE_UP_colourmap_ptex_z_lit_single_pixel_2-4       ; \
		movl %edx, REWRITE_UP_colourmap_ptex_z_lit_four_pixels_0-4        ; \
		movl %edx, REWRITE_UP_colourmap_ptex_z_lit_four_pixels_1-4        ; \
		movl %edx, REWRITE_UP_colourmap_ptex_z_lit_four_pixels_2-4        ; \
		movl %edx, REWRITE_UP_colourmap_ptex_z_lit_four_pixels_3-4        ; \
      movl %eax, C


   #define SINGLE_PIXEL(n)                                                   \
																			; \
		movl ONEOVERZ_0, %ecx    /* Read existing value from  Z_BUFFER */	; \
		cmpl %ecx, (%esi)  /* If existing_value >= ONEOVERZ: skip */	   ; \
		jge ptex_z_lit_skip_left_##n  										; \
		   																	; \
      xorl %eax, %eax                                                      ; \
		movb REWRITE_DUMMY(%edx), %al     /* read texel */                 ; \
		REWRITE_UP_texture_ptex_z_lit_single_pixel_##n:                    ; \
									   ; \
		movb 1+C, %ah              /* lookup in lighting table */          ; \
		movb REWRITE_DUMMY(%eax), %al                                      ; \
		REWRITE_UP_colourmap_ptex_z_lit_single_pixel_##n:                  ; \
									   ; \
      movb %al, %fs:(%edi)       /* write the pixel */                     ; \
		movl %ecx,(%esi)	/* Record the new value in the Z_BUFFER */     ; \
		   		                                                           ; \
   ptex_z_lit_skip_left_##n:                                               ; \
      addl $REWRITE_DUMMY, C    	/* Update C with DC */                 ; \
		REWRITE_UP_dc_ptex_z_lit_single_pixel_##n:                         ; \
									   ; \
      incl %edi                                                            ; \
		addl $4, %esi


   #define PREPARE_FOUR_PIXELS()                                             \
      /* noop */


   #define FOUR_PIXELS(n)                                                    \
		   												                	; \
 		movl ONEOVERZ_##n, %ecx                                           	; \
		cmpl %ecx, (%esi)  /* If existing_value >= ONEOVERZ: skip */	   ; \
		jge ptex_z_lit_skip_##n  										    ; \
		   																	; \
		movb REWRITE_DUMMY(%eax), %al     /* read texel */                 ; \
		REWRITE_UP_texture_ptex_z_lit_four_pixels_##n:			           ; \
									   ; \
		movb 1+C, %ah       /* lookup in lighting table */                 ; \
		andl $0x0000FFFF, %eax                                             ; \
		movb REWRITE_DUMMY(%eax), %al                                      ; \
		REWRITE_UP_colourmap_ptex_z_lit_four_pixels_##n:                   ; \
									   ; \
      movb %al, %fs:(%edi)       /* write the pixel */                     ; \
		movl %ecx,(%esi)	/* Record the new value in the Z_BUFFER */     ; \
		   		                                                           ; \
   ptex_z_lit_skip_##n:                                                    ; \
      addl $REWRITE_DUMMY, C    	/* Update C with DC */                 ; \
		REWRITE_UP_dc_ptex_z_lit_four_pixels_##n:                          ; \
									   ; \
      incl %edi                                                            ; \
		addl $4, %esi


   #define WRITE_FOUR_PIXELS()                                               \
      /* noop */


   DO_ZB_PTEX(ptex_lit)


   #undef INIT
   #undef SINGLE_PIXEL
   #undef PREPARE_FOUR_PIXELSS
   #undef FOUR_PIXELS
   #undef WRITE_FOUR_PIXELSS

   ret                           /* end of _poly_scanline_ptex_lit() */



/* void _zb_poly_scanline_ptex_mask_lit(ulong addr, int w, POLYGON_SEGMENT *info);
 *  Fills a lit perspective correct texture mapped polygon scanline.
 */
.globl __zb_poly_scanline_ptex_mask_lit
   .align 4
__zb_poly_scanline_ptex_mask_lit:

   #define INIT()                                                            \
      movl POLYSEG_C(%esi), %eax                                           ; \
      movl POLYSEG_DC(%esi), %ebx                                          ; \
      movl POLYSEG_TEXTURE(%esi), %ecx                                     ; \
      sarl $8, %eax                                                        ; \
      sarl $8, %ebx                                                        ; \
 		movl _color_map, %edx                                              ; \
		movl %ebx, REWRITE_UP_dc_ptex_z_mask_lit_single_pixel_1-4          ; \
		movl %ebx, REWRITE_UP_dc_ptex_z_mask_lit_single_pixel_2-4          ; \
		movl %ebx, REWRITE_UP_dc_ptex_z_mask_lit_four_pixels_0-4           ; \
		movl %ebx, REWRITE_UP_dc_ptex_z_mask_lit_four_pixels_1-4           ; \
		movl %ebx, REWRITE_UP_dc_ptex_z_mask_lit_four_pixels_2-4           ; \
		movl %ebx, REWRITE_UP_dc_ptex_z_mask_lit_four_pixels_3-4           ; \
		movl %ecx, REWRITE_UP_texture_ptex_z_mask_lit_single_pixel_1-4     ; \
		movl %ecx, REWRITE_UP_texture_ptex_z_mask_lit_single_pixel_2-4     ; \
		movl %ecx, REWRITE_UP_texture_ptex_z_mask_lit_four_pixels_0-4      ; \
		movl %ecx, REWRITE_UP_texture_ptex_z_mask_lit_four_pixels_1-4      ; \
		movl %ecx, REWRITE_UP_texture_ptex_z_mask_lit_four_pixels_2-4      ; \
		movl %ecx, REWRITE_UP_texture_ptex_z_mask_lit_four_pixels_3-4      ; \
		movl %edx, REWRITE_UP_colourmap_ptex_z_mask_lit_single_pixel_1-4   ; \
		movl %edx, REWRITE_UP_colourmap_ptex_z_mask_lit_single_pixel_2-4   ; \
		movl %edx, REWRITE_UP_colourmap_ptex_z_mask_lit_four_pixels_0-4    ; \
		movl %edx, REWRITE_UP_colourmap_ptex_z_mask_lit_four_pixels_1-4    ; \
		movl %edx, REWRITE_UP_colourmap_ptex_z_mask_lit_four_pixels_2-4    ; \
		movl %edx, REWRITE_UP_colourmap_ptex_z_mask_lit_four_pixels_3-4    ; \
      movl %eax, C


   #define SINGLE_PIXEL(n)                                                   \
																			; \
		movl ONEOVERZ_0, %ecx                                               ; \
		cmpl %ecx, (%esi)  /* If existing_value >= ONEOVERZ: skip */	    ; \
		jge ptex_z_mask_lit_skip_left_##n  									; \
		   																	; \
      xorl %eax, %eax                                                      ; \
		movb REWRITE_DUMMY(%edx), %al     /* read texel */                 ; \
		REWRITE_UP_texture_ptex_z_mask_lit_single_pixel_##n:                ; \
									   ; \
      orb %al, %al               /* Is there a zero pixel? */               ; \
      jz ptex_z_mask_lit_skip_left_##n                                      ; \
		   																	; \
		movb 1+C, %ah              /* lookup in lighting table */          ; \
		movb REWRITE_DUMMY(%eax), %al                                      ; \
		REWRITE_UP_colourmap_ptex_z_mask_lit_single_pixel_##n:              ; \
									   ; \
      movb %al, %fs:(%edi)       /* write the pixel */                     ; \
		movl %ecx,(%esi)	/* Record the new value in the Z_BUFFER */     ; \
		   		                                                           ; \
   ptex_z_mask_lit_skip_left_##n:                                           ; \
      addl $REWRITE_DUMMY, C    	/* Update C with DC */                 ; \
		REWRITE_UP_dc_ptex_z_mask_lit_single_pixel_##n:                     ; \
									   ; \
      incl %edi                                                             ; \
		addl $4, %esi


   #define PREPARE_FOUR_PIXELS()                                             \
      /* noop */


   #define FOUR_PIXELS(n)                                                    \
		   												                	; \
 		movl ONEOVERZ_##n, %ecx                                           	; \
		cmpl %ecx, (%esi)  /* If existing_value >= ONEOVERZ: skip */	    ; \
		jge ptex_z_mask_lit_skip_##n  										; \
		   																	; \
		movb REWRITE_DUMMY(%eax), %al     /* read texel */                 ; \
		REWRITE_UP_texture_ptex_z_mask_lit_four_pixels_##n:			        ; \
									   ; \
      orb %al, %al               /* Is there a zero pixel? */               ; \
      jz ptex_z_mask_lit_skip_##n                                           ; \
		   																	; \
		movb 1+C, %ah       /* lookup in lighting table */                 ; \
		andl $0x0000FFFF, %eax                                             ; \
		movb REWRITE_DUMMY(%eax), %al                                      ; \
		REWRITE_UP_colourmap_ptex_z_mask_lit_four_pixels_##n:              ; \
									   ; \
      movb %al, %fs:(%edi)       /* write the pixel */                     ; \
		movl %ecx,(%esi)	/* Record the new value in the Z_BUFFER */     ; \
		   		                                                           ; \
   ptex_z_mask_lit_skip_##n:                                               ; \
      addl $REWRITE_DUMMY, C    	/* Update C with DC */                 ; \
		REWRITE_UP_dc_ptex_z_mask_lit_four_pixels_##n:                     ; \
									   ; \
      incl %edi                                                            ; \
		addl $4, %esi


   #define WRITE_FOUR_PIXELS()                                               \
      /* noop */


   DO_ZB_PTEX(ptex_mask_lit)


   #undef INIT
   #undef SINGLE_PIXEL
   #undef PREPARE_FOUR_PIXELSS
   #undef FOUR_PIXELS
   #undef WRITE_FOUR_PIXELSS

   ret                           /* end of _poly_scanline_ptex_mask_lit() */


