/*
zbsprite.s
Original non-z-buffered versions by Shawn Hargreaves.
Modified by Andrei Ellman to use Z buffers.

Note: sometimes, "sprite" means the "zprite->bitmap".

NOTE: A lot of this code uses self-modifying code, which on Pentiums, is generally not a good idea, as it causes the code-cache to re-load. Instead, try to only use self-modifying code as little as possible (even once per poly is too much). Perhaps use it to hard-code the screensize and the screen address / z-buffer address.

NOTE: I have deliberately avoided using %ebp as a register to help maintain consistency with Shawn's code, but he tells me it's OK to use %ebp (the reason having something to do with a faulty assumption made about Windows).

Can I use the FPU registers to store ints?I have 23 bits of accuracy for the mantissa, so I should be able to use one to store a byte or a word.
*/

/*
For flat_trans_zprite, I could try to place the counter in a register, by freeing %edi by writing in the Z value. But we've established that SRW code is bad.

For all flat_zprites, Dont forget to skip masked pixels. This is so that Z-information for masked pixels is not recorded.

REM: Have I remembered to remove all masking code from regular zprites?

For flat_regular_zprite, does it use enough SRW code?
Trans re-writes video_rw_offset and _colourmap.
*/



#include "asmdefs.h"

.text

/* The below defines should really be in asmdef.h */
#define ZB_DAT 8
#define ZB_LINE	12


/* Dummy value for use in self-re-writing code */
#define REWRITE_DUMMY 0x12345678



#define SPRITE	ARG1
#define ZBUF	ARG2
#define ZVALUE	ARG3
#define COLOUR	ARG4

/*
_zprite_make_z(the_sprite, the_zbuffer, z, colour);
Automatically generates a flat surface for the sprite.
Given a zprite's bitmap and zbuffer, all areas not the mask colour are turned into the 1/Z value.
The zprite must be a memory bitmap, not the screen or a sub-bitmap.
This code will have to change if sub-zbuffers are implemented (ie. It would have to check zbuffer->line[y] for each line).
*/
.globl __zprite_make_z
	.align 4
__zprite_make_z:
	pushl %ebp
	movl %esp,%ebp
	pushl %esi
	pushl %ebx

	movl SPRITE, %ebx
	movl BMP_W(%ebx), %eax
	imull BMP_H(%ebx), %eax	/* %eax = w * h */
	movl ZBUF, %ecx
	movl ZB_DAT(%ecx), %ecx	/* %ecx = zbuffer->line[0] */
	movl BMP_LINE(%ebx), %esi	/* %esi = bmp->line[0] */

	/*
	Self-modifying code below:
	Hint for finding out correct offset of dummy value:
	movl label+offset, %reg_to_save (ie. the reverse of the direction you would move)
	When program crashes, examine the registers
	If %reg_to_save contains the dummy longword, you have the right offset.
	BTW: should the dummy value always be positive (ie 00000000 - 7FFFFFFF)
	Note: a 32 bit displacement or a 32 bit constant could make the
	instruction longer than the equivalent with a register value.
	Are there such instances where replacing a register that remains
	constant slows the code down? Longer instructions can slow the system
	down. Also, instructions with both a displacement and immediate data
	are not pairable on Pentiums. (ie. movl $0,DISP(%reg)).
	However, self-re-writing code should not be used when the value
	changes in the loop.
	What effect does all this have on the code-cache? (The code-cache should be updated when code is modified).
	For some reason, if I try to change %ecx into a constant, I get a
	floating-point exception.
	+ Should I have the jmp +2 to clear the pre-fetch queue, or are there enough bytes between the modidied and the modifier? Does the Pentium's branch-prediction tell the prefetch-queue what to load, or will a jmp always clear it?
	*/

	movl %esi, REWRITE_DOWN_zprite_make_z_adr+2	/* Write bitmap start address. */
	movl COLOUR, %ebx
	movl ZVALUE, %esi
	decl %eax	/* Set %eax to index the last pixel */
	movl %esi, REWRITE_UP_zprite_make_z_zvalue-4	/* Write the Z value */


.align 4,0x90
loop_zprite_make_z:
REWRITE_DOWN_zprite_make_z_adr:
	cmpb %ebx, REWRITE_DUMMY(%eax)	/* Is the pixel equal to the mask value? */
	je write0_zprite_make_z

	movl $REWRITE_DUMMY, (%ecx, %eax, 4)	/* Store Z value in Z buffer */
REWRITE_UP_zprite_make_z_zvalue:
	decl %eax
	jns loop_zprite_make_z
	jmp end_zprite_make_z

write0_zprite_make_z:
	movl $0,(%ecx, %eax, 4)	/* Store 0 in Z buffer */
	decl %eax
	jns loop_zprite_make_z

/* Cleanup */
end_zprite_make_z:
	leal -12(%ebp),%esp
	popl %ebx
	popl %esi
	movl %ebp,%esp
	popl %ebp
	ret


#undef SPRITE
#undef ZBUF
#undef ZVALUE
#undef COLOUR







#define BMP        ARG1
#define DEST_Z_B   ARG2	/* Address of destination Z_BUFFER*/
#define SPRITE     ARG3
#define SOURCE_Z_B ARG4	/* Address of Z_BUFFER for sprite's depth-map */
#define X          ARG5
#define Y          ARG6
#define COLOR      ARG7

#define TGAP   -4(%ebp)		/* How much of the top of the sprite is missing */
#define LGAP   -8(%ebp)		/* How much of the left of the sprite is missing */
#define SGAP   -12(%ebp)	/* sprite->w - W. ie. size of the rightmost part not drawn */
#define W      -16(%ebp)	/* Width of portion of sprite to be blitted */
#define H      -20(%ebp)	/* Height of portion of sprite to hbe blitted */
#define C      -24(%ebp)	/* Used as loop-counter in translucent sprite drawing routines */





/* sets up a sprite draw operation and handles the clipping */
/*
This is used by the following functions:
_linear_draw_zprite
_linear_draw_zprite_v_flip
_linear_draw_zprite_h_flip
_linear_draw_zprite_vh_flip
_linear_draw_trans_zprite
_linear_draw_lit_zprite
*/
#define START_ZPRITE_DRAW(name)                                              \
   pushl %ebp                                                              ; \
   movl %esp, %ebp                                                         ; \
   subl $28, %esp                         /* seven local variables */        ; \
									   ; \
   pushl %edi                                                              ; \
   pushl %esi                                                              ; \
   pushl %ebx                                                              ; \
   pushw %es                                                               ; \
									   ; \
									   ; \
   movl BMP, %edx                         /* edx = dest_bitmap pointer */  ; \
									   ; \
	movl SPRITE, %esi                /* esi = zprite's sprite pointer */ ; \
									   ; \
   movw BMP_SEG(%edx), %es                /* es = segment selector */      ; \
									   ; \
   cmpl $0, BMP_CLIP(%edx)                /* test bmp->clip */             ; \
   jz name##_no_clip                                                       ; \
									   ; \
   movl BMP_CT(%edx), %eax                /* bmp->ct Checking top cliping */  ; \
   subl Y, %eax                           /* eax -= y */                   ; \
   jge name##_tgap_ok                                                      ; \
   xorl %eax, %eax                                                         ; \
name##_tgap_ok:                                                            ; \
   movl %eax, TGAP                        /* set tgap */                   ; \
									   ; \
   movl BMP_H(%esi), %ebx                 /* sprite->h */                  ; \
   movl BMP_CB(%edx), %ecx                /* bmp->cb */                    ; \
   subl Y, %ecx                           /* ecx -= y */                   ; \
   cmpl %ebx, %ecx                        /* check bottom clipping */      ; \
   jg name##_height_ok                                                     ; \
   movl %ecx, %ebx                                                         ; \
name##_height_ok:                                                          ; \
   subl %eax, %ebx                        /* height -= tgap */             ; \
   jle name##_done                /* Sprite vertically offrange */         ; \
   movl %ebx, H                           /* set h */                      ; \
									   ; \
   movl BMP_CL(%edx), %eax                /* bmp->cl */                    ; \
   subl X, %eax                           /* eax -= x */                   ; \
   jge name##_lgap_ok                                                      ; \
   xorl %eax, %eax                                                         ; \
name##_lgap_ok:                                                            ; \
   movl %eax, LGAP                        /* set lgap */                   ; \
									   ; \
   movl BMP_W(%esi), %ebx                 /* sprite->w */                  ; \
   movl BMP_CR(%edx), %ecx                /* bmp->cr */                    ; \
   subl X, %ecx                           /* ecx -= x */                   ; \
   cmpl %ebx, %ecx                        /* check left clipping */        ; \
   jg name##_width_ok                     /* Width will be sprite->w */    ; \
   movl %ecx, %ebx                        /* ELSE Width will be cr-x */    ; \
name##_width_ok:                                                           ; \
   subl %eax, %ebx                        /* width -= lgap */              ; \
   jle name##_done                /* Sprite horizontally offrange */       ; \
   movl %ebx, W                           /* set w */                      ; \
									   ; \
   jmp name##_clip_done                                                    ; \
									   ; \
   .align 4, 0x90                                                          ; \
name##_no_clip:                                                            ; \
   movl $0, TGAP                                                           ; \
   movl $0, LGAP                                                           ; \
   movl BMP_W(%esi), %eax                                                  ; \
   movl %eax, W                           /* w = sprite->w */              ; \
   movl BMP_H(%esi), %eax                                                  ; \
   movl %eax, H                           /* h = sprite->h */              ; \
									   ; \
   .align 4, 0x90                                                          ; \
name##_clip_done:


/* cleans up the stack after a sprite draw operation */
#define END_ZPRITE_DRAW()                                                    \
   popw %es                                                                ; \
   popl %ebx                                                               ; \
   popl %esi                                                               ; \
   popl %edi                                                               ; \
   movl %ebp, %esp                                                         ; \
   popl %ebp



/* sets up the inner sprite drawing loop, loads registers, etc */
#define ZPRITE_LOOP(name)                                                    \
zprite_y_loop_##name:                                                      ; \
   movl BMP, %edx                     /* edx = dest_bitmap pointer */      ; \
   movl Y, %eax                           /* load line */                  ; \
	movl DEST_Z_B, %ecx          /* load &dest_zbuffer */                  ; \
	movl ZB_LINE(%ecx, %eax, 4), %ecx          /* load ZB_line[y]  */      ; \
   WRITE_BANK()                       /* select bank. USES %edx and outputs to %eax */ ; \
	movl X, %ebx                      /* x-offset */                       ; \
   addl %ebx, %eax                    /* add x offset to destbmppos */ ; \
    leal (%ecx, %ebx, 4), %edx        /* edx = ZB write adr; add x offset to destZBpos */  ; \
   movl W, %ecx                       /* x loop counter (could it be %cx instead?) */ ; \
   .align 4, 0x90                                                          ; \
zprite_x_loop_##name:



/* ends the inner (x) part of a sprite drawing loop */
#define ZPRITE_END_X(name)                                                   \
	/* Shared by flat and non-flat zprites. Same as sprites' version */	; \
   decl %ecx                                                               ; \
   jg zprite_x_loop_##name



/* ends the outer (y) part of a sprite drawing loop */
#define ZPRITE_END_Y(name)                                                   \
	movl SGAP, %ebx									; \
   addl %ebx, %esi                        /* skip sprite bytes */          ; \
	leal (%edi, %ebx, 4), %edi			/* skip zbuffer lwords */          ; \
   incl Y                                 /* next line */                  ; \
   decl H                                 /* loop counter */               ; \
   jg zprite_y_loop_##name




/* void _linear_draw_zprite(BITMAP *destbmp, Z_BUFFER *dest_zbuffer, ZPRITE *the_zprite, int x, int y);
 *  Draws a sprite onto a linear bitmap at the specified x, y position,
 *  using a masked drawing mode where zero pixels are not output.
 */
.globl __linear_draw_zprite

   /* inner loop that copies just the one byte */
   #define Z_LOOP_ONLY_ONE_BYTE                                                \
      ZPRITE_LOOP(only_one_byte)                                           ; \
		movl (%edi), %ebx           /* read new 1/z value */               ; \
		cmpl (%edx), %ebx			/* Compare with existing 1/z */	       ; \
		jl only_one_byte_skip	/* If new value < existing value, skip */ ; \
		movl %ebx, (%edx)                   /* Record new 1/z value */     ; \
      movb (%esi), %bl                       /* read pixel */              ; \
      movb %bl, %es:(%eax)                   /* write */                   ; \
   only_one_byte_skip:                                                     ; \
      incl %esi                                                            ; \
      incl %eax                                                            ; \
		addl $4, %edx                                                      ; \
		addl $4, %edi                                                      ; \
      /* no x loop */                                                      ; \
      ZPRITE_END_Y(only_one_byte)


   /* inner loop that copies just the one word */
   #define Z_LOOP_ONLY_ONE_WORD                                                \
      ZPRITE_LOOP(only_one_word)                                           ; \
		movl (%edi), %ebx           /* read new 1/z value */               ; \
		cmpl (%edx), %ebx			/* Compare with existing 1/z */	       ; \
		jl only_one_word_skip_1	/* If new value < existing value, skip */ ; \
		movl %ebx, (%edx)                   /* Record new 1/z value */     ; \
      movb (%esi), %bl                       /* read two pixels */         ; \
      movb %bl, %es:(%eax)                   /* write */                   ; \
   only_one_word_skip_1:                                                   ; \
		movl 4(%edi), %ebx           /* read new 1/z value */               ; \
		cmpl 4(%edx), %ebx			/* Compare with existing 1/z */	       ; \
		jl only_one_word_skip_2	/* If new value < existing value, skip */ ; \
		movl %ebx, 4(%edx)                   /* Record new 1/z value */     ; \
      movb 1(%esi), %bl                       /* read two pixels */         ; \
      movb %bl, %es:1(%eax)                  /* write */                   ; \
   only_one_word_skip_2:                                                   ; \
      addl $2, %esi                                                        ; \
      addl $2, %eax                                                        ; \
		addl $8, %edx                                                      ; \
		addl $8, %edi                                                      ; \
      /* no x loop */                                                      ; \
      ZPRITE_END_Y(only_one_word)


   /* inner loop that copies a word at a time, plus a leftover byte */
   #define Z_LOOP_WORDS_AND_BYTE                                               \
      ZPRITE_LOOP(zprite_words_and_byte)                                          ; \
		movl (%edi), %ebx           /* read new 1/z value */               ; \
		cmpl (%edx), %ebx			/* Compare with existing 1/z */	       ; \
		jl zprite_words_and_byte_skip_1	/* If new value < existing value, skip */ ; \
		movl %ebx, (%edx)                   /* Record new 1/z value */     ; \
      movb (%esi), %bl                       /* read pixel */              ; \
      movb %bl, %es:(%eax)                   /* write */                   ; \
   zprite_words_and_byte_skip_1:                                                  ; \
		movl 4(%edi), %ebx           /* read new 1/z value */              ; \
		cmpl 4(%edx), %ebx			/* Compare with existing 1/z */	       ; \
		jl zprite_words_and_byte_skip_2	/* If new value < existing value, skip */ ; \
		movl %ebx, 4(%edx)                   /* Record new 1/z value */     ; \
      movb 1(%esi), %bl                       /* read pixel */             ; \
      movb %bl, %es:1(%eax)                  /* write */                   ; \
   zprite_words_and_byte_skip_2:                                                  ; \
      addl $2, %esi                                                        ; \
      addl $2, %eax                                                        ; \
		addl $8, %edx                                                      ; \
		addl $8, %edi                                                      ; \
      ZPRITE_END_X(zprite_words_and_byte)           /* end of x loop */           ; \
		movl (%edi), %ebx           /* read new 1/z value */               ; \
		cmpl (%edx), %ebx			/* Compare with existing 1/z */	       ; \
		jl zprite_words_and_byte_end_skip	/* If new value < existing value, skip */ ; \
		movl %ebx, (%edx)                   /* Record new 1/z value */     ; \
      movb (%esi), %bl                       /* read pixel */              ; \
      movb %bl, %es:(%eax)                   /* write */                   ; \
   zprite_words_and_byte_end_skip:                                                ; \
      incl %esi                                                            ; \
      incl %eax                                                            ; \
		addl $4, %edx                                                      ; \
		addl $4, %edi                                                      ; \
      ZPRITE_END_Y(zprite_words_and_byte)


   /* inner loop that copies a long at a time */
   #define Z_LOOP_LONGS_ONLY                                                   \
      ZPRITE_LOOP(zprite_longs_only)                                              ; \
		movl (%edi), %ebx           /* read new 1/z value */               ; \
		cmpl (%edx), %ebx			/* Compare with existing 1/z */	       ; \
		jl zprite_longs_only_skip_1		/* If new value < existing value, skip */ ; \
		movl %ebx, (%edx)                   /* Record new 1/z value */     ; \
	  movb (%esi), %bl                      /* read pixel */               ; \
      movb %bl, %es:(%eax)                   /* write */                   ; \
   zprite_longs_only_skip_1:                                                      ; \
		movl 4(%edi), %ebx           /* read new 1/z value */              ; \
		cmpl 4(%edx), %ebx			/* Compare with existing 1/z */	       ; \
		jl zprite_longs_only_skip_2		/* If new value < existing value, skip */ ; \
		movl %ebx, 4(%edx)                   /* Record new 1/z value */    ; \
      movb 1(%esi), %bl                      /* read pixel */        ; \
      movb %bl, %es:1(%eax)                  /* write */                   ; \
   zprite_longs_only_skip_2:                                                      ; \
		movl 8(%edi), %ebx           /* read new 1/z value */              ; \
		cmpl 8(%edx), %ebx			/* Compare with existing 1/z */	       ; \
		jl zprite_longs_only_skip_3		/* If new value < existing value, skip */ ; \
		movl %ebx, 8(%edx)                   /* Record new 1/z value */    ; \
      movb 2(%esi), %bl                      /* read pixel */        ; \
      movb %bl, %es:2(%eax)                  /* write */                   ; \
   zprite_longs_only_skip_3:                                                      ; \
		movl 12(%edi), %ebx           /* read new 1/z value */             ; \
		cmpl 12(%edx), %ebx			/* Compare with existing 1/z */	       ; \
		jl zprite_longs_only_skip_4		/* If new value < existing value, skip */ ; \
		movl %ebx, 12(%edx)                   /* Record new 1/z value */   ; \
      movb 3(%esi), %bl                      /* read pixel */        ; \
      movb %bl, %es:3(%eax)                  /* write */                   ; \
   zprite_longs_only_skip_4:                                                      ; \
      addl $4, %esi                                                        ; \
      addl $4, %eax                                                        ; \
		addl $16, %edx                                                     ; \
		addl $16, %edi                                                     ; \
      ZPRITE_END_X(zprite_longs_only)                                             ; \
      /* no cleanup at end of line */                                      ; \
      ZPRITE_END_Y(zprite_longs_only)


   /* inner loop that copies a long at a time, plus a leftover word */
   #define Z_LOOP_LONGS_AND_WORD                                               \
      ZPRITE_LOOP(zprite_longs_and_word)                                          ; \
		movl (%edi), %ebx           /* read new 1/z value */               ; \
		cmpl (%edx), %ebx			/* Compare with existing 1/z */	       ; \
		jl zprite_longs_and_word_skip_1		/* If new value < existing value, skip */ ; \
		movl %ebx, (%edx)                   /* Record new 1/z value */     ; \
      movb (%esi), %bl                       /* read pixel */              ; \
      movb %bl, %es:(%eax)                   /* write */                   ; \
   zprite_longs_and_word_skip_1:                                                  ; \
		movl 4(%edi), %ebx           /* read new 1/z value */              ; \
		cmpl 4(%edx), %ebx			/* Compare with existing 1/z */	       ; \
		jl zprite_longs_and_word_skip_2		/* If new value < existing value, skip */ ; \
		movl %ebx, 4(%edx)                   /* Record new 1/z value */    ; \
      movb 1(%esi), %bl                       /* read pixel */             ; \
      movb %bl, %es:1(%eax)                  /* write */                   ; \
   zprite_longs_and_word_skip_2:                                                  ; \
		movl 8(%edi), %ebx           /* read new 1/z value */              ; \
		cmpl 8(%edx), %ebx			/* Compare with existing 1/z */	       ; \
		jl zprite_longs_and_word_skip_3		/* If new value < existing value, skip */ ; \
		movl %ebx, 8(%edx)                   /* Record new 1/z value */    ; \
      movb 2(%esi), %bl                       /* read pixel */             ; \
      movb %bl, %es:2(%eax)                  /* write */                   ; \
   zprite_longs_and_word_skip_3:                                                  ; \
		movl 12(%edi), %ebx           /* read new 1/z value */             ; \
		cmpl 12(%edx), %ebx			/* Compare with existing 1/z */	       ; \
		jl zprite_longs_and_word_skip_4		/* If new value < existing value, skip */ ; \
		movl %ebx, 12(%edx)                   /* Record new 1/z value */   ; \
      movb 3(%esi), %bl                       /* read pixel */             ; \
      movb %bl, %es:3(%eax)                  /* write */                   ; \
   zprite_longs_and_word_skip_4:                                                  ; \
      addl $4, %esi                                                        ; \
      addl $4, %eax                                                        ; \
		addl $16, %edx                                                     ; \
		addl $16, %edi                                                     ; \
      ZPRITE_END_X(zprite_longs_and_word)           /* end of x loop */           ; \
		movl (%edi), %ebx           /* read new 1/z value */               ; \
		cmpl (%edx), %ebx			/* Compare with existing 1/z */	       ; \
		jl zprite_longs_and_word_end_skip_1		/* If new value < existing value, skip */ ; \
		movl %ebx, (%edx)                   /* Record new 1/z value */     ; \
      movb (%esi), %bl                       /* read pixel */              ; \
      movb %bl, %es:(%eax)                   /* write */                   ; \
   zprite_longs_and_word_end_skip_1:                                              ; \
		movl 4(%edi), %ebx           /* read new 1/z value */              ; \
		cmpl 4(%edx), %ebx			/* Compare with existing 1/z */	       ; \
		jl zprite_longs_and_word_end_skip_2		/* If new value < existing value, skip */ ; \
		movl %ebx, 4(%edx)                   /* Record new 1/z value */    ; \
      movb 1(%esi), %bl                       /* read pixel */             ; \
      movb %bl, %es:1(%eax)                  /* write */                   ; \
   zprite_longs_and_word_end_skip_2:                                              ; \
      addl $2, %esi                                                        ; \
      addl $2, %eax                                                        ; \
		addl $8, %edx                                                      ; \
		addl $8, %edi                                                      ; \
      ZPRITE_END_Y(zprite_longs_and_word)


   /* the actual zprite drawing routine... */
   .align 4
__linear_draw_zprite:
   START_ZPRITE_DRAW(zprite)

   movl BMP_W(%esi), %eax        /* sprite->w */
	movl SOURCE_Z_B, %edi          /* load &source_zbuffer */
   subl W, %eax                  /* - w */
   movl %eax, SGAP               /* store sprite gap */

   movl LGAP, %eax
   addl %eax, X                  /* X += lgap */

   movl TGAP, %eax
   addl %eax, Y                  /* Y += tgap */

   movl BMP_LINE(%esi, %eax, 4), %esi	/* esi = start of sprite line */
	movl ZB_LINE(%edi, %eax, 4), %edi          /* load ZB_line[y]  */

	movl LGAP, %eax
   addl %eax, %esi               /* esi = source sprite data ptr */
	leal (%edi, %eax, 4), %edi	/* edi = source_zb data ptr */

   shrl $1, W                    /* halve counter for word copies */
   jz zprite_only_one_byte
   jnc zprite_even_bytes

   .align 4, 0x90
   Z_LOOP_WORDS_AND_BYTE           /* word at a time, plus leftover byte */
   jmp zprite_done

   .align 4, 0x90
zprite_even_bytes:
   shrl $1, W                    /* halve counter again, for long copies */
   jz zprite_only_one_word
   jnc zprite_even_words

   .align 4, 0x90
   Z_LOOP_LONGS_AND_WORD           /* long at a time, plus leftover word */
   jmp zprite_done

   .align 4, 0x90
zprite_even_words:
   Z_LOOP_LONGS_ONLY               /* copy a long at a time */
   jmp zprite_done

   .align 4, 0x90
zprite_only_one_byte:
   Z_LOOP_ONLY_ONE_BYTE            /* copy just the one byte */
   jmp zprite_done

   .align 4, 0x90
zprite_only_one_word:
   Z_LOOP_ONLY_ONE_WORD            /* copy just the one word */

   .align 4, 0x90
zprite_done:
   END_ZPRITE_DRAW()
   ret                           /* end of _linear_draw_zprite() */

.globl __linear_draw_zprite_end
   .align 4
__linear_draw_zprite_end:
   ret



/* void _linear_draw_zprite_v_flip(BITMAP *destbmp, Z_BUFFER *dest_zbuffer, ZPRITE *the_zprite, int x, int y);
 *  Draws a sprite to a linear bitmap, flipping vertically.
 */
.globl __linear_draw_zprite_v_flip
   .align 4
__linear_draw_zprite_v_flip:
   START_ZPRITE_DRAW(zprite_v_flip)

   movl BMP_W(%esi), %eax        /* sprite->w */
	movl SOURCE_Z_B, %edi          /* load &source_zbuffer */
   addl W, %eax                  /* + w */
   negl %eax
   movl %eax, SGAP               /* store sprite gap */

   movl LGAP, %eax
   addl %eax, X                  /* X += lgap */

   movl TGAP, %eax
   addl %eax, Y                  /* Y += tgap */

   negl %eax                     /* - tgap */
   addl BMP_H(%esi), %eax        /* + sprite->h */
   decl %eax
   movl BMP_LINE(%esi, %eax, 4), %esi
	movl ZB_LINE(%edi, %eax, 4), %edi          /* load ZB_line[y]  */

	movl LGAP, %eax
   addl %eax, %esi               /* esi = source sprite data ptr */
	leal (%edi, %eax, 4), %edi	/* edi = source_zb data ptr */

   .align 4, 0x90
   ZPRITE_LOOP(v_flip)
	movl (%edi), %ebx           /* read new 1/z value */
	cmpl (%edx), %ebx			/* Compare with existing 1/z */
	jl zprite_v_flip_skip	/* If new value < existing value, skip */
	movl %ebx, (%edx)                   /* Record new 1/z value */
   movb (%esi), %bl              /* read pixel */
   movb %bl, %es:(%eax)          /* write */
zprite_v_flip_skip:
   incl %esi
   incl %eax
	addl $4, %edi
	addl $4, %edx
   ZPRITE_END_X(v_flip)
   ZPRITE_END_Y(v_flip)

zprite_v_flip_done:
   END_ZPRITE_DRAW()
   ret                           /* end of _linear_draw_zprite_v_flip() */



/* void _linear_draw_zprite_h_flip(BITMAP *destbmp, Z_BUFFER *dest_zbuffer, ZPRITE *the_zprite, int x, int y);
 *  Draws a sprite to a linear bitmap, flipping horizontally.
 */
.globl __linear_draw_zprite_h_flip
   .align 4
__linear_draw_zprite_h_flip:
   START_ZPRITE_DRAW(zprite_h_flip)

   movl BMP_W(%esi), %eax        /* sprite->w */
	movl SOURCE_Z_B, %edi          /* load &source_zbuffer */
   addl W, %eax                  /* + w */
   movl %eax, SGAP               /* store sprite gap */

   movl LGAP, %eax
   addl %eax, X                  /* X += lgap */

   movl TGAP, %eax
   addl %eax, Y                  /* Y += tgap */

   movl BMP_W(%esi), %ecx
   movl BMP_LINE(%esi, %eax, 4), %esi
	movl ZB_LINE(%edi, %eax, 4), %edi          /* load ZB_line[y]  */
   addl %ecx, %esi
	leal (%edi, %ecx, 4), %edi
	movl LGAP, %eax
   subl %eax, %esi
	sal $2, %eax
   subl %eax, %edi
   decl %esi                     /* esi = sprite data ptr */
	subl $4, %edi                     /* edi = source_zb data ptr */

   .align 4, 0x90
   ZPRITE_LOOP(h_flip)
	movl (%edi), %ebx           /* read new 1/z value */
	cmpl (%edx), %ebx			/* Compare with existing 1/z */
	jl zprite_h_flip_skip	/* If new value < existing value, skip */
	movl %ebx, (%edx)                   /* Record new 1/z value */
   movb (%esi), %bl              /* read pixel */
   movb %bl, %es:(%eax)          /* write */
zprite_h_flip_skip:
   decl %esi
   incl %eax
	subl $4, %edi
	addl $4, %edx
   ZPRITE_END_X(h_flip)
   ZPRITE_END_Y(h_flip)

zprite_h_flip_done:
   END_ZPRITE_DRAW()
   ret                           /* end of _linear_draw_zprite_h_flip() */




/* void _linear_draw_zprite_vh_flip(BITMAP *destbmp, Z_BUFFER *dest_zbuffer, ZPRITE *the_zprite, int x, int y);
 *  Draws a sprite to a linear bitmap, flipping both vertically and horizontally.
 */
.globl __linear_draw_zprite_vh_flip
   .align 4
__linear_draw_zprite_vh_flip:
   START_ZPRITE_DRAW(zprite_vh_flip)

   movl W, %eax                  /* w */
	movl SOURCE_Z_B, %edi          /* load &source_zbuffer */
   subl BMP_W(%esi), %eax        /* - sprite->w */
   movl %eax, SGAP               /* store sprite gap */

   movl LGAP, %eax
   addl %eax, X                  /* X += lgap */

   movl TGAP, %eax
   addl %eax, Y                  /* Y += tgap */

   negl %eax                     /* - tgap */
   addl BMP_H(%esi), %eax        /* + sprite->h */
   decl %eax
   movl BMP_W(%esi), %ecx
   movl BMP_LINE(%esi, %eax, 4), %esi
	movl ZB_LINE(%edi, %eax, 4), %edi          /* load ZB_line[y]  */
   addl %ecx, %esi
	leal (%edi, %ecx, 4), %edi
	movl LGAP, %eax
   subl %eax, %esi
	sal $2, %eax
   subl %eax, %edi
   decl %esi                     /* esi = sprite data ptr */
	subl $4, %edi                     /* edi = source_zb data ptr */

   .align 4, 0x90
   ZPRITE_LOOP(vh_flip)
	movl (%edi), %ebx           /* read new 1/z value */
	cmpl (%edx), %ebx			/* Compare with existing 1/z */
	jl zprite_vh_flip_skip	/* If new value < existing value, skip */
	movl %ebx, (%edx)                   /* Record new 1/z value */
   movb (%esi), %bl              /* read pixel */
   movb %bl, %es:(%eax)          /* write */
zprite_vh_flip_skip:
   decl %esi
   incl %eax
	subl $4, %edi
	addl $4, %edx
   ZPRITE_END_X(vh_flip)
   ZPRITE_END_Y(vh_flip)

zprite_vh_flip_done:
   END_ZPRITE_DRAW()
   ret                           /* end of _linear_draw_zprite_vh_flip() */





/* sets up the inner translucent sprite drawing loop, loads registers, etc */
#define T_ZPRITE_LOOP(name)                                                  \
zprite_y_loop_##name:                                                      ; \
   movl BMP, %edx                         /* load bitmap pointer */        ; \
   movl Y, %eax                           /* load line */                  ; \
   READ_BANK()                            /* select read bank */           ; \
   movl %eax, %ebx                        /* read address in ebx */        ; \
   movl Y, %eax                           /* reload line */                ; \
	movl DEST_Z_B, %ecx          /* load &dest_zbuffer */                  ; \
	movl ZB_LINE(%ecx, %eax, 4), %ecx          /* load ZB_line[y]  */      ; \
   WRITE_BANK()                           /* select write bank */          ; \
   subl %eax, %ebx                        /* convert ebx to offset */      ; \
   movl W, %edx                           /* x loop counter */             ; \
   movl %ebx, REWRITE_UP_readoffset_##name-4          ; \
   movl %ebx, REWRITE_UP_readoffset2_##name-4          ; \
   /* Question: Is the read-write offset always going to be the same for each line? */      ; \
   /* If it is, I could do the above two re-writes outside the loop */      ; \
	movl X, %ebx                      /* x-offset */                       ; \
   movl %edx, C                           /* store x loop counter */                      ; \
   addl %ebx, %eax                           /* add x offset */               ; \
    leal (%ecx, %ebx, 4), %edx        /* edx = ZB write adr; add x offset to destZBpos */  ; \
									             ; \
   .align 4, 0x90                                                          ; \
zprite_x_loop_##name:                                                      ; \


/* ends the inner (x) part of a translucent sprite drawing loop */
#define T_ZPRITE_END_X(name)                                                 \
   decl C                                                                  ; \
   jg zprite_x_loop_##name




/* void _linear_draw_trans_zprite(BITMAP *destbmp, Z_BUFFER *dest_zbuffer, ZPRITE *the_zprite, int x, int y);
 *  Draws a translucent sprite onto a linear bitmap.
 */
.globl __linear_draw_trans_zprite


/*
eax = screen_mem_address
ebx = where data to lookup table is assembled (top word MUST be 0) and Z buffer compare
ecx = four pixels from source that get turned into dest pixels.
edx = dest Z-buffer
esi = zprite_mem_address
edi = source Z-buffer
*/

   /* inner loop that copies just the one byte */
   #define TRANS_LOOP_ONLY_ONE_BYTE                                          \
      T_ZPRITE_LOOP(trans_only_one_byte)                                   ; \
      movb %es:REWRITE_DUMMY(%eax), %cl             /* read a pixel */            ; \
      REWRITE_UP_readoffset_trans_only_one_byte:                            ; \
      REWRITE_UP_readoffset2_trans_only_one_byte:                            ; \
                                  ; \
		movl (%edi), %ebx           /* read new 1/z value */               ; \
		cmpl (%edx), %ebx			/* Compare with existing 1/z */	       ; \
		jl REWRITE_UP_colourmap1_trans_only_one_byte	/* If new < existing , skip */ ; \
		movl %ebx, (%edx)                   /* Record new 1/z value */     ; \
		xorl %ebx, %ebx             /* upper 16 bits of %ebx must be 0 */  ; \
      movb %cl, %bl                          /* lookup pixel 1 */          ; \
      movb (%esi), %bh                       /* lookup pixel */            ; \
      movb REWRITE_DUMMY(%ebx), %cl                                               ; \
      REWRITE_UP_colourmap1_trans_only_one_byte:                           ; \
                                  ; \
      movb %cl, %es:(%eax)                   /* write the pixel */         ; \
      incl %esi                                                            ; \
      incl %eax                                                            ; \
		addl $4, %edx                                                      ; \
		addl $4, %edi                                                      ; \
      /* no x loop */                                                      ; \
      ZPRITE_END_Y(trans_only_one_byte)


   /* inner loop that copies just the one word */
   #define TRANS_LOOP_ONLY_ONE_WORD                                          \
      T_ZPRITE_LOOP(trans_only_one_word)                                   ; \
      movw %es:REWRITE_DUMMY(%eax), %cx             /* read two pixels */         ; \
      REWRITE_UP_readoffset_trans_only_one_word:                            ; \
      REWRITE_UP_readoffset2_trans_only_one_word:                            ; \
                                  ; \
		movl (%edi), %ebx           /* read new 1/z value */               ; \
		cmpl (%edx), %ebx			/* Compare with existing 1/z */	       ; \
		jl REWRITE_UP_colourmap1_trans_only_one_word	/* If new < existing , skip */ ; \
		movl %ebx, (%edx)                   /* Record new 1/z value */     ; \
		xorl %ebx, %ebx             /* upper 16 bits of %ebx must be 0 */  ; \
      movb %cl, %bl                          /* lookup pixel 1 */          ; \
      movb (%esi), %bh                                                     ; \
      movb REWRITE_DUMMY(%ebx), %cl                                               ; \
      REWRITE_UP_colourmap1_trans_only_one_word:                           ; \
                                  ; \
		movl 4(%edi), %ebx           /* read new 1/z value */               ; \
		cmpl 4(%edx), %ebx			/* Compare with existing 1/z */	       ; \
		jl REWRITE_UP_colourmap2_trans_only_one_word	/* If new < existing , skip */ ; \
		movl %ebx, 4(%edx)                   /* Record new 1/z value */     ; \
		xorl %ebx, %ebx             /* upper 16 bits of %ebx must be 0 */  ; \
      movb %ch, %bl                          /* lookup pixel 2 */          ; \
      movb 1(%esi), %bh                                                     ; \
      movb REWRITE_DUMMY(%ebx), %ch                                               ; \
      REWRITE_UP_colourmap2_trans_only_one_word:                           ; \
                                  ; \
      movw %cx, %es:(%eax)                   /* write two pixels */        ; \
		addl $2, %esi                                                            ; \
      addl $2, %eax                                                        ; \
		addl $8, %edx                                                      ; \
		addl $8, %edi                                                      ; \
      /* no x loop */                                                      ; \
      ZPRITE_END_Y(trans_only_one_word)


   /* inner loop that copies a word at a time, plus a leftover byte */
   #define TRANS_LOOP_WORDS_AND_BYTE                                         \
      T_ZPRITE_LOOP(trans_words_and_byte)                                  ; \
      movw %es:REWRITE_DUMMY(%eax), %cx             /* read two pixels */         ; \
      REWRITE_UP_readoffset_trans_words_and_byte:                            ; \
                                  ; \
		movl (%edi), %ebx           /* read new 1/z value */               ; \
		cmpl (%edx), %ebx			/* Compare with existing 1/z */	       ; \
		jl REWRITE_UP_colourmap1_trans_words_and_byte	/* If new < existing , skip */ ; \
		movl %ebx, (%edx)                   /* Record new 1/z value */     ; \
		xorl %ebx, %ebx             /* upper 16 bits of %ebx must be 0 */  ; \
      movb %cl, %bl                          /* lookup pixel 1 */          ; \
      movb (%esi), %bh                                                     ; \
      movb REWRITE_DUMMY(%ebx), %cl                                               ; \
      REWRITE_UP_colourmap1_trans_words_and_byte:                            ; \
                                  ; \
		movl 4(%edi), %ebx           /* read new 1/z value */               ; \
		cmpl 4(%edx), %ebx			/* Compare with existing 1/z */	       ; \
		jl REWRITE_UP_colourmap2_trans_words_and_byte	/* If new < existing , skip */ ; \
		movl %ebx, 4(%edx)                   /* Record new 1/z value */     ; \
		xorl %ebx, %ebx             /* upper 16 bits of %ebx must be 0 */  ; \
      movb %ch, %bl                          /* lookup pixel 2 */          ; \
      movb 1(%esi), %bh                                                     ; \
      movb REWRITE_DUMMY(%ebx), %ch                                               ; \
      REWRITE_UP_colourmap2_trans_words_and_byte:                            ; \
	                              ; \
      movw %cx, %es:(%eax)                   /* write two pixels */        ; \
		addl $2, %esi                                                            ; \
      addl $2, %eax                                                        ; \
		addl $8, %edx                                                      ; \
		addl $8, %edi                                                      ; \
      T_ZPRITE_END_X(trans_words_and_byte)   /* end of x loop */           ; \
      movb %es:REWRITE_DUMMY(%eax), %cl             /* read a pixel */            ; \
      REWRITE_UP_readoffset2_trans_words_and_byte:                            ; \
                                  ; \
		movl (%edi), %ebx           /* read new 1/z value */               ; \
		cmpl (%edx), %ebx			/* Compare with existing 1/z */	       ; \
		jl REWRITE_UP_colourmap3_trans_words_and_byte	/* If new < existing , skip */ ; \
		movl %ebx, (%edx)                   /* Record new 1/z value */     ; \
		xorl %ebx, %ebx             /* upper 16 bits of %ebx must be 0 */  ; \
      movb %cl, %bl                          /* lookup pixel 2 */          ; \
      movb (%esi), %bh                       /* lookup pixel */            ; \
      movb REWRITE_DUMMY(%ebx), %cl                                               ; \
      REWRITE_UP_colourmap3_trans_words_and_byte:                            ; \
	           ; \
      movb %cl, %es:(%eax)                   /* write the pixel */         ; \
      incl %esi                                                            ; \
      incl %eax                                                            ; \
		addl $4, %edx                                                      ; \
		addl $4, %edi                                                      ; \
      ZPRITE_END_Y(trans_words_and_byte)


   /* inner loop that copies a long at a time */
   #define TRANS_LOOP_LONGS_ONLY                                             \
      T_ZPRITE_LOOP(trans_longs_only)                                      ; \
      movl %es:REWRITE_DUMMY(%eax), %ecx  /* read four pixels to memory */ ; \
      REWRITE_UP_readoffset_trans_longs_only:                            ; \
      REWRITE_UP_readoffset2_trans_longs_only:                            ; \
                                  ; \
		movl (%edi), %ebx           /* read new 1/z value */               ; \
		cmpl (%edx), %ebx			/* Compare with existing 1/z */	       ; \
		jl REWRITE_UP_colourmap1_trans_longs_only	/* If new < existing , skip */ ; \
		movl %ebx, (%edx)                   /* Record new 1/z value */     ; \
		xorl %ebx, %ebx             /* upper 16 bits of %ebx must be 0 */  ; \
      movb %cl, %bl                          /* lookup pixel 1 */          ; \
      movb (%esi), %bh                                                     ; \
      movb REWRITE_DUMMY(%ebx), %cl                                               ; \
      REWRITE_UP_colourmap1_trans_longs_only:                            ; \
                                  ; \
		movl 4(%edi), %ebx           /* read new 1/z value */               ; \
		cmpl 4(%edx), %ebx			/* Compare with existing 1/z */	       ; \
		jl REWRITE_UP_colourmap2_trans_longs_only	/* If new < existing , skip */ ; \
		movl %ebx, 4(%edx)                   /* Record new 1/z value */     ; \
		xorl %ebx, %ebx             /* upper 16 bits of %ebx must be 0 */  ; \
      movb %ch, %bl                          /* lookup pixel 2 */          ; \
      movb 1(%esi), %bh                                                     ; \
      movb REWRITE_DUMMY(%ebx), %ch                                               ; \
      REWRITE_UP_colourmap2_trans_longs_only:                            ; \
                                  ; \
      roll $16, %ecx                                                       ; \
                                  ; \
		movl 8(%edi), %ebx           /* read new 1/z value */               ; \
		cmpl 8(%edx), %ebx			/* Compare with existing 1/z */	       ; \
		jl REWRITE_UP_colourmap3_trans_longs_only	/* If new < existing , skip */ ; \
		movl %ebx, 8(%edx)                   /* Record new 1/z value */     ; \
		xorl %ebx, %ebx             /* upper 16 bits of %ebx must be 0 */  ; \
      movb %cl, %bl                          /* lookup pixel 3 */          ; \
      movb 2(%esi), %bh                                                     ; \
      movb REWRITE_DUMMY(%ebx), %cl                                               ; \
      REWRITE_UP_colourmap3_trans_longs_only:                            ; \
                                  ; \
		movl 12(%edi), %ebx           /* read new 1/z value */               ; \
		cmpl 12(%edx), %ebx			/* Compare with existing 1/z */	       ; \
		jl REWRITE_UP_colourmap4_trans_longs_only	/* If new < existing , skip */ ; \
		movl %ebx, 12(%edx)                   /* Record new 1/z value */     ; \
		xorl %ebx, %ebx             /* upper 16 bits of %ebx must be 0 */  ; \
      movb %ch, %bl                          /* lookup pixel 4 */          ; \
      movb 3(%esi), %bh                                                     ; \
      movb REWRITE_DUMMY(%ebx), %ch                                               ; \
      REWRITE_UP_colourmap4_trans_longs_only:                            ; \
                                  ; \
      roll $16, %ecx                                                       ; \
      movl %ecx, %es:(%eax)                  /* write four pixels */       ; \
		addl $4, %esi                                                            ; \
      addl $4, %eax                                                        ; \
		addl $16, %edx                                                      ; \
		addl $16, %edi                                                      ; \
      T_ZPRITE_END_X(trans_longs_only)                                     ; \
      /* no cleanup at end of line */                                      ; \
      ZPRITE_END_Y(trans_longs_only)


   /* inner loop that copies a long at a time, plus a leftover word */
   #define TRANS_LOOP_LONGS_AND_WORD                                         \
      T_ZPRITE_LOOP(trans_longs_and_word)                                  ; \
      movl %es:REWRITE_DUMMY(%eax), %ecx           /* read four pixels */        ; \
      REWRITE_UP_readoffset_trans_longs_and_word:                            ; \
                                  ; \
		movl (%edi), %ebx           /* read new 1/z value */               ; \
		cmpl (%edx), %ebx			/* Compare with existing 1/z */	       ; \
		jl REWRITE_UP_colourmap1_trans_longs_and_word	/* If new < existing , skip */ ; \
		movl %ebx, (%edx)                   /* Record new 1/z value */     ; \
		xorl %ebx, %ebx             /* upper 16 bits of %ebx must be 0 */  ; \
      movb %cl, %bl                          /* lookup pixel 1 */          ; \
      movb (%esi), %bh                                                     ; \
      movb REWRITE_DUMMY(%ebx), %cl                                               ; \
      REWRITE_UP_colourmap1_trans_longs_and_word:                            ; \
                                  ; \
		movl 4(%edi), %ebx           /* read new 1/z value */               ; \
		cmpl 4(%edx), %ebx			/* Compare with existing 1/z */	       ; \
		jl REWRITE_UP_colourmap2_trans_longs_and_word	/* If new < existing , skip */ ; \
		movl %ebx, 4(%edx)                   /* Record new 1/z value */     ; \
		xorl %ebx, %ebx             /* upper 16 bits of %ebx must be 0 */  ; \
      movb %ch, %bl                          /* lookup pixel 2 */          ; \
      movb 1(%esi), %bh                                                     ; \
      movb REWRITE_DUMMY(%ebx), %ch                                               ; \
      REWRITE_UP_colourmap2_trans_longs_and_word:                            ; \
                                  ; \
      roll $16, %ecx                                                       ; \
                                  ; \
		movl 8(%edi), %ebx           /* read new 1/z value */               ; \
		cmpl 8(%edx), %ebx			/* Compare with existing 1/z */	       ; \
		jl REWRITE_UP_colourmap3_trans_longs_and_word	/* If new < existing , skip */ ; \
		movl %ebx, 8(%edx)                   /* Record new 1/z value */     ; \
		xorl %ebx, %ebx             /* upper 16 bits of %ebx must be 0 */  ; \
      movb %cl, %bl                          /* lookup pixel 3 */          ; \
      movb 2(%esi), %bh                                                     ; \
      movb REWRITE_DUMMY(%ebx), %cl                                               ; \
      REWRITE_UP_colourmap3_trans_longs_and_word:                            ; \
                                  ; \
		movl 12(%edi), %ebx           /* read new 1/z value */               ; \
		cmpl 12(%edx), %ebx			/* Compare with existing 1/z */	       ; \
		jl REWRITE_UP_colourmap4_trans_longs_and_word	/* If new < existing , skip */ ; \
		movl %ebx, 12(%edx)                   /* Record new 1/z value */     ; \
		xorl %ebx, %ebx             /* upper 16 bits of %ebx must be 0 */  ; \
      movb %ch, %bl                          /* lookup pixel 4 */          ; \
      movb 3(%esi), %bh                                                     ; \
      movb REWRITE_DUMMY(%ebx), %ch                                               ; \
      REWRITE_UP_colourmap4_trans_longs_and_word:                            ; \
                                  ; \
      roll $16, %ecx                                                       ; \
      movl %ecx, %es:(%eax)                  /* write four pixels */       ; \
		addl $4, %esi                                                            ; \
      addl $4, %eax                                                        ; \
		addl $16, %edx                                                      ; \
		addl $16, %edi                                                      ; \
      T_ZPRITE_END_X(trans_longs_and_word)   /* end of x loop */           ; \
                                  ; \
      movw %es:REWRITE_DUMMY(%eax), %cx             /* read two pixels */         ; \
      REWRITE_UP_readoffset2_trans_longs_and_word:                            ; \
                                  ; \
		movl (%edi), %ebx           /* read new 1/z value */               ; \
		cmpl (%edx), %ebx			/* Compare with existing 1/z */	       ; \
		jl REWRITE_UP_colourmap5_trans_longs_and_word	/* If new < existing , skip */ ; \
		movl %ebx, (%edx)                   /* Record new 1/z value */     ; \
		xorl %ebx, %ebx             /* upper 16 bits of %ebx must be 0 */  ; \
      movb %cl, %bl                          /* lookup pixel 1 */          ; \
      movb (%esi), %bh                                                     ; \
      movb REWRITE_DUMMY(%ebx), %cl                                               ; \
      REWRITE_UP_colourmap5_trans_longs_and_word:                            ; \
                                  ; \
		movl 4(%edi), %ebx           /* read new 1/z value */               ; \
		cmpl 4(%edx), %ebx			/* Compare with existing 1/z */	       ; \
		jl REWRITE_UP_colourmap6_trans_longs_and_word	/* If new < existing , skip */ ; \
		movl %ebx, 4(%edx)                   /* Record new 1/z value */     ; \
		xorl %ebx, %ebx             /* upper 16 bits of %ebx must be 0 */  ; \
      movb %ch, %bl                          /* lookup pixel 2 */          ; \
      movb 1(%esi), %bh                                                     ; \
      movb REWRITE_DUMMY(%ebx), %ch                                               ; \
      REWRITE_UP_colourmap6_trans_longs_and_word:                            ; \
                                  ; \
      movw %cx, %es:(%eax)                   /* write two pixels */        ; \
		addl $2, %esi                                                            ; \
      addl $2, %eax                                                        ; \
		addl $8, %edx                                                      ; \
		addl $8, %edi                                                      ; \
      ZPRITE_END_Y(trans_longs_and_word)


   /* the actual translucent sprite drawing routine... */
   .align 4
__linear_draw_trans_zprite:
   START_ZPRITE_DRAW(trans_zprite)

	/* Re-write the code with the colour_mapping table as a displacement */
   movl _color_map, %eax
	movl %eax, REWRITE_UP_colourmap1_trans_only_one_byte-4
	movl %eax, REWRITE_UP_colourmap1_trans_only_one_word-4
	movl %eax, REWRITE_UP_colourmap2_trans_only_one_word-4
	movl %eax, REWRITE_UP_colourmap1_trans_words_and_byte-4
	movl %eax, REWRITE_UP_colourmap2_trans_words_and_byte-4
	movl %eax, REWRITE_UP_colourmap3_trans_words_and_byte-4
	movl %eax, REWRITE_UP_colourmap1_trans_longs_only-4
	movl %eax, REWRITE_UP_colourmap2_trans_longs_only-4
	movl %eax, REWRITE_UP_colourmap3_trans_longs_only-4
	movl %eax, REWRITE_UP_colourmap4_trans_longs_only-4
	movl %eax, REWRITE_UP_colourmap1_trans_longs_and_word-4
	movl %eax, REWRITE_UP_colourmap2_trans_longs_and_word-4
	movl %eax, REWRITE_UP_colourmap3_trans_longs_and_word-4
	movl %eax, REWRITE_UP_colourmap4_trans_longs_and_word-4
	movl %eax, REWRITE_UP_colourmap5_trans_longs_and_word-4
	movl %eax, REWRITE_UP_colourmap6_trans_longs_and_word-4

   movl BMP_W(%esi), %eax        /* sprite->w */
	movl SOURCE_Z_B, %edi          /* load &source_zbuffer */
   subl W, %eax                  /* - w */
   movl %eax, SGAP               /* store sprite gap */

   movl LGAP, %eax
   addl %eax, X                  /* X += lgap */

   movl TGAP, %eax
   addl %eax, Y                  /* Y += tgap */

   movl BMP_LINE(%esi, %eax, 4), %esi	/* esi = start of sprite line */
	movl ZB_LINE(%edi, %eax, 4), %edi          /* load ZB_line[y]  */

	movl LGAP, %eax
   addl %eax, %esi               /* esi = source sprite data ptr */
	leal (%edi, %eax, 4), %edi	/* edi = source_zb data ptr */

   shrl $1, W                    /* halve counter for word copies */
   jz trans_zprite_only_one_byte
   jnc trans_zprite_even_bytes

   .align 4, 0x90
   TRANS_LOOP_WORDS_AND_BYTE     /* word at a time, plus leftover byte */
   jmp trans_zprite_done

   .align 4, 0x90
trans_zprite_even_bytes:
   shrl $1, W                    /* halve counter again, for long copies */
   jz trans_zprite_only_one_word
   jnc trans_zprite_even_words

   .align 4, 0x90
   TRANS_LOOP_LONGS_AND_WORD     /* long at a time, plus leftover word */
   jmp trans_zprite_done

   .align 4, 0x90
trans_zprite_even_words:
   TRANS_LOOP_LONGS_ONLY         /* copy a long at a time */
   jmp trans_zprite_done

   .align 4, 0x90
trans_zprite_only_one_byte:
   TRANS_LOOP_ONLY_ONE_BYTE      /* copy just the one byte */
   jmp trans_zprite_done

   .align 4, 0x90
trans_zprite_only_one_word:
   TRANS_LOOP_ONLY_ONE_WORD      /* copy just the one word */

   .align 4, 0x90
trans_zprite_done:
   END_ZPRITE_DRAW()
   ret                           /* end of _linear_draw_trans_zprite() */



/* void _linear_draw_lit_zprite(BITMAP *bmp, BITMAP *sprite, int x, y, color);
 *  Draws a lit sprite onto a linear bitmap.
 */
.globl __linear_draw_lit_zprite

   .align 4
__linear_draw_lit_zprite:
   START_ZPRITE_DRAW(lit_zprite)

	/* Re-write the code with the colour_mapping table as a displacement */
   movl _color_map, %eax
	movl %eax, REWRITE_UP_colourmap1_lit_zprite-4

	movl COLOR, %eax	/* Color was passed as an int, so upper 24 bits are 0 */
	shll $8, %eax	/* Make sure colour is in the 'h' byte */
	movl %eax, REWRITE_UP_light_value_lit_zprite-4

   movl BMP_W(%esi), %eax        /* sprite->w */
	movl SOURCE_Z_B, %edi          /* load &source_zbuffer */
   subl W, %eax                  /* - w */
   movl %eax, SGAP               /* store sprite gap */

   movl LGAP, %eax
   addl %eax, X                  /* X += lgap */

   movl TGAP, %eax
   addl %eax, Y                  /* Y += tgap */

   movl BMP_LINE(%esi, %eax, 4), %esi
	movl ZB_LINE(%edi, %eax, 4), %edi          /* load ZB_line[y]  */

	movl LGAP, %eax
   addl %eax, %esi               /* esi = source sprite data ptr */
	leal (%edi, %eax, 4), %edi	/* edi = source_zb data ptr */

   .align 4, 0x90
   ZPRITE_LOOP(lit_zprite)
	movl (%edi), %ebx           /* read new 1/z value */
	cmpl (%edx), %ebx			/* Compare with existing 1/z */
	jl lit_zprite_skip	/* If new value < existing value, skip */
	movl %ebx, (%edx)                   /* Record new 1/z value */


	movl $REWRITE_DUMMY,%ebx     /* store light_value in %bh */
	REWRITE_UP_light_value_lit_zprite:

   movb (%esi), %bl              /* read pixel into low byte */
   movb REWRITE_DUMMY(%ebx), %bl        /* color table lookup */
	REWRITE_UP_colourmap1_lit_zprite:
   movb %bl, %es:(%eax)          /* write pixel */
lit_zprite_skip:
   incl %esi
   incl %eax
	addl $4, %edx
	addl $4, %edi
   ZPRITE_END_X(lit_zprite)
   ZPRITE_END_Y(lit_zprite)

lit_zprite_done:
   END_ZPRITE_DRAW()
   ret                           /* end of _linear_draw_lit_zprite() */



#undef BMP
#undef DEST_Z_B
#undef SPRITE
#undef SOURCE_Z_B
#undef X
#undef Y
#undef COLOR





/* *** ***
 * Now, we draw flat sprites that use a given Z value instead of a z-map
 *** *** */

#define BMP       ARG1
#define DEST_Z_B  ARG2	/* Address of Z_BUFFER datastructure for dest */
#define SPRITE    ARG3
#define X         ARG4
#define Y         ARG5
#define DEPTH     ARG6
#define COLOR     ARG7





/* sets up a sprite draw operation and handles the clipping */
#define START_FLAT_ZPRITE_DRAW(name)                                              \
   pushl %ebp                                                              ; \
   movl %esp, %ebp                                                         ; \
   subl $28, %esp                         /* seven local variables */        ; \
									   ; \
   pushl %edi                                                              ; \
   pushl %esi                                                              ; \
   pushl %ebx                                                              ; \
   pushw %es                                                               ; \
									   ; \
   /* %edi and %edx are used more with the trans sprites. */                   ; \
   /* Lit sprites don't use %edx during the loop */                        ; \
   /* Regular sprites don't use %edx or %edi during the loop */               ; \
   /* The colour mapping table is a good candidate for self-modifying code */ ; \
									   ; \
   movl BMP, %edx                         /* edx = dest_bitmap pointer */  ; \
	movl SPRITE, %esi                                                      ; \
									   ; \
   movw BMP_SEG(%edx), %es                /* es = segment selector */      ; \
									   ; \
   cmpl $0, BMP_CLIP(%edx)                /* test bmp->clip */             ; \
   jz name##_no_clip                                                       ; \
									   ; \
   movl BMP_CT(%edx), %eax                /* bmp->ct Checking top cliping */  ; \
   subl Y, %eax                           /* eax -= y */                   ; \
   jge name##_tgap_ok                                                      ; \
   xorl %eax, %eax                                                         ; \
name##_tgap_ok:                                                            ; \
   movl %eax, TGAP                        /* set tgap */                   ; \
									   ; \
   movl BMP_H(%esi), %ebx                 /* sprite->h */                  ; \
   movl BMP_CB(%edx), %ecx                /* bmp->cb */                    ; \
   subl Y, %ecx                           /* ecx -= y */                   ; \
   cmpl %ebx, %ecx                        /* check bottom clipping */      ; \
   jg name##_height_ok                                                     ; \
   movl %ecx, %ebx                                                         ; \
name##_height_ok:                                                          ; \
   subl %eax, %ebx                        /* height -= tgap */             ; \
   jle name##_done                /* Sprite vertically offrange */         ; \
   movl %ebx, H                           /* set h */                      ; \
									   ; \
   movl BMP_CL(%edx), %eax                /* bmp->cl */                    ; \
   subl X, %eax                           /* eax -= x */                   ; \
   jge name##_lgap_ok                                                      ; \
   xorl %eax, %eax                                                         ; \
name##_lgap_ok:                                                            ; \
   movl %eax, LGAP                        /* set lgap */                   ; \
									   ; \
   movl BMP_W(%esi), %ebx                 /* sprite->w */                  ; \
   movl BMP_CR(%edx), %ecx                /* bmp->cr */                    ; \
   subl X, %ecx                           /* ecx -= x */                   ; \
   cmpl %ebx, %ecx                        /* check left clipping */        ; \
   jg name##_width_ok                     /* Width will be sprite->w */    ; \
   movl %ecx, %ebx                        /* ELSE Width will be cr-x */    ; \
name##_width_ok:                                                           ; \
   subl %eax, %ebx                        /* width -= lgap */              ; \
   jle name##_done                /* Sprite horizontally offrange */       ; \
   movl %ebx, W                           /* set w */                      ; \
									   ; \
   jmp name##_clip_done                                                    ; \
									   ; \
   .align 4, 0x90                                                          ; \
name##_no_clip:                                                            ; \
   movl $0, TGAP                                                           ; \
   movl $0, LGAP                                                           ; \
   movl BMP_W(%esi), %eax                                                  ; \
   movl %eax, W                           /* w = sprite->w */              ; \
   movl BMP_H(%esi), %eax                                                  ; \
   movl %eax, H                           /* h = sprite->h */              ; \
									   ; \
   .align 4, 0x90                                                          ; \
name##_clip_done:



#define FLAT_ZPRITE_LOOP(name)                                                    \
zprite_y_loop_##name:                                                      ; \
   movl BMP, %edx                     /* edx = dest_bitmap pointer */      ; \
   movl Y, %eax                           /* load line */                  ; \
	movl DEST_Z_B, %ecx          /* load &dest_zbuffer */                  ; \
	movl ZB_LINE(%ecx, %eax, 4), %ecx          /* load ZB_line[y]  */      ; \
   WRITE_BANK()                       /* select bank. USES %edx and outputs to %eax */ ; \
	movl X, %ebx                      /* x-offset */                       ; \
   addl %ebx, %eax                    /* add x offset to destbmppos */ ; \
    leal (%ecx, %ebx, 4), %edx        /* edx = ZB write adr; add x offset to destZBpos */  ; \
   movl W, %ecx                       /* x loop counter (could it be %cx instead?) */ ; \
   .align 4, 0x90                                                          ; \
zprite_x_loop_##name:



/* ends the outer (y) part of a sprite drawing loop */
#define FLAT_ZPRITE_END_Y(name)                                                   \
	/* At the mo', it's identical to SPRITE_END_Y */				; \
   addl SGAP, %esi                        /* skip sprite bytes */          ; \
   incl Y                                 /* next line */                  ; \
   decl H                                 /* loop counter */               ; \
   jg zprite_y_loop_##name



/* void _linear_draw_flat_zprite(BITMAP *destbmp, Z_BUFFER *dest_zbuffer, ZPRITE *the_zprite, int x, int y);
 *  Draws a sprite onto a linear bitmap at the specified x, y position,
 *  using a masked drawing mode where zero pixels are not output.
 */
.globl __linear_draw_flat_zprite

   /* inner loop that copies just the one byte */
   #define FLAT_Z_LOOP_ONLY_ONE_BYTE                                                \
      FLAT_ZPRITE_LOOP(flat_zprite_only_one_byte)                                           ; \
		cmpl (%edx), %edi			/* Compare with existing 1/z */	       ; \
		jl flat_zprite_only_one_byte_skip	/* If new value < existing value, skip */ ; \
      movb (%esi), %bl                       /* read pixel */              ; \
		orb %bl, %bl							; \
		jz flat_zprite_only_one_word_skip_1	/* Skip masked pixels */		; \
		movl %edi, (%edx)                   /* Record new 1/z value */     ; \
      movb %bl, %es:(%eax)                   /* write */                   ; \
   flat_zprite_only_one_byte_skip:                                                     ; \
      incl %esi                                                            ; \
      incl %eax                                                            ; \
		addl $4, %edx                                                      ; \
      /* no x loop */                                                      ; \
      FLAT_ZPRITE_END_Y(flat_zprite_only_one_byte)


   /* inner loop that copies just the one word */
   #define FLAT_Z_LOOP_ONLY_ONE_WORD                                                \
      FLAT_ZPRITE_LOOP(flat_zprite_only_one_word)                                           ; \
		cmpl (%edx), %edi			/* Compare with existing 1/z */	       ; \
		jl flat_zprite_only_one_word_skip_1	/* If new value < existing value, skip */ ; \
      movb (%esi), %bl                       /* read two pixels */         ; \
		orb %bl, %bl							; \
		jz flat_zprite_only_one_word_skip_1	/* Skip masked pixels */		; \
		movl %edi, (%edx)                   /* Record new 1/z value */     ; \
      movb %bl, %es:(%eax)                   /* write */                   ; \
   flat_zprite_only_one_word_skip_1:                                                   ; \
		cmpl 4(%edx), %edi			/* Compare with existing 1/z */	       ; \
		jl flat_zprite_only_one_word_skip_2	/* If new value < existing value, skip */ ; \
      movb 1(%esi), %bl                       /* read two pixels */         ; \
		orb %bl, %bl							; \
		jz flat_zprite_only_one_word_skip_2	/* Skip masked pixels */		; \
		movl %edi, 4(%edx)                   /* Record new 1/z value */     ; \
      movb %bl, %es:1(%eax)                  /* write */                   ; \
   flat_zprite_only_one_word_skip_2:                                                   ; \
      addl $2, %esi                                                        ; \
      addl $2, %eax                                                        ; \
		addl $8, %edx                                                      ; \
      /* no x loop */                                                      ; \
      FLAT_ZPRITE_END_Y(flat_zprite_only_one_word)


   /* inner loop that copies a word at a time, plus a leftover byte */
   #define FLAT_Z_LOOP_WORDS_AND_BYTE                                               \
      FLAT_ZPRITE_LOOP(flat_zprite_words_and_byte)                                          ; \
		cmpl (%edx), %edi			/* Compare with existing 1/z */	       ; \
		jl flat_zprite_words_and_byte_skip_1	/* If new value < existing value, skip */ ; \
      movb (%esi), %bl                       /* read pixel */              ; \
		orb %bl, %bl							; \
		jz flat_zprite_words_and_byte_skip_1	/* Skip masked pixels */		; \
		movl %edi, (%edx)                   /* Record new 1/z value */     ; \
      movb %bl, %es:(%eax)                   /* write */                   ; \
   flat_zprite_words_and_byte_skip_1:                                                  ; \
		cmpl 4(%edx), %edi			/* Compare with existing 1/z */	       ; \
		jl flat_zprite_words_and_byte_skip_2	/* If new value < existing value, skip */ ; \
      movb 1(%esi), %bl                       /* read pixel */             ; \
		orb %bl, %bl							; \
		jz flat_zprite_words_and_byte_skip_2	/* Skip masked pixels */		; \
		movl %edi, 4(%edx)                   /* Record new 1/z value */     ; \
      movb %bl, %es:1(%eax)                  /* write */                   ; \
   flat_zprite_words_and_byte_skip_2:                                                  ; \
      addl $2, %esi                                                        ; \
      addl $2, %eax                                                        ; \
		addl $8, %edx                                                      ; \
		addl $8, %edi                                                      ; \
      ZPRITE_END_X(flat_zprite_words_and_byte)           /* end of x loop */           ; \
		cmpl (%edx), %edi			/* Compare with existing 1/z */	       ; \
		jl flat_zprite_words_and_byte_end_skip	/* If new value < existing value, skip */ ; \
      movb (%esi), %bl                       /* read pixel */              ; \
		orb %bl, %bl							; \
		jz flat_zprite_words_and_byte_end_skip	/* Skip masked pixels */		; \
		movl %edi, (%edx)                   /* Record new 1/z value */     ; \
      movb %bl, %es:(%eax)                   /* write */                   ; \
   flat_zprite_words_and_byte_end_skip:                                                ; \
      incl %esi                                                            ; \
      incl %eax                                                            ; \
		addl $4, %edx                                                      ; \
      FLAT_ZPRITE_END_Y(flat_zprite_words_and_byte)


   /* inner loop that copies a long at a time */
   #define FLAT_Z_LOOP_LONGS_ONLY                                                   \
      FLAT_ZPRITE_LOOP(flat_zprite_longs_only)                                              ; \
		cmpl (%edx), %edi			/* Compare with existing 1/z */	       ; \
		jl flat_zprite_longs_only_skip_1		/* If new value < existing value, skip */ ; \
	  movb (%esi), %bl                      /* read pixel */               ; \
		orb %bl, %bl							; \
		jz flat_zprite_longs_only_skip_1	/* Skip masked pixels */		; \
		movl %edi, (%edx)                   /* Record new 1/z value */     ; \
      movb %bl, %es:(%eax)                   /* write */                   ; \
   flat_zprite_longs_only_skip_1:                                                      ; \
		cmpl 4(%edx), %edi			/* Compare with existing 1/z */	       ; \
		jl flat_zprite_longs_only_skip_2		/* If new value < existing value, skip */ ; \
      movb 1(%esi), %bl                      /* read pixel */        ; \
		orb %bl, %bl							; \
		jz flat_zprite_longs_only_skip_2	/* Skip masked pixels */		; \
		movl %edi, 4(%edx)                   /* Record new 1/z value */    ; \
      movb %bl, %es:1(%eax)                  /* write */                   ; \
   flat_zprite_longs_only_skip_2:                                                      ; \
		cmpl 8(%edx), %edi			/* Compare with existing 1/z */	       ; \
		jl flat_zprite_longs_only_skip_3		/* If new value < existing value, skip */ ; \
      movb 2(%esi), %bl                      /* read pixel */        ; \
		orb %bl, %bl							; \
		jz flat_zprite_longs_only_skip_3	/* Skip masked pixels */		; \
		movl %edi, 8(%edx)                   /* Record new 1/z value */    ; \
      movb %bl, %es:2(%eax)                  /* write */                   ; \
   flat_zprite_longs_only_skip_3:                                                      ; \
		cmpl 12(%edx), %edi			/* Compare with existing 1/z */	       ; \
		jl flat_zprite_longs_only_skip_4		/* If new value < existing value, skip */ ; \
      movb 3(%esi), %bl                      /* read pixel */        ; \
		orb %bl, %bl							; \
		jz flat_zprite_longs_only_skip_4	/* Skip masked pixels */		; \
		movl %edi, 12(%edx)                   /* Record new 1/z value */   ; \
      movb %bl, %es:3(%eax)                  /* write */                   ; \
   flat_zprite_longs_only_skip_4:                                                      ; \
      addl $4, %esi                                                        ; \
      addl $4, %eax                                                        ; \
		addl $16, %edx                                                     ; \
      ZPRITE_END_X(flat_zprite_longs_only)                                             ; \
      /* no cleanup at end of line */                                      ; \
      FLAT_ZPRITE_END_Y(flat_zprite_longs_only)


   /* inner loop that copies a long at a time, plus a leftover word */
   #define FLAT_Z_LOOP_LONGS_AND_WORD                                               \
      FLAT_ZPRITE_LOOP(flat_zprite_longs_and_word)                                          ; \
		cmpl (%edx), %edi			/* Compare with existing 1/z */	       ; \
		jl flat_zprite_longs_and_word_skip_1		/* If new value < existing value, skip */ ; \
      movb (%esi), %bl                       /* read pixel */              ; \
		orb %bl, %bl							; \
		jz flat_zprite_longs_and_word_skip_1	/* Skip masked pixels */		; \
		movl %edi, (%edx)                   /* Record new 1/z value */     ; \
      movb %bl, %es:(%eax)                   /* write */                   ; \
   flat_zprite_longs_and_word_skip_1:                                                  ; \
		cmpl 4(%edx), %edi			/* Compare with existing 1/z */	       ; \
		jl flat_zprite_longs_and_word_skip_2		/* If new value < existing value, skip */ ; \
      movb 1(%esi), %bl                       /* read pixel */             ; \
		orb %bl, %bl							; \
		jz flat_zprite_longs_and_word_skip_2	/* Skip masked pixels */		; \
		movl %edi, 4(%edx)                   /* Record new 1/z value */    ; \
      movb %bl, %es:1(%eax)                  /* write */                   ; \
   flat_zprite_longs_and_word_skip_2:                                                  ; \
		cmpl 8(%edx), %edi			/* Compare with existing 1/z */	       ; \
		jl flat_zprite_longs_and_word_skip_3		/* If new value < existing value, skip */ ; \
      movb 2(%esi), %bl                       /* read pixel */             ; \
		orb %bl, %bl							; \
		jz flat_zprite_longs_and_word_skip_3	/* Skip masked pixels */		; \
		movl %edi, 8(%edx)                   /* Record new 1/z value */    ; \
      movb %bl, %es:2(%eax)                  /* write */                   ; \
   flat_zprite_longs_and_word_skip_3:                                                  ; \
		cmpl 12(%edx), %edi			/* Compare with existing 1/z */	       ; \
		jl flat_zprite_longs_and_word_skip_4		/* If new value < existing value, skip */ ; \
      movb 3(%esi), %bl                       /* read pixel */             ; \
		orb %bl, %bl							; \
		jz flat_zprite_longs_and_word_skip_4	/* Skip masked pixels */		; \
		movl %edi, 12(%edx)                   /* Record new 1/z value */   ; \
      movb %bl, %es:3(%eax)                  /* write */                   ; \
   flat_zprite_longs_and_word_skip_4:                                                  ; \
      addl $4, %esi                                                        ; \
      addl $4, %eax                                                        ; \
		addl $16, %edx                                                     ; \
      ZPRITE_END_X(flat_zprite_longs_and_word)           /* end of x loop */           ; \
		cmpl (%edx), %edi			/* Compare with existing 1/z */	       ; \
		jl flat_zprite_longs_and_word_end_skip_1		/* If new value < existing value, skip */ ; \
      movb (%esi), %bl                       /* read pixel */              ; \
		orb %bl, %bl							; \
		jz flat_zprite_longs_and_word_end_skip_1	/* Skip masked pixels */		; \
		movl %edi, (%edx)                   /* Record new 1/z value */     ; \
      movb %bl, %es:(%eax)                   /* write */                   ; \
   flat_zprite_longs_and_word_end_skip_1:                                              ; \
		cmpl 4(%edx), %edi			/* Compare with existing 1/z */	       ; \
		jl flat_zprite_longs_and_word_end_skip_2		/* If new value < existing value, skip */ ; \
      movb 1(%esi), %bl                       /* read pixel */             ; \
		orb %bl, %bl							; \
		jz flat_zprite_longs_and_word_end_skip_2	/* Skip masked pixels */		; \
		movl %edi, 4(%edx)                   /* Record new 1/z value */    ; \
      movb %bl, %es:1(%eax)                  /* write */                   ; \
   flat_zprite_longs_and_word_end_skip_2:                                              ; \
      addl $2, %esi                                                        ; \
      addl $2, %eax                                                        ; \
		addl $8, %edx                                                      ; \
      FLAT_ZPRITE_END_Y(flat_zprite_longs_and_word)


   /* the actual zprite drawing routine... */
   .align 4
__linear_draw_flat_zprite:
   START_FLAT_ZPRITE_DRAW(flat_zprite)

	movl DEPTH, %edi				/* %edi = the value we are using */
	/* Turn into SRW code so %edi can be used to hold W */

   movl BMP_W(%esi), %eax        /* sprite->w */
   subl W, %eax                  /* - w */
   movl %eax, SGAP               /* store sprite gap */

   movl LGAP, %eax
   addl %eax, X                  /* X += lgap */

   movl TGAP, %eax
   addl %eax, Y                  /* Y += tgap */

   movl BMP_LINE(%esi, %eax, 4), %esi
   addl LGAP, %esi               /* esi = sprite data ptr */

   shrl $1, W                    /* halve counter for word copies */
   jz flat_zprite_only_one_byte
   jnc flat_zprite_even_bytes

   .align 4, 0x90
   FLAT_Z_LOOP_WORDS_AND_BYTE           /* word at a time, plus leftover byte */
   jmp flat_zprite_done

   .align 4, 0x90
flat_zprite_even_bytes:
   shrl $1, W                    /* halve counter again, for long copies */
   jz flat_zprite_only_one_word
   jnc flat_zprite_even_words

   .align 4, 0x90
   FLAT_Z_LOOP_LONGS_AND_WORD           /* long at a time, plus leftover word */
   jmp flat_zprite_done

   .align 4, 0x90
flat_zprite_even_words:
   FLAT_Z_LOOP_LONGS_ONLY               /* copy a long at a time */
   jmp flat_zprite_done

   .align 4, 0x90
flat_zprite_only_one_byte:
   FLAT_Z_LOOP_ONLY_ONE_BYTE            /* copy just the one byte */
   jmp flat_zprite_done

   .align 4, 0x90
flat_zprite_only_one_word:
   FLAT_Z_LOOP_ONLY_ONE_WORD            /* copy just the one word */

   .align 4, 0x90
flat_zprite_done:
   END_ZPRITE_DRAW()
   ret                           /* end of _linear_draw_zprite() */

.globl __linear_draw_flat_zprite_end
   .align 4
__linear_draw_flat_zprite_end:
   ret



/* void _linear_draw_flat_zprite_v_flip(BITMAP *destbmp, Z_BUFFER *dest_zbuffer, ZPRITE *the_flat_zprite, int x, int y);
 *  Draws a sprite to a linear bitmap, flipping vertically.
 */
.globl __linear_draw_flat_zprite_v_flip
   .align 4
__linear_draw_flat_zprite_v_flip:
   START_FLAT_ZPRITE_DRAW(flat_zprite_v_flip)

	movl DEPTH, %edi				/* %edi = the value we are using */
	/* Turn into SRW code so %edi can be used to hold W */

   movl BMP_W(%esi), %eax        /* sprite->w */
   addl W, %eax                  /* + w */
   negl %eax
   movl %eax, SGAP               /* store sprite gap */

   movl LGAP, %eax
   addl %eax, X                  /* X += lgap */

   movl TGAP, %eax
   addl %eax, Y                  /* Y += tgap */

   negl %eax                     /* - tgap */
   addl BMP_H(%esi), %eax        /* + sprite->h */
   decl %eax
   movl BMP_LINE(%esi, %eax, 4), %esi

   addl LGAP, %esi               /* esi = source sprite data ptr */

   .align 4, 0x90
   FLAT_ZPRITE_LOOP(flat_zprite_v_flip)
	cmpl (%edx), %edi			/* Compare with existing 1/z */
	jl flat_zprite_v_flip_skip	/* If new value < existing value, skip */
   movb (%esi), %bl              /* read pixel */
	orb %bl, %bl							; \
	jz flat_zprite_v_flip_skip	/* Skip masked pixels */		; \
	movl %edi, (%edx)                   /* Record new 1/z value */
   movb %bl, %es:(%eax)          /* write */
flat_zprite_v_flip_skip:
   incl %esi
   incl %eax
	addl $4, %edx
   ZPRITE_END_X(flat_zprite_v_flip)
   FLAT_ZPRITE_END_Y(flat_zprite_v_flip)

flat_zprite_v_flip_done:
   END_ZPRITE_DRAW()
   ret                           /* end of _linear_draw_flat_zprite_v_flip() */


/* void _linear_draw_flat_zprite_h_flip(BITMAP *destbmp, Z_BUFFER *dest_zbuffer, ZPRITE *the_flat_zprite, int x, int y);
 *  Draws a sprite to a linear bitmap, flipping horizontally.
 */
.globl __linear_draw_flat_zprite_h_flip
   .align 4
__linear_draw_flat_zprite_h_flip:
   START_FLAT_ZPRITE_DRAW(flat_zprite_h_flip)

	movl DEPTH, %edi				/* %edi = the value we are using */
	/* Turn into SRW code so %edi can be used to hold W */

   movl BMP_W(%esi), %eax        /* sprite->w */
   addl W, %eax                  /* + w */
   movl %eax, SGAP               /* store sprite gap */

   movl LGAP, %eax
   addl %eax, X                  /* X += lgap */

   movl TGAP, %eax
   addl %eax, Y                  /* Y += tgap */

   movl BMP_W(%esi), %ecx
   movl BMP_LINE(%esi, %eax, 4), %esi
   addl %ecx, %esi

   subl LGAP, %esi
   decl %esi                     /* esi = sprite data ptr */

   .align 4, 0x90
   FLAT_ZPRITE_LOOP(flat_zprite_h_flip)
	cmpl (%edx), %edi			/* Compare with existing 1/z */
	jl flat_zprite_h_flip_skip	/* If new value < existing value, skip */
   movb (%esi), %bl              /* read pixel */
	orb %bl, %bl							; \
	jz flat_zprite_h_flip_skip	/* Skip masked pixels */		; \
	movl %edi, (%edx)                   /* Record new 1/z value */
   movb %bl, %es:(%eax)          /* write */
flat_zprite_h_flip_skip:
   decl %esi
   incl %eax
	addl $4, %edx
   ZPRITE_END_X(flat_zprite_h_flip)
   FLAT_ZPRITE_END_Y(flat_zprite_h_flip)

flat_zprite_h_flip_done:
   END_ZPRITE_DRAW()
   ret                           /* end of _linear_draw_flat_zprite_h_flip() */




/* void _linear_draw_flat_zprite_vh_flip(BITMAP *destbmp, Z_BUFFER *dest_zbuffer, ZPRITE *the_flat_zprite, int x, int y);
 *  Draws a sprite to a linear bitmap, flipping both vertically and horizontally.
 */
.globl __linear_draw_flat_zprite_vh_flip
   .align 4
__linear_draw_flat_zprite_vh_flip:
   START_FLAT_ZPRITE_DRAW(flat_zprite_vh_flip)

   movl W, %eax                  /* w */

	movl DEPTH, %edi				/* %edi = the value we are using */
	/* Turn into SRW code so %edi can be used to hold W */

   subl BMP_W(%esi), %eax        /* - sprite->w */
   movl %eax, SGAP               /* store sprite gap */

   movl LGAP, %eax
   addl %eax, X                  /* X += lgap */

   movl TGAP, %eax
   addl %eax, Y                  /* Y += tgap */

   negl %eax                     /* - tgap */
   addl BMP_H(%esi), %eax        /* + sprite->h */
   decl %eax
   movl BMP_W(%esi), %ecx
   movl BMP_LINE(%esi, %eax, 4), %esi
   addl %ecx, %esi

   subl LGAP, %esi
   decl %esi                     /* esi = sprite data ptr */

   .align 4, 0x90
   ZPRITE_LOOP(flat_zprite_vh_flip)
	cmpl (%edx), %edi			/* Compare with existing 1/z */
	jl flat_zprite_vh_flip_skip	/* If new value < existing value, skip */
   movb (%esi), %bl              /* read pixel */
	orb %bl, %bl							; \
	jz flat_zprite_vh_flip_skip	/* Skip masked pixels */		; \
	movl %edi, (%edx)                   /* Record new 1/z value */
   movb %bl, %es:(%eax)          /* write */
flat_zprite_vh_flip_skip:
   decl %esi
   incl %eax
	addl $4, %edx
   ZPRITE_END_X(flat_zprite_vh_flip)
   FLAT_ZPRITE_END_Y(flat_zprite_vh_flip)

flat_zprite_vh_flip_done:
   END_ZPRITE_DRAW()
   ret                           /* end of _linear_draw_flat_zprite_vh_flip() */






/* sets up the inner translucent sprite drawing loop, loads registers, etc */
#define FLAT_T_ZPRITE_LOOP(name)                                                  \
zprite_y_loop_##name:                                                      ; \
   movl BMP, %edx                         /* load bitmap pointer */        ; \
   movl Y, %eax                           /* load line */                  ; \
   READ_BANK()                            /* select read bank */           ; \
   movl %eax, %ebx                        /* read address in ebx */        ; \
   movl Y, %eax                           /* reload line */                ; \
	movl DEST_Z_B, %ecx          /* load &dest_zbuffer */                  ; \
	movl ZB_LINE(%ecx, %eax, 4), %ecx          /* load ZB_line[y]  */      ; \
   WRITE_BANK()                           /* select write bank */          ; \
   subl %eax, %ebx                        /* convert ebx to offset */      ; \
   movl W, %edx                           /* x loop counter */             ; \
   movl %ebx, REWRITE_UP_readoffset_##name-4          ; \
   movl %ebx, REWRITE_UP_readoffset2_##name-4          ; \
   /* Question: Is the read-write offset always going to be the same for each line? */      ; \
   /* If it is, I could do the above two re-writes outside the loop */      ; \
	movl X, %ebx                      /* x-offset */                       ; \
   movl %edx, C                           /* store x loop counter */                      ; \
   addl %ebx, %eax                           /* add x offset */               ; \
    leal (%ecx, %ebx, 4), %edx        /* edx = ZB write adr; add x offset to destZBpos */  ; \
	xorl %ebx, %ebx             /* upper 16 bits of %ebx must be 0 */  ; \
									             ; \
   .align 4, 0x90                                                          ; \
zprite_x_loop_##name:                                                      ; \


/* ends the inner (x) part of a translucent sprite drawing loop */
#define FLAT_T_ZPRITE_END_X(name)                                                 \
   decl C                                                                  ; \
   jg zprite_x_loop_##name





/* void _linear_draw_flat_trans_zprite(BITMAP *destbmp, Z_BUFFER *dest_zbuffer, ZPRITE *the_zprite, int x, int y);
 *  Draws a translucent sprite onto a linear bitmap.
 */
.globl __linear_draw_flat_trans_zprite


/*
eax = screen_mem_address
ebx = where data to lookup table is assembled (top word MUST be 0) and Z buffer compare
ecx = four pixels from source that get turned into dest pixels.
edx = dest Z-buffer
esi = zprite_mem_address
edi = the Z value
*/

   /* inner loop that copies just the one byte */
   #define FLAT_TRANS_LOOP_ONLY_ONE_BYTE                                          \
      FLAT_T_ZPRITE_LOOP(flat_trans_only_one_byte)                                   ; \
      movb %es:REWRITE_DUMMY(%eax), %cl             /* read a pixel */            ; \
      REWRITE_UP_readoffset_flat_trans_only_one_byte:                            ; \
      REWRITE_UP_readoffset2_flat_trans_only_one_byte:                            ; \
                                  ; \
		cmpl (%edx), %edi			/* Compare with existing 1/z */		; \
		jl REWRITE_UP_colourmap1_flat_trans_only_one_byte	/* If new < existing , skip */ ; \
      movb (%esi), %bh                       /* lookup pixel */            ; \
		orb %bh, %bh							; \
		jz REWRITE_UP_colourmap1_flat_trans_only_one_byte	/* Skip masked pixels */		; \
		movl %edi, (%edx)                   /* Record new 1/z value */		; \
      movb %cl, %bl                          /* lookup pixel 1 */          ; \
      movb REWRITE_DUMMY(%ebx), %cl                                               ; \
      REWRITE_UP_colourmap1_flat_trans_only_one_byte:                           ; \
                                  ; \
      movb %cl, %es:(%eax)                   /* write the pixel */         ; \
      incl %esi                                                            ; \
      incl %eax                                                            ; \
		addl $4, %edx                                                      ; \
      /* no x loop */                                                      ; \
      FLAT_ZPRITE_END_Y(flat_trans_only_one_byte)


   /* inner loop that copies just the one word */
   #define FLAT_TRANS_LOOP_ONLY_ONE_WORD                                          \
      FLAT_T_ZPRITE_LOOP(flat_trans_only_one_word)                                   ; \
      movw %es:REWRITE_DUMMY(%eax), %cx             /* read two pixels */         ; \
      REWRITE_UP_readoffset_flat_trans_only_one_word:                            ; \
      REWRITE_UP_readoffset2_flat_trans_only_one_word:                            ; \
                                  ; \
		cmpl (%edx), %edi			/* Compare with existing 1/z */		; \
		jl REWRITE_UP_colourmap1_flat_trans_only_one_word	/* If new < existing , skip */ ; \
      movb (%esi), %bh                                                     ; \
		orb %bh, %bh							; \
		jz REWRITE_UP_colourmap1_flat_trans_only_one_word	/* Skip masked pixels */		; \
		movl %edi, (%edx)                   /* Record new 1/z value */		; \
      movb %cl, %bl                          /* lookup pixel 1 */          ; \
      movb REWRITE_DUMMY(%ebx), %cl                                               ; \
      REWRITE_UP_colourmap1_flat_trans_only_one_word:                           ; \
                                  ; \
		cmpl 4(%edx), %edi			/* Compare with existing 1/z */		; \
		jl REWRITE_UP_colourmap2_flat_trans_only_one_word	/* If new < existing , skip */ ; \
      movb 1(%esi), %bh                                                     ; \
		orb %bh, %bh							; \
		jz REWRITE_UP_colourmap2_flat_trans_only_one_word	/* Skip masked pixels */		; \
		movl %edi, 4(%edx)                   /* Record new 1/z value */     ; \
      movb %ch, %bl                          /* lookup pixel 2 */          ; \
      movb REWRITE_DUMMY(%ebx), %ch                                               ; \
      REWRITE_UP_colourmap2_flat_trans_only_one_word:                           ; \
                                  ; \
      movw %cx, %es:(%eax)                   /* write two pixels */        ; \
		addl $2, %esi                                                            ; \
      addl $2, %eax                                                        ; \
		addl $8, %edx                                                      ; \
      /* no x loop */                                                      ; \
      FLAT_ZPRITE_END_Y(flat_trans_only_one_word)


   /* inner loop that copies a word at a time, plus a leftover byte */
   #define FLAT_TRANS_LOOP_WORDS_AND_BYTE                                         \
      FLAT_T_ZPRITE_LOOP(flat_trans_words_and_byte)                                  ; \
      movw %es:REWRITE_DUMMY(%eax), %cx             /* read two pixels */         ; \
      REWRITE_UP_readoffset_flat_trans_words_and_byte:                            ; \
                                  ; \
		cmpl (%edx), %edi			/* Compare with existing 1/z */		; \
		jl REWRITE_UP_colourmap1_flat_trans_words_and_byte	/* If new < existing , skip */ ; \
      movb (%esi), %bh                                                     ; \
		orb %bh, %bh							; \
		jz REWRITE_UP_colourmap1_flat_trans_words_and_byte	/* Skip masked pixels */		; \
		movl %edi, (%edx)                   /* Record new 1/z value */     ; \
      movb %cl, %bl                          /* lookup pixel 1 */          ; \
      movb REWRITE_DUMMY(%ebx), %cl                                               ; \
      REWRITE_UP_colourmap1_flat_trans_words_and_byte:                            ; \
                                  ; \
		cmpl 4(%edx), %edi			/* Compare with existing 1/z */		; \
		jl REWRITE_UP_colourmap2_flat_trans_words_and_byte	/* If new < existing , skip */ ; \
      movb 1(%esi), %bh                                                     ; \
		orb %bh, %bh							; \
		jz REWRITE_UP_colourmap2_flat_trans_words_and_byte	/* Skip masked pixels */		; \
		movl %edi, 4(%edx)                   /* Record new 1/z value */     ; \
      movb %ch, %bl                          /* lookup pixel 2 */          ; \
      movb REWRITE_DUMMY(%ebx), %ch                                               ; \
      REWRITE_UP_colourmap2_flat_trans_words_and_byte:                            ; \
	                              ; \
      movw %cx, %es:(%eax)                   /* write two pixels */        ; \
		addl $2, %esi                                                            ; \
      addl $2, %eax                                                        ; \
		addl $8, %edx                                                      ; \
      FLAT_T_ZPRITE_END_X(flat_trans_words_and_byte)   /* end of x loop */           ; \
      movb %es:REWRITE_DUMMY(%eax), %cl             /* read a pixel */            ; \
      REWRITE_UP_readoffset2_flat_trans_words_and_byte:                            ; \
                                  ; \
		cmpl (%edx), %edi			/* Compare with existing 1/z */		; \
		jl REWRITE_UP_colourmap3_flat_trans_words_and_byte	/* If new < existing , skip */ ; \
      movb (%esi), %bh                       /* lookup pixel */            ; \
		orb %bh, %bh							; \
		jz REWRITE_UP_colourmap3_flat_trans_words_and_byte	/* Skip masked pixels */		; \
		movl %edi, (%edx)                   /* Record new 1/z value */     ; \
      movb %cl, %bl                          /* lookup pixel 1 */          ; \
      movb REWRITE_DUMMY(%ebx), %cl                                               ; \
      REWRITE_UP_colourmap3_flat_trans_words_and_byte:                            ; \
	           ; \
      movb %cl, %es:(%eax)                   /* write the pixel */         ; \
      incl %esi                                                            ; \
      incl %eax                                                            ; \
		addl $4, %edx                                                      ; \
      FLAT_ZPRITE_END_Y(flat_trans_words_and_byte)


   /* inner loop that copies a long at a time */
   #define FLAT_TRANS_LOOP_LONGS_ONLY                                             \
      FLAT_T_ZPRITE_LOOP(flat_trans_longs_only)                                      ; \
      movl %es:REWRITE_DUMMY(%eax), %ecx  /* read four pixels to memory */ ; \
      REWRITE_UP_readoffset_flat_trans_longs_only:                            ; \
      REWRITE_UP_readoffset2_flat_trans_longs_only:                            ; \
                                  ; \
		cmpl (%edx), %edi			/* Compare with existing 1/z */		; \
		jl REWRITE_UP_colourmap1_flat_trans_longs_only	/* If new < existing , skip */ ; \
      movb (%esi), %bh                                                     ; \
		orb %bh, %bh							; \
		jz REWRITE_UP_colourmap1_flat_trans_longs_only	/* Skip masked pixels */		; \
		movl %edi, (%edx)                   /* Record new 1/z value */     ; \
      movb %cl, %bl                          /* lookup pixel 1 */          ; \
      movb REWRITE_DUMMY(%ebx), %cl                                               ; \
      REWRITE_UP_colourmap1_flat_trans_longs_only:                            ; \
                                  ; \
		cmpl 4(%edx), %edi			/* Compare with existing 1/z */		; \
		jl REWRITE_UP_colourmap2_flat_trans_longs_only	/* If new < existing , skip */ ; \
      movb 1(%esi), %bh                                                     ; \
		orb %bh, %bh							; \
		jz REWRITE_UP_colourmap2_flat_trans_longs_only	/* If new < existing , skip */ ; \
		movl %edi, 4(%edx)                   /* Record new 1/z value */     ; \
      movb %ch, %bl                          /* lookup pixel 2 */          ; \
      movb REWRITE_DUMMY(%ebx), %ch                                               ; \
      REWRITE_UP_colourmap2_flat_trans_longs_only:                            ; \
                                  ; \
      roll $16, %ecx                                                       ; \
                                  ; \
		cmpl 8(%edx), %edi			/* Compare with existing 1/z */		; \
		jl REWRITE_UP_colourmap3_flat_trans_longs_only	/* If new < existing , skip */ ; \
      movb 2(%esi), %bh                                                     ; \
		orb %bh, %bh							; \
		jz REWRITE_UP_colourmap3_flat_trans_longs_only	/* If new < existing , skip */ ; \
		movl %edi, 8(%edx)                   /* Record new 1/z value */     ; \
      movb %cl, %bl                          /* lookup pixel 3 */          ; \
      movb REWRITE_DUMMY(%ebx), %cl                                               ; \
      REWRITE_UP_colourmap3_flat_trans_longs_only:                            ; \
                                  ; \
		cmpl 12(%edx), %edi			/* Compare with existing 1/z */		; \
		jl REWRITE_UP_colourmap4_flat_trans_longs_only	/* If new < existing , skip */ ; \
      movb 3(%esi), %bh                                                     ; \
		orb %bh, %bh							; \
		jz REWRITE_UP_colourmap4_flat_trans_longs_only	/* If new < existing , skip */ ; \
		movl %edi, 12(%edx)                   /* Record new 1/z value */     ; \
      movb %ch, %bl                          /* lookup pixel 4 */          ; \
      movb REWRITE_DUMMY(%ebx), %ch                                               ; \
      REWRITE_UP_colourmap4_flat_trans_longs_only:                            ; \
                                  ; \
      roll $16, %ecx                                                       ; \
      movl %ecx, %es:(%eax)                  /* write four pixels */       ; \
		addl $4, %esi                                                            ; \
      addl $4, %eax                                                        ; \
		addl $16, %edx                                                      ; \
      FLAT_T_ZPRITE_END_X(flat_trans_longs_only)                                     ; \
      /* no cleanup at end of line */                                      ; \
      FLAT_ZPRITE_END_Y(flat_trans_longs_only)


   /* inner loop that copies a long at a time, plus a leftover word */
   #define FLAT_TRANS_LOOP_LONGS_AND_WORD                                         \
      FLAT_T_ZPRITE_LOOP(flat_trans_longs_and_word)                                  ; \
      movl %es:REWRITE_DUMMY(%eax), %ecx           /* read four pixels */        ; \
      REWRITE_UP_readoffset_flat_trans_longs_and_word:                            ; \
                                  ; \
		cmpl (%edx), %edi			/* Compare with existing 1/z */		; \
		jl REWRITE_UP_colourmap1_flat_trans_longs_and_word	/* If new < existing , skip */ ; \
      movb (%esi), %bh                                                     ; \
		orb %bh, %bh							; \
		jz REWRITE_UP_colourmap1_flat_trans_longs_and_word	/* If new < existing , skip */ ; \
		movl %edi, (%edx)                   /* Record new 1/z value */     ; \
      movb %cl, %bl                          /* lookup pixel 1 */          ; \
      movb REWRITE_DUMMY(%ebx), %cl                                               ; \
      REWRITE_UP_colourmap1_flat_trans_longs_and_word:                            ; \
                                  ; \
		cmpl 4(%edx), %edi			/* Compare with existing 1/z */		; \
		jl REWRITE_UP_colourmap2_flat_trans_longs_and_word	/* If new < existing , skip */ ; \
      movb 1(%esi), %bh                                                     ; \
		orb %bh, %bh							; \
		jz REWRITE_UP_colourmap2_flat_trans_longs_and_word	/* If new < existing , skip */ ; \
		movl %edi, 4(%edx)                   /* Record new 1/z value */     ; \
      movb %ch, %bl                          /* lookup pixel 2 */          ; \
      movb REWRITE_DUMMY(%ebx), %ch                                               ; \
      REWRITE_UP_colourmap2_flat_trans_longs_and_word:                            ; \
                                  ; \
      roll $16, %ecx                                                       ; \
                                  ; \
		cmpl 8(%edx), %edi			/* Compare with existing 1/z */		; \
		jl REWRITE_UP_colourmap3_flat_trans_longs_and_word	/* If new < existing , skip */ ; \
      movb 2(%esi), %bh                                                     ; \
		orb %bh, %bh							; \
		jz REWRITE_UP_colourmap3_flat_trans_longs_and_word	/* If new < existing , skip */ ; \
		movl %edi, 8(%edx)                   /* Record new 1/z value */     ; \
      movb %cl, %bl                          /* lookup pixel 3 */          ; \
      movb REWRITE_DUMMY(%ebx), %cl                                               ; \
      REWRITE_UP_colourmap3_flat_trans_longs_and_word:                            ; \
                                  ; \
		cmpl 12(%edx), %edi			/* Compare with existing 1/z */		; \
		jl REWRITE_UP_colourmap4_flat_trans_longs_and_word	/* If new < existing , skip */ ; \
      movb 3(%esi), %bh                                                     ; \
		orb %bh, %bh							; \
		jz REWRITE_UP_colourmap4_flat_trans_longs_and_word	/* If new < existing , skip */ ; \
		movl %edi, 12(%edx)                   /* Record new 1/z value */     ; \
      movb %ch, %bl                          /* lookup pixel 4 */          ; \
      movb REWRITE_DUMMY(%ebx), %ch                                               ; \
      REWRITE_UP_colourmap4_flat_trans_longs_and_word:                            ; \
                                  ; \
      roll $16, %ecx                                                       ; \
      movl %ecx, %es:(%eax)                  /* write four pixels */       ; \
		addl $4, %esi                                                            ; \
      addl $4, %eax                                                        ; \
		addl $16, %edx                                                      ; \
      FLAT_T_ZPRITE_END_X(flat_trans_longs_and_word)   /* end of x loop */           ; \
                                  ; \
      movw %es:REWRITE_DUMMY(%eax), %cx             /* read two pixels */         ; \
      REWRITE_UP_readoffset2_flat_trans_longs_and_word:                            ; \
                                  ; \
		cmpl (%edx), %edi			/* Compare with existing 1/z */		; \
		jl REWRITE_UP_colourmap5_flat_trans_longs_and_word	/* If new < existing , skip */ ; \
      movb (%esi), %bh                                                     ; \
		orb %bh, %bh							; \
		jz REWRITE_UP_colourmap5_flat_trans_longs_and_word	/* If new < existing , skip */ ; \
		movl %edi, (%edx)                   /* Record new 1/z value */     ; \
      movb %cl, %bl                          /* lookup pixel 1 */          ; \
      movb REWRITE_DUMMY(%ebx), %cl                                               ; \
      REWRITE_UP_colourmap5_flat_trans_longs_and_word:                            ; \
                                  ; \
		cmpl 4(%edx), %edi			/* Compare with existing 1/z */		; \
		jl REWRITE_UP_colourmap6_flat_trans_longs_and_word	/* If new < existing , skip */ ; \
      movb 1(%esi), %bh                                                     ; \
		orb %bh, %bh							; \
		jz REWRITE_UP_colourmap6_flat_trans_longs_and_word	/* If new < existing , skip */ ; \
		movl %edi, 4(%edx)                   /* Record new 1/z value */     ; \
      movb %ch, %bl                          /* lookup pixel 2 */          ; \
      movb REWRITE_DUMMY(%ebx), %ch                                               ; \
      REWRITE_UP_colourmap6_flat_trans_longs_and_word:                            ; \
                                  ; \
      movw %cx, %es:(%eax)                   /* write two pixels */        ; \
		addl $2, %esi                                                            ; \
      addl $2, %eax                                                        ; \
		addl $8, %edx                                                      ; \
      FLAT_ZPRITE_END_Y(flat_trans_longs_and_word)


   /* the actual translucent sprite drawing routine... */
   .align 4
__linear_draw_flat_trans_zprite:
   START_FLAT_ZPRITE_DRAW(flat_trans_zprite)

	/* Re-write the code with the colour_mapping table as a displacement */
   movl _color_map, %eax
	movl %eax, REWRITE_UP_colourmap1_flat_trans_only_one_byte-4
	movl %eax, REWRITE_UP_colourmap1_flat_trans_only_one_word-4
	movl %eax, REWRITE_UP_colourmap2_flat_trans_only_one_word-4
	movl %eax, REWRITE_UP_colourmap1_flat_trans_words_and_byte-4
	movl %eax, REWRITE_UP_colourmap2_flat_trans_words_and_byte-4
	movl %eax, REWRITE_UP_colourmap3_flat_trans_words_and_byte-4
	movl %eax, REWRITE_UP_colourmap1_flat_trans_longs_only-4
	movl %eax, REWRITE_UP_colourmap2_flat_trans_longs_only-4
	movl %eax, REWRITE_UP_colourmap3_flat_trans_longs_only-4
	movl %eax, REWRITE_UP_colourmap4_flat_trans_longs_only-4
	movl %eax, REWRITE_UP_colourmap1_flat_trans_longs_and_word-4
	movl %eax, REWRITE_UP_colourmap2_flat_trans_longs_and_word-4
	movl %eax, REWRITE_UP_colourmap3_flat_trans_longs_and_word-4
	movl %eax, REWRITE_UP_colourmap4_flat_trans_longs_and_word-4
	movl %eax, REWRITE_UP_colourmap5_flat_trans_longs_and_word-4
	movl %eax, REWRITE_UP_colourmap6_flat_trans_longs_and_word-4

	movl DEPTH, %edi				/* %edi = the value we are using */
	/* Turn into SRW code so %edi can be used to hold W */

   movl BMP_W(%esi), %eax        /* sprite->w */
   subl W, %eax                  /* - w */
   movl %eax, SGAP               /* store sprite gap */

   movl LGAP, %eax
   addl %eax, X                  /* X += lgap */

   movl TGAP, %eax
   addl %eax, Y                  /* Y += tgap */

   movl BMP_LINE(%esi, %eax, 4), %esi
   addl LGAP, %esi               /* esi = sprite data ptr */

   shrl $1, W                    /* halve counter for word copies */
   jz flat_trans_zprite_only_one_byte
   jnc flat_trans_zprite_even_bytes

   .align 4, 0x90
   FLAT_TRANS_LOOP_WORDS_AND_BYTE     /* word at a time, plus leftover byte */
   jmp flat_trans_zprite_done

   .align 4, 0x90
flat_trans_zprite_even_bytes:
   shrl $1, W                    /* halve counter again, for long copies */
   jz flat_trans_zprite_only_one_word
   jnc flat_trans_zprite_even_words

   .align 4, 0x90
   FLAT_TRANS_LOOP_LONGS_AND_WORD     /* long at a time, plus leftover word */
   jmp flat_trans_zprite_done

   .align 4, 0x90
flat_trans_zprite_even_words:
   FLAT_TRANS_LOOP_LONGS_ONLY         /* copy a long at a time */
   jmp flat_trans_zprite_done

   .align 4, 0x90
flat_trans_zprite_only_one_byte:
   FLAT_TRANS_LOOP_ONLY_ONE_BYTE      /* copy just the one byte */
   jmp flat_trans_zprite_done

   .align 4, 0x90
flat_trans_zprite_only_one_word:
   FLAT_TRANS_LOOP_ONLY_ONE_WORD      /* copy just the one word */

   .align 4, 0x90
flat_trans_zprite_done:
   END_ZPRITE_DRAW()
   ret                           /* end of _linear_draw_flat_trans_zprite() */







/* void _linear_draw_flat_lit_zprite(BITMAP *bmp, BITMAP *sprite, int x, y, color);
 *  Draws a lit sprite onto a linear bitmap.
 */
.globl __linear_draw_flat_lit_zprite

   .align 4
__linear_draw_flat_lit_zprite:
   START_FLAT_ZPRITE_DRAW(flat_lit_zprite)

	/* Re-write the code with the colour_mapping table as a displacement */
   movl _color_map, %eax
	movl %eax, REWRITE_UP_colourmap1_flat_lit_zprite-4

	movl COLOR, %eax	/* Color was passed as an int, so upper 24 bits are 0 */
	shll $8, %eax	/* Make sure colour is in the 'h' byte */
	movl %eax, REWRITE_UP_light_value_flat_lit_zprite-4

	movl DEPTH, %edi				/* %edi = the value we are using */

   movl BMP_W(%esi), %eax        /* sprite->w */
   subl W, %eax                  /* - w */
   movl %eax, SGAP               /* store sprite gap */

   movl LGAP, %eax
   addl %eax, X                  /* X += lgap */

   movl TGAP, %eax
   addl %eax, Y                  /* Y += tgap */

   movl BMP_LINE(%esi, %eax, 4), %esi
   addl LGAP, %esi               /* esi = sprite data ptr */


   .align 4, 0x90
   ZPRITE_LOOP(flat_lit_zprite)
	cmpl (%edx), %edi			/* Compare with existing 1/z */
	jl flat_lit_zprite_skip	/* If new value < existing value, skip */

	movl $REWRITE_DUMMY,%ebx     /* store light value in %bh */
	REWRITE_UP_light_value_flat_lit_zprite:

   movb (%esi), %bl              /* read pixel into low byte */
   orb %bl, %bl
   jz flat_lit_zprite_skip		/* Skip masked pixels */
	movl %edi, (%edx)                   /* Record new 1/z value */

   movb REWRITE_DUMMY(%ebx), %bl        /* color table lookup */
	REWRITE_UP_colourmap1_flat_lit_zprite:
   movb %bl, %es:(%eax)          /* write pixel */
flat_lit_zprite_skip:
   incl %esi
   incl %eax
	addl $4, %edx
   ZPRITE_END_X(flat_lit_zprite)
   FLAT_ZPRITE_END_Y(flat_lit_zprite)

flat_lit_zprite_done:
   END_ZPRITE_DRAW()
   ret                           /* end of _linear_draw_flat_lit_sprite() */



#undef DEST_Z_B








/* Z-buffered sprite stretchers */


#define SOURCE       ARG1
#define DEST         ARG2
#define DRAWER       ARG3
#define SX           ARG4
#define SY           ARG5
#define SYD          ARG6
#define DX           ARG7
#define DY           ARG8
#define DH           ARG9



#define SOURCE_Z_B   ARG10
#define DEST_Z_B     ARG11
#define ZVALUE       ARG12

/* void _do_zb_displace(BITMAP *source, BITMAP *dest, void *drawer, int sx, fixed sy, fixed syd, int dx, int dy, int dh, Z_BUFFER *source_zbuffer, Z_BUFFER *dest_zbuffer, float z);
 *
 *  Helper function for displace_flat_zprite() (and possibly for stretch_flat_zprite() ), calls the compiled line drawer.
 *
 * Originally written by Shawn Hargreaves
 * Modified by Andrei Ellman to handle Z-buffering.
 */
.globl __do_zb_displace

   .align 4
__do_zb_displace:
   pushl %ebp
   movl %esp, %ebp
   pushw %es
   pushl %edi
   pushl %esi
   pushl %ebx
	flds ZVALUE		/* Store the Z displacement on the FPU */

   movl DEST, %edx
   movw BMP_SEG(%edx), %es       /* load destination segment */

   .align 4, 0x90
zb_displace_y_loop:
   movl SOURCE, %edx             /* get source line (in esi) and bank */
   movl SY, %eax
   shrl $16, %eax
	movl SOURCE_Z_B, %ebx          /* load &source_zbuffer */
	movl ZB_LINE(%ebx, %eax, 4), %ebx          /* load ZB_line[y]  */
   READ_BANK()
   movl %eax, %esi
	movl SX, %eax
	addl %eax, %esi
    leal (%ebx, %eax, 4), %ebx   /* ebx = ZB read adr; add x offset to sourceZBpos */

/* Question, With sprites, is it just a question of moving the address of the start at the begining, and adding the gap each time? */

	/* Start the first calculation */
	fld %st
	flds (%ebx)
	fmul %st,%st(1)

   movl DEST, %edx               /* get dest line (in edi) and bank */
   movl DY, %eax
	movl DEST_Z_B, %ecx          /* load &dest_zbuffer */
	movl ZB_LINE(%ecx, %eax, 4), %ecx          /* load ZB_line[y]  */
   WRITE_BANK()
   movl %eax, %edi
	movl DX, %eax
	addl %eax, %edi
    leal (%ecx, %eax, 4), %ecx   /* ecx = ZB write adr; add x offset to destZBpos */

	/* Continue the first calculation */
	fld1
	faddp %st,%st(2)
	fdivp %st,%st(1)

	xorl %eax, %eax	/* For transparent zprites, upper 16 bits of %eax must be 0 */
   call *DRAWER                   /* draw the line (clobbers eax, ecx, edx) */

	/* Discard the result of the last divide */
	fstp %st(0)

   movl SYD, %eax                /* next line in source bitmap */
   addl %eax, SY
   incl DY                       /* next line in dest bitmap */
   decl DH
   jg zb_displace_y_loop

	fstp %st(0)
   popl %ebx
   popl %esi
   popl %edi
   popw %es
   movl %ebp, %esp
   popl %ebp
   ret                           /* end of _do_zb_displace() */


#undef SOURCE_Z_B
#undef DEST_Z_B
#undef ZVALUE




#define DEST_Z_B     ARG10

/* void _do_flat_zb_displace(BITMAP *source, BITMAP *dest, void *drawer, int sx, fixed sy, fixed syd, int dx, int dy, int dh, Z_BUFFER *dest_zbuffer);
 *
 *  Helper function for displace_flat_zprite() (and possibly for stretch_flat_zprite() ), calls the compiled line drawer.
 *
 * Here the Z value is written into the scanline filler, so we don't need to pass it here.
 *
 * Originally written by Shawn Hargreaves
 * Modified by Andrei Ellman to handle Z-buffering.
 */
.globl __do_flat_zb_displace

   .align 4
__do_flat_zb_displace:
   pushl %ebp
   movl %esp, %ebp
   pushw %es
   pushl %edi
   pushl %esi
   pushl %ebx

   movl DEST, %edx
   movw BMP_SEG(%edx), %es       /* load destination segment */
   movl DRAWER, %ebx             /* the actual line drawer */

   .align 4, 0x90
flat_zb_displace_y_loop:
   movl SOURCE, %edx             /* get source line (in esi) and bank */
   movl SY, %eax
   shrl $16, %eax
   READ_BANK()
   movl %eax, %esi
   addl SX, %esi

/* Question, With sprites, is it just a question of moving the address of the start at the begining, and adding the gap each time? */

   movl DEST, %edx               /* get dest line (in edi) and bank */
   movl DY, %eax
	movl DEST_Z_B, %ecx          /* load &dest_zbuffer */
	movl ZB_LINE(%ecx, %eax, 4), %ecx          /* load ZB_line[y]  */
   WRITE_BANK()
   movl %eax, %edi
	movl DX, %eax
	addl %eax, %edi
    leal (%ecx, %eax, 4), %ecx   /* ecx = ZB write adr; add x offset to destZBpos */

	xorl %eax, %eax	/* For transparent zprites, upper 16 bits of %eax must be 0 */

   call *%ebx                    /* draw the line (clobbers eax and ecx) */

   movl SYD, %eax                /* next line in source bitmap */
   addl %eax, SY
   incl DY                       /* next line in dest bitmap */
   decl DH
   jg flat_zb_displace_y_loop

   popl %ebx
   popl %esi
   popl %edi
   popw %es
   movl %ebp, %esp
   popl %ebp
   ret                           /* end of _do_flat_zb_displace() */

