/*
zbuf.c
C code for z buffers.

By Andrei Ellman
*/

/*
Can store both Z, MAX_VALUE-Z and 1/Z values or whatever you like in them.
Wether it's Z or 1/Z depends on how the routines that use the Z buffer treat the Z buffer.
The z-buffered sprites used in zprite.c rely on a 32-bit float 1/z buffer.
The z-buffered polygon functions in zbpoly.c also use 32-bit float 1/z buffers.
My z-buffered spheres demo also uses 32-bit float 1/z buffers.
(In this type of Z-buffer, the values range from infinity (Z=0 (the plane of the screen)) to 0 (Z=infinity). 1/z values were chosen to make interpolating the z-value accrosss a polygon easier (1/z values in a perspective-distorted space can be interpolated linearly). No negative Zs used because Allegro's co-ordinate system has Z increasing away from the viewer, and when the value is always positive, you can use an integer compare instruction in 80x86 assembly to compare 2 floats. (Is it better to define the Z-buffer as type "unsigned-float"?)

At the moment, this file only supports 32-bit z-buffers, but support for 16-bit z-buffers (or even 8-bit ones) may be added in the future.
Also, this could be re-done in case certain 3D cards require their own special z-buffer memory (this would probarbly involve the use of far pointer simulation functions). Perhaps I should make some preparations for this (ie. calculate sizeof(float) earlier).
*/

/*
sizeof(fixed)=sizeof(float)=4

A z-buffer can takeup large ammounts of ram (x*y*size)bytes. Perhaps have a Z buffer only for the part of the screen where drawing is done. If Z-buffers must be the same size as bitmaps (like they are now), then I could use a double-buffer smaller than the screen.
*/

/*
KNOWN BUGS AND ISSUES:

The zoomer uses an optimisation that reads the next entry of the z-buffer and starts the divide. If the next entry is a certain value, this could cause a divde by zero. To work around this one, I have added an extra entry to each Z-buffer. See the comments of the appropriate functions for more details. Also, do FPU underflows and overflows cause a crash?
*/

/*
TODO:

A more flexible z-buffer filling routine that can fill/clear an area of a z-buffer: zb_fill(), zb_clear()

Z-buffered Blit (Does not check existing Z values - just writes new ones). Then, I could do dirty rectangles. Would it be worth adding support for overlapping source and destination?
*/

/*
FUTURE:
Support many sizes of Z-buffer

Should I have Sub-Z-buffers? (similar concept to Allegro's sub-bitmaps). I may then have problems if I've already assumed that next_line[start] == this_line[end]+1.

About the extra lword at the end of each zbuffer.
This is so the optimised zoomer doesn't crash by dividing by zero under certain circumstances when zooming depth-maps (see create_zbuffer() for more info). Do I really need to have a pentium-optimisation in a 386-optimised zprite zoomer? Have two versions, a 386 version (scanline fillers and no next-divide optimisation), and a 586 version (next-divide optimisation, and done in a loop). If the 586 version does not need an extra lword, then I can get rid of the lword at the end of the zbuffer.
*/


#include <stdlib.h>			// malloc()

#include "zbuf.h"


/*
Create a new Zbuffer
*/
Z_BUFFER *create_zbuffer(int width, int height)
{
	Z_BUFFER *new_zbuffer;
	int i;

	size_t zb_data_size = sizeof(float);	// When support for other sizes is implemented, pass this as a paramater.

	/* Allocate memory for the z-buffer structure */
	new_zbuffer = malloc(sizeof(Z_BUFFER) + (sizeof(void *) * height));
	if (!new_zbuffer) {
		return NULL;
	}

	/*
	The displace_[trnas_]zprite() functions optimise the code by
	starting the next divide inside the current iteration.
	When it comes to the end of the zprite, it would read in whatever is
	stored after the zprite's depth-map. There is a 1/2^32 probability of the
	value that is read causing the routine to crash. When this value is 0.0,
	it is guaranteed not to crash (1-0.0*K = 1).

	Cause: the optimisation used to make divides occur in paralell means that when the end of the zprite is reached, it reads the next byte, and if D = -1.0 / *((float *)next_location), you get a /0 error.

	To work-around this, z-buffers have an extra float after the z-buffer data which is set to 0.0. That way, there should be no risk of dividing by zero. (hence "zb_data_size * (width*height + 1)" )

	Perhaps I should make it so that it only allocates an extra sizeof(float) bytes when allocating memory for a zprite.

	Note: An algorithm that predicted when not to start the next divide would add too much overhead (or can this be pipelined with the divide?)
	*/

	/* Allocate memory for the z-buffer data */
	new_zbuffer->dat = malloc(zb_data_size * (width*height + 1) );
	if (!new_zbuffer->dat) {
		free(new_zbuffer);
		return NULL;
	}
	// Fill the extra lword with 0.0 to prevent a divide by zero when the last byte of the depth-map is processed
	*( ((float *)new_zbuffer->dat) + (width*height) ) = 0.0;

	new_zbuffer->w=width;
	new_zbuffer->h=height;

	/*
	An array of pointers to the satrt of each line (similar to the BITMAP structure). In all cases, line[i+1] is just after line[i]+width. However, the line[] array makes it easier to find the y value.
	*/
	new_zbuffer->line[0] = new_zbuffer->dat;
	for (i=1; i<height; i++) {
		new_zbuffer->line[i] = new_zbuffer->line[i-1] + width;
	}

	return new_zbuffer;
}


/*
Destroy a Z-buffer
*/
void destroy_zbuffer(Z_BUFFER *old_zbuffer)
{
	if (old_zbuffer) {
		if (old_zbuffer->dat) {
			free(old_zbuffer->dat);
		}
		free(old_zbuffer);
	}

}


/* Fill the Z_buffer with an arbitary value */
// Perhaps could just fill part of a Zbuffer.
// This only works with 32-bit z-buffers. If it's a 16-bit z-buffer, use fill_memw() instead.
// Should I use void instead of float for the value? But what if someone passes a large datatype instead of a simple value?
void fill_zbuffer(Z_BUFFER *fill_zbuffer, float value)
{
	fill_meml(value, fill_zbuffer->dat, fill_zbuffer->h*fill_zbuffer->w);
}


/*
Clear the Z_buffer (set it to 0. Ie. The back is infinity in the distance)
In the DJGPP implementation, it just calls fill_zbuffer with 0, but when code is ported to a machine where clearing memory is faster than filling memory, this could be re-written to be much faster than fill_zbuffer.
This should be re-written to clear various different types of Z-buffer (ie, with Z values instead of 1/Z values, use MAX_FLOAT)
*/
// Perhaps could just clear part of a Zbuffer.
void clear_zbuffer(Z_BUFFER *clear_zbuffer)
{
	fill_zbuffer(clear_zbuffer, 0.0);
}
