/*
zprite.c

Z-buffered sprites with (zprites) and without (flat_zprites) depth-maps.

All these functions are for 32-bit floating-point z-buffers with 1/Z values ranging from infinity(Z=0 (the plane of the screen)) to 0(Z=infinity)

Original sprite routines written by Shawn Hargreaves.
Modified by Andrei Ellman to include Z-buffering
*/


/*
For now, when writing to a bitmap with corresponding Z-buffer, the Z-buffer size must be the same dimensions as the destination bitmap.
Later, I could mix different bitmap and z_buffer sizes to reduce memory usage. This would involve re-writing all routines that read and write z-buffers. What about speed? I think it's best to use double-buffering and have a z-buffer and a bitmap the same size as the viewport, and blit the viewport to the screen.

NOTE: Having a C function that does nothing other than call an ASM function has the added advantage that you will be warned if passing arguments of the wrong type. In the future, it could decide on wether to call the linear or modex version of the function.
*/

/*
zprite = sprite + zbuffer.
flat sprites draw sprite onto z-buffered screen using a constant Z for the entire sprite.

With zprites, depth-maps cannot have negative values. This is because I use integer compares to compare floats. This only works if both have the same sign. No problem for the destination, as <=0 behind or on viewplane, but for the source, must always be >0

zprites: no mask info used.
flat_sprites: mask info used (ie. masked (zero) pixels skipped). This is so correct 1/z value can be recorded. FYI, with regular zprites, the background is defined to be infinitely far away.

Remember, It may be a good idea to clear the z-buffer to a value less than infinity if you don't want the masked part to mess up the background. When set to infinity (1/z=0.0), zprites' backgrounds are drawn, but not flat_zprites' backgrounds (background = area not part of image)
*/

/*
TODO:

[flat]_lit_zprite_[v|h|vh]_flip()

MODE X version.

A general-purpose function for X-Y size calculation for displace_zprite (see function's rems). I have already included such a function in my 3dspdemo (main3dsp.c), but it may not be general enough for all purposes.

In translucent Zprites, re-write so that the offset is only done once per zprite. See Shawn's e-mail.

Should I Merge the flat and non_flat functions so that if 'NULL' is passed as the depth-map, then use the flat version? Not a good idea, coz then, people would have to pass a dummy value for the depth-vaule when using depth-maps.
*/

/*
FUTURE:

A seperate function called "stretch_zprite" does not make the zprite move in the 3rd dimension, but it stretches the Zs accordingly. Multiply the Z values by some stretch factor. X and Y are stretched by whatever the user tells them to be stretched by (which should correspond to the Z stretch). Should I make a command to stretch and displace a depth-map at the same time?

In displace_zprite: does it always have to stretch the whome BITMAP/ZB?, or can it stretch just part (displace_zprite is the wrapper to some function that has 0,0 in the corners).

[draw|displace|stretch]_[flat_]trans_lit_zprite()

Versions of all functions for flat_zprite, that don't record the z-value (for static scenes, this can save a lot of time, as the buffer needn't be cleared).
Fow when the background is not moving (ie. the camera isn't moving), if all zprites are depth-sorted, there needn't be any need to record the new Z values in the Z-buffer, and the z-buffer needn't be cleared (but a copy of the screen-background must exist). However, zprites with depth-maps must not overlap other zprites.
Also have versions that only record and don't check (quick if you're 100% certain the zprite is nearer to the viewer than the background).

What about RLE sprites and Compiled z-buffered sprites?
With mostly flat depth-maps, an RLE Z-buffer would save lots of RAM.
However, Allegro's regular RLE sprites cannot be stretched or rotated (or at least not without a significant slowdown).
I could try compiled trans sprites.

80586 optimised version of the zoomer that doesn't use scanline filler.
Then, the version with the scanline filler would be for 80386s, and I could get rid of the next-divide optimisation.
THIS would mean that I no longer need to add an extra entry after the end of the Z-buffer.
*/




#include <sys/farptr.h>		// C functions that write to the screen directly

#include <allegro.h>

#include "intern.h"
#include "opcodes.h"		// For writing scanline fillers

#include "zbuf.h"

#include "zprite.h"



/*
Automatically generates a flat surface for the sprite. Using a zprite generated from a bitmap in this fasion is equivalent to using a flat zprite. However, the user may want to later apply some kind of distortion to the generated depth-map.

Note: the Z value is in fact a 1/z value.
Given a zprite, all areas not the mask colour are turned into the 1/Z value.
All areas that are the mask colour are turned into 1/z = 0.

This is also useful (but expensive) for sprite drawing where you want to assign an arbitary colour as the mask. It does however have the bonus that you can assign a depth to the sprites.

The zprite's bitmap must be a memory bitmap, not the screen or a sub-bitmap.
This code will have to change if sub-zbuffers are implemented (ie. It would have to check zbuffer->line[y] for each line).
*/
void zprite_make_z(BITMAP *the_sprite, Z_BUFFER *the_zbuffer, float z, int colour)
{
	_zprite_make_z(the_sprite, the_zbuffer, z, colour);
}



/*
draw_zprite()

Draws a sprite with depth-map (zprite) onto the destination bitmap and z-biffer using z-buffering.

The source bitmap and zbuffer must be of the same dimensions. The destination bitmap and zbuffer must also be the same, but the source and destination can of course be of different sizes.

The source zprite must be a memory bitmap, not the screen or a sub-bitmap. The destination can be any bitmap (but not MODEX ... yet).

ZPRITEs obey clipping rectangles.

At the moment, all this does is call the assembler function.
In the future it could decide wehter to call the linear or MODEX version.
*/
void draw_zprite(BITMAP *destbmp, Z_BUFFER *dest_zbuffer, BITMAP *the_sprite, Z_BUFFER *the_zbuffer, int xpos, int ypos)
{
	_linear_draw_zprite(destbmp, dest_zbuffer, the_sprite, the_zbuffer, xpos, ypos);
}


/*
Flip zprite vertically
*/
void draw_zprite_v_flip(BITMAP *destbmp, Z_BUFFER *dest_zbuffer, BITMAP *the_sprite, Z_BUFFER *the_zbuffer, int xpos, int ypos)
{
	_linear_draw_zprite_v_flip(destbmp, dest_zbuffer, the_sprite, the_zbuffer, xpos, ypos);
}


/*
Flip zprite horizontally
*/
void draw_zprite_h_flip(BITMAP *destbmp, Z_BUFFER *dest_zbuffer, BITMAP *the_sprite, Z_BUFFER *the_zbuffer, int xpos, int ypos)
{
	_linear_draw_zprite_h_flip(destbmp, dest_zbuffer, the_sprite, the_zbuffer, xpos, ypos);
}


/*
Flip zprite vertically and horizontally
*/
void draw_zprite_vh_flip(BITMAP *destbmp, Z_BUFFER *dest_zbuffer, BITMAP *the_sprite, Z_BUFFER *the_zbuffer, int xpos, int ypos)
{
	_linear_draw_zprite_vh_flip(destbmp, dest_zbuffer, the_sprite, the_zbuffer, xpos, ypos);
}


/*
Z-buffered version of draw_trans_sprite()
*/
void draw_trans_zprite(BITMAP *destbmp, Z_BUFFER *dest_zbuffer, BITMAP *the_sprite, Z_BUFFER *the_zbuffer, int xpos, int ypos)
{
	_linear_draw_trans_zprite(destbmp, dest_zbuffer, the_sprite, the_zbuffer, xpos, ypos);
}


/*
Z-buffered version of draw_lit_sprite()
*/
void draw_lit_zprite(BITMAP *destbmp, Z_BUFFER *dest_zbuffer, BITMAP *the_sprite, Z_BUFFER *the_zbuffer, int xpos, int ypos, int color)
{
	_linear_draw_lit_zprite(destbmp, dest_zbuffer, the_sprite, the_zbuffer, xpos, ypos, color);
}




/*
Note: the Z value is in fact a 1/z value.
*/
void draw_flat_zprite(BITMAP *destbmp, Z_BUFFER *dest_zbuffer, BITMAP *the_sprite, int xpos, int ypos, float z)
{
	_linear_draw_flat_zprite(destbmp, dest_zbuffer, the_sprite, xpos, ypos, z);
}


/*
Note: the Z value is in fact a 1/z value.
*/
void draw_flat_zprite_h_flip(BITMAP *destbmp, Z_BUFFER *dest_zbuffer, BITMAP *the_sprite, int xpos, int ypos, float z)
{
	_linear_draw_flat_zprite_h_flip(destbmp, dest_zbuffer, the_sprite, xpos, ypos, z);
}


/*
Note: the Z value is in fact a 1/z value.
*/
void draw_flat_zprite_v_flip(BITMAP *destbmp, Z_BUFFER *dest_zbuffer, BITMAP *the_sprite, int xpos, int ypos, float z)
{
	_linear_draw_flat_zprite_v_flip(destbmp, dest_zbuffer, the_sprite, xpos, ypos, z);
}


/*
Note: the Z value is in fact a 1/z value.
*/
void draw_flat_zprite_vh_flip(BITMAP *destbmp, Z_BUFFER *dest_zbuffer, BITMAP *the_sprite, int xpos, int ypos, float z)
{
	_linear_draw_flat_zprite_vh_flip(destbmp, dest_zbuffer, the_sprite, xpos, ypos, z);
}


/*
Note: the Z value is in fact a 1/z value.
*/
void draw_flat_trans_zprite(BITMAP *destbmp, Z_BUFFER *dest_zbuffer, BITMAP *the_sprite, int xpos, int ypos, float z)
{
	_linear_draw_flat_trans_zprite(destbmp, dest_zbuffer, the_sprite, xpos, ypos, z);
}


/*
Note: the Z value is in fact a 1/z value.
*/
void draw_flat_lit_zprite(BITMAP *destbmp, Z_BUFFER *dest_zbuffer, BITMAP *the_sprite, int xpos, int ypos, float z, int color)
{
	_linear_draw_flat_lit_zprite(destbmp, dest_zbuffer, the_sprite, xpos, ypos, z, color);
}






/*
Helper function for void do_displace_zprite(). Builds a machine code stretching routine in the scratch memory area.
Originally by Shawn Hargreaves.
Extended by Andrei Ellman to handle Z-buffering, depth-maps and translucency.

Note: The STOSB and LODSB macros (without REP) when expanded are the equivalent movb/inc instructions. That way, they're faster on 80586s.

Masked=true for flat, false for non-flat.

Future work:
+ Write to the screen one long at a time (with translucent zprites, read as well).
(See draw_trans_zprite). However, this requires an extra register.
+ With masked pixels, For little -> big, Calculate the number of dest pixels to skip for each masked pixel. Problem if dest_size not an integral multiple of source_size.

The opcode macros make use of compiler_pos.
*/
static int make_zb_stretcher(int compiler_pos, fixed sx, fixed sxd, int dest_width, int masked, int trans, float z)
{
	int x, x2;
	int c;

	static long tmp;	/* The ASM stretcher uses this to pass from FPU to regs */


	if (dest_width > 0) {
		if (sxd == itofix(1)) {				 /* easy case for 1 -> 1 scaling */
			if (masked) {
				for (c=0; c<dest_width; c++) {
					COMPILER_LODSB();
					if (trans) {
						COMPILER_MASKED_FLAT_TRANS_ZB_STOSB(z);
					} else {
						COMPILER_MASKED_FLAT_ZB_STOSB(z);
					}
				}
			} else {
				for (c=0; c<dest_width; c++) {
					COMPILER_ZB_END_THIS_CALC();
					COMPILER_ZB_LODSB();
					COMPILER_ZB_START_NEXT_CALC();
					if (trans) {
						COMPILER_TRANS_ZB_STOSB();
					} else {
						COMPILER_ZB_STOSB();
					}
				}
	 		}
		}
		else if (sxd > itofix(1)) {			/* big -> little scaling */
			for (x=0; x<dest_width; x++) {
				if (masked) {
					COMPILER_LODSB();
					if (trans) {
						COMPILER_MASKED_FLAT_TRANS_ZB_STOSB(z);
					} else {
						COMPILER_MASKED_FLAT_ZB_STOSB(z);
					}
				} else {
					COMPILER_ZB_END_THIS_CALC();
					COMPILER_ZB_LODSB();
					COMPILER_ZB_START_NEXT_CALC();
					if (trans) {
						COMPILER_TRANS_ZB_STOSB();
					} else {
						COMPILER_ZB_STOSB();
					}
				}
				x2 = (sx >> 16) + 1;
				sx += sxd;
				x2 = (sx >> 16) - x2;
				if (x2 > 1) {
					COMPILER_ADD_ESI(x2);
					if (!masked) {
						COMPILER_ADD_EBX(x2<<2)
					}
				} else if (x2 == 1) {
					COMPILER_INC_ESI();
					if (!masked) {
						COMPILER_ADD_EBX_4()
					}
				}
			}
		} else  {									  /* little -> big scaling */
	 		x2 = sx >> 16;
			if (masked) {
				COMPILER_LODSB();
			} else {
				COMPILER_ZB_END_THIS_CALC();
				COMPILER_ZB_LODSB();
				COMPILER_ZB_START_NEXT_CALC();
			}
			for (x=0; x<dest_width; x++) {
				if (masked) {
					if (trans) {
						COMPILER_MASKED_FLAT_TRANS_ZB_STOSB(z);
					} else {
						COMPILER_MASKED_FLAT_ZB_STOSB(z);
					}
				} else {
					if (trans) {
						COMPILER_TRANS_ZB_STOSB();
					} else {
						COMPILER_ZB_STOSB();
					}
				}
				sx += sxd;
				if ((sx >> 16) > x2) {
					if (masked) {
						COMPILER_LODSB();
					} else {
						COMPILER_ZB_END_THIS_CALC();
						COMPILER_ZB_LODSB();
						COMPILER_ZB_START_NEXT_CALC();
					}
					x2++;
				}
			}

		}
	}

	return compiler_pos;
}



/*
Given the sizes and z value, similar to do_stretch_sprite, but uses z-buffers with the given z value.
*/
static void do_displace_zprite(BITMAP *source, Z_BUFFER *source_zbuffer, BITMAP *dest, Z_BUFFER *dest_zbuffer, int source_x, int source_y, int source_width, int source_height, int dest_x, int dest_y, int dest_width, int dest_height, int masked, int trans, float z)
{
	fixed sx, sy, sxd, syd;
	int compiler_pos = 0;	/* The opcode macros make use of this */

	/* trivial reject for zero sizes */
	if ((source_width <= 0) || (source_height <= 0) ||
		 (dest_width <= 0) || (dest_height <= 0))
			return;

	/* convert to fixed point */
	sx = itofix(source_x);
	sy = itofix(source_y);

	/* calculate delta values */
	sxd = itofix(source_width) / dest_width;
	syd = itofix(source_height) / dest_height;

	/* clip? */
	if (dest->clip) {
		while (dest_x < dest->cl) {
		 dest_x++;
		 dest_width--;
		 sx += sxd;
		}

		while (dest_y < dest->ct) {
		 dest_y++;
		 dest_height--;
		 sy += syd;
		}

		if (dest_x+dest_width > dest->cr)
		 dest_width = dest->cr - dest_x;

		if (dest_y+dest_height > dest->cb)
		 dest_height = dest->cb - dest_y;

		if ((dest_width <= 0) || (dest_height <= 0))
		 return;
	}

	/* build a simple linear stretcher */
	compiler_pos = make_zb_stretcher(0, sx, sxd, dest_width, masked, trans, z);

	COMPILER_RET();

	/*
	call the stretcher function for each line
	ZB: masked=false. FLAT_ZB: masked=true
	*/
	if (masked) {
		_do_flat_zb_displace(source, dest, _scratch_mem, sx>>16, sy, syd, dest_x, dest_y, dest_height, dest_zbuffer);
	} else {
		_do_zb_displace(source, dest, _scratch_mem, sx>>16, sy, syd, dest_x, dest_y, dest_height, source_zbuffer, dest_zbuffer, z);
	}
}




/*
Stretches the zprite's sprite and depth-map to size w, h and displaces the depth of the zprite with the given Z value.

Zooms the entire zprite to the destination size. The zprite's bitmap and zbuffer must be the same size

Given the right formula for calculating w,h for a given Zchange, this can re-position the sprites anywhere in the 3D world.

This has the effect of displacing the zprite in 3D on the destination.

This is different from stretch_zprite (not yet implemented) which stretches the sprite and depth-map to w,h and given a central depth-value, stretches the depth-map by a certain ammount.

Zchange is a Z value, not a 1/Z valure.
new z = old z + Zchange
new 1/z = 1 / ( 1/ (old 1/z) + Zchange )
new 1/z = old 1/z / (1 + Zchange *  old 1/z)

Z can be the absolute Z value if the depth-map remains the same.

When displacing by Z, use the following to get x and y, w, h.
+ Assume sprite's position in viewspace is the center of the sprite.
+ project the width and height of the sprite (as they would be when the sprite is on the z=1 plane) to screenspace with persp_project() using the Z position relative to the camera. Call the output w,h
+ For the sprite stretch: {top-left} = {center}-(0.5{w,h}), {bottom-right} = {center}+(0.5{w,h})

That is:
Assuming x and y are the center of the sprite:

	persp_project[_f](SPRITE_XSIZE*XSCALE, (SPRITE_YSIZE*YSCALE)*(SCREEN_W/SCREEN_H), Z, &projected_sprite_w, &projected_sprite_h);

	projected_sprite_w -= (SCREEN_W>>1);
	projected_sprite_h -= (SCREEN_H>>1);

	stretch_sprite(dest, src, x-(projected_sprite_w>>1), y-(projected_sprite_h>>1), projected_sprite_w, projected_sprite_h);

SEE main3dsp.c for an example.

Uses the entire zprite as the source.
Takes Z values. Should it take 1/Z values instead?
Z corresponds to where z=0 plane of the depthmap maps to in screenspace.
Note: The depth-map shouldn't contain negative Z values, so the given Z value is the minimum Z value that will be written.
Users of isometric 3D might make this take MAXINT-Z
*/
void displace_zprite(BITMAP *destbmp, Z_BUFFER *dest_zbuffer, BITMAP *the_sprite, Z_BUFFER *the_zbuffer, int xpos, int ypos, int w, int h, float z)
{
   do_displace_zprite(the_sprite, the_zbuffer, destbmp, dest_zbuffer, 0, 0, the_sprite->w, the_sprite->h, xpos, ypos, w, h, FALSE, FALSE, z);	/* Uses entire zprite as source, hence 0,0,the_sprite->w,the_sprite->h */
}


/*
Trans version of displace_zprite
Uses the entire zprite as the source.
Takes Z values. Should it take 1/Z values instead?
Z corresponds to where z=0 plane of the depthmap maps to in screenspace.
Users of isometric 3D might make this take MAXINT-Z
*/
void displace_trans_zprite(BITMAP *destbmp, Z_BUFFER *dest_zbuffer, BITMAP *the_sprite, Z_BUFFER *the_zbuffer, int xpos, int ypos, int w, int h, float z)
{
   do_displace_zprite(the_sprite, the_zbuffer, destbmp, dest_zbuffer, 0, 0, the_sprite->w, the_sprite->h, xpos, ypos, w, h, FALSE, TRUE, z);
}


/*
Flat version of displace_zprite. Takes an absolute Z value.
Uses the entire zprite's bitmap as the source.
Takes 1/Z, values. Should it take Z values instead?
Users of isometric 3D might make this take MAXINT-Z
*/
void displace_flat_zprite(BITMAP *destbmp, Z_BUFFER *dest_zbuffer, BITMAP *sprite, int xpos, int ypos, int w, int h, float z)
{
   do_displace_zprite(sprite, NULL, destbmp, dest_zbuffer, 0, 0, sprite->w, sprite->h, xpos, ypos, w, h, TRUE, FALSE, z);
}


/*
Flat_trans version of displace_zprite. of above function. Takes an absolute Z value.
Uses the entire zprite's bitmap as the source.
Takes 1/Z, values. Should it take Z valuse instead?
Users of isometric 3D might make this take MAXINT-Z
*/
void displace_flat_trans_zprite(BITMAP *destbmp, Z_BUFFER *dest_zbuffer, BITMAP *sprite, int xpos, int ypos, int w, int h, float z)
{
   do_displace_zprite(sprite, NULL, destbmp, dest_zbuffer, 0, 0, sprite->w, sprite->h, xpos, ypos, w, h, TRUE, TRUE, z);
}






/*
Helper for rotating zprites and flat_zprites.

Everything except the z-buffering was copied from Allegro's original code.
Must be a linear bitmap.
This should be re-written in assembly.
*/
void do_rotate_zprite(BITMAP *bmp, Z_BUFFER *dest_zb, BITMAP *source_bmp, Z_BUFFER *source_zb, int x, int y, fixed angle, float zvalue)
{
	float *dest_zb_line_adr;

	fixed f1x, f1y, f1xd, f1yd;
	fixed f2x, f2y, f2xd, f2yd;
	fixed w, h;
	fixed dir, dir2, dist;
	int x1, y1, x2, y2;
	int dx, dy;
	int sx, sy;
	unsigned long addr;
	int wgap, hgap;

	wgap = source_bmp->w;
	hgap = source_bmp->h;

	/* rotate the top left pixel of the source_bmp */
	w = itofix(wgap/2);
	h = itofix(hgap/2);
	dir = fatan2(h, w);
	dir2 = fatan2(h, -w);
	dist = fsqrt((wgap*wgap + hgap*hgap) << 6) << 4;
	f1x = w - fmul(dist, fcos(dir - angle));
	f1y = h - fmul(dist, fsin(dir - angle));

	/* map the destination y axis onto the source_bmp */
	f1xd = fcos(itofix(64) - angle);
	f1yd = fsin(itofix(64) - angle);

	/* map the destination x axis onto the source_bmp */
	f2xd = fcos(-angle);
	f2yd = fsin(-angle);

	/* adjust the size of the region to be scanned */
	x1 = fixtoi(fmul(dist, fcos(dir + angle)));
	y1 = fixtoi(fmul(dist, fsin(dir + angle)));

	x2 = fixtoi(fmul(dist, fcos(dir2 + angle)));
	y2 = fixtoi(fmul(dist, fsin(dir2 + angle)));

	x1 = MAX(ABS(x1), ABS(x2)) - wgap/2;
	y1 = MAX(ABS(y1), ABS(y2)) - hgap/2;

	x -= x1;
	wgap += x1 * 2;
	f1x -= f2xd * x1;
	f1y -= f2yd * x1;

	y -= y1;
	hgap += y1 * 2;
	f1x -= f1xd * y1;
	f1y -= f1yd * y1;

	/* clip the output rectangle */
	if (bmp->clip) {
		while (x < bmp->cl) {
			x++;
			wgap--;
			f1x += f2xd;
			f1y += f2yd;
		}

		while (y < bmp->ct) {
			y++;
			hgap--;
			f1x += f1xd;
			f1y += f1yd;
		}

		while (x+wgap >= bmp->cr)
			wgap--;

		while (y+hgap >= bmp->cb)
			hgap--;

		if ((wgap <= 0) || (hgap <= 0))
			return;
	}

	/* select the destination segment */
	_farsetsel(bmp->seg);

	/* and trace a bunch of lines through the bitmaps */
	for (dy=0; dy<hgap; dy++) {
		f2x = f1x;
		f2y = f1y;

		addr = bmp_write_line(bmp, y+dy) + x;
		dest_zb_line_adr = dest_zb->line[y+dy] + x;

		if (source_zb==NULL) {

			/* Flat */
			for (dx=0; dx<wgap; dx++) {
				sx = fixtoi(f2x);
				sy = fixtoi(f2y);

				if ((sx >= 0) && (sx < source_bmp->w) && (sy >= 0) && (sy < source_bmp->h)) {

					if (zvalue > *dest_zb_line_adr) {
						if (source_bmp->line[sy][sx]) {
							_farnspokeb(addr, source_bmp->line[sy][sx]);	// Write pixel
							*dest_zb_line_adr = zvalue;	// Record new 1/z
						}
					}

				}

				addr++;
				dest_zb_line_adr++;
				f2x += f2xd;
				f2y += f2yd;
			}

		} else {

			/* Non-Flat */
			for (dx=0; dx<wgap; dx++) {
				sx = fixtoi(f2x);
				sy = fixtoi(f2y);

				if ((sx >= 0) && (sx < source_bmp->w) && (sy >= 0) && (sy < source_bmp->h)) {

					if (source_zb->line[sy][sx] > *dest_zb_line_adr) {
						_farnspokeb(addr, source_bmp->line[sy][sx]);	// Write pixel
						*dest_zb_line_adr = source_zb->line[sy][sx];	// Record new 1/z
					}
				}

				addr++;
				dest_zb_line_adr++;
				f2x += f2xd;
				f2y += f2yd;
			}

		}


		f1x += f1xd;
		f1y += f1yd;
	}
}


/*
Rotates a sprite and it's depth-mask and uses z-buffering to write the result.
Should the angle be a float value instead of fixed?
*/
void rotate_zprite(BITMAP *bmp, Z_BUFFER *dest_zb, BITMAP *sprite, Z_BUFFER *zbuffer, int x, int y, fixed angle)
{
	do_rotate_zprite(bmp, dest_zb, sprite, zbuffer, x, y, angle, 0.0);
}


/*
Rotates a sprite and uses z-buffering to write the result.
z is a 1/z value
*/
void rotate_flat_zprite(BITMAP *bmp, Z_BUFFER *dest_zb, BITMAP *sprite, int x, int y, fixed angle, float z)
{
	do_rotate_zprite(bmp, dest_zb, sprite, NULL, x, y, angle, z);
}
