/*
zpriteds.c

Some aditional helper functions for facilitating working with Z-buffered sprites, AKA zprites by creating the ZPRITE Data Structure. You can of course ignore this if you want to create an alternative datastructure.

These are a group of functions for facilitating the creation and deletion of zprites. There's also one for clearing zprites.

C++ users may want to write a zprite class instead of using my zpriteds.[c|h] files

Note: This file only contains functions for working with ZPRITE structures. The actual functions for working with z-buffered sprites are in zprite.[c|h]

This won't officially be part of AllegroPak, but instead, it would be part of the Allegropak demo (keeping in the spirit of Allegro: Avoid complex datastructures).

Now that the ZPRITE structure has been separated from the core library, is it still necessary to have to dynamically allocate the memory for the ZPRITE's members?
*/



#include <stdlib.h>			// malloc()

#include <allegro.h>

#include "zbuf.h"

#include "zpriteds.h"


ZPRITE *create_zprite(int width, int height) {
	ZPRITE *new_zprite;

	new_zprite = malloc(sizeof(ZPRITE));
	if(!new_zprite){
		return NULL;
	}

	new_zprite->bitmap = create_bitmap(width, height);
	if(!new_zprite->bitmap) {
		free(new_zprite);
		return NULL;
	}

	new_zprite->z_buffer = create_zbuffer(width, height);
	if(!new_zprite->z_buffer) {
		free(new_zprite);
		destroy_bitmap(new_zprite->bitmap);
		return NULL;
	}

	new_zprite->w=width;
	new_zprite->h=height;

	return new_zprite;
}


ZPRITE *create_zprite_clear(int width, int height)
/*
Similar to above, bit clears the bitmap and the z_buffer.
*/
{
	ZPRITE *new_zprite;

	new_zprite = malloc(sizeof(ZPRITE));
	if(!new_zprite){
		return NULL;
	}

	new_zprite->bitmap = create_bitmap(width, height);
	if(!new_zprite->bitmap) {
		free(new_zprite);
		return NULL;
	}

	new_zprite->z_buffer = create_zbuffer(width, height);
	if(!new_zprite->z_buffer) {
		free(new_zprite);
		destroy_bitmap(new_zprite->bitmap);
		return NULL;
	}

	new_zprite->w=width;
	new_zprite->h=height;

	clear(new_zprite->bitmap);
	clear_zbuffer(new_zprite->z_buffer);

	return new_zprite;
}

ZPRITE *set_zprite_to_bitmap(BITMAP *existing_bitmap)
/*
Similar to create_zprite(), but uses an existing bitmap for the zprite.
This is useful for turning sprites stored in datafiles into zprites.

Be careful to use destroy_set_to_bitmap_zprite instead of destroy_zprite.
Destroying a zprite created this way with destroy_zprite would then destroy the original sprite bitmap.

Should I add a flag to the zprite so that it knows which function created it? Then, I could use the appropriate dedstroy_[set_to_*]zprite() function. This would be more elegant in C++
*/
{
	ZPRITE *new_zprite;

	new_zprite = malloc(sizeof(ZPRITE));
	if(!new_zprite){
		return NULL;
	}

	new_zprite->z_buffer = create_zbuffer(existing_bitmap->w, existing_bitmap->h);
	if(!new_zprite->z_buffer) {
		free(new_zprite);
		return NULL;
	}

	new_zprite->bitmap = existing_bitmap;	// Must be same size as bitmap

	new_zprite->w=existing_bitmap->w;
	new_zprite->h=existing_bitmap->h;

	return new_zprite;
}


ZPRITE *set_zprite_to_bitmap_clear(BITMAP *existing_bitmap)
/*
Same as above, but clears the newly-created z-buffer
*/
{
	ZPRITE *new_zprite;

	new_zprite = malloc(sizeof(ZPRITE));
	if(!new_zprite){
		return NULL;
	}

	new_zprite->z_buffer = create_zbuffer(existing_bitmap->w, existing_bitmap->h);
	if(!new_zprite->z_buffer) {
		free(new_zprite);
		return NULL;
	}

	new_zprite->bitmap = existing_bitmap;	// Must be same size as bitmap

	new_zprite->w=existing_bitmap->w;
	new_zprite->h=existing_bitmap->h;

	clear_zbuffer(new_zprite->z_buffer);

	return new_zprite;
}


ZPRITE *set_zprite_to_zbuffer(Z_BUFFER *existing_zbuffer)
/*
Similar to create_zprite(), but uses an existing z-buffer for the zprite.
This is useful for re-using existing depth-maps with different bitmaps.
(make sure that the masked pixels are the same for both bitmaps).

Be careful to use destroy_set_to_zbuffer_zprite instead of destroy_zprite.
Destroying a zprite created this way with destroy_zprite would then destroy the original depth-map.

Should I add a flag to the zprite so that it knows which function created it? Then, I could use the appropriate dedstroy_[set_to_*]zprite() function.
*/
{
	ZPRITE *new_zprite;

	new_zprite = malloc(sizeof(ZPRITE));
	if(!new_zprite){
		return NULL;
	}

	new_zprite->bitmap = create_bitmap(existing_zbuffer->w, existing_zbuffer->h);
	if(!new_zprite->bitmap) {
		free(new_zprite);
		return NULL;
	}

	new_zprite->z_buffer = existing_zbuffer;	// Must be same size as z_buffer

	new_zprite->w=existing_zbuffer->w;
	new_zprite->h=existing_zbuffer->h;

	return new_zprite;
}


ZPRITE *set_zprite_to_zbuffer_clear(Z_BUFFER *existing_zbuffer)
/*
Same as above, but clears the newly-created bitmap.
*/
{
	ZPRITE *new_zprite;

	new_zprite = malloc(sizeof(ZPRITE));
	if(!new_zprite){
		return NULL;
	}

	new_zprite->bitmap = create_bitmap(existing_zbuffer->w, existing_zbuffer->h);
	if(!new_zprite->bitmap) {
		free(new_zprite);
		return NULL;
	}

	new_zprite->z_buffer = existing_zbuffer;	// Must be same size as z_buffer

	new_zprite->w=existing_zbuffer->w;
	new_zprite->h=existing_zbuffer->h;

	clear(new_zprite->bitmap);

	return new_zprite;
}


ZPRITE *set_zprite_to_both(BITMAP *existing_bitmap, Z_BUFFER *existing_zbuffer)
/*
Similar to create_zprite(), but uses an existing z-buffer and an existing bitmap for the zprite.
This is useful for re-using existing depth-maps with existing bitmaps.

Be careful to use destroy_set_to_both_zprite instead of destroy_zprite.
Destroying a zprite created this way with destroy_zprite would then destroy the original bitmap and depth-map.

WARNING: The zprite's bitmap and depth-map must be of equal dimensions to the size passed to the function. It uses the bitmap to get the dimensions.

Should I add a flag to the zprite so that it knows which function created it? Then, I could use the appropriate dedstroy_[set_to_*]zprite() function.
*/
{
	ZPRITE *new_zprite;

	new_zprite = malloc(sizeof(ZPRITE));
	if(!new_zprite){
		return NULL;
	}

	new_zprite->bitmap = existing_bitmap;	// Must be same size as z_buffer

	new_zprite->z_buffer = existing_zbuffer;	// Must be same size as bitmap

	/* Get the size from the bitmap. The Z-buffer MUST be of the same size */
	new_zprite->w=existing_bitmap->w;
	new_zprite->h=existing_bitmap->h;

	return new_zprite;
}


void destroy_zprite(ZPRITE *old_zprite)
/*
Destroys a zprite created with create_zprite() or create_zprite_clear()
*/
{
	if(old_zprite->bitmap) {
		destroy_bitmap(old_zprite->bitmap);
	}
	if(old_zprite->z_buffer) {
		destroy_zbuffer(old_zprite->z_buffer);
	}
	free(old_zprite);
}


void destroy_set_to_bitmap_zprite(ZPRITE *old_zprite)
/*
Destroys a zprite created with set_zprite_to_bitmap()
Destroys the zprite's Z_BUFFER, but not the BITMAP.
Useful for when a zprite was made from a datafile bitmap.
*/
{
	if (old_zprite->z_buffer) {
		destroy_zbuffer(old_zprite->z_buffer);
	}
	free(old_zprite);
}


void destroy_set_to_zbuffer_zprite(ZPRITE *old_zprite)
/*
Destroys a zprite created with set_zprite_to_z_buffer()
Destroys the zprite's BITMAP, but not the Z_BUFFER.
Useful for when a zprite was made from an existing depth-map.
*/
{
	if (old_zprite->bitmap) {
		destroy_bitmap(old_zprite->bitmap);
	}
	free(old_zprite);
}


void destroy_set_to_both_zprite(ZPRITE *old_zprite)
/*
Destroys a zprite created with set_zprite_to_both()
Destroys only the ZPRITE datastructure, and not it's BITMAP or Z_BUFFER.
Useful for grouping existing bitmaps with existing depth-maps
*/
{
	free(old_zprite);
}


void clear_zprite(ZPRITE *the_zprite)
/*
Clears both the bitmap and the z-buffer of the zprite
*/
{
	clear(the_zprite->bitmap);
	clear_zbuffer(the_zprite->z_buffer);
}
